/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2;

import dk.nsi.fmk.admin.Admin;
import dk.nsi.fmk.advis.AdvisService;
import dk.sds.ecpr2.advis.AdvisModuleClientFactory;
import dk.sds.ecpr2.advis.AsyncAdvisFacade;
import dk.sds.ecpr2.advis.AsyncAdvisFacadeModuleImpl;
import dk.sds.ecpr2.datagenerator.DataGeneratorApplication;
import dk.sds.ecpr2.service.admin.AdministrationFacade;
import dk.sds.ecpr2.service.admin.impl.AdminModuleAdministrationFacadeImpl;
import dk.sds.ecpr2.service.admin.impl.AdminModuleClientFactory;
import dk.sds.ecpr2.service.admin.impl.AdministrationFacadeWrapper;
import dk.sds.ecpr2.util.CustomValidationInterceptor;
import dk.sds.ecpr2.security.RoleValidationInterceptor;
import dk.sds.ecpr2.util.TimeHelper;
import lombok.extern.log4j.Log4j2;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.Banner;
import org.springframework.boot.SpringApplication;
import org.springframework.boot.WebApplicationType;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.context.ConfigurableApplicationContext;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.ComponentScan;
import org.springframework.context.annotation.FilterType;
import org.springframework.core.env.ConfigurableEnvironment;
import org.springframework.core.io.ClassPathResource;
import org.springframework.ws.config.annotation.WsConfigurerAdapter;
import org.springframework.ws.server.EndpointInterceptor;
import org.springframework.xml.xsd.SimpleXsdSchema;
import org.springframework.xml.xsd.XsdSchema;

import javax.annotation.PostConstruct;
import java.security.Security;
import java.util.Date;
import java.util.List;


@SpringBootApplication
@ComponentScan(excludeFilters={@ComponentScan.Filter(type=FilterType.ASSIGNABLE_TYPE, value= DataGeneratorApplication.class)})
@Log4j2
public class ECPR2Application extends WsConfigurerAdapter {

    @Autowired
    dk.sds.ecpr2.util.Properties properties;

    @Autowired
    private ConfigurableEnvironment springEnv;

    public static Date DEPLOY_TIME = TimeHelper.getCurrentTime();

    public static void main(String[] args) {
        if (args.length > 0 && args[0].toLowerCase().equals("generatedata")) {
            runAsDataGenerator(args);
        } else {
            runAsService(args);
        }
    }

    private static void runAsService(String[] args) {
        Security.addProvider(new BouncyCastleProvider());
        SpringApplication.run(ECPR2Application.class, args);
    }

    private static void runAsDataGenerator(String[] args) {
        SpringApplication app = new SpringApplication(DataGeneratorApplication.class);
        app.setWebApplicationType(WebApplicationType.NONE);
        app.setBannerMode(Banner.Mode.OFF);
        ConfigurableApplicationContext context = app.run(args);
        context.close();
    }

    /*@Bean
    public WsdlDefinition serviceDefinition_20171001(XsdSchema schema) {
        final DefaultWsdl11Definition bean = new DefaultWsdl11Definition();
        bean.setSchema(schema);
        bean.setPortTypeName("PersonalDataCardPort");
        final Properties soapActions = new Properties();
        soapActions.put("GetPersonalDataCard",      "http://sds.dk/ecpr2/2017/10/01/GetPersonalDataCard");
        soapActions.put("UpdateContactInformation", "http://sds.dk/ecpr2/2017/10/01/UpdateContactInformation");
        bean.setSoapActions(soapActions);
        return bean;
    }*/

    @Bean
    public RoleValidationInterceptor roleValidationInterceptor() {
        return new RoleValidationInterceptor();
    }

    @Bean
    public AsyncAdvisFacade asyncAdvisFacade(AdvisService advisService) {
        return new AsyncAdvisFacadeModuleImpl(advisService);
    }

    @Bean
    public AdvisService advisService(dk.sds.ecpr2.util.Properties properties) {
        return new AdvisModuleClientFactory(properties).getAdvisServiceClient();
    }

    @Bean
    public XsdSchema schema() {
        return new SimpleXsdSchema(new ClassPathResource("schema/2017_10_01/ecpr2service.xsd"));
    }

    @Bean
    public AdminModuleClientFactory adminModuleClientFactory(){
        AdminModuleClientFactory adminModuleClientFactory = new AdminModuleClientFactory(false);
        return adminModuleClientFactory;
    }

    @Bean
    public Admin adminClient(){
        return adminModuleClientFactory().getAdminServiceClient();
    }

    @Bean
    public AdministrationFacade adminModuleAdministrationFacade(){
        return new AdminModuleAdministrationFacadeImpl();
    }

    @Bean
    public AdministrationFacade administrationFacade(){
        return new AdministrationFacadeWrapper();
    }

    @Override
    public void addInterceptors(List<EndpointInterceptor> interceptors) {
        final CustomValidationInterceptor validatingInterceptor = new CustomValidationInterceptor(properties);
        validatingInterceptor.setValidateRequest(properties.schemavalidationValidateRequests);
        validatingInterceptor.setValidateResponse(properties.schemavalidationValidateResponse);
        validatingInterceptor.setXsdSchema(schema());
        interceptors.add(validatingInterceptor);

        interceptors.add(roleValidationInterceptor());

        super.addInterceptors(interceptors);
    }

    @PostConstruct
    public void logSettings() {
        log.info("Info logging is enabled.");
        log.debug("Debug logging is enabled.");
        log.trace("Trace logging is enabled.");
    }
    

}
