/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao;

import dk.sds.ecpr2.security.AuthorisationVO;
import dk.sds.ecpr2.security.Permission;
import dk.sds.ecpr2.util.TimeHelper;
import org.joda.time.DateTime;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.stereotype.Component;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Date;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

@Component
public class AuthorisationDAOImpl implements AuthorisationDAO {

    @Autowired
    JdbcTemplate jdbcTemplate;

    // Must match system name on bemyndigelse service, se etc/bemyndigelse/metadata.xml
    private static final String BEMYNDIGELSE_SYSTEM = "eCPR2";
    // Bemyndigelser must have this status to be valid (see bemyndigelses service documentation for details)
    private static final String BEMYNDIGELSE_GODKENDT = "Godkendt";

    private static final String EDUCATION_CODES = "'7170', '5433', '5175', '5166', '5152', 'B511', '0'";

    @Value("${sdm.schemaname}")
    private String sdmSchema;

    private static final RowMapper<AuthorisationVO> rowMapper = (resultSet, i) -> {
        AuthorisationVO authorisationVO = new AuthorisationVO();
        authorisationVO.setAuthorisationId(resultSet.getString("autorisationsNummer"));
        authorisationVO.setEducationCode(resultSet.getString("uddannelsesKode"));
        authorisationVO.setPersonIdentifier(resultSet.getString("cpr"));
        return authorisationVO;
    };

    @Override
    public List<AuthorisationVO> getAuthorisationsForCpr(String userCpr) {
        Date now = TimeHelper.getCurrentTime();
        String sql = "select * from " + sdmSchema + ".Autorisation " + //
                "where cpr = ? and validFrom <= ? and validTo > ? " + //
                "and UddannelsesKode in (" + EDUCATION_CODES + ") " + //
                "order by UddannelsesKode DESC ";

        return jdbcTemplate.query(sql, rowMapper, userCpr, now, now);
    }

    @Override
    public boolean isChemist(String userCpr) {
        Date now = TimeHelper.getCurrentTime();
        String sql = "select count(*) from " + sdmSchema + ".Apotekere " + //
                "where CPR = ? and (ValidFrom < ? AND ValidTo >= ?) order by LastReplicated desc";
        return jdbcTemplate.queryForObject(sql, Long.class, userCpr, now, now) >= 1;
    }

    @Override
    public Set<String> getAllApotekCvr() {
        Date now = TimeHelper.getCurrentTime();
        String sql = "SELECT DISTINCT(SorParent.CVRnumber) FROM " + sdmSchema + ".SorOrganisation SorParent INNER JOIN " + sdmSchema + ".SorOrganisation SorPharmacy ON SorPharmacy.ParentSorNumber = SorParent.SorNumber " +
                "WHERE (SorPharmacy.SorType = 'apotek' OR SorPharmacy.SorType = 'apoteksfilial') " +
                "AND SorParent.CVRnumber IS NOT NULL " +
                "AND SorPharmacy.EanLocationCode IS NOT NULL " +
                "AND SorPharmacy.ValidFrom <= ? AND SorPharmacy.ValidTo > ? AND SorParent.ValidFrom <= ? AND SorPharmacy.ValidTo > ?";
        return jdbcTemplate.queryForList(sql, now, now, now, now).stream().map(a -> String.valueOf(a.get("CVRnumber"))).collect(Collectors.toSet());
    }

    private static final RowMapper<Permission> rowMapperPermission = (rs, row) -> {
        String delegationPermission = rs.getString("rettighed");

        // Bem uses danish ids. Need to convert them to english
        if ("Læs".equals(delegationPermission)) {
            return Permission.Read;
        } else if ("Skriv".equals(delegationPermission)) {
            return Permission.Write;
        }
        return Permission.valueOf(delegationPermission);
    };

    @Override
    public List<Permission> getBemyndigetPermissions(String assistantCpr, String doctorCpr, String cvr, String authorizedByRole) {
        boolean hasCvr = (cvr != null && cvr.length() > 0);

        StringBuilder sqlBuilder = new StringBuilder();
        sqlBuilder.append("SELECT bem.rettighed FROM " + sdmSchema + ".Bemyndigelse bem")
                .append(" WHERE bem.validFrom <= ? AND bem.validTo > ?")
                .append(" AND bem.gyldig_fra_dato <= ? AND bem.gyldig_til_dato > ?")
                .append(" AND bem.system='" + BEMYNDIGELSE_SYSTEM + "'")
                .append(" AND bem.Status='" + BEMYNDIGELSE_GODKENDT + "'")
                .append(" AND bem.Arbejdsfunktion=?");

        // If bemyndigede_cvr is null or empty it works like a wildcard
        sqlBuilder.append(" AND (bem.bemyndigede_cvr IS NULL OR bem.bemyndigede_cvr=''");
        if (hasCvr) {
            sqlBuilder.append(" OR bem.bemyndigede_cvr=?)");
        } else {
            sqlBuilder.append(")");
        }
        sqlBuilder.append(" AND bem.bemyndigede_cpr=? AND bem.bemyndigende_cpr=?");
        sqlBuilder.append(" ORDER BY BemyndigelsePID DESC");
        String sql = sqlBuilder.toString();

        // I am not very fond of using Date() here, but this is whats being done in the other functions as well
        // it should be fixed in all functions at some point...
        Date now = TimeHelper.getCurrentTime();
        String bemStringDate = ZonedDateTime.now(ZoneOffset.UTC).format(DateTimeFormatter.ISO_INSTANT);

        if (hasCvr) {
            return jdbcTemplate.query(sql, rowMapperPermission, now, now, bemStringDate, bemStringDate,
                    authorizedByRole, cvr, assistantCpr, doctorCpr);
        } else {
            return jdbcTemplate.query(sql, rowMapperPermission, now, now, bemStringDate, bemStringDate,
                    authorizedByRole, assistantCpr, doctorCpr);
        }
    }

    @Override
    public AuthorisationVO getAuthorisation(String onBehalfOfAuthorisationNo) {
        Date now = TimeHelper.getCurrentTime();
        String sql = "select * from " + sdmSchema + ".Autorisation " + //
                "where autorisationsNummer = ? and validFrom <= ? and validTo > ? " + //
                "and UddannelsesKode in (" + EDUCATION_CODES + ") " + //
                "order by UddannelsesKode DESC ";

        List<AuthorisationVO> result = jdbcTemplate.query(sql, rowMapper, onBehalfOfAuthorisationNo, now, now);
        if (result == null) {
            return null;
        } else {
            return result.get(0);
        }
    }

    @Override
    public boolean isPersonDead(String personIdentifier, Date now) {
        String sql = "select status from " + sdmSchema + ".Person " +//
                "where CPR = ? and validFrom <= ? and validTo > ?";
        String status = jdbcTemplate.queryForObject(sql, new Object[]{personIdentifier, now, now}, String.class);
        return "90".equals(status);
    }
}
