/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.dao.AddressDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.abstr.AbstractECPRDAOImpl;
import dk.sds.ecpr2.domain.AddressVO;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.stereotype.Repository;

import java.time.Instant;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.UUID;

@Repository
public class AddressDAOImpl extends AbstractECPRDAOImpl implements AddressDAO {

    public static final String TABLE_NAME = "Address";
    final String INSERT_SQL = "INSERT INTO Address (AddressPID, Uuid, PersonPID, ValidFrom, ValidTo, `Use`, Type, StreetAddressLine1, StreetAddressLine2, StreetAddressLine3, StreetAddressLine4, StreetAddressLine5, StreetAddressLine6, StreetAddressLine7, StreetAddressLine8, StreetAddressLine9, StreetAddressLine10, City, District, State, PostalCode, Country, PostBox, Expiry, ModifiedTime)" +
            "VALUES(:addressPID, :uuid, :personPID, :validFrom, :validTo, :use, :type, :streetAddressLine1, :streetAddressLine2, :streetAddressLine3, :streetAddressLine4, :streetAddressLine5, :streetAddressLine6, :streetAddressLine7, :streetAddressLine8, :streetAddressLine9, :streetAddressLine10, :city, :district, :state, :postalCode, :country, :postBox, :expiry, :modified)";

    @Override
    public List<AddressVO> getWithServiceActor(List<String> personUuids, List<Integer> personPids) throws DAOException {
        return getWithServiceActorByPersonUUIDs(personUuids, personPids);
    }

    @Override
    public List<AddressVO> getWithServiceActor(String personUuid, Date date) throws DAOException {
        return getWithServiceActorByPersonUUIDAndDate(personUuid, date);
    }

    @Override
    public List<AddressVO> getAllAddressesBetweenPersonPIDForSearch(int minPersonPid, int maxPersonPid) throws DAOException {
        return getSelectedFieldsBetweenPersonPIDWithoutServiceActor(Arrays.asList("StreetAddressLine1", "StreetAddressLine2", "StreetAddressLine3", "StreetAddressLine4", "StreetAddressLine5", "StreetAddressLine6", "StreetAddressLine7", "StreetAddressLine8", "StreetAddressLine9", "StreetAddressLine10", "City", "District", "State", "PostalCode", "Country", "PostBox", "Uuid", "ValidFrom"),
                minPersonPid, maxPersonPid);
    }

    @Override
    public List<AddressVO> getAllAddressesBetweenPersonPID(int minPersonPID, int maxPersonPID) throws DAOException {
        return getAllFieldsBetweenPersonPIDWithServiceActor(minPersonPID, maxPersonPID);
    }

    @Override
    public List<AddressVO> getAllAddressesAfterModified(Instant modifiedTime) throws DAOException {
        return getAllFieldsAfterModifiedTimeWithServiceActor(modifiedTime);
    }

    @Override
    public List<AddressVO> getAllAddressesAfterModifiedForSearch(Instant modifiedTime) throws DAOException {
        List<String> fields = Arrays.asList("StreetAddressLine1", "StreetAddressLine2", "StreetAddressLine3", "StreetAddressLine4", "StreetAddressLine5", "StreetAddressLine6", "StreetAddressLine7", "StreetAddressLine8", "StreetAddressLine9", "StreetAddressLine10", "City", "District", "State", "PostalCode", "Country", "PostBox", "Uuid", "ValidFrom");
        return getSelectedFieldsAfterModifiedTimeWithoutServiceActor(fields, modifiedTime);
    }

    @Override
    public Integer createAddress(AddressVO address, Date nowTimestamp, int personId) throws DAOException {
        assignFieldsForNew(address, nowTimestamp, personId);
        final Integer pid = createSingleObject(INSERT_SQL, address);
        address.setAddressPID(pid);
        return pid;
    }

    @Override
    public void createAddresses(List<AddressVO> addresses) throws DAOException {
        addresses.forEach(id -> id.setUuid(UUID.randomUUID().toString()));
        executeBatchUpdate(INSERT_SQL, addresses);
    }

    @Override
    public void updateCurrentAndCreateNew(AddressVO vo, Date nowTimestamp, int personId) throws DAOException {
        updateValidToTimestamp(vo.getUuid(), nowTimestamp);
        vo.setValidFrom(nowTimestamp);
        vo.setValidTo(AbstractECPRDAOImpl.VALID_TO_DEFAULT);
        vo.setPersonPID(personId);
        createAddress(vo, nowTimestamp, personId);
    }

    @Override
    public void updateTimeStamp(AddressVO vo, Date nowTimestamp) throws DAOException {
        updateValidToTimestamp(vo.getUuid(), nowTimestamp);
    }

    @Override
    public String getTableName() {
        return TABLE_NAME;
    }

    @Override
    protected RowMapper getDefaultRowMapper() {
        return new AddressRowMapper();
    }

}
