/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.dao.*;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.abstr.AbstractCrossTableDAOImpl;
import dk.sds.ecpr2.domain.*;
import dk.sds.ecpr2.domain.abstr.PersonDescriptorVO;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import java.time.Instant;
import java.util.*;
import java.util.stream.Collectors;

@Repository
@Log4j2
public class ExportCrossTableDAOImpl extends AbstractCrossTableDAOImpl {

    @Autowired
    PersonDAO personDAO;

    @Autowired
    private IdentifierDAO identifierDAO;

    @Autowired
    private AddressDAO addressDAO;

    @Autowired
    private BirthDateDAO birthDateDAO;

    @Autowired
    private ContactDAO contactDAO;

    @Autowired
    private GenderDAO genderDAO;

    @Autowired
    private NameDAO nameDAO;

    @Override
    public List<PersonVO> getPersonsWithRelationsAfterPID(int minPersonPid,
                                                          int batchSize,
                                                          boolean excludeReservedUntilFilter,
                                                          boolean excludeDuplicatePersonsFilter) throws DAOException {
        return getPersonsWithRelations(minPersonPid, batchSize, excludeReservedUntilFilter);
    }

    @Override
    public List<PersonVO> getPersonsWithRelationsAfterModifiedTime(Instant modifiedTime, boolean excludeDuplicatePersons) throws DAOException {
        // For export we do not need excludeDuplicatePerson, because we want to export everything
        List<IdentifierVO> allIdentifierRelations = identifierDAO.getAllIdentifiersAfterModified(modifiedTime);
        List<NameVO> allNameRelations = nameDAO.getAllNamesAfterModified(modifiedTime);
        List<GenderVO> allGenderRelations = genderDAO.getAllGendersAfterModified(modifiedTime);
        List<BirthDateVO> allBirthdateRelations = birthDateDAO.getAllBirthdatesAfterModified(modifiedTime);
        List<ContactVO> allContactRelations = contactDAO.getAllContactsAfterModified(modifiedTime);
        List<AddressVO> allAddressRelations = addressDAO.getAllAddressesAfterModified(modifiedTime);

        ArrayList<PersonDescriptorVO> personDescriptorVOS = new ArrayList<>();
        personDescriptorVOS.addAll(allIdentifierRelations);
        personDescriptorVOS.addAll(allNameRelations);
        personDescriptorVOS.addAll(allGenderRelations);
        personDescriptorVOS.addAll(allBirthdateRelations);
        personDescriptorVOS.addAll(allContactRelations);
        personDescriptorVOS.addAll(allAddressRelations);

        Set<Integer> uniquePersonPIDS = personDescriptorVOS.stream().map(PersonDescriptorVO::getPersonPID).collect(Collectors.toSet());

        // TODO: ONLY RELEVANT WHEN USED IN EXPORT
        // TODO: This only gets undeleted persons, therefore Egon Olsen is exported to replicate 1 million times
        // TODO: Research the side effects of removing deleted, as each identifier/name/contact/address/birthday/gender need to have their ValidTo changed

        List<PersonVO> persons = personDAO.getPersonByPIDs(new ArrayList<>(uniquePersonPIDS));

        Map<Integer, PersonVO> personMap = new HashMap<>();
        for (PersonVO p : persons) {
            personMap.put(p.getPersonPID(), p);
        }

        for (IdentifierVO identifierVO : allIdentifierRelations) {
            final PersonVO person = personMap.get(identifierVO.getPersonPID());
            if (person != null) {
                person.getIdentifiers().add(identifierVO);
            }
        }

        for (NameVO nameVO : allNameRelations) {
            final PersonVO person = personMap.get(nameVO.getPersonPID());
            if (person != null) {
                person.getNames().add(nameVO);
            }
        }

        for (GenderVO genderVO : allGenderRelations) {
            final PersonVO person = personMap.get(genderVO.getPersonPID());
            if (person != null) {
                personMap.get(genderVO.getPersonPID()).getGenders().add(genderVO);
            }
        }

        for (BirthDateVO birthDateVO : allBirthdateRelations) {
            final PersonVO person = personMap.get(birthDateVO.getPersonPID());
            if (person != null) {
                personMap.get(birthDateVO.getPersonPID()).getBirthDates().add(birthDateVO);
            }
        }

        for (ContactVO contactVO : allContactRelations) {
            final PersonVO person = personMap.get(contactVO.getPersonPID());
            if (person != null) {
                personMap.get(contactVO.getPersonPID()).getContacts().add(contactVO);
            }
        }

        for (AddressVO addressVO : allAddressRelations) {
            final PersonVO person = personMap.get(addressVO.getPersonPID());
            if (person != null) {
                personMap.get(addressVO.getPersonPID()).getAddresses().add(addressVO);
            }
        }
        return persons;
    }

    private List<PersonVO> getPersonsWithRelations(int minPersonPid,
                                                   int batchSize,
                                                   boolean excludeReservedUntil) throws DAOException {

        // Remove persons with duplicate UUID's
        List<PersonVO> persons = personDAO.getPersonsAfterPID(minPersonPid, batchSize, excludeReservedUntil);

        Map<Integer, PersonVO> personMap = new HashMap<>();
        for (PersonVO p : persons) {
            personMap.put(p.getPersonPID(), p);
        }

        final int maxPersonPID = PersonVO.getMaxPersonPID(persons);
        final int minPersonPID = PersonVO.getMinPersonPID(persons);

        // Not possible to make these generic?
        List<IdentifierVO> allIdentifierRelations = identifierDAO.getAllIdentifiersBetweenPersonPID(minPersonPID, maxPersonPID);
        for (IdentifierVO identifierVO : allIdentifierRelations) {
            final PersonVO person = personMap.get(identifierVO.getPersonPID());
            if (person != null) {
                person.getIdentifiers().add(identifierVO);
            }
        }
        List<NameVO> allNameRelations = nameDAO.getAllNamesBetweenPersonPID(minPersonPID, maxPersonPID);
        for (NameVO nameVO : allNameRelations) {
            final PersonVO person = personMap.get(nameVO.getPersonPID());
            if (person != null) {
                person.getNames().add(nameVO);
            }
        }
        List<GenderVO> allGenderRelations = genderDAO.getAllGendersBetweenPersonPID(minPersonPID, maxPersonPID);
        for (GenderVO genderVO : allGenderRelations) {
            final PersonVO person = personMap.get(genderVO.getPersonPID());
            if (person != null) {
                personMap.get(genderVO.getPersonPID()).getGenders().add(genderVO);
            }
        }
        List<BirthDateVO> allBirthdateRelations = birthDateDAO.getAllBirthdatesBetweenPersonPID(minPersonPID, maxPersonPID);
        for (BirthDateVO birthDateVO : allBirthdateRelations) {
            final PersonVO person = personMap.get(birthDateVO.getPersonPID());
            if (person != null) {
                personMap.get(birthDateVO.getPersonPID()).getBirthDates().add(birthDateVO);
            }
        }
        List<ContactVO> allContactRelations = contactDAO.getAllContactsBetweenPersonPID(minPersonPID, maxPersonPID);
        for (ContactVO contactVO : allContactRelations) {
            final PersonVO person = personMap.get(contactVO.getPersonPID());
            if (person != null) {
                personMap.get(contactVO.getPersonPID()).getContacts().add(contactVO);
            }
        }
        List<AddressVO> allAddressRelations = addressDAO.getAllAddressesBetweenPersonPID(minPersonPID, maxPersonPID);
        for (AddressVO addressVO : allAddressRelations) {
            final PersonVO person = personMap.get(addressVO.getPersonPID());
            if (person != null) {
                personMap.get(addressVO.getPersonPID()).getAddresses().add(addressVO);
            }
        }

        return persons;
    }
}
