/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.dao.GenderDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.abstr.AbstractECPRDAOImpl;
import dk.sds.ecpr2.domain.GenderVO;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.stereotype.Repository;

import java.time.Instant;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.UUID;

@Repository
public class GenderDAOImpl extends AbstractECPRDAOImpl implements GenderDAO {

    public static final String TABLE_NAME = "Gender";
    final String INSERT_SQL = "INSERT INTO Gender (GenderPID, Uuid, PersonPID, ValidFrom, ValidTo, Value, ModifiedTime)" +
            "VALUES(:genderPID, :uuid, :personPID, :validFrom, :validTo, :value, :modified)";

    @Override
    public List<GenderVO> getWithServiceActor(List<String> personUuids, List<Integer> personPids) throws DAOException {
        return getWithServiceActorByPersonUUIDs(personUuids, personPids);
    }

    @Override
    public List<GenderVO> getWithServiceActor(String personUuid, Date date) throws DAOException {
        return getWithServiceActorByPersonUUIDAndDate(personUuid, date);
    }

    @Override
    public List<GenderVO> getAllGendersBetweenPersonPIDForSearch(int minPersonPid, int maxNoOfRows) throws DAOException {
        return getSelectedFieldsBetweenPersonPIDWithoutServiceActor(Arrays.asList("Value", "Uuid", "ValidFrom"), minPersonPid, maxNoOfRows);
    }

    @Override
    public List<GenderVO> getAllGendersBetweenPersonPID(int minPersonPID, int maxPersonPID) throws DAOException {
        return getAllFieldsBetweenPersonPIDWithServiceActor(minPersonPID, maxPersonPID);
    }

    @Override
    public List<GenderVO> getAllGendersAfterModified(Instant modifiedTime) throws DAOException {
        return getAllFieldsAfterModifiedTimeWithServiceActor(modifiedTime);
    }

    @Override
    public List<GenderVO> getAllGendersAfterModifiedForSearch(Instant modifiedTime) throws DAOException {
        return getSelectedFieldsAfterModifiedTimeWithoutServiceActor(
                Arrays.asList("Value", "Uuid", "ValidFrom"),
                modifiedTime
        );
    }

    @Override
    public Integer createGender(GenderVO gender, Date nowTimestamp, int personId) throws DAOException {
        assignFieldsForNew(gender, nowTimestamp, personId);
        final Integer pid = createSingleObject(INSERT_SQL, gender);
        gender.setGenderPID(pid);
        return pid;
    }

    @Override
    public void createGenders(List<GenderVO> genders) throws DAOException {
        genders.forEach(id -> id.setUuid(UUID.randomUUID().toString()));
        executeBatchUpdate(INSERT_SQL, genders);
    }

    @Override
    public void updateCurrentAndCreateNew(GenderVO vo, Date nowTimestamp, int personId) throws DAOException {
        updateValidToTimestamp(vo.getUuid(), nowTimestamp);
        vo.setValidFrom(nowTimestamp);
        vo.setValidTo(AbstractECPRDAOImpl.VALID_TO_DEFAULT);
        vo.setPersonPID(personId);
        createGender(vo, nowTimestamp, personId);
    }

    @Override
    public void updateTimeStamp(GenderVO vo, Date nowTimestamp) throws DAOException {
        updateValidToTimestamp(vo.getUuid(), nowTimestamp);
    }

    @Override
    public String getTableName() {
        return TABLE_NAME;
    }

    @Override
    protected RowMapper getDefaultRowMapper() {
        return new GenderRowMapper();
    }

}
