/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.dao.IdentifierDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.abstr.AbstractECPRDAOImpl;
import dk.sds.ecpr2.domain.IdentifierVO;
import dk.sds.ecpr2.domain.IdentifierWithoutServiceActorVO;
import dk.sds.ecpr2.domain.KeyIdentifierVO;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.stereotype.Repository;

import java.time.Instant;
import java.util.*;

@Repository
public class IdentifierDAOImpl extends AbstractECPRDAOImpl implements IdentifierDAO {

    private static final String TABLE_NAME = "Identifier";
    private final String INSERT_SQL = "INSERT INTO Identifier (IdentifierPID, Uuid, OID, PersonPID, ValidFrom, ValidTo, Value, Validity, Expiry, ModifiedTime) " +
            "VALUES (:IdentifierPID, :uuid, :Oid, :PersonPID, :ValidFrom, :ValidTo, :Value, :Validity, :Expiry, :modified) ";

    @Override
    public List<IdentifierVO> getWithServiceActor(List<String> personUuids, List<Integer> personPids) throws DAOException {
        return getWithServiceActorByPersonUUIDs(personUuids, personPids);
    }

    @Override
    public List<IdentifierVO> getWithServiceActor(String personUuid, Date date) throws DAOException {
        return getWithServiceActorByPersonUUIDAndDate(personUuid, date);
    }

    @Override
    public IdentifierWithoutServiceActorVO getCurrentIdentifierByValueAndOID(String value, String oid) throws DAOException {
        return fetchSingleRowMatchingTwoColumns("OID", oid, "Value", value, new IdentifierWithoutServiceActorRowMapper(), true);
    }

    @Override
    public IdentifierWithoutServiceActorVO getIdentifierByValueAndOID(String value, String oid) throws DAOException {
        return fetchSingleRowMatchingTwoColumns("OID", oid, "Value", value, new IdentifierWithoutServiceActorRowMapper(), false);
    }

    @Override
    public List<IdentifierWithoutServiceActorVO> getIdentifiersByValueAndOID(String value, String oid) throws DAOException {
        return fetchRowsMatchingTwoColumns("OID", oid, "Value", value, new IdentifierWithoutServiceActorRowMapper(), false);
    }

    @Override
    public List<IdentifierWithoutServiceActorVO> getCurrentIdentifiersByValueAndOID(String value, String oid) throws DAOException {
        return fetchRowsMatchingTwoColumns("OID", oid, "Value", value, new IdentifierWithoutServiceActorRowMapper(), true);
    }

    @Override
    public List<IdentifierVO> getAllIdentifiersBetweenPersonPIDForSearch(int minPersonPid, int maxNoOfRows) throws DAOException {
        return getSelectedFieldsBetweenPersonPIDWithoutServiceActor(Arrays.asList("OID", "Value", "Uuid", "ValidFrom"),
                minPersonPid, maxNoOfRows);
    }

    @Override
    public List<IdentifierVO> getAllIdentifiersBetweenPersonPID(int minPersonPID, int maxNoOfRows) throws DAOException {
        return getAllFieldsBetweenPersonPIDWithServiceActor(minPersonPID, maxNoOfRows);
    }

    @Override
    public List<IdentifierVO> getAllIdentifiersAfterModified(Instant modifiedTime) throws DAOException {
        return getAllFieldsAfterModifiedTimeWithServiceActor(modifiedTime);
    }

    @Override
    public List<IdentifierVO> getAllIdentifiersAfterModifiedForSearch(Instant modifiedTime) throws DAOException {
        return getSelectedFieldsAfterModifiedTimeWithoutServiceActor(
                Arrays.asList("OID", "Value", "Uuid", "ValidFrom"),
                modifiedTime
        );
    }

    @Override
    public List<IdentifierWithoutServiceActorVO> getMatchingIdentifiersWithoutServiceActor(String pattern, String oid) throws DAOException {
        final String sql = "SELECT * FROM Identifier WHERE Oid = :oid AND Value LIKE :pattern";
        final MapSqlParameterSource parameterMap = new MapSqlParameterSource().addValue("oid", oid)
                .addValue("pattern", pattern);
        return queryForList(sql, parameterMap, new IdentifierWithoutServiceActorRowMapper());
    }

    @Override
    public List<IdentifierWithoutServiceActorVO> getIdentifiersByUuids(List<String> uuids, Date now) throws DAOException {
        if (uuids.size() == 0) return new ArrayList<>();
        final String sql = "SELECT * FROM Identifier WHERE Uuid IN (:uuids) AND ValidFrom <= :now AND ValidTo > :now";
        final MapSqlParameterSource parms = new MapSqlParameterSource()
                .addValue("uuids", uuids)
                .addValue("now", now);
        return queryForList(sql, parms, new IdentifierWithoutServiceActorRowMapper());
    }

    @Override
    public Integer createIdentifier(IdentifierVO identifier, Date now, int personId) throws DAOException {
        assignFieldsForNew(identifier, now, personId);
        final Integer pid = createSingleObject(INSERT_SQL, identifier);
        identifier.setIdentifierPID(pid);
        return pid;
    }

    @Override
    public IdentifierWithoutServiceActorVO getCurrentIdentifierByKeyIdentifier(KeyIdentifierVO keyIdentifier) throws DAOException {
        if (keyIdentifier == null || keyIdentifier.getValue() == null) {
            return null;
        } else {
            return getCurrentIdentifierByValueAndOID(keyIdentifier.getValue(), keyIdentifier.getOid());
        }
    }

    @Override
    public void updateCurrentAndCreateNew(IdentifierVO vo, Date now, int personId) throws DAOException {
        updateValidToTimestamp(vo.getUuid(), now);
        vo.setValidFrom(now);
        vo.setValidTo(AbstractECPRDAOImpl.VALID_TO_DEFAULT);
        vo.setPersonPID(personId);
        createIdentifier(vo, now, personId);
    }

    public void updateIdentifier(IdentifierVO vo, Date now) throws DAOException {
        updateValidToTimestamp(vo.getUuid(), now);

    }

    @Override
    public void updateTimeStamp(IdentifierVO vo, Date now) throws DAOException {
        updateValidToTimestamp(vo.getUuid(), now);
    }

    @Override
    public void createIdentifiers(List<IdentifierVO> identifiers) throws DAOException {
        identifiers.forEach(id -> id.setUuid(UUID.randomUUID().toString()));
        executeBatchUpdate(INSERT_SQL, identifiers);
    }

    @Override
    public String getTableName() {
        return TABLE_NAME;
    }


    @Override
    protected RowMapper getDefaultRowMapper() {
        return new IdentifierRowMapper();
    }
}
