/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.*;
import dk.sds.ecpr2.dao.ModifierDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.abstr.AbstractDAOImpl;
import dk.sds.ecpr2.domain.ModifierWithId;
import lombok.extern.log4j.Log4j2;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.support.GeneratedKeyHolder;
import org.springframework.stereotype.Repository;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;

@Log4j2
@Repository
public class ModifierDAOImpl extends AbstractDAOImpl implements ModifierDAO {

    private final String INSERT_SQL = "INSERT INTO Modifier (AuthorisationIdentifier, PersonName, PersonIdentifier, PersonIdentifierSource, SpecialityCode, Role, " +
            "OrganisationName, OrganisationAddressLine, OrganisationTelephoneNumber, OrganisationEmailAddress, OrganisationType, OrganisationIdentifier, OrganisationIdentifierSource, AuthorisedByModifierPID, ModifiedTime) " +
            "VALUES (:authorisationIdentifier, :personName, :personIdentifier, :personIdentifierSource, :specialityCode, :role, " +
            ":organisationName, :organisationAddressLine, :organisationTelephoneNumber, :organisationEmailAddress, :organisationType, :organisationIdentifier, :organisationIdentifierSource, :authorisedByModifierPID, :modified)";

    @Override
    public ModifiedBy getModifier(long modifierPID, boolean includePersonIdentifier) throws DAOException {
        ModifierWithId modifier = fetchSingleRowMatchingSingleColumn("ModifierPID", modifierPID + "");
        if (modifier.getAuthorisedById() == null || modifier.getAuthorisedById() == 0) {
            if (!includePersonIdentifier) {
                modifier.getModifier().getPerson().setPersonIdentifier(null);
            }
            return modifier.getModifier();
        }
        ModifiedBy authorisedBy = getAuthorisedBy(modifier.getAuthorisedById());
        AuthorisedByType authorisedByType = new AuthorisedByType();
        authorisedByType.setOrganisation(authorisedBy.getOrganisation());
        authorisedByType.setPerson(authorisedBy.getPerson());
        modifier.getModifier().setAuthorisedBy(authorisedByType);
        if (!includePersonIdentifier) {
            modifier.getModifier().getAuthorisedBy().getPerson().setPersonIdentifier(null);
            modifier.getModifier().getPerson().setPersonIdentifier(null);
        }
        return modifier.getModifier();
    }

    private ModifiedBy getAuthorisedBy(long modifierPID) throws DAOException {
        ModifierWithId modifier = fetchSingleRowMatchingSingleColumn("ModifierPID", modifierPID + "");
        return modifier.getModifier();
    }

    private List<ModifierWithId> getModifiersForPerson(ModificatorPersonType person) throws DAOException {
        String authorisationIdentifier = person.getAuthorisationIdentifier();
        if (authorisationIdentifier != null) {
            return getModifiersForPersonByAuthNumber(authorisationIdentifier);
        }
        PersonIdentifierType personIdentifier = person.getPersonIdentifier();
        if (personIdentifier != null) {
            return getModifiersForPersonByPersonIdentifier(personIdentifier);
        }
        return Collections.emptyList();
    }

    private List<ModifierWithId> getModifiersForPersonByPersonIdentifier(PersonIdentifierType personIdentifier) throws DAOException {
        if (personIdentifier == null) {
            return new ArrayList<>();
        }
        String personId = personIdentifier.getPersonIdentifier();
        String source = personIdentifier.getSource();
        List<ModifierWithId> results = fetchRowsMatchingTwoColumns("PersonIdentifier", personId, "PersonIdentifierSource", source, getDefaultRowMapper(), false);
        for (ModifierWithId modifier : results) {
            if (modifier.getAuthorisedById() != null && modifier.getAuthorisedById() != 0) {
                ModifiedBy authorisedBy = getAuthorisedBy(modifier.getAuthorisedById());
                AuthorisedByType authorisedByType = new AuthorisedByType();
                authorisedByType.setOrganisation(authorisedBy.getOrganisation());
                authorisedByType.setPerson(authorisedBy.getPerson());
                modifier.getModifier().setAuthorisedBy(authorisedByType);
            }
        }
        return results;
    }

    private List<ModifierWithId> getModifiersForPersonByAuthNumber(String authId) throws DAOException {
        if (authId == null) {
            return new ArrayList<>();
        }
        MapSqlParameterSource mapSqlParameterSource = new MapSqlParameterSource()
                .addValue("authId", authId);
        List<ModifierWithId> results = queryForList("SELECT * FROM " + getTableName() + " WHERE AuthorisationIdentifier = :authId", mapSqlParameterSource, getDefaultRowMapper());
        for (ModifierWithId modifier : results) {
            if (modifier.getAuthorisedById() != null && modifier.getAuthorisedById() != 0) {
                ModifiedBy authorisedBy = getAuthorisedBy(modifier.getAuthorisedById());
                AuthorisedByType authorisedByType = new AuthorisedByType();
                authorisedByType.setOrganisation(authorisedBy.getOrganisation());
                authorisedByType.setPerson(authorisedBy.getPerson());
                modifier.getModifier().setAuthorisedBy(authorisedByType);
            }
        }
        return results;
    }

    @Override
    public ModifierWithId getOrCreateModified(ModifiedBy newModifier, Date now) throws DAOException {
        ModificatorPersonType person = newModifier.getPerson();
        List<ModifierWithId> modifiersForPerson = getModifiersForPerson(person);
        for (ModifierWithId existingModifier : modifiersForPerson) {
            if (sameModifier(newModifier, existingModifier.getModifier())) {
                return existingModifier;
            }
        }
        return createModifier(newModifier, now);
    }


    private ModifierWithId createModifier(ModifiedBy modifiedBy, Date now) throws DAOException {
        final GeneratedKeyHolder generatedKeyHolder = new GeneratedKeyHolder();
        ModificatorPersonType person = modifiedBy.getPerson();
        OrganisationType organisation = modifiedBy.getOrganisation();
        final MapSqlParameterSource parameterSource = new MapSqlParameterSource()
                .addValue("authorisationIdentifier", person.getAuthorisationIdentifier())
                .addValue("personName", person.getName())
                // Don't add CPR if AuthNumber exists, when working onBehalfOf someone, only the AuthNumber is available
                // So the personIdentifier will result in NullPointerException
                .addValue("personIdentifier", person.getAuthorisationIdentifier() == null
                        ? person.getPersonIdentifier().getPersonIdentifier()
                        : null)
                .addValue("personIdentifierSource", person.getAuthorisationIdentifier() == null
                        ? person.getPersonIdentifier().getSource()
                        : null)
                .addValue("specialityCode", person.getSpecialityCode())
                .addValue("role", modifiedBy.getRole())
                .addValue("organisationName", organisation.getName())
                .addValue("organisationAddressLine", organisation.getAddressLine())
                .addValue("organisationTelephoneNumber", organisation.getTelephoneNumber())
                .addValue("organisationEmailAddress", organisation.getEmailAddress())
                .addValue("organisationType", organisation.getType())
                .addValue("organisationIdentifier", organisation.getIdentifier().getIdentifier())
                .addValue("organisationIdentifierSource", organisation.getIdentifier().getSource())
                .addValue("modified", now);

        Integer authorisedById = null;
        if (modifiedBy.getAuthorisedBy() != null) {
            ModifierWithId authorisedBy = getOrCreateModified(getAuthorizedByAsModifiedBy(modifiedBy), now);
            parameterSource.addValue("authorisedByModifierPID", authorisedBy.getId());
            authorisedById = authorisedBy.getAuthorisedById();
        } else {
            parameterSource.addValue("authorisedByModifierPID", null);
        }

        namedParamJdbcTemplate.update(INSERT_SQL, parameterSource, generatedKeyHolder);
        Integer id = generatedKeyHolder.getKey().intValue();
        return new ModifierWithId(modifiedBy, id, authorisedById);
    }

    private ModifiedBy getAuthorizedByAsModifiedBy(ModifiedBy currentModificator) {
        AuthorisedByType authorisedBy = currentModificator.getAuthorisedBy();
        ModifiedBy modifiedBy = new ModifiedBy();
        modifiedBy.setOrganisation(currentModificator.getOrganisation());
        modifiedBy.setPerson(authorisedBy.getPerson());
        return modifiedBy;
    }

    private boolean sameModifier(ModifiedBy newModifier, ModifiedBy existingModifier) {
        if (newModifier == null || existingModifier == null) {
            return false;
        }
        if (newModifier.getRole() != null && existingModifier.getRole() != null) {
            if (!newModifier.getRole().equalsIgnoreCase(existingModifier.getRole())) {
                return false;
            }
        }
        if (!samePerson(newModifier.getPerson(), existingModifier.getPerson())) {
            return false;
        }
        if (!sameOrganisation(newModifier.getOrganisation(), existingModifier.getOrganisation())) {
            return false;
        }
        if (newModifier.getAuthorisedBy() == null && existingModifier.getAuthorisedBy() != null) {
            return false;
        }
        if (existingModifier.getAuthorisedBy() == null && newModifier.getAuthorisedBy() != null) {
            return false;
        }
        if (newModifier.getAuthorisedBy() != null && !samePerson(newModifier.getAuthorisedBy().getPerson(), existingModifier.getAuthorisedBy().getPerson())) {
            return false;
        }
        if (newModifier.getAuthorisedBy() != null && !sameOrganisation(newModifier.getAuthorisedBy().getOrganisation(), existingModifier.getAuthorisedBy().getOrganisation())) {
            return false;
        }
        return true;
    }

    private boolean sameOrganisation(OrganisationType organisation1, OrganisationType organisation2) {
        if (!organisation1.getType().equalsIgnoreCase(organisation2.getType())) {
            return false;
        }
        if (organisation1.getTelephoneNumber() == null && organisation2.getTelephoneNumber() != null) {
            return false;
        }
        if (organisation2.getTelephoneNumber() == null && organisation1.getTelephoneNumber() != null) {
            return false;
        }
        if (organisation1.getTelephoneNumber() != null && !organisation1.getTelephoneNumber().equalsIgnoreCase(organisation2.getTelephoneNumber())) {
            return false;
        }
        if (organisation1.getAddressLine() == null && organisation2.getAddressLine() != null) {
            return false;
        }
        if (organisation2.getAddressLine() == null && organisation1.getAddressLine() != null) {
            return false;
        }
        if (organisation1.getAddressLine() != null && !organisation1.getAddressLine().equalsIgnoreCase(organisation2.getAddressLine())) {
            return false;
        }
        if (organisation1.getEmailAddress() == null && organisation2.getEmailAddress() != null) {
            return false;
        }
        if (organisation2.getEmailAddress() == null && organisation1.getEmailAddress() != null) {
            return false;
        }
        if (organisation1.getEmailAddress() != null && !organisation1.getEmailAddress().equalsIgnoreCase(organisation2.getEmailAddress())) {
            return false;
        }
        if (!organisation1.getName().equalsIgnoreCase(organisation2.getName())) {
            return false;
        }
        if (!organisation1.getIdentifier().getSource().equalsIgnoreCase(organisation2.getIdentifier().getSource())) {
            return false;
        }
        if (!organisation1.getIdentifier().getIdentifier().equalsIgnoreCase(organisation2.getIdentifier().getIdentifier())) {
            return false;
        }
        return true;
    }

    private boolean samePerson(ModificatorPersonType newPerson, ModificatorPersonType existingPerson) {
        if (newPerson.getAuthorisationIdentifier() == null && existingPerson.getAuthorisationIdentifier() != null) {
            return false;
        }
        if (existingPerson.getAuthorisationIdentifier() == null && newPerson.getAuthorisationIdentifier() != null) {
            return false;
        }
        if (newPerson.getAuthorisationIdentifier() != null && !newPerson.getAuthorisationIdentifier().equalsIgnoreCase(existingPerson.getAuthorisationIdentifier())) {
            return false;
        }
        if (!newPerson.getName().equalsIgnoreCase(existingPerson.getName())) {
            return false;
        }
        if (newPerson.getSpecialityCode() == null && existingPerson.getSpecialityCode() != null) {
            return false;
        }
        if (existingPerson.getSpecialityCode() == null && newPerson.getSpecialityCode() != null) {
            return false;
        }
        if (newPerson.getSpecialityCode() != null && !newPerson.getSpecialityCode().equalsIgnoreCase(existingPerson.getSpecialityCode())) {
            return false;
        }

        // Only compare PersonIdentifier, if Auth Number does not exist
        if (newPerson.getAuthorisationIdentifier() == null) {
            PersonIdentifierType newPersonPersonIdentifier = newPerson.getPersonIdentifier();
            PersonIdentifierType existingPersonPersonIdentifier = existingPerson.getPersonIdentifier();
            if (!newPersonPersonIdentifier.getSource().equalsIgnoreCase(existingPersonPersonIdentifier.getSource())) {
                return false;
            }
            if (!newPersonPersonIdentifier.getPersonIdentifier().equalsIgnoreCase(existingPersonPersonIdentifier.getPersonIdentifier())) {
                return false;
            }
        }

        return true;
    }

    @Override
    public String getTableName() {
        return "Modifier";
    }

    @Override
    protected String getIdColumnName() {
        return "ModifierPID";
    }

    @Override
    protected RowMapper getDefaultRowMapper() {
        return (resultSet, i) -> {
            ModifiedBy modifiedBy = new ModifiedBy();
            modifiedBy.setRole(resultSet.getString("Role"));
            ModificatorPersonType person = new ModificatorPersonType();
            person.setName(resultSet.getString("PersonName"));
            person.setAuthorisationIdentifier(resultSet.getNString("AuthorisationIdentifier"));
            person.setSpecialityCode(resultSet.getNString("SpecialityCode"));

            String personIdentifier = resultSet.getNString("PersonIdentifier");
            String personIdentifierSource = resultSet.getNString("PersonIdentifierSource");

            if (personIdentifier != null && personIdentifierSource != null) {
                PersonIdentifierType personIdentifierType = new PersonIdentifierType();
                personIdentifierType.setPersonIdentifier(personIdentifier);
                personIdentifierType.setSource(personIdentifierSource);
                person.setPersonIdentifier(personIdentifierType);
            }

            modifiedBy.setPerson(person);
            OrganisationType organisation = new OrganisationType();
            organisation.setType(resultSet.getNString("OrganisationType"));
            organisation.setTelephoneNumber(resultSet.getNString("OrganisationTelephoneNumber"));
            organisation.setName(resultSet.getNString("OrganisationName"));
            organisation.setEmailAddress(resultSet.getNString("OrganisationEmailAddress"));
            organisation.setAddressLine(resultSet.getNString("OrganisationAddressLine"));
            OrganisationIdentifierType organisationId = new OrganisationIdentifierType();
            organisationId.setSource(resultSet.getNString("OrganisationIdentifierSource"));
            organisationId.setIdentifier(resultSet.getNString("OrganisationIdentifier"));
            organisation.setIdentifier(organisationId);
            modifiedBy.setOrganisation(organisation);
            Integer modifiedId = resultSet.getInt("ModifierPID");
            Integer authorisedById = resultSet.getInt("AuthorisedByModifierPID");
            return new ModifierWithId(modifiedBy, modifiedId, authorisedById);
        };
    }
}
