/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.dao.NameDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.abstr.AbstractECPRDAOImpl;
import dk.sds.ecpr2.domain.NameVO;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.stereotype.Repository;

import java.time.Instant;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.UUID;

@Repository
public class NameDAOImpl extends AbstractECPRDAOImpl implements NameDAO {

    public static final String TABLE_NAME = "Name";
    private final String INSERT_SQL = "INSERT INTO Name (NamePID, Uuid, PersonPID, ValidFrom, ValidTo, `Use`, FamilyName, GivenName, Expiry, ModifiedTime)" +
            "VALUES (:namePID, :uuid, :personPID, :validFrom, :validTo, :use, :familyName, :givenName, :expiry, :modified)";

    @Override
    public List<NameVO> getWithServiceActor(List<String> personUuids, List<Integer> personPids) throws DAOException {
        return getWithServiceActorByPersonUUIDs(personUuids, personPids);
    }

    @Override
    public List<NameVO> getWithServiceActor(String personUuid, Date date) throws DAOException {
        return getWithServiceActorByPersonUUIDAndDate(personUuid, date);
    }

    @Override
    public List<NameVO> getAllNamesBetweenPersonPIDForSearch(int minPersonPid, int maxNoOfRows) throws DAOException {
        return getSelectedFieldsBetweenPersonPIDWithoutServiceActor(Arrays.asList("GivenName", "FamilyName", "Uuid", "ValidFrom"),
                minPersonPid, maxNoOfRows);
    }

    @Override
    public List<NameVO> getAllNamesBetweenPersonPID(int minPersonPID, int maxPersonPID) throws DAOException {
        return getAllFieldsBetweenPersonPIDWithServiceActor(minPersonPID, maxPersonPID);
    }

    @Override
    public List<NameVO> getAllNamesAfterModified(Instant modifiedTime) throws DAOException {
        return getAllFieldsAfterModifiedTimeWithServiceActor(modifiedTime);
    }

    @Override
    public List<NameVO> getAllNamesAfterModifiedForSearch(Instant modifiedTime) throws DAOException {
        return getSelectedFieldsAfterModifiedTimeWithoutServiceActor(
                Arrays.asList("GivenName", "FamilyName", "Uuid", "ValidFrom"),
                modifiedTime
        );
    }

    @Override
    public Integer createName(NameVO name, Date now, int personId) throws DAOException {
        assignFieldsForNew(name, now, personId);
        final Integer pid = createSingleObject(INSERT_SQL, name);
        name.setNamePID(pid);
        return pid;
    }

    @Override
    public void createNames(List<NameVO> names) throws DAOException {
        names.forEach(id -> id.setUuid(UUID.randomUUID().toString()));
        executeBatchUpdate(INSERT_SQL, names);
    }

    @Override
    public void updateCurrentAndCreateNew(NameVO vo, Date now, int personId) throws DAOException {
        updateValidToTimestamp(vo.getUuid(), now);
        vo.setValidFrom(now);
        vo.setValidTo(AbstractECPRDAOImpl.VALID_TO_DEFAULT);
        vo.setPersonPID(personId);
        createName(vo, now, personId);
    }

    @Override
    public void updateTimeStamp(NameVO vo, Date now) throws DAOException {
        updateValidToTimestamp(vo.getUuid(), now);
    }

    @Override
    public String getTableName() {
        return TABLE_NAME;
    }

    @Override
    protected RowMapper getDefaultRowMapper() {
        return new NameRowMapper();
    }
}
