/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.dao.OIDDAO;
import dk.sds.ecpr2.dao.OIDFacade;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.CountryCode;
import dk.sds.ecpr2.domain.OIDVO;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

@Repository
public class OIDFacadeImpl implements OIDFacade {
    @Autowired OIDDAO oiddao;

    @Override
    public List<OIDVO> getAllOIDs() throws DAOException {
        List<OIDVO> oids = oiddao.getAllOIDs();
        oids = oids.stream().map(this::multiplyVOByCountries).flatMap(Collection::stream).collect(Collectors.toList());
        return oids;
    }

    @Override
    public OIDVO getByOIDValue(String oid) throws DAOException {
        OIDVO oidVO = oiddao.getByOIDValue(oid);
        if (oidVO == null) {
            String partialOid = oid.substring(0, oid.lastIndexOf("."));
            oidVO = oiddao.getByOIDValue(partialOid);
            if (oidVO == null) {
                return null;
            }
            String countryOid = oid.substring(oid.lastIndexOf(".") + 1);
            CountryCode countryCode = CountryCode.getByNumericValue(Integer.parseInt(countryOid));
            oidVO.setPrimaryOidLabel(oidVO.getPrimaryOidLabel());
            oidVO.setSecondaryOidLabel(countryCode.getFullName());
        }
        return oidVO;
    }

    @Override
    public OIDVO getByOIDValueAndType(String oid, String type) throws DAOException {
        OIDVO oidVO = oiddao.getByOIDValueAndType(oid, type);
        if (oidVO == null) {
            for (String countryOid : multiplyOIDByCountries(oid)) {
                oidVO = oiddao.getByOIDValueAndType(countryOid, type);
                if (oidVO != null) {
                    return oidVO;
                }
            }
        }
        return oidVO;
    }

    private List<String> multiplyOIDByCountries(String oid) {
        List<String> result = new ArrayList<>();
        for (CountryCode countryCode : CountryCode.values()) {
            result.add(oid + "." + String.format("%03d", countryCode.getNumericValue()));
        }
        return result;
    }

    private List<OIDVO> multiplyVOByCountries(OIDVO oidvo) {
        if (!oidvo.isCountryBased()) {
            return Collections.singletonList(oidvo);
        }

        List<OIDVO> result = new ArrayList<>();
        for (CountryCode countryCode : CountryCode.values()) {
            OIDVO tmp = new OIDVO();
            tmp.setModifiedTime(oidvo.getModifiedTime());
            tmp.setRegex(oidvo.getRegex());
            tmp.setType(oidvo.getType());
            tmp.setOidPID(oidvo.getOidPID());
            tmp.setOid(oidvo.getOid() + "." + String.format("%03d", countryCode.getNumericValue()));
            tmp.setPrimaryOidLabel(oidvo.getPrimaryOidLabel());
            tmp.setSecondaryOidLabel(countryCode.getFullName());
            result.add(tmp);
        }
        return result;
    }
}
