/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.dao.PersonDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.abstr.AbstractECPRDAOImpl;
import dk.sds.ecpr2.domain.IdentifierWithoutServiceActorVO;
import dk.sds.ecpr2.domain.PersonIdVO;
import dk.sds.ecpr2.domain.PersonVO;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.stereotype.Repository;
import org.springframework.util.CollectionUtils;

import java.time.Instant;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

@Repository
public class PersonDAOImpl extends AbstractECPRDAOImpl implements PersonDAO {

    private final String TABLE_NAME = "Person";
    private final String personStatus = "Deleted";
    private final String INSERT_SQL = "INSERT INTO `ecpr2`.`Person` (`PersonPID`, `ServiceActorPID`, `ModifierPID`, `PersonUUID`, `Version`, `VersionDescription`, `ReservedUntil`, `ModifiedTime`) " +
            "VALUES (:PersonPID, :ServiceActorPID, :ModifierPID, :PersonUUID, :Version, :VersionDescription, :ReservedUntil, :modified)";
    private final String NOTDELETED_SQL = "(Status <> :status OR Status IS NULL)";

    @Override
    public List<PersonVO> getPersonsByUUIDs(List<String> personUuids) throws DAOException {
        final String sql = "SELECT * FROM Person WHERE PersonUUID IN (:personUUIDs) AND " + NOTDELETED_SQL + " ORDER BY `Version` ASC";
        final MapSqlParameterSource params = new MapSqlParameterSource()
                .addValue("personUUIDs", personUuids)
                .addValue("status", personStatus);
        return queryForList(sql, params);
    }

    @Override
    public PersonVO getCurrentPerson(String personUuid) throws DAOException {
        final String sql = "SELECT * FROM Person p WHERE PersonUUID = :personUUID ORDER BY p.version DESC LIMIT 1";
        final MapSqlParameterSource params = new MapSqlParameterSource().addValue("personUUID", personUuid);
        List<PersonVO> list = queryForList(sql, params);
        return list.isEmpty() ? null : list.get(0);
    }

    @Override
    public void createPersons(List<PersonVO> persons) throws DAOException {
        executeBatchUpdate(INSERT_SQL, persons);
    }

    @Override
    public Integer createPerson(PersonVO p) throws DAOException {
        final Integer pid = createSingleObject(INSERT_SQL, p);
        p.setPersonPID(pid);
        return pid;
    }

    @Override
    public List<PersonVO> getPersonsModifiedAfter(Instant modifiedDateTimeUtc, boolean excludeReservedUntil) throws DAOException {
        String sql = "SELECT * FROM Person p WHERE " + NOTDELETED_SQL;
        final MapSqlParameterSource parameters = new MapSqlParameterSource();
        parameters.addValue("status", personStatus);
        if (modifiedDateTimeUtc != null) {
            sql += " AND p.ModifiedTime >= :modified";
            parameters.addValue("modified", getTimestampInCopenhagenTime(modifiedDateTimeUtc));
        }
        if (excludeReservedUntil) {
            sql += " AND p.ReservedUntil IS NULL";
        }
        return queryForList(sql, parameters);
    }

    @Override
    public List<PersonVO> getPersonsAfterPID(int minPersonPID, int batchSize, boolean excludeReservedUntil) throws DAOException {
        String sql = "SELECT * FROM Person p WHERE " + NOTDELETED_SQL;
        final MapSqlParameterSource parameters = new MapSqlParameterSource();
        parameters.addValue("status", personStatus);
        if (excludeReservedUntil) {
            sql += " AND p.ReservedUntil IS NULL";
        }
        if (minPersonPID > -1 && batchSize > 0) {
            sql += " AND p.PersonPID >= :minPersonPid ORDER BY p.PersonPID LIMIT :batchSize";
            parameters.addValue("minPersonPid", minPersonPID);
            parameters.addValue("batchSize", batchSize);
        }
        return queryForList(sql, parameters);
    }

    @Override
    public PersonVO getPersonByPID(Integer personPID) throws DAOException {
        List<PersonVO> personList = getPersonByPIDs(Collections.singletonList(personPID));
        if (personList.isEmpty()) {
            return null;
        }
        return personList.get(0);
    }

    public List<PersonVO> getPersonByPIDs(List<Integer> personPIDs) throws DAOException {
        if (personPIDs == null || personPIDs.isEmpty()) {
            return new ArrayList<>();
        }
        final String sql = "SELECT * FROM Person WHERE PersonPID IN (:personPID) AND " + NOTDELETED_SQL;
        final MapSqlParameterSource parms = new MapSqlParameterSource();
        parms.addValue("personPID", personPIDs);
        parms.addValue("status", personStatus);
        return queryForList(sql, parms);
    }

    @Override
    public void deletePerson(String personuuid) throws DAOException {
        try {
            final String sql = "UPDATE " + TABLE_NAME + " SET Status = :status WHERE PersonUUID = :personUUID";
            final MapSqlParameterSource parameters = new MapSqlParameterSource();
            parameters.addValue("status", personStatus);
            parameters.addValue("personUUID", personuuid);
            executeUpdate(sql, parameters);
        } catch (DAOException e) {
            throw new DAOException("An error occurred deleting person with personuuid: " + personuuid, e);
        }
    }

    @Override
    public List<PersonVO> getPersonUuidByPersonIds(List<PersonIdVO> identifiers) throws DAOException {
        if (CollectionUtils.isEmpty(identifiers)) {
            return Collections.emptyList();
        }
        final String sql = "SELECT i.* FROM Identifier i WHERE i.value in (:values)";
        final MapSqlParameterSource parameters = new MapSqlParameterSource();
        List<String> identifierValues = identifiers.stream().map(PersonIdVO::getValue).collect(Collectors.toList());
        parameters.addValue("values", identifierValues);
        List<IdentifierWithoutServiceActorVO> result = queryForList(sql, parameters, new IdentifierWithoutServiceActorRowMapper());
        List<Integer> identifier = removeInvalidIdentifierCombi(identifiers, result);
        if (CollectionUtils.isEmpty(identifier)) {
            return Collections.emptyList();
        }
        return getPersonByPIDs(identifier);
    }

    private List<Integer> removeInvalidIdentifierCombi(List<PersonIdVO> identifiers, List<IdentifierWithoutServiceActorVO> result) {
        List<IdentifierWithoutServiceActorVO> identifierResult = new ArrayList<>();
        List<Integer> personPIDs = new ArrayList<>();
        for (IdentifierWithoutServiceActorVO identifierVO : result) {
            for (PersonIdVO identifier : identifiers) {
                if (identifierVO.getOid().equals(identifier.getOid()) && identifierVO.getValue().equals(identifier.getValue())) {
                    identifierResult.add(identifierVO);
                }
            }
        }
        for (IdentifierWithoutServiceActorVO identifier : identifierResult) {
            personPIDs.add(identifier.getPersonPID());
        }
        return personPIDs;
    }

    @Override
    public List<PersonVO> getPersonByReservedUntil(Instant date) throws DAOException {
        final String sql = "SELECT * FROM Person WHERE ReservedUntil < :date AND " + NOTDELETED_SQL;
        final MapSqlParameterSource parameters = new MapSqlParameterSource();
        parameters.addValue("date", getTimestampInCopenhagenTime(date));
        parameters.addValue("status", personStatus);

        return queryForList(sql, parameters);
    }

    @Override
    public List<PersonVO> getReservedPersonByModifiedTime(Instant date) throws DAOException {
        final String sql = "SELECT * FROM Person WHERE ReservedUntil is not null AND ModifiedTime > :date AND " + NOTDELETED_SQL;
        final MapSqlParameterSource parameters = new MapSqlParameterSource();
        parameters.addValue("date", getTimestampInCopenhagenTime(date));
        parameters.addValue("status", personStatus);

        return queryForList(sql, parameters);
    }


    @Override
    public String getTableName() {
        return TABLE_NAME;
    }

    @Override
    protected RowMapper getDefaultRowMapper() {
        return new PersonRowMapper();
    }
}
