/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.dao.*;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.abstr.AbstractCrossTableDAOImpl;
import dk.sds.ecpr2.domain.*;
import dk.sds.ecpr2.domain.abstr.PersonDescriptorVO;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import java.time.Instant;
import java.util.*;
import java.util.stream.Collectors;

@Repository
@Log4j2
public class SearchCrossTableDAOImpl extends AbstractCrossTableDAOImpl {

    @Autowired
    PersonDAO personDAO;

    @Autowired
    private IdentifierDAO identifierDAO;

    @Autowired
    private AddressDAO addressDAO;

    @Autowired
    private BirthDateDAO birthDateDAO;

    @Autowired
    private ContactDAO contactDAO;

    @Autowired
    private GenderDAO genderDAO;

    @Autowired
    private NameDAO nameDAO;


    @Override
    public List<PersonVO> getPersonsWithRelationsAfterPID(int minPersonPid,
                                                          int batchSize,
                                                          boolean excludeReservedUntilFilter,
                                                          boolean excludeDuplicatePersonsFilter) throws DAOException {
        return getPersonsWithRelations(minPersonPid, batchSize, excludeReservedUntilFilter, excludeDuplicatePersonsFilter);
    }

    @Override
    public List<PersonVO> getPersonsWithRelationsAfterModifiedTime(Instant modifiedDateTime, boolean excludeDuplicatePersonsFilter) throws DAOException {
        return getPersonsWithRelations(modifiedDateTime, excludeDuplicatePersonsFilter);
    }

    private List<PersonVO> getPersonsWithRelations(int minPersonPid,
                                                   int batchSize,
                                                   boolean excludeReservedUntil,
                                                   boolean excludeDuplicatePersonsFilter) throws DAOException {
        // Remove persons with duplicate UUID's
        List<PersonVO> persons = personDAO.getPersonsAfterPID(minPersonPid, batchSize, excludeReservedUntil);
        List<PersonVO> filteredPersons = excludeDuplicatePersonsFilter
                ? getNewestPersons(persons)
                : persons;

        Map<Integer, PersonVO> personMap = new HashMap<>();
        for (PersonVO p : filteredPersons) {
            personMap.put(p.getPersonPID(), p);
        }

        final int maxPersonPID = PersonVO.getMaxPersonPID(filteredPersons);
        final int minPersonPID = PersonVO.getMinPersonPID(filteredPersons);

        // Not possible to make these generic?
        List<IdentifierVO> allIdentifierRelations = identifierDAO.getAllIdentifiersBetweenPersonPIDForSearch(minPersonPID, maxPersonPID);
        List<IdentifierVO> identifierVOS = getNewestPersonDescriptors(allIdentifierRelations, excludeDuplicatePersonsFilter);
        for (IdentifierVO vo : identifierVOS) {
            final PersonVO person = personMap.get(vo.getPersonPID());
            if (person != null) {
                person.getIdentifiers().add(vo);
            }
        }
        List<NameVO> allNameRelations = nameDAO.getAllNamesBetweenPersonPIDForSearch(minPersonPID, maxPersonPID);
        List<NameVO> nameVOS = getNewestPersonDescriptors(allNameRelations, excludeDuplicatePersonsFilter);
        for (NameVO vo : nameVOS) {
            final PersonVO person = personMap.get(vo.getPersonPID());
            if (person != null) {
                person.getNames().add(vo);
            }
        }
        List<GenderVO> allGenderRelations = genderDAO.getAllGendersBetweenPersonPIDForSearch(minPersonPID, maxPersonPID);
        List<GenderVO> genderVOS = getNewestPersonDescriptors(allGenderRelations, excludeDuplicatePersonsFilter);
        for (GenderVO vo : genderVOS) {
            final PersonVO person = personMap.get(vo.getPersonPID());
            if (person != null) {
                personMap.get(vo.getPersonPID()).getGenders().add(vo);
            }
        }
        List<BirthDateVO> allBirthdateRelations = birthDateDAO.getAllBirthdatesBetweenPersonPIDForSearch(minPersonPID, maxPersonPID);
        List<BirthDateVO> birthDateVOS = getNewestPersonDescriptors(allBirthdateRelations, excludeDuplicatePersonsFilter);
        for (BirthDateVO vo : birthDateVOS) {
            final PersonVO person = personMap.get(vo.getPersonPID());
            if (person != null) {
                personMap.get(vo.getPersonPID()).getBirthDates().add(vo);
            }
        }
        List<ContactVO> allContactRelations = contactDAO.getAllContactsBetweenPersonPIDForSearch(minPersonPID, maxPersonPID);
        List<ContactVO> contactVOS = getNewestPersonDescriptors(allContactRelations, excludeDuplicatePersonsFilter);
        for (ContactVO vo : contactVOS) {
            final PersonVO person = personMap.get(vo.getPersonPID());
            if (person != null) {
                personMap.get(vo.getPersonPID()).getContacts().add(vo);
            }
        }
        List<AddressVO> allAddressRelations = addressDAO.getAllAddressesBetweenPersonPIDForSearch(minPersonPID, maxPersonPID);
        List<AddressVO> addressVOS = getNewestPersonDescriptors(allAddressRelations, excludeDuplicatePersonsFilter);
        for (AddressVO vo : addressVOS) {
            final PersonVO person = personMap.get(vo.getPersonPID());
            if (person != null) {
                personMap.get(vo.getPersonPID()).getAddresses().add(vo);
            }
        }

        return filteredPersons;
    }

    private List<PersonVO> getPersonsWithRelations(Instant modifiedTime, boolean excludeDuplicatePersons) throws DAOException {
        List<IdentifierVO> allIdentifierRelations = identifierDAO.getAllIdentifiersAfterModifiedForSearch(modifiedTime);
        List<NameVO> allNameRelations = nameDAO.getAllNamesAfterModifiedForSearch(modifiedTime);
        List<GenderVO> allGenderRelations = genderDAO.getAllGendersAfterModifiedForSearch(modifiedTime);
        List<BirthDateVO> allBirthdateRelations = birthDateDAO.getAllBirthdatesAfterModifiedForSearch(modifiedTime);
        List<ContactVO> allContactRelations = contactDAO.getAllContactsAfterModifiedForSearch(modifiedTime);
        List<AddressVO> allAddressRelations = addressDAO.getAllAddressesAfterModifiedForSearch(modifiedTime);

        List<IdentifierVO> identifierVOS = getNewestPersonDescriptors(allIdentifierRelations, excludeDuplicatePersons);
        List<NameVO> nameVOS = getNewestPersonDescriptors(allNameRelations, excludeDuplicatePersons);
        List<GenderVO> genderVOS = getNewestPersonDescriptors(allGenderRelations, excludeDuplicatePersons);
        List<BirthDateVO> birthDateVOS = getNewestPersonDescriptors(allBirthdateRelations, excludeDuplicatePersons);
        List<ContactVO> contactVOS = getNewestPersonDescriptors(allContactRelations, excludeDuplicatePersons);
        List<AddressVO> addressVOS = getNewestPersonDescriptors(allAddressRelations, excludeDuplicatePersons);

        ArrayList<PersonDescriptorVO> personDescriptorVOS = new ArrayList<>();
        personDescriptorVOS.addAll(identifierVOS);
        personDescriptorVOS.addAll(nameVOS);
        personDescriptorVOS.addAll(genderVOS);
        personDescriptorVOS.addAll(birthDateVOS);
        personDescriptorVOS.addAll(contactVOS);
        personDescriptorVOS.addAll(addressVOS);

        Set<Integer> uniquePersonPIDS = personDescriptorVOS.stream().map(PersonDescriptorVO::getPersonPID).collect(Collectors.toSet());

        List<PersonVO> persons = personDAO.getPersonByPIDs(new ArrayList<>(uniquePersonPIDS));

        Map<Integer, PersonVO> personMap = new HashMap<>();
        for (PersonVO p : persons) {
            personMap.put(p.getPersonPID(), p);
        }

        for (IdentifierVO vo : identifierVOS) {
            final PersonVO person = personMap.get(vo.getPersonPID());
            if (person != null) {
                person.getIdentifiers().add(vo);
            }
        }

        for (NameVO vo : nameVOS) {
            final PersonVO person = personMap.get(vo.getPersonPID());
            if (person != null) {
                person.getNames().add(vo);
            }
        }

        for (GenderVO vo : genderVOS) {
            final PersonVO person = personMap.get(vo.getPersonPID());
            if (person != null) {
                personMap.get(vo.getPersonPID()).getGenders().add(vo);
            }
        }

        for (BirthDateVO vo : birthDateVOS) {
            final PersonVO person = personMap.get(vo.getPersonPID());
            if (person != null) {
                personMap.get(vo.getPersonPID()).getBirthDates().add(vo);
            }
        }

        for (ContactVO vo : contactVOS) {
            final PersonVO person = personMap.get(vo.getPersonPID());
            if (person != null) {
                personMap.get(vo.getPersonPID()).getContacts().add(vo);
            }
        }

        for (AddressVO vo : addressVOS) {
            final PersonVO person = personMap.get(vo.getPersonPID());
            if (person != null) {
                personMap.get(vo.getPersonPID()).getAddresses().add(vo);
            }
        }
        if (!excludeDuplicatePersons) {
            return persons;
        }
        return getNewestPersons(persons);
    }
}
