/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.dao.SearchPersonDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.abstr.AbstractSearchDAOImpl;
import dk.sds.ecpr2.domain.LuceneKeyValue;
import dk.sds.ecpr2.domain.PersonVO;
import lombok.extern.log4j.Log4j2;
import org.apache.lucene.document.Document;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.Term;
import org.apache.lucene.search.*;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Repository;
import org.springframework.util.CollectionUtils;

import java.io.IOException;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

@Repository
@Log4j2
public class SearchPersonDAOImpl extends AbstractSearchDAOImpl implements SearchPersonDAO {
    @Value("${jobs.searchindex.maxResults:100}")
    protected int MAXNUMBEROFLUCENEMATCHES;

    @Value("${jobs.searchindex.alwaysIndexFullDatabase:false}")
    protected boolean alwaysIndexFullDatabase;

    protected static final String PERSONUUID = "personuuid";
    public static boolean firstExecutionCompleted = false;

    @Value("${jobs.searchindex.initialLoad.batchSize:20000}")
    private int batchSize;

    @Value("${jobs.searchindex.initialLoad.msBetweenBatches:500}")
    private int SLEEP_TIME_MS_BETWEEN_BATCHES;

    @Override
    public int buildIndex(Instant date, IndexWriter writer, boolean excludeDuplicatePersons) throws Exception {
        try {
            int count = 0;
            count += buildIndexWithoutErrorHandling(writer, date, excludeDuplicatePersons);
            writer.commit();
            return count;
        } catch (Exception e) {
            if (writer != null)
                writer.rollback();
            throw e;
        }
    }

    private int buildIndexWithoutErrorHandling(IndexWriter writer, Instant date, boolean excludeDuplicatePersons) throws DAOException, IOException, InterruptedException {
        int count = 0;
        log.debug("Starting to fetch data for search index... Last fetched timestamp: " + date);

        final boolean isInitialLoad = (date == null);
        if (isInitialLoad || alwaysIndexFullDatabase) {
            count = indexCompleteDatabase(writer, excludeDuplicatePersons);
            firstExecutionCompleted = true;
        } else {
            count = indexDatabaseChanges(writer, date, excludeDuplicatePersons);
        }

        log.debug("Done updating index.");
        return count;
    }

    private int indexDatabaseChanges(IndexWriter writer, Instant date, boolean excludeDuplicatePersons) throws DAOException, IOException {
        List<PersonVO> deltaPersons = getDAO().getPersonsWithRelationsAfterModifiedTime(date, excludeDuplicatePersons);
        log.debug("Done fetching data incrementally for search index (" + deltaPersons.size() + " rows). Indexing...");
        updateLuceneIndex(writer, deltaPersons);
        return deltaPersons.size();
    }

    private int indexCompleteDatabase(IndexWriter writer, boolean excludeDuplicatePersons) throws DAOException, InterruptedException, IOException {
        int count = 0;
        List<PersonVO> personInfoBatch;
        int minPersonPID = 0;
        log.info("Starting to index all existing data.");
        writer.deleteAll();
        do {
            personInfoBatch = getDAO().getPersonsWithRelationsAfterPID(minPersonPID,
                    batchSize,
                    false,
                    excludeDuplicatePersons);
            log.debug("Fetched " + personInfoBatch.size() + " rows for search index. First person PID: " + minPersonPID);
            updateLuceneIndex(writer, personInfoBatch);
            count += personInfoBatch.size();
            minPersonPID = PersonVO.getMaxPersonPID(personInfoBatch) + 1;
            Thread.sleep(SLEEP_TIME_MS_BETWEEN_BATCHES); // Don't overload the machine when loading large amounts of data.
        } while (personInfoBatch.size() > 0);
        writer.commit();
        writer.flush();
        log.info("Finished indexing all existing data. Row count: " + count);
        return count;
    }


    private void updateLuceneIndex(IndexWriter writer, List<PersonVO> allPersonInfo) throws IOException {
        for (PersonVO person : allPersonInfo) {
            Document document = new Document();
            addPersonIndexFields(document, person);
            writer.addDocument(document);
        }

        if (allPersonInfo.size() > 0) {
            log.debug("Updated Lucene index with " + allPersonInfo.size() + " database rows.");
        }
    }

    private void addPersonIndexFields(Document document, PersonVO person) {
        for (LuceneKeyValue luceneList : person.getLuceneList()) {
            addIndexedField(document, luceneList.getFieldName(), luceneList.getValue());
        }
    }

    @Override
    public List<PersonVO> searchPerson(List<String> searchStr) throws Exception {
        return searchPerson(getFullIndexSearcher(), searchStr);
    }

    @Override
    public void addPersonToIndex(List<PersonVO> persons) throws IOException {
        IndexWriter indexWriter = getFullIndexWriter();
        addIndex(persons, indexWriter);

        indexWriter = getOnlyCurrentIndexWriter();
        addIndex(persons, indexWriter);
    }

    @Override
    public void updatePersonsInIndex(List<PersonVO> persons) throws IOException {
        IndexWriter indexWriter = getFullIndexWriter();
        updateIndex(persons, indexWriter);

        indexWriter = getOnlyCurrentIndexWriter();
        deleteIndex(persons, indexWriter);
        updateIndex(persons, indexWriter);
    }

    private void addIndex(List<PersonVO> persons, IndexWriter indexWriter) throws IOException {
        for (PersonVO person : persons) {
            Document document = new Document();
            addPersonIndexFields(document, person);
            indexWriter.addDocument(document);
            indexWriter.commit();
            indexWriter.flush();
        }
    }

    private void updateIndex(List<PersonVO> persons, IndexWriter indexWriter) throws IOException {
        for (PersonVO person : persons) {
            Document document = new Document();
            addPersonIndexFields(document, person);
            indexWriter.updateDocument(new Term(PERSONUUID, person.getPersonUUID()), document);
            indexWriter.commit();
            indexWriter.flush();
        }
    }

    @Override
    public void deleteIndex(List<PersonVO> personVO) throws IOException {
        IndexWriter indexWriter = getFullIndexWriter();
        deleteIndex(personVO, indexWriter);
        indexWriter = getOnlyCurrentIndexWriter();
        deleteIndex(personVO, indexWriter);
    }

    private void deleteIndex(List<PersonVO> person, IndexWriter indexWriter) throws IOException {
        for (PersonVO persons : person) {
            indexWriter.deleteDocuments(new Term(PERSONUUID, persons.getPersonUUID()));
            indexWriter.commit();
            indexWriter.flush();
        }
    }

    public List<PersonVO> searchPerson(IndexSearcher searcher, List<String> searchStr) throws Exception {
        BooleanQuery.Builder booleanQueryBuilder = new BooleanQuery.Builder();

        for (String sStr : searchStr) {
            List<String> keyValue = Arrays.asList(sStr.split(":"));
            String fieldName = keyValue.get(0).trim();
            String searchString = keyValue.get(1).trim();
            Query query = parseQuery(searchString, fieldName);
            log.info(query);
            booleanQueryBuilder.add(query, BooleanClause.Occur.MUST);
        }

        TopDocs topDocs = searcher.search(booleanQueryBuilder.build(), MAXNUMBEROFLUCENEMATCHES);
        List<String> personUUIDs = new ArrayList<>();

        for (int i = 0; i < topDocs.scoreDocs.length; i++) {
            ScoreDoc scoreDoc = topDocs.scoreDocs[i];
            Document doc = searcher.doc(scoreDoc.doc);
            personUUIDs.add(getPersonUUID(doc));

        }
        return getMatchesInDAOByUUID(personUUIDs);
    }

    List<PersonVO> getMatchesInDAOByUUID(List<String> personUUIDs) throws DAOException {
        if (!CollectionUtils.isEmpty(personUUIDs)) {
            return getDAO().getPersonsWithRelations(personUUIDs);
        }
        return Collections.emptyList();
    }

    private String getPersonUUID(Document doc) {
        return doc.get(PERSONUUID);
    }

    public boolean getFirstExecutionStatus() {
        return firstExecutionCompleted;
    }

    @Override
    public int countPerson(List<String> searchQuery) throws Exception {
        return countPerson(getOnlyCurrentIndexSearcher(), searchQuery);
    }

    private int countPerson(IndexSearcher searcher, List<String> searchQuery) throws Exception {
        BooleanQuery.Builder booleanQuery = new BooleanQuery.Builder();

        for (String sStr : searchQuery) {
            String[] split = sStr.split(":");
            String fieldName = split[0].trim();
            String searchString = split[1].trim();
            Query query = parseQuery(searchString, fieldName);
            log.info(query);
            booleanQuery.add(query, BooleanClause.Occur.MUST);
        }

        return searcher.count(booleanQuery.build());
    }
}

