/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.dao.ServiceActorDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.abstr.AbstractECPRDAOImpl;
import dk.sds.ecpr2.domain.ServiceActorVO;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.stereotype.Repository;

import java.util.Date;
import java.util.List;

import static dk.sds.ecpr2.util.ArgumentChecker.checkNotNull;

@Repository
public class ServiceActorDAOImpl extends AbstractECPRDAOImpl implements ServiceActorDAO {


    private static final String TABLE_NAME = "ServiceActor";
    final String INSERT_SQL = "INSERT INTO ServiceActor (PersonIdentifier, PersonIdentifierOID, FirstName, LastName, Role, AuthorisationNumber, OrganizationId, OrganizationName, HashValue, ModifiedTime) "
            + "VALUES (:personIdentifier, :personIdentifierOID, :firstName, :lastName, :role, :authorisationNumber, :organizationId, :organizationName, :hashCode, :modifiedTime)";

    @Override
    public ServiceActorVO getServiceActor(Integer serviceActorPID) throws DAOException {
        checkNotNull(serviceActorPID, "ServiceActorPID");
        final String sql = "SELECT " + ACTOR_FIELDS + " FROM ServiceActor a WHERE a.ServiceActorPID = :pid";
        final MapSqlParameterSource parms = new MapSqlParameterSource().addValue("pid", serviceActorPID);
        return namedParamJdbcTemplate.queryForObject(sql, parms, new ServiceActorRowMapper());
    }

    @Override
    public boolean exists(ServiceActorVO serviceActorVO) throws DAOException {
        final int hashCode = serviceActorVO.hashCode();
        if (hashCode == 0) {
            return false;
        }

        final String sql = "SELECT " + ACTOR_FIELDS + " FROM ServiceActor a WHERE a.HashValue = :hashCode";
        final MapSqlParameterSource parms = new MapSqlParameterSource().addValue("hashCode", hashCode);
        final List<ServiceActorVO> list = namedParamJdbcTemplate.query(sql, parms, getDefaultRowMapper());

        for (ServiceActorVO vo : list) {
            if (vo.equals(serviceActorVO)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public void createServiceActors(List<ServiceActorVO> serviceActors) throws DAOException {
        executeBatchUpdate(INSERT_SQL, serviceActors);
    }

    public Integer createServiceActor(ServiceActorVO sa, Date now) throws DAOException {
        sa.setModifiedTime(now);
        final Integer pid = createSingleObject(INSERT_SQL, sa);
        sa.setServiceActorPID(pid);
        return pid;
    }

    @Override
    public Integer getServiceActorFromHashCode(ServiceActorVO serviceActorVO) throws DAOException {
        final int hashCode = serviceActorVO.hashCode();

        try {
            if (hashCode != 0) {
                final String sql = "SELECT " + ACTOR_FIELDS + " FROM ServiceActor a WHERE a.HashValue = :hashCode";
                final MapSqlParameterSource parms = new MapSqlParameterSource().addValue("hashCode", hashCode);
                final List<ServiceActorVO> list = namedParamJdbcTemplate.query(sql, parms, new ServiceActorRowMapper());

                for (ServiceActorVO vo : list) {
                    if (vo.equals(serviceActorVO)) {
                        return vo.getServiceActorPID();
                    }
                }
            }
            return null;
        } catch (Exception e) {
            throw new DAOException("An error occurred loading ServiceActors by hashcode", e);
        }
    }

    @Override
    public String getTableName() {
        return TABLE_NAME;
    }

    @Override
    protected RowMapper getDefaultRowMapper() {
        return new ServiceActorRowMapper();
    }
}
