/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl.abstr;

import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.util.TimeHelper;
import io.micrometer.core.instrument.Metrics;
import org.apache.catalina.util.ParameterMap;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.core.namedparam.SqlParameterSource;
import org.springframework.jdbc.core.namedparam.SqlParameterSourceUtils;
import org.springframework.jdbc.support.GeneratedKeyHolder;

import javax.annotation.PostConstruct;
import javax.sql.DataSource;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

/**
 * This class contains general purpose database operations. They should all make sense outside this project.
 */
public abstract class AbstractDAOImpl {
    protected NamedParameterJdbcTemplate namedParamJdbcTemplate;
    @Autowired
    private DataSource dataSource;

    @PostConstruct
    private void init() {
        namedParamJdbcTemplate = new NamedParameterJdbcTemplate(dataSource);
    }

    /**
     * This is based on search time and has nothing to do with updating
     * @param instant
     * @return
     */
    protected String getTimestampInCopenhagenTime(Instant instant) {
        ZoneId zoneId = ZoneId.of("Europe/Copenhagen");
        ZoneOffset zoneOffset = zoneId.getRules().getOffset(instant);
        LocalDateTime localDateTime = LocalDateTime.ofInstant(instant, zoneOffset);
        return localDateTime.toString();
    }

    public abstract String getTableName();

    protected abstract String getIdColumnName();

    protected abstract RowMapper getDefaultRowMapper();

    public int rowCount() throws DAOException {
        final String sql = "SELECT COUNT(*) FROM " + getTableName();
        return queryForInteger(sql);
    }

    /**
     * Careful! Deletes rows from the table. Typically only used for testing.
     */
    public void deleteRowsById(int fromId, int toId) throws DAOException {
        final MapSqlParameterSource parms = new MapSqlParameterSource()
                .addValue("from", fromId)
                .addValue("to", toId);
        executeUpdate("DELETE FROM " + getTableName() + " WHERE "
                        + getIdColumnName() + " >= :from AND " + getIdColumnName() + " <= :to",
                parms);
    }

    public int getMaxId() throws DAOException {
        return queryForInteger("SELECT Max(" + getIdColumnName() + ") FROM " + getTableName());
    }

    protected int queryForInteger(String sql) throws DAOException {
        try {
            final long startTime = System.nanoTime();
            int rowCount = namedParamJdbcTemplate.queryForObject(sql, new ParameterMap<>(), Integer.class);
            recordMetrics(sql, startTime);
            return rowCount;
        } catch (Exception e) {
            throw new DAOException("An error occurred executing count query: " + sql, e);
        }
    }

    protected <T> T queryForObject(String sql, SqlParameterSource parms) throws DAOException {
        List<T> result = queryForList(sql, parms);
        if (result == null || result.size() == 0) {
            return null;
        }
        if (result.size() > 1) {
            throw new DAOException("More than one row returned when only one was expected.");
        }
        return result.get(0);
    }

    /**
     * @return null if no data is found.
     * @throws DAOException if multiple rows are found.
     */
    protected <T> T fetchSingleRowMatchingSingleColumn(String columnName, String filterValue) throws DAOException {
        List<T> result = fetchRowsMatchingSingleColumn(columnName, filterValue);
        if (result.size() == 0) {
            return null;
        } else if (result.size() > 1) {
            throw new DAOException("Multiple rows were found in table " + getTableName()
                    + " looking in column " + columnName + " for the value " + filterValue);
        } else {
            return result.get(0);
        }
    }

    protected <T> List<T> fetchRowsMatchingSingleColumn(String columnName, String filterValue) throws DAOException {
        final MapSqlParameterSource parameters = new MapSqlParameterSource()
                .addValue("val", filterValue);
        String sql = "SELECT * FROM " + getTableName() + " WHERE " + columnName + "=:val";
        return queryForList(sql, parameters);
    }

    protected <T> List<T> fetchAllRows() throws DAOException {
        return queryForList("SELECT * FROM " + getTableName());
    }

    protected <T> List<T> queryForList(String sql) throws DAOException {
        return queryForList(sql, new MapSqlParameterSource());
    }

    protected <T> T fetchSingleRowMatchingTwoColumns(String columnName1, String filterValue1, String columnName2, String filterValue2) throws DAOException {
        return fetchSingleRowMatchingTwoColumns(columnName1, filterValue1, columnName2, filterValue2, getDefaultRowMapper());
    }

    protected <T> T fetchSingleRowMatchingTwoColumns(String columnName1, String filterValue1, String columnName2, String filterValue2, RowMapper rowMapper) throws DAOException {
        return fetchSingleRowMatchingTwoColumns(columnName1, filterValue1, columnName2, filterValue2, rowMapper, false);
    }

    protected <T> T fetchSingleRowMatchingTwoColumns(String columnName1, String filterValue1, String columnName2, String filterValue2, RowMapper rowMapper, boolean requireCurrentValue) throws DAOException {
        List<T> result = fetchRowsMatchingTwoColumns(columnName1, filterValue1, columnName2, filterValue2, rowMapper, requireCurrentValue);
        if (result.size() == 0) {
            return null;
        } else if (result.size() > 1) {
            throw new DAOException("Multiple rows were found in table " + getTableName()
                    + " looking in column " + columnName1 + " for the value " + filterValue1 + " and column " + columnName2 + " for the value " + filterValue2);
        } else {
            return result.get(0);
        }
    }

    protected <T> List<T> fetchRowsMatchingTwoColumns(String columnName1, String filterValue1, String columnName2, String filterValue2, RowMapper rowMapper, boolean requireCurrentValue) throws DAOException {
        final MapSqlParameterSource parms = new MapSqlParameterSource()
                .addValue("val1", filterValue1)
                .addValue("val2", filterValue2);
        String sql = "SELECT * FROM " + getTableName() + " WHERE " + columnName1 + "=:val1 AND " + columnName2 + "=:val2";
        if (requireCurrentValue) {
            sql += " AND ValidFrom <= :now AND :now < ValidTo";
            parms.addValue("now", TimeHelper.getCurrentTime());
        }
        return queryForList(sql, parms, rowMapper);
    }

    protected <T> List<T> queryForList(String sql, SqlParameterSource parms) throws DAOException {
        return queryForList(sql, parms, getDefaultRowMapper());
    }

    protected <T> List<T> queryForList(String sql, SqlParameterSource parms, RowMapper rowMapper) throws DAOException {
        try {
            final long startTime = System.nanoTime();
            final List<T> list = namedParamJdbcTemplate.query(sql, parms, rowMapper);
            recordMetrics(sql, startTime);
            return list;
        } catch (Exception e) {
            throw new DAOException("An error occurred executing query: " + sql, e);
        }
    }

    protected void executeUpdate(String updateSql) throws DAOException {
        executeUpdate(updateSql, new MapSqlParameterSource());
    }

    protected void executeUpdate(String updateSql, SqlParameterSource parameters) throws DAOException {
        try {
            final long startTime = System.nanoTime();
            namedParamJdbcTemplate.update(updateSql, parameters);
            recordMetrics(updateSql, startTime);
        } catch (Exception e) {
            throw new DAOException("An error occurred executing update: " + updateSql, e);
        }
    }

    protected <T> Integer createSingleObject(String insertSql, T object) throws DAOException {
        if (object == null) {
            return null;
        }
        try {
            final GeneratedKeyHolder generatedKeyHolder = new GeneratedKeyHolder();
            ArrayList<T> a = new ArrayList<>();
            a.add(object);
            SqlParameterSource[] params = SqlParameterSourceUtils.createBatch(a.toArray());
            namedParamJdbcTemplate.update(insertSql, params[0], generatedKeyHolder);
            return generatedKeyHolder.getKey().intValue();
        } catch (Exception e) {
            throw new DAOException("An error occurred creating object: " + insertSql, e);
        }
    }

    protected <T> void executeBatchUpdate(String updateSql, List<T> batch) throws DAOException {
        if (batch.isEmpty()) {
            return;
        }
        try {
            final long startTime = System.nanoTime();
            namedParamJdbcTemplate.batchUpdate(updateSql, SqlParameterSourceUtils.createBatch(batch.toArray()));
            recordMetrics(updateSql, startTime);
        } catch (Exception e) {
            throw new DAOException("An error occurred executing batch update: " + updateSql, e);
        }
    }

    private void recordMetrics(String sql, long startTime) {
        final String metricName = sql.substring(0, Math.min(50, sql.length()));
        final long durationMillis = (System.nanoTime() - startTime) / 1000000;
        Metrics.timer(metricName).record(durationMillis, TimeUnit.MILLISECONDS);
    }
}
