/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl.abstr;

import dk.sds.ecpr2.dao.ECPR2DAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.PIDRowMapper;
import dk.sds.ecpr2.domain.abstr.PersonDescriptorVO;
import dk.sds.ecpr2.util.TimeHelper;
import lombok.extern.log4j.Log4j2;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;

import java.time.Instant;
import java.util.*;


/**
 * This class contains general purpose database operations that only make sense within the E-CPR project.
 */
@Log4j2
public abstract class AbstractECPRDAOImpl extends AbstractDAOImpl implements ECPR2DAO {

    public static final Date VALID_TO_DEFAULT = TimeHelper.createNew(2999, Calendar.DECEMBER, 31);
    protected static final String ACTOR_FIELDS = "a.ServiceActorPID, a.PersonIdentifier as ActorPersonIdentifier, a.PersonIdentifierOID as ActorPersonIdentifierOID, a.FirstName, a.LastName, a.Role, a.AuthorisationNumber, a.OrganizationId, a.OrganizationName, a.HashValue, a.ModifiedTime";
    protected static final String TIMESTAMP_AND_ACTOR_FIELDS = "p.ModifiedTime, " + ACTOR_FIELDS;

    @Override
    protected String getIdColumnName() {
        return getTableName() + "PID";
    }

    protected <T extends PersonDescriptorVO> List<T> getWithServiceActorByPersonUUIDs(List<String> personUuids, List<Integer> personPids) throws DAOException {
        if (personUuids.isEmpty() && personPids.isEmpty()) {
            return new ArrayList<T>();
        }
        try {
            String columnName = null;
            final boolean uuidIsInitialized = (!personUuids.isEmpty());
            final boolean pidIsInitialized = (!personPids.isEmpty());
            List columnList = null;
            if (uuidIsInitialized) {
                columnName = "PersonUUID";
                columnList = personUuids;
            } else if (pidIsInitialized) {
                columnName = "PersonPid";
                columnList = personPids;
            }
            final String tableName = getTableName();
            String sql = "SELECT t.*, p.PersonUUID, " + TIMESTAMP_AND_ACTOR_FIELDS + " FROM Person p, ServiceActor a, "
                    + tableName + " t WHERE p." + columnName + " IN (:list)" +
                    " AND t.PersonPID = p.PersonPID AND p.ServiceActorPID = a.ServiceActorPID";

            final MapSqlParameterSource parms = new MapSqlParameterSource()
                    .addValue("list", columnList);
            return queryForList(sql, parms);
        } catch (Exception e) {
            throw new DAOException("An error occurred loading " + getTableName() + " by UUIDs", e);
        }
    }

    protected <T extends PersonDescriptorVO> List<T> getAllFieldsBetweenPersonPIDWithServiceActor(int minPersonPID, int maxPersonPID) throws DAOException {
        if (minPersonPID < 0 || maxPersonPID < 0) {
            throw new UnsupportedOperationException("Min and Max Person PID must be specified.");
        }
        try {
            final String tableName = getTableName();
            String sql = "SELECT t.*, p.PersonUUID, " + ACTOR_FIELDS + " FROM Person p, ServiceActor a, " + tableName + " t ";
            final MapSqlParameterSource parms = new MapSqlParameterSource();
            sql += " WHERE t.PersonPID >= :minPersonPid AND t.PersonPID <= :maxPersonPid " +
                    "AND t.PersonPID = p.PersonPID AND p.ServiceActorPID = a.ServiceActorPID";
            parms.addValue("minPersonPid", minPersonPID);
            parms.addValue("maxPersonPid", maxPersonPID);

            final PersonDescriptorRowMapper rowMapper = (PersonDescriptorRowMapper) getDefaultRowMapper().getClass().newInstance();
            rowMapper.useOnlySearchFields = false;
            return queryForList(sql, parms, rowMapper);
        } catch (Exception e) {
            throw new DAOException("An error occurred loading selected fields from " + getTableName() + "", e);
        }
    }

    protected <T extends PersonDescriptorVO> List<T> getSelectedFieldsBetweenPersonPIDWithoutServiceActor(List<String> fields,
                                                                                                          int minPersonPID,
                                                                                                          int maxPersonPID) throws DAOException {
        if (minPersonPID < 0 || maxPersonPID < 0) {
            throw new UnsupportedOperationException("Min and Max Person PID must be specified.");
        }

        try {
            final String tableName = getTableName();
            final String commaSeparatedFields = String.join(",", fields);
            String sql = "SELECT " + getIdColumnName() + ", PersonPID, " + commaSeparatedFields + " FROM " + tableName;

            final MapSqlParameterSource parms = new MapSqlParameterSource();
            sql += " WHERE PersonPID >= :minPersonPid AND PersonPID <= :maxPersonPid";
            parms.addValue("minPersonPid", minPersonPID);
            parms.addValue("maxPersonPid", maxPersonPID);

            final PersonDescriptorRowMapper rowMapper = (PersonDescriptorRowMapper) getDefaultRowMapper().getClass().newInstance();
            rowMapper.useOnlySearchFields = true;
            return queryForList(sql, parms, rowMapper);
        } catch (Exception e) {
            throw new DAOException("An error occurred loading selected fields from " + getTableName() + "", e);
        }
    }

    protected <T extends PersonDescriptorVO> List<T> getAllFieldsAfterModifiedTimeWithServiceActor(Instant modifiedTime) throws DAOException {
        if (modifiedTime == null) {
            throw new UnsupportedOperationException("ModifiedTime must not be null");
        }
        try {
            final String tableName = getTableName();
            String sql = "SELECT t.*, p.PersonUUID, " + ACTOR_FIELDS + " FROM Person p, ServiceActor a, " + tableName + " t ";
            final MapSqlParameterSource parms = new MapSqlParameterSource();
            sql += " WHERE t.ModifiedTime >= :modifiedTime " +
                    "AND t.PersonPID = p.PersonPID AND p.ServiceActorPID = a.ServiceActorPID";
            parms.addValue("modifiedTime", modifiedTime);

            final PersonDescriptorRowMapper rowMapper = (PersonDescriptorRowMapper) getDefaultRowMapper().getClass().newInstance();
            rowMapper.useOnlySearchFields = false;
            return queryForList(sql, parms, rowMapper);
        } catch (Exception e) {
            throw new DAOException("An error occurred loading selected fields from " + getTableName() + "", e);
        }
    }

    protected <T extends PersonDescriptorVO> List<T> getSelectedFieldsAfterModifiedTimeWithoutServiceActor(List<String> fields,
                                                                                                           Instant modifiedTime) throws DAOException {
        if (modifiedTime == null) {
            throw new UnsupportedOperationException("ModifiedTime must not be null");
        }

        try {
            final String tableName = getTableName();
            final String commaSeparatedFields = String.join(",", fields);
            String sql = "SELECT " + getIdColumnName() + ", PersonPID, " + commaSeparatedFields + " FROM " + tableName;

            final MapSqlParameterSource parms = new MapSqlParameterSource();
            sql += " WHERE ModifiedTime >= :modifiedTime";
            parms.addValue("modifiedTime", modifiedTime);

            final PersonDescriptorRowMapper rowMapper = (PersonDescriptorRowMapper) getDefaultRowMapper().getClass().newInstance();
            rowMapper.useOnlySearchFields = true;
            return queryForList(sql, parms, rowMapper);
        } catch (Exception e) {
            throw new DAOException("An error occurred loading selected fields from " + getTableName() + "", e);
        }
    }

    protected <T extends PersonDescriptorVO> List<T> getWithServiceActorByPersonUUIDAndDate(String personUuid, Date date) throws DAOException {
        try {
            final String tableName = getTableName();
            String sql = "SELECT t.*, p.PersonUUID," + TIMESTAMP_AND_ACTOR_FIELDS + " FROM Person p, ServiceActor a, "
                    + tableName + " t WHERE p.PersonUUID = :personUuid" +
                    " AND t.PersonPID = p.PersonPID AND t.ValidFrom <= :date AND t.ValidTo > :date AND p.ServiceActorPID = a.ServiceActorPID";

            final MapSqlParameterSource parameterMap = new MapSqlParameterSource()
                    .addValue("personUuid", personUuid)
                    .addValue("date", date);
            return queryForList(sql, parameterMap);
        } catch (Exception e) {
            throw new DAOException("An error occurred loading " + getTableName() + " by UUID and date", e);
        }
    }

    protected void updateValidToTimestamp(String idColumnName, List<Integer> idList, Date timestamp) throws DAOException {
        if (idList.isEmpty()) {
            return;
        }

        try {
            final String tableName = getTableName();

            String sql = "UPDATE " + tableName + " SET ValidTo = :ValidTo, ModifiedTime = :ModifiedTime";
            sql += " WHERE " + idColumnName + " IN (:Ids)";
            final MapSqlParameterSource parameters = new MapSqlParameterSource();
            parameters.addValue("ValidTo", timestamp);
            parameters.addValue("ModifiedTime", timestamp);
            parameters.addValue("Ids", idList);
            executeUpdate(sql, parameters);
        } catch (Exception e) {
            throw new DAOException("An error occurred updating valid-to timestamps", e);
        }
    }

    @Override
    public void deleteReservedIds(List<Integer> idList) throws DAOException {
        deleteReservedPersonIds(idList);
    }

    protected void deleteReservedPersonIds(List<Integer> idList) throws DAOException {
        if (idList.isEmpty()) {
            return;
        }
        try {
            final String tableName = getTableName();
            String sql = "DELETE FROM " + tableName + " WHERE personPID" + " IN (:Ids)";
            final MapSqlParameterSource parameterMap = new MapSqlParameterSource();
            parameterMap.addValue("Ids", idList);

            executeUpdate(sql, parameterMap);

        } catch (Exception e) {
            throw new DAOException("An error occurred updating valid-to timestamps", e);
        }
    }

    protected void assignFieldsForNew(PersonDescriptorVO vo, Date nowTimestamp, int personId) {
        if (vo.getUuid() == null) {
            vo.setUuid(UUID.randomUUID().toString());
        }
        if (personId != 0) {
            vo.setPersonPID(personId);
        }
        if (nowTimestamp != null) {
            vo.setValidFrom(nowTimestamp);
            vo.setModified(nowTimestamp);
        }
        if (vo.getValidTo() == null) {
            vo.setValidTo(AbstractECPRDAOImpl.VALID_TO_DEFAULT);
        }
    }

    protected Integer getNewestRecordForUUID(String uuid) throws DAOException {
        final String sql = "SELECT " + getIdColumnName() + " FROM " + getTableName() + " WHERE Uuid = :uuid ORDER BY ValidFrom DESC LIMIT 1";
        final MapSqlParameterSource parameters = new MapSqlParameterSource().addValue("uuid", uuid);
        List<Integer> result = queryForList(sql, parameters, new PIDRowMapper(getIdColumnName()));
        if (result.size() != 1) {
            throw new DAOException("Expected single result for latest record from " + getTableName() + " and uuid " + uuid);
        }
        return result.get(0);
    }

    @Override
    public void delete(PersonDescriptorVO vo, Date timestamp) throws DAOException {
        Integer pidToUpdate = getNewestRecordForUUID(vo.getUuid());
        updateValidToTimestamp(getIdColumnName(), Collections.singletonList(pidToUpdate), timestamp);
    }

    public void updateValidToTimestamp(String uuid, Date timestamp) throws DAOException {
        Integer pidToUpdate = getNewestRecordForUUID(uuid);
        updateValidToTimestamp(getIdColumnName(), Collections.singletonList(pidToUpdate), timestamp);
    }
}
