/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl.abstr;

import dk.sds.ecpr2.constants.SearchConstants;
import dk.sds.ecpr2.dao.impl.SearchCrossTableDAOImpl;
import dk.sds.ecpr2.domain.PersonVO;
import lombok.extern.log4j.Log4j2;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.StringField;
import org.apache.lucene.document.TextField;
import org.apache.lucene.index.DirectoryReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.queryparser.classic.QueryParser;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.similarities.BM25Similarity;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;

import java.io.IOException;
import java.time.Instant;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Timer;
import java.util.TimerTask;

@Log4j2
public abstract class AbstractSearchDAOImpl {
    protected static final String PERSONUUID = "personuuid";
    protected static final String BIRTHDATE = "birthdate";

    @Autowired
    private SearchCrossTableDAOImpl dao;

    @Autowired
    @Qualifier("FullIndexWriter")
    ECPRIndexWriter fullIndexWriter;
    @Autowired
    @Qualifier("OnlyCurrentIndexWriter")
    ECPRIndexWriter onlyCurrentIndexWriter;

    private IndexSearcher fullIndexSearcher;
    private DirectoryReader fullIndexDirectoryReader;
    private IndexSearcher onlyCurrentIndexSearcher;
    private DirectoryReader onlyCurrentDirectoryReader;

    protected void addIndexedField(Document document, String fieldName, String value) {
        if (value != null && value.length() > 0) {
            if (PERSONUUID.equals(fieldName) || BIRTHDATE.equals(fieldName)) {
                document.add(new StringField(fieldName, value, Field.Store.YES));
            } else {
                document.add(new TextField(fieldName, value, Field.Store.YES));
            }
        }
    }

    protected Query parseQuery(String queryStr, String fieldName) throws Exception {
        StringBuilder stringBuilder = new StringBuilder();
        StringTokenizer tokenizer = new StringTokenizer(queryStr, " +-*&?|!(){}[]/#;.,¤><@£$§¶€'%\\\"^~:");
        int originalTokenCount = tokenizer.countTokens();
        int count = 0;

        while (tokenizer.hasMoreTokens()) {
            String token = tokenizer.nextToken();

            if (token.length() > 0) {
                if (count == 0) {
                    if (SearchConstants.canBeWildCard(fieldName)) {
                        stringBuilder.append("*");
                    }
                    count++;
                }
                stringBuilder.append(token);
                if (SearchConstants.canBeWildCard(fieldName)) {
                    stringBuilder.append("*");
                }
            }
        }
        // If the string had to be split, then we need to wrap the string in quotes, otherwise it will not find results with special characters
        if (originalTokenCount > 1 && !BIRTHDATE.equals(fieldName)) {
            stringBuilder.insert(0, "\"");
            stringBuilder.insert(stringBuilder.length(), "\"");
        }
        QueryParser queryParser = new QueryParser(fieldName, new StandardAnalyzer());
        queryParser.setAllowLeadingWildcard(true);
        queryParser.setDefaultOperator(QueryParser.Operator.AND);
        return queryParser.parse(stringBuilder.toString());
    }

    protected IndexWriter getFullIndexWriter() throws IOException {
        return fullIndexWriter.getIndexWriter();
    }

    protected IndexWriter getOnlyCurrentIndexWriter() throws IOException {
        return onlyCurrentIndexWriter.getIndexWriter();
    }

    protected IndexSearcher getFullIndexSearcher() throws IOException {
        // We don't want to enter a synchronized block unless necessary
        if (fullIndexSearcher == null) {
            synchronized (this) {
                // Once in the synchronized block - did another thread already allocate an IndexSearcher?
                if (fullIndexSearcher == null) {
                    this.fullIndexDirectoryReader = DirectoryReader.open(fullIndexWriter.getDirectory());

                    try {
                        fullIndexSearcher = new IndexSearcher(this.fullIndexDirectoryReader);
                        fullIndexSearcher.setSimilarity(new BM25Similarity() {
                        });
                    } finally {
                        //reader.close();
                    }
                }

            }
        } else {
            refreshFullReader();
        }

        return fullIndexSearcher;
    }

    private void refreshFullReader() throws IOException {
        DirectoryReader newReader = DirectoryReader.openIfChanged(this.fullIndexDirectoryReader);
        if (newReader != null && newReader != this.fullIndexDirectoryReader) {
            this.fullIndexDirectoryReader.close();
            this.fullIndexDirectoryReader = newReader;
            try {
                this.fullIndexSearcher = new IndexSearcher(this.fullIndexDirectoryReader);
                this.fullIndexSearcher.setSimilarity(new BM25Similarity() {
                });
            } finally {
                //reader.close();
            }
        }
    }


    protected IndexSearcher getOnlyCurrentIndexSearcher() throws IOException {
        // We don't want to enter a synchronized block unless necessary
        if (onlyCurrentIndexSearcher == null) {
            synchronized (this) {
                // Once in the synchronized block - did another thread already allocate an IndexSearcher?
                if (onlyCurrentIndexSearcher == null) {
                    this.onlyCurrentDirectoryReader = DirectoryReader.open(onlyCurrentIndexWriter.getDirectory());
                    onlyCurrentIndexSearcher = new IndexSearcher(this.onlyCurrentDirectoryReader);
                    onlyCurrentIndexSearcher.setSimilarity(new BM25Similarity() {
                    });
                }
            }
        } else {
            refreshCurrentReader();
        }

        return onlyCurrentIndexSearcher;
    }

    private void refreshCurrentReader() throws IOException {
        DirectoryReader newReader = DirectoryReader.openIfChanged(this.onlyCurrentDirectoryReader);
        if (newReader != null && newReader != this.onlyCurrentDirectoryReader) {
            this.onlyCurrentDirectoryReader.close();
            this.onlyCurrentDirectoryReader = newReader;
            this.onlyCurrentIndexSearcher = new IndexSearcher(this.onlyCurrentDirectoryReader);
            this.onlyCurrentIndexSearcher.setSimilarity(new BM25Similarity() {
            });
        }
    }

    public void buildIndex(Instant date) throws Exception {
        try {
            buildIndex(date, getFullIndexWriter(), false);
            buildIndex(date, getOnlyCurrentIndexWriter(), true);
        } finally {
            final IndexSearcher oldFullIndexSearcher = fullIndexSearcher;
            fullIndexSearcher = null;
            final IndexSearcher oldOnlyCurrentIndexSearcher = onlyCurrentIndexSearcher;
            onlyCurrentIndexSearcher = null;
            if (oldFullIndexSearcher != null || oldOnlyCurrentIndexSearcher != null) {
                // New searches will generate a new instance of IndexSearcher.
                // Give active searches a chance to finalize before closing the old IndexSearcher.
                Timer timer = new Timer();
                timer.schedule(new TimerTask() {
                    @Override
                    public void run() {
                        try {
                            if (oldFullIndexSearcher != null) {
                                oldFullIndexSearcher.getIndexReader().close();
                            }
                            if (oldOnlyCurrentIndexSearcher != null) {
                                oldOnlyCurrentIndexSearcher.getIndexReader().close();
                            }
                        } catch (Exception e) {
                            log.error("Error while closing old IndexSearcher", e);
                        }
                    }
                }, 20000);
            }
        }
    }

    protected abstract int buildIndex(Instant date, IndexWriter writer, boolean excludeDuplicatePersons) throws Exception;

    public abstract List<PersonVO> searchPerson(List<String> searchStr) throws Exception;

    public abstract void addPersonToIndex(List<PersonVO> personUUID) throws Exception;

    public abstract void updatePersonsInIndex(List<PersonVO> personUUID) throws Exception;

    public abstract void deleteIndex(List<PersonVO> personVO) throws Exception;

    public void setCrossTableDAO(SearchCrossTableDAOImpl crossTableDAO) {
        this.dao = crossTableDAO;
    }

    protected SearchCrossTableDAOImpl getDAO() {
        return dao;
    }

}






