/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.datagenerator;

import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.abstr.AbstractDAOImpl;
import lombok.AllArgsConstructor;
import lombok.Getter;
import lombok.Setter;
import org.springframework.stereotype.Component;

import java.time.LocalDate;
import java.time.ZoneId;
import java.util.Date;
import java.util.Map;
import java.util.Random;

@Component
public abstract class AbstractGenerator {

    @AllArgsConstructor
    class IdRange {

        int min;
        int max;
    }

    private Random rnd = new Random();

    protected abstract AbstractDAOImpl getDao();

    protected abstract int getNumberOfRowsToGenerate();

    protected int getMinimumIdToRemove() {
        return 0;
    }

    protected String getTableName() {
        return getDao().getTableName();
    }

    protected abstract IdRange generate(Map<String, IdRange> previouslyGeneratedIds) throws DAOException;

    protected void removeRows() throws DAOException {
        getDao().deleteRowsById(getMinimumIdToRemove(), Integer.MAX_VALUE);
    }

    protected int rowCount() throws DAOException {
        return getDao().rowCount();
    }

    protected int getRandomIntBetween(IdRange range) {
        return getRandomIntBetween(range.min, range.max);
    }

    protected int getRandomIntBetween(int min, int max) {
        return rnd.nextInt(max - min + 1) + min;
    }

    protected int getSkewedRandomIntBetween(int min, int max, int mostFrequent) {
        int random = getRandomIntBetween(min, max);
        int skew = (int) Math.round(rnd.nextGaussian() * 2 * (max - mostFrequent));
        return Math.min(random + skew, max);
    }

    protected char getRandomChar() {
        final String alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
        return alphabet.charAt(rnd.nextInt(alphabet.length()));
    }

    protected Date getRandomDateBetween(int minYear, int maxYear) {
        return Date.from(getRandomLocalDateBetween(minYear, maxYear).atStartOfDay(ZoneId.systemDefault()).toInstant());
    }

    protected LocalDate getRandomLocalDateBetween(int minYear, int maxYear) {
        int day = getRandomIntBetween(1, 28);
        int month = getRandomIntBetween(1, 12);
        int year = getRandomIntBetween(minYear, maxYear);
        return LocalDate.of(year, month, day);
    }

    protected int getRandomRowCount(double average, double standardDeviation) {
        final double randomValueWithAverageZero = rnd.nextGaussian() * standardDeviation;
        final double randomValueWithCorrectAverage = randomValueWithAverageZero + average;
        final int randomIntWithSameAverage = (int) Math.rint(randomValueWithCorrectAverage);
        return randomIntWithSameAverage;
    }

}
