/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.datagenerator;

import dk.sds.ecpr2.*;
import dk.sds.ecpr2.util.ValueSets;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import java.io.StringWriter;
import java.util.List;

@Component
public class DataFactory {

    @Autowired
    IdentifierGenerator identifierGenerator;

    public static String asString(Object pObject) throws JAXBException {
        JAXBContext pContext = JAXBContext.newInstance(pObject.getClass());
        java.io.StringWriter sw = new StringWriter();
        Marshaller marshaller = pContext.createMarshaller();
        marshaller.setProperty(Marshaller.JAXB_ENCODING, "UTF-8");
        marshaller.marshal(pObject, sw);
        return sw.toString();
    }

    public PersonType createFullPerson2() {
        PersonType person = new PersonType();
        person.getIdentifiers().add(createXECPRIdentifier());
        person.getAddresses().add(createAddress("Gadeveh 42", "Aarhus", "8000", "DK"));
        person.getAddresses().add(createAddress("Gadeveh 42", "Aarhus", "8000", "DK"));
        person.getBirthDates().add(createBirthDate("1980"));
        person.getContacts().add(createContact("88888888"));
        person.getContacts().add(createContact("88888888"));
        person.getNames().add(createName("Hans", "Hansen", ValueSets.nameUseValues[0]));
        person.getNames().add(createName("Hans", "Hansen", ValueSets.nameUseValues[1]));
        person.getGenders().add(createGender(ValueSets.genderValues[0]));
        person.getModifieds().add(createModified("12345", "Jens Jensen", "1111111118", "CPR", "Læge", "Hospitalsvej 1", "hospital@hospital.dk", "Hospitalets hospital", "12345678", "hospital", "12345678901", "SOR"));
        return person;
    }

    public PersonType createPerson(String givenName, String familyName, String streetAddressLine, String city, String postalCode, String birthDate, String countryCode) {
        return createPersonDetails(ValueSets.nameUseValues[0], givenName, familyName, streetAddressLine, city, postalCode, birthDate, countryCode, "88888888",
                ValueSets.genderValues[0], "12345", "Jens Jensen", "1111111118", "CPR", "Læge", "Hospitalsvej 1", "hospital@hospital.dk", "Hospitalets hospital", "12345678", "hospital", "12345678901", "SOR");
    }

    public PersonType createPersonDetails(String nameUseValue, String givenName, String familyName, String streetAddressLine, String city, String postalCode, String birthDate, String countryCode, String contactNumber, String genderValue, String authId, String name, String personIdentifier, String personIdentifierSource, String role, String addressLine, String email, String orgName, String telephone, String orgType, String organisationIdentifier, String organisationIdentifierSource) {
        PersonType person = new PersonType();
        person.getIdentifiers().add(createXECPRIdentifier());
        person.getAddresses().add(createAddress(streetAddressLine, city, postalCode, countryCode));
        person.getBirthDates().add(createBirthDate(birthDate));
        person.getContacts().add(createContact(contactNumber));
        person.getNames().add(createName(givenName, familyName, nameUseValue));
        person.getGenders().add(createGender(genderValue));
        person.getModifieds().add(createModified(authId, name, personIdentifier, personIdentifierSource, role, addressLine, email, orgName, telephone, orgType, organisationIdentifier, organisationIdentifierSource));
        return person;
    }

    public Identifier createCprIdentifier(String cpr) {
        Identifier identifier = new Identifier();
        identifier.setValidity(0);
        identifier.setOID(OID.CPR);
        identifier.setValue(cpr);
        return identifier;
    }

    public Identifier createXECPRIdentifier() {
        Identifier identifier = new Identifier();
        identifier.setValidity(0);
        identifier.setOID(OID.X_ECPR);
        identifier.setValue(null); // To create C-eCPR, just don't supply a value
        return identifier;
    }

    public Identifier createMigrantIdentifier() {
        Identifier identifier = new Identifier();
        identifier.setValidity(0);
        identifier.setOID(OID.MIGRANT_IDENTIFIER);
        identifier.setValue("12345678");
        return identifier;
    }

    public Identifier createRegionIdentifier() {
        Identifier identifier = new Identifier();
        identifier.setValidity(0);
        identifier.setOID(OID.ECPR_REGION_SOUTH);
        identifier.setValue(identifierGenerator.getRandomXECPR());
        return identifier;
    }

    public AddressType createAddress(String streetAddressLine, String city, String postalCode, String countryCode) {
        AddressType address = new AddressType();
        address.setCity(city);
        address.setCountry(countryCode);
        address.setDistrict("Dummy District");
        address.setPostalCode(postalCode);
        address.setPostBox("Postbox 1000");
        address.setState("Dummy State");
        address.getTexts().add(streetAddressLine);
        address.setType(ValueSets.addressTypeValues[0]);
        address.setUse(ValueSets.addressUseValues[0]);
        return address;
    }

    public void populatePersonType(PersonType person) {
        person.getIdentifiers().add(createXECPRIdentifier());
        person.getIdentifiers().add(createRegionIdentifier());
        person.getAddresses().add(createAddress("Gadevej 42", "Aarhus", "8000", "DK"));
        person.getBirthDates().add(createBirthDate("1980"));
        person.getContacts().add(createContact("88888888"));
        person.getNames().add(createName("Hans", "Hansen", ValueSets.nameUseValues[0]));
        person.getGenders().add(createGender(ValueSets.genderValues[0]));
        person.getModifieds().add(createModified("12345", "Jens Jensen", "1111111118", "CPR", "Læge", "Hospitalsvej 1", "hospital@hospital.dk", "Hospitalets hospital", "12345678", "hospital", "12345678901", "SOR"));
    }

    public PersonType createMinimalPerson(String givenName, String familyName, String birthDate) {
        PersonType person = new PersonType();
        person.getBirthDates().add(createBirthDate(birthDate));
        person.getNames().add(createName(givenName, familyName, ValueSets.nameUseValues[0]));
        person.getGenders().add(createGender(ValueSets.genderValues[0]));
        person.getModifieds().add(createModified("12345", "Jens Jensen", "1111111118", "CPR", "Læge", "Hospitalsvej 1", "hospital@hospital.dk", "Hospitalets hospital", "12345678", "hospital", "12345678901", "SOR"));
        Identifier ceCPR = new Identifier();
        ceCPR.setOID(OID.C_ECPR);
        person.getIdentifiers().add(ceCPR);
        return person;
    }

    public Modified createModified() {
        return createModified("12345", "Jens Jensen", "1111111118", "CPR", "Læge", "Hospitalsvej 1", "hospital@hospital.dk", "Hospitalets hospital", "12345678", "hospital", "12345678901", "SOR");
    }

    public Modified createModified(String authId, String name, String personIdentifier, String personIdentifierSource, String role, String addressLine, String email, String orgName, String telephone, String orgType, String organisationIdentifier, String organisationIdentifierSource) {
        Modified modified = new Modified();
        ModifiedBy modifiedBy = new ModifiedBy();
        ModificatorPersonType person = new ModificatorPersonType();
        person.setAuthorisationIdentifier(authId);
        person.setName(name);
        PersonIdentifierType personId = new PersonIdentifierType();
        personId.setPersonIdentifier(personIdentifier);
        personId.setSource(personIdentifierSource);
        person.setPersonIdentifier(personId);
        modifiedBy.setPerson(person);
        modifiedBy.setRole(role);
        OrganisationType organisation = new OrganisationType();
        organisation.setAddressLine(addressLine);
        organisation.setEmailAddress(email);
        organisation.setName(orgName);
        organisation.setTelephoneNumber(telephone);
        organisation.setType(orgType);
        OrganisationIdentifierType organisationId = new OrganisationIdentifierType();
        organisationId.setIdentifier(organisationIdentifier);
        organisationId.setSource(organisationIdentifierSource);
        organisation.setIdentifier(organisationId);
        modifiedBy.setOrganisation(organisation);
        modified.setBy(modifiedBy);
        return modified;
    }

    public GenderType createGender(String genderValue) {
        GenderType gender = new GenderType();
        gender.setValue(genderValue);
        return gender;
    }

    public NameType createName(String givenName, String familyName, String use) {
        NameType name = new NameType();
        name.setFamilyName(familyName);
        name.setGivenName(givenName);
        name.setUse(use);
        return name;
    }

    public ContactType createContact(String value) {
        ContactType contact = new ContactType();
        contact.setSystem(ValueSets.contactSystemValues[0]);
        contact.setUse(ValueSets.contactUseValues[0]);
        contact.setValue(value);
        return contact;
    }

    public void clearValidity(List<? extends EntityType> list) {
        list.forEach(o -> o.setValidFrom(null));
        list.forEach(o -> o.setValidTo(null));
    }

    public void clearAllExceptKeyIdentifier(PersonType personForUpdate) {
        personForUpdate.getIdentifiers().clear();
        personForUpdate.getBirthDates().clear();
        personForUpdate.getContacts().clear();
        personForUpdate.getGenders().clear();
        personForUpdate.getNames().clear();
        personForUpdate.getAddresses().clear();
    }

    public BirthDateType createBirthDate(String value) {
        BirthDateType birthDate = new BirthDateType();
        birthDate.setValue(value);
        return birthDate;
    }

    public KeyIdentifier createKeyIdentifier(String oid, String value) {
        KeyIdentifier ki = new KeyIdentifier();
        ki.setValue(value);
        ki.setOID(oid);
        return ki;
    }

    public PersonWithKeyIdentifierType createPersonForUpdate(PersonType createdPerson) {
        String oid = createdPerson.getIdentifiers().get(0).getOID();
        String value = createdPerson.getIdentifiers().get(0).getValue();
        PersonWithKeyIdentifierType person = new PersonWithKeyIdentifierType();
        person.setKeyIdentifier(createKeyIdentifier(oid, value));
        person.getAddresses().addAll(createdPerson.getAddresses());
        person.getBirthDates().addAll(createdPerson.getBirthDates());
        person.getNames().addAll(createdPerson.getNames());
        person.getContacts().addAll(createdPerson.getContacts());
        person.getGenders().addAll(createdPerson.getGenders());
        person.getIdentifiers().addAll(createdPerson.getIdentifiers());
        person.getModifieds().addAll(createdPerson.getModifieds());
        clearValidity(person.getAddresses());
        clearValidity(person.getBirthDates());
        clearValidity(person.getNames());
        clearValidity(person.getContacts());
        clearValidity(person.getGenders());
        clearValidity(person.getIdentifiers());
        person.getIdentifiers().forEach(i -> i.setPrimaryOIDLabel(null));
        person.getNames().forEach(n -> n.setText(null));
        person.getAddresses().forEach(a -> a.setCountryName(null));
        person.getModifieds().forEach(v -> v.setValidFrom(null));
        person.getModifieds().forEach(v -> v.setValidTo(null));
        person.getIdentifiers().remove(0); // C-eCPR. Must not exist in update
        return person;
    }

}
