/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.datagenerator;

import dk.nsi.fmk.admin.Admin;
import dk.nsi.fmk.advis.AdvisService;
import dk.sds.ecpr2.*;
import dk.sds.ecpr2._2017._10._01.CreatePersonRequest;
import dk.sds.ecpr2._2017._10._01.CreatePersonResponse;
import dk.sds.ecpr2._2017._10._01.UpdatePersonRequest;
import dk.sds.ecpr2._2017._10._01.UpdatePersonResponse;
import dk.sds.ecpr2.advis.AdvisModuleClientFactory;
import dk.sds.ecpr2.advis.AsyncAdvisFacade;
import dk.sds.ecpr2.advis.AsyncAdvisFacadeModuleImpl;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.service.admin.AdministrationFacade;
import dk.sds.ecpr2.service.admin.impl.AdminModuleAdministrationFacadeImpl;
import dk.sds.ecpr2.service.admin.impl.AdminModuleClientFactory;
import dk.sds.ecpr2.service.admin.impl.AdministrationFacadeWrapper;
import dk.sds.ecpr2.util.Properties;
import dk.sds.ecpr2.util.ValueSets;
import dk.sds.ecpr2.webservice.ECPR2WebServiceImpl;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.CommandLineRunner;
import org.springframework.boot.autoconfigure.EnableAutoConfiguration;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.ComponentScan;
import org.springframework.stereotype.Component;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

@Component
@ComponentScan(basePackages = {"dk.sds.ecpr2.dao", "dk.sds.ecpr2.webservice", "dk.sds.ecpr2.util", "dk.sds.ecpr2.service", "dk.sds.ecpr2.datagenerator"})
@EnableAutoConfiguration
public class DataGeneratorApplication implements CommandLineRunner {

    public static final double ACCEPTABLE_NUMBER_OF_ROWS_DIFFERENCE_IN_PERCENT = 5.0;

    @Autowired
    DataFactory factory;

    @Autowired
    protected ECPR2WebServiceImpl service;

    @Autowired
    private List<AbstractGenerator> generatorsInReferentialOrder;

    private Map<String, AbstractGenerator.IdRange> generatedIdRanges = new HashMap<>();

    private double startTime;

    @Override
    public void run(String... args) throws DAOException {
        System.out.println("Running data generator application.");

        if (!allTablesAlreadyHaveAcceptableRowCount()) {
            removeAllData();
            generateData();
        }
    }

    public void generateData() throws DAOException {
        printAndTime("Starting data generation.");

        for (AbstractGenerator generator : generatorsInReferentialOrder) {
            AbstractGenerator.IdRange generatedIds = generator.generate(generatedIdRanges);
            generatedIdRanges.put(generator.getTableName(), generatedIds);
            printAndTime(generator.getNumberOfRowsToGenerate() + " " + generator.getTableName() + " rows generated.");
        }
        createRealisticPerson();
        printAndTime("Successfully completed data generation.");
    }

    private void createRealisticPerson() {
        try {
            PersonWithKeyIdentifierType personForUpdate = setupPersonForUpdate();

            AddressType addedEntity = createAddress("Sommerhusvej 44", "Thisted", "7620", "DK");
            addedEntity.setType(ValueSets.addressTypeValues[0]);
            addedEntity.setUse(ValueSets.addressUseValues[2]);
            personForUpdate.getAddresses().add(addedEntity);
            UpdatePersonResponse updatedPerson = callUpdatePerson(personForUpdate);

            // Add name
            personForUpdate = factory.createPersonForUpdate(updatedPerson.getPerson());
            factory.clearAllExceptKeyIdentifier(personForUpdate);
            NameType addedName = factory.createName("Hans", "Thomsen", ValueSets.nameUseValues[1]);
            personForUpdate.getNames().add(addedName);
            updatedPerson = callUpdatePerson(personForUpdate);

            // Add address
            personForUpdate = factory.createPersonForUpdate(updatedPerson.getPerson());
            factory.clearAllExceptKeyIdentifier(personForUpdate);
            AddressType addedAddress = createAddress("Strassenbahn 46", "Dortmund", "12550", "DE");
            personForUpdate.getAddresses().add(addedAddress);
            updatedPerson = callUpdatePerson(personForUpdate);

            // Update address
            AddressType addr = updatedPerson.getPerson().getAddresses().stream().filter(a -> a.getCountry().equals("DE")).collect(Collectors.toList()).get(0);
            personForUpdate = factory.createPersonForUpdate(updatedPerson.getPerson());
            factory.clearAllExceptKeyIdentifier(personForUpdate);
            AddressType updatedAddress = createAddress("Strassenbahn 146", "Dortmund", "12550", "DE");
            updatedAddress.setPID(addr.getPID());
            personForUpdate.getAddresses().add(updatedAddress);
            updatedPerson = callUpdatePerson(personForUpdate);

            // Add contact
            personForUpdate = factory.createPersonForUpdate(updatedPerson.getPerson());
            factory.clearAllExceptKeyIdentifier(personForUpdate);
            ContactType addedContact = factory.createContact("myemail@company.com");
            addedContact.setUse(ValueSets.contactUseValues[2]);
            addedContact.setSystem(ValueSets.contactSystemValues[1]);
            personForUpdate.getContacts().add(addedContact);
            updatedPerson = callUpdatePerson(personForUpdate);

            // Add identifier
            personForUpdate = factory.createPersonForUpdate(updatedPerson.getPerson());
            factory.clearAllExceptKeyIdentifier(personForUpdate);
            Identifier addedIdentifier = factory.createMigrantIdentifier();
            personForUpdate.getIdentifiers().add(addedIdentifier);
            updatedPerson = callUpdatePerson(personForUpdate);

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public AddressType createAddress(String streetAddressLine, String city, String postalCode, String countryCode) {
        AddressType address = new AddressType();
        address.setCity(city);
        address.setCountry(countryCode);
        address.setPostalCode(postalCode);
        address.getTexts().add(streetAddressLine);
        address.setType(ValueSets.addressTypeValues[0]);
        address.setUse(ValueSets.addressUseValues[0]);
        return address;
    }


    private UpdatePersonResponse callUpdatePerson(PersonWithKeyIdentifierType personForUpdate) throws Exception {
        personForUpdate.getModifieds().clear();
        personForUpdate.getModifieds().add(factory.createModified("12345", "Jens Jensen", "1111111118", "CPR", "Læge", "Hospitalsvej 1", "hospital@hospital.dk", "Hospitalets hospital", "12345678", "hospital", "12345678901", "SOR"));
        UpdatePersonRequest request = new UpdatePersonRequest();
        request.setPerson(personForUpdate);
        Thread.sleep(1000);
        System.out.println("=================================================== UPDATE REQUEST ==================================================");
        System.out.println(factory.asString(request));
        System.out.println("=====================================================================================================================");
        return service.updatePerson(request);
    }

    private PersonWithKeyIdentifierType setupPersonForUpdate() throws Exception {
        PersonType newPerson = createPerson();
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(newPerson);
        factory.clearAllExceptKeyIdentifier(personForUpdate);
        return personForUpdate;
    }

    protected PersonType createPerson() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        PersonType person = new PersonType();
        factory.populatePersonType(person);
        person.getAddresses().get(0).setDistrict(null);
        person.getAddresses().get(0).setPostBox(null);
        person.getAddresses().get(0).setState(null);
        request.setPerson(person);
        CreatePersonResponse response = service.createPerson(request);
        return response.getPerson();
    }

    public void removeAllData() throws DAOException {
        printAndTime("Deleting data.");
        for (int i = generatorsInReferentialOrder.size() - 1; i >= 0; i--) {
            AbstractGenerator generator = generatorsInReferentialOrder.get(i);
            generator.removeRows();
            printAndTime("Removed data from " + generator.getTableName() + " table starting from id " + generator.getMinimumIdToRemove());
        }
        printAndTime("Successfully completed deletion.");
    }

    private boolean allTablesAlreadyHaveAcceptableRowCount() throws DAOException {
        boolean amountOfDataIsOk = true;
        printAndTime("Checking current amount of data.");

        for (AbstractGenerator generator : generatorsInReferentialOrder) {
            int actualRows = generator.rowCount();
            int neededRows = generator.getNumberOfRowsToGenerate();
            double percentOff = 100.0 * Math.abs(actualRows - neededRows) / neededRows;
            if (percentOff > ACCEPTABLE_NUMBER_OF_ROWS_DIFFERENCE_IN_PERCENT) {
                printAndTime("Wrong amount of data in the " + generator.getTableName() + "table. Found: " + actualRows + " Expected: " + neededRows + " (" + percentOff + "% off)");
                amountOfDataIsOk = false;
                break;
            } else {
                printAndTime("Table " + generator.getTableName() + " had " + actualRows + ". Expected " + neededRows + " (" + percentOff + "% off) OK");
            }

        }

        printAndTime("Done checking. Amount of data was " + (amountOfDataIsOk ? "ok." : "not ok."));
        return amountOfDataIsOk;
    }

    protected void printAndTime(String message) {
        if (startTime > 0) {
            final double durationMillis = (System.nanoTime() - startTime) / 1000000;
            String durationText = durationMillis > 1000
                    ? Math.round(durationMillis / 1000) + " s"
                    : Math.round(durationMillis) + " ms";
            System.out.println(message + "  (Time since last log: " + durationText + ")");
        } else {
            System.out.println(message);
        }
        startTime = System.nanoTime();
    }

    @Bean
    public AsyncAdvisFacade asyncAdvisFacade(AdvisService advisService) {
        return new AsyncAdvisFacadeModuleImpl(advisService);
    }

    @Bean
    public AdvisService advisService(Properties properties) {
        return new AdvisModuleClientFactory(properties).getAdvisServiceClient();
    }

    @Bean
    public AdminModuleClientFactory adminModuleClientFactory() {
        return new AdminModuleClientFactory(false);
    }

    @Bean
    public Admin adminClient(AdminModuleClientFactory adminModuleClientFactory) {
        return adminModuleClientFactory.getAdminServiceClient();
    }

    @Bean
    public AdministrationFacade adminModuleAdministrationFacade() {
        return new AdminModuleAdministrationFacadeImpl();
    }

    @Bean
    public AdministrationFacade administrationFacade() {
        return new AdministrationFacadeWrapper();
    }

}
