/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.security;

import dk.sds.ecpr2.dao.StamdataFacade;
import dk.sds.ecpr2.service.admin.impl.Role;
import dk.sds.ecpr2.util.TimeHelper;
import lombok.extern.log4j.Log4j2;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Arrays;
import java.util.List;

@Component
@Log4j2
public class AssistantAuthenticator implements Authenticator {
    @Autowired StamdataFacade stamdataFacade;

    private boolean delegationCheck = true;

    /**
     * Delegation isn't properly set up for test of eCPR
     */
    public void skipDelegationCheckInTest() {
        delegationCheck = false;
    }

    @Override
    public boolean authenticate(SecurityCredentials sc) {
        // Rules:
        // - Must be a MOCES certificate, eg. there has to be a cvr number
        // - Must specify a valid authorisation id in on behalf of
        // - Must have a bemyndigelse

        List<Permission> bemyndigetPermissions = null;
        // MOCES
        if (sc.getCvr() == null || sc.getCvr().trim().length() == 0 || (sc.getIdCardType() != null && !"user".equals(sc.getIdCardType()))) {
            log.debug("AssistantAuthenticator failed to authenticate: MOCES");
            return false;
        }

        if (sc.getOnBehalfOfAuthorisationNo() == null || sc.getOnBehalfOfAuthorisationNo().trim().length() == 0) {
            log.debug("AssistantAuthenticator failed to authenticate: getOnBehalfOfAuthorisationNo is null or empty");
            return false;
        }

        AuthorisationVO authorisation = stamdataFacade.getAuthorisation(sc.getOnBehalfOfAuthorisationNo());
        if (authorisation == null) {
            log.debug("AssistantAuthenticator failed to authenticate: Could not find authorisation");
            return false;
        }

        // Is authorised person still alive?
        if(!assistantHasAccessOnBehalfOfDeceasedPerson(authorisation.getPersonIdentifier())) {
            log.debug("AssistantAuthenticator failed to authenticate: authorized person is deceased");
            return false;
        }

        Role authorisingRole = Role.roleForEducationCode(authorisation.getEducationCode());
        if (authorisingRole == null) {
            log.debug("AssistantAuthenticator failed to authenticate: Could not find role based on education code");
            return false;
        }

        // Try using bemyndigelse from NSP
        if(delegationCheck) {
            bemyndigetPermissions = stamdataFacade.getBemyndigetPermissions(sc.getUserCpr(), authorisation.getPersonIdentifier(), sc.getCvr(), authorisingRole.getSchemaName());
            if (bemyndigetPermissions == null || bemyndigetPermissions.size() == 0) {
                log.debug("AssistantAuthenticator failed to authenticate: Could not find any permissions in BEM");
                // Validated authorisation/bemyndigelse
                return false;
            }
        }

        return true;
    }

    private boolean assistantHasAccessOnBehalfOfDeceasedPerson(String personIdentifier) {
        boolean isDead = stamdataFacade.isPersonDead(personIdentifier, TimeHelper.getCurrentTime());
        if (isDead) {
            return stamdataFacade.isPersonDead(personIdentifier, TimeHelper.getBeforeCurrentTime(30 * (60*60*24)));
        } else {
            return true;
        }
    }

    @Override
    public List<Permission> getPermissions() {
        return Arrays.asList(Permission.Read, Permission.Write);
    }
}
