/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service;

import dk.nsi._2011._09._23.stamdatacpr.PersonLookupRequestType;
import dk.nsi._2014._03._01.stamdatacpr.PersonLookupResponse;
import dk.sds.ecpr2.KeyIdentifier;
import dk.sds.ecpr2.domain.CprPersonVO;
import dk.sds.ecpr2.util.ECPRException;
import dk.sds.ecpr2.util.FaultCode;
import dk.sds.ecpr2.util.Properties;
import dk.sds.ecpr2.util.SosiUtil;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.xml.bind.*;
import javax.xml.namespace.QName;
import java.io.StringReader;
import java.io.StringWriter;
import java.text.SimpleDateFormat;

@Component
public class CprService {
    private static Logger logger = LogManager.getLogger(CprService.class);

    @Autowired
    SosiUtil sosiUtil;

    @Autowired
    Properties properties;

    SimpleDateFormat birthdateFormat = new SimpleDateFormat("yyyy-MM-dd");

    private static class JaxbObject {
        private final Marshaller marshaller;
        private final Unmarshaller unmarshaller;

        private JaxbObject(Marshaller marshaller, Unmarshaller unmarshaller) {
            this.marshaller = marshaller;
            this.unmarshaller = unmarshaller;
        }

        private static JaxbObject init() {
            try {
                JAXBContext jaxbContext = JAXBContext.newInstance(
                        dk.nsi._2011._09._23.stamdatacpr.PersonLookupRequestType.class,
                        dk.nsi._2014._03._01.stamdatacpr.PersonLookupResponse.class);
                return new JaxbObject(jaxbContext.createMarshaller(), jaxbContext.createUnmarshaller());
            } catch (JAXBException e) {
                throw new RuntimeException(e);
            }
        }
    }

    private static ThreadLocal<JaxbObject> jaxb = ThreadLocal.withInitial(JaxbObject::init);

    public CprPersonVO getCprPerson(KeyIdentifier keyIdentifier) {
        try {
            PersonLookupRequestType request = new PersonLookupRequestType();
            request.setCivilRegistrationNumberPersonQuery(keyIdentifier.getValue());
            StringWriter writer = new StringWriter();
            QName qName = new QName("http://nsi.dk/2011/09/23/StamdataCpr/", PersonLookupRequestType.class.getSimpleName());
            jaxb.get().marshaller.marshal(new JAXBElement<>(qName, PersonLookupRequestType.class, request), writer);
            String requestString = writer.toString();
            requestString = requestString.replaceAll("PersonLookupRequestType", "PersonLookupRequest"); // This is a hack, because wsimport seemingly only wants to include PersonLookupRequestType and not PersonLookupRequest
            String responseString = sosiUtil.sendServiceRequestStr(properties.cprUrl, "http://nsi.dk/2011/09/23/#StamdataCpr", requestString);
            responseString = findUniqueElementInXml(responseString, "PersonLookupResponse");
            PersonLookupResponse response = (PersonLookupResponse) jaxb.get().unmarshaller.unmarshal(new StringReader(responseString));
            return mapResponse(response);
        } catch (Exception e) {
            logger.error("Fejl i kald til stamdata", e);
            throw new ECPRException(FaultCode.GET_CPR_PERSON_STAMDATA_ERROR);
        }
    }

    private String findUniqueElementInXml(String xmlString, String elementName) {
        String startTag = ":" + elementName;
        String endTag = ":" + elementName + ">";

        int startPosition = xmlString.indexOf(startTag);
        int endPosition = xmlString.lastIndexOf(endTag);
        if (endPosition == -1) {
            endPosition = xmlString.indexOf("/>", startPosition) - endTag.length() + "/>".length();
        }
        startPosition = xmlString.lastIndexOf("<", startPosition);
        endPosition += endTag.length();
        xmlString = xmlString.substring(startPosition, endPosition);
        return xmlString;
    }

    private CprPersonVO mapResponse(PersonLookupResponse response) {
        if (response.getPersonInformationStructures().size() >= 1) {
            CprPersonVO cprPersonVO = new CprPersonVO();
            cprPersonVO.setCpr(response.getPersonInformationStructures().get(0).getCurrentPersonCivilRegistrationIdentifier());
            cprPersonVO.setGivenName(response.getPersonInformationStructures().get(0).getRegularCPRPerson().getSimpleCPRPerson().getPersonNameStructure().getPersonGivenName());
            cprPersonVO.setMiddleName(response.getPersonInformationStructures().get(0).getRegularCPRPerson().getSimpleCPRPerson().getPersonNameStructure().getPersonMiddleName());
            cprPersonVO.setSurName(response.getPersonInformationStructures().get(0).getRegularCPRPerson().getSimpleCPRPerson().getPersonNameStructure().getPersonSurnameName());
            cprPersonVO.setBirthdate(birthdateFormat.format(response.getPersonInformationStructures().get(0).getRegularCPRPerson().getPersonBirthDateStructure().getBirthDate().toGregorianCalendar().getTime()));
            cprPersonVO.setGender(response.getPersonInformationStructures().get(0).getRegularCPRPerson().getPersonGenderCode().value());
            return cprPersonVO;
        } else {
            return null;
        }
    }
}
