/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service;

import dk.sds.ecpr2.KeyIdentifier;
import dk.sds.ecpr2.Modified;
import dk.sds.ecpr2.OID;
import dk.sds.ecpr2.advis.AsyncAdvisFacade;
import dk.sds.ecpr2.dao.BlacklistingDAO;
import dk.sds.ecpr2.dao.IdentifierDAO;
import dk.sds.ecpr2.dao.OIDDAO;
import dk.sds.ecpr2.dao.OIDFacade;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.*;
import dk.sds.ecpr2.kafka.KafkaService;
import dk.sds.ecpr2.security.Permission;
import dk.sds.ecpr2.security.RequestContext;
import dk.sds.ecpr2.security.RoleValidationInterceptor;
import dk.sds.ecpr2.service.admin.AdministrationFacade;
import dk.sds.ecpr2.service.admin.impl.Role;
import dk.sds.ecpr2.service.handler.*;
import dk.sds.ecpr2.util.FaultCode;
import dk.sds.ecpr2.util.FaultUtil;
import dk.sds.ecpr2.util.SosiUtil;
import dk.sds.ecpr2.webservice.mapping.FromXMLMapper;
import dk.sdsd.dgws._2010._08.PredefinedRequestedRole;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;

@Service
@Log4j2
public class ECPR2ServiceImpl implements ECPR2Service {

    @Autowired
    RoleValidationInterceptor roleValidationInterceptor;

    @Autowired
    protected Validator validator;

    @Autowired
    FromXMLMapper requestMapper;

    @Autowired
    AsyncAdvisFacade advisService;

    @Autowired
    CreatePersonHandler createPersonHandler;

    @Autowired
    UpdatePersonHandler updatePersonHandler;

    @Autowired
    DeletePersonHandler deletePersonHandler;

    @Autowired
    ReservePersonIdHandler reservePersonIdHandler;

    @Autowired
    GetPersonIdHandler getPersonIdHandler;

    @Autowired
    CountPersonHandler countPersonHandler;

    @Autowired
    SearchPersonHandler searchPersonHandler;

    @Autowired
    CreatePersonIdHandler createPersonIdHandler;

    @Autowired
    MergePersonHandler mergePersonHandler;

    @Autowired
    CompareNameHandler compareNameHandler;

    @Autowired
    IdentifierDAO identifierDAO;

    @Autowired
    OIDFacade oidFacade;

    @Autowired
    OIDDAO oidDao;

    @Autowired
    BlacklistingDAO blacklistingDAO;

    @Autowired
    AdministrationFacade administrationFacade;

    @Autowired
    KafkaService kafkaService;

    @Autowired
    SosiUtil sosiUtil;

    @Autowired
    CprService cprService;

    @Override
    public PersonVO createPerson(PersonVO personVO, ServiceActorVO actor, RequestContextVO context) {
        PersonVO result = null;
        try {
            result = createPersonHandler.execute(personVO, actor, context);
        } catch (DAOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.CREATE_PERSON_INTERNAL_SERVER_ERROR, e);
            return null;
        } catch (IOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.CREATE_PERSON_LUCENE_INDEX_INTERNAL_SERVER_ERROR, e);
        }
        return result;
    }

    @Override
    public PersonVO createPersonId(PersonVO personVO, ServiceActorVO actor, RequestContextVO context) {
        PersonVO person = null;
        try {
            person = createPersonIdHandler.execute(personVO, actor, context);
        } catch (DAOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.CREATE_PERSONID_INTERNAL_SERVER_ERROR, e);
        }
        return person;
    }

    @Override
    public PersonVO mergePerson(KeyIdentifierVO from, KeyIdentifierVO to, Modified modified, ServiceActorVO actor, RequestContextVO context) {
        PersonVO person = null;
        try {
            person = mergePersonHandler.execute(from, to, modified, actor, context);
        } catch (DAOException | IOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.MERGE_PERSON_INTERNAL_SERVER_ERROR, e);
        }
        return person;
    }

    @Override
    public List<OIDVO> getOIDs() throws DAOException {
        try {
            return oidDao.getAllOIDs();
        } catch (DAOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.GET_OIDS_INTERNAL_SERVER_ERROR, e);
        }
        return new ArrayList<>();
    }

    @Override
    public Map<String, List<Permission>> getPermissions(boolean getAllPermissions, boolean getCallersPermissions) {
        try {
            if (getAllPermissions) {
                Map<String, List<Permission>> result = new HashMap<>();
                for (Role role : Role.values()) {
                    PredefinedRequestedRole predefinedRequestedRole = PredefinedRequestedRole.fromValue(role.getSchemaName());
                    result.put(role.getSchemaName(), roleValidationInterceptor.getRoleToAuthenticatorMap().get(predefinedRequestedRole).getPermissions());
                }
                return result;
            }
            if (getCallersPermissions) {
                Map<String, List<Permission>> result = new HashMap<>();
                Role role = Role.forSchemaName(RequestContext.get().getRequestedRole());
                if (role != null) {
                    log.debug(role.getSchemaName());
                    PredefinedRequestedRole predefinedRequestedRole = PredefinedRequestedRole.fromValue(role.getSchemaName());
                    result.put(role.getSchemaName(), roleValidationInterceptor.getRoleToAuthenticatorMap().get(predefinedRequestedRole).getPermissions());
                } else {
                    result.put(RequestContext.get().getRequestedRole(), Collections.EMPTY_LIST);
                }
                return result;
            }
        } catch (Exception e) {
            FaultUtil.throwECPR2Fault(FaultCode.GET_PERMISSIONS_INTERNAL_SERVER_ERROR, e);
        }
        return new HashMap<>();
    }

    @Override
    public int countPersons(String searchQuery) throws Exception {
        int count = 0;
        try {
            count = countPersonHandler.execute(searchQuery);
        } catch (DAOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.SEARCH_PERSON_INTERNAL_SERVICE_ERROR, e);
        }
        return count;
    }

    @Override
    public void blacklistDeCPR(String oid, String deCPR) {
        blacklistingDAO.blacklistDeCPR(oid, deCPR);
    }

    @Override
    public BlacklistedDeCPRVO isBlacklistedDeCPR(String oid, String deCPR) {
        return blacklistingDAO.isBlacklistedDeCPR(oid, deCPR);
    }

    @Override
    public List<PersonVO> searchPersons(String searchStr) throws Exception {
        List<PersonVO> personList = null;
        try {
            personList = searchPersonHandler.execute(searchStr);
        } catch (DAOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.SEARCH_PERSON_INTERNAL_SERVICE_ERROR, e);
        }
        return personList;
    }

    @Override
    public List<List<PersonVO>> reservePersonIds(List<ReserveRequestVO> reserveRequest, ServiceActorVO actor, RequestContextVO context) {
        List<List<PersonVO>> personsList = new ArrayList<>();
        try {
            for (ReserveRequestVO request : reserveRequest) {
                personsList.add(reservePersonIdHandler.execute(request, actor, context));
            }
        } catch (DAOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.CREATE_PERSON_INTERNAL_SERVER_ERROR, e);
        }
        return personsList;
    }

    @Override
    public PersonVO updatePerson(PersonVO personUpdate, ServiceActorVO actor, RequestContextVO context) {
        PersonVO result = null;
        try {
            List<IdentifierWithoutServiceActorVO> deletions = getIdentifiersForDeletion(personUpdate, context);
            result = updatePersonHandler.execute(personUpdate, actor, context);
            advisUpdate(context, result.getIdentifiers(), deletions);
        } catch (DAOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.UPDATE_PERSON_INTERNAL_SERVER_ERROR, e);
        } catch (IOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.UPDATE_PERSON_LUCENE_INDEX_INTERNAL_SERVER_ERROR, e);
        }
        return result;
    }

    private List<IdentifierWithoutServiceActorVO> getIdentifiersForDeletion(PersonVO person, RequestContextVO context) throws DAOException {
        List<IdentifierWithoutServiceActorVO> returnValue = new ArrayList<>();
        List<IdentifierVO> identifiers = person.getIdentifiers();
        if (identifiers == null || identifiers.size() == 0) {
            return returnValue;
        }
        Date now = Date.from(context.getRequestTimeStamp());
        List<String> toBeDeleted = identifiers.stream().filter(i -> validator.toBeDeleted(i)).map(IdentifierVO::getOid).collect(Collectors.toList());
        returnValue = identifierDAO.getIdentifiersByUuids(toBeDeleted, now);
        return returnValue;
    }

    @Override
    public List<PersonVO> getPersonsById(List<PersonVO> persons, ServiceActorVO actor, RequestContextVO context) {
        List<PersonVO> personList = new ArrayList<>();
        try {
            personList = getPersonIdHandler.execute(persons, actor);
        } catch (DAOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.GET_PERSON_BY_PERSONID_INTERNAL_SERVER_ERROR, e);
        }
        return personList;
    }

    @Override
    public void deletePerson(List<IdentifierVO> identifiers, Modified modified, ServiceActorVO actor, RequestContextVO context) {
        try {
            deletePersonHandler.execute(identifiers, modified, actor, context);
            advisDelete(context, identifiers);
        } catch (DAOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.DELETE_PERSON_INTERNAL_SERVER_ERROR, e);
        } catch (IOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.DELETE_PERSON_LUCENE_INDEX_INTERNAL_SERVER_ERROR, e);
        }
    }

    @Override
    public Integer compareName(KeyIdentifier keyIdentifier, String name, ServiceActorVO actor, RequestContextVO context) {
        Integer score;
        List<PersonVO> tmpPersonList = requestMapper.mapIdentifierToPersonFromRequest(Collections.singletonList(keyIdentifier));
        List<PersonVO> personList = getPersonsById(tmpPersonList, actor, context);
        score = compareNameHandler.execute(personList, name, actor, context);
        return score;
    }

    // Note: for testing only
    public void setAdvisService(AsyncAdvisFacade mock) {
        this.advisService = mock;
    }

    private void advisUpdate(RequestContextVO context, List<IdentifierVO> identifiers, List<IdentifierWithoutServiceActorVO> deletions) {
        if (identifiers != null) {
            for (IdentifierVO identifier : identifiers) {
                if (identifier.getOid() != null && identifier.getValue() != null) {
                    advisService.advisPersonIdUpdated(context, identifier.getOid(), identifier.getValue());
                }
            }
            if (identifiers.stream().anyMatch(id -> id.getOid().equals(OID.CPR) && id.getValidTo().after(Date.from(context.getRequestTimeStamp())))) {
                IdentifierVO xeCPR = identifiers.stream().filter(id ->
                        id.getOid().equals(OID.X_ECPR) &&
                        id.getValidTo().after(Date.from(context.getRequestTimeStamp()))
                ).findFirst().orElse(
                        identifiers.stream().filter(id ->
                                id.getOid().equals(OID.C_ECPR) &&
                                id.getValidTo().after(Date.from(context.getRequestTimeStamp()))
                        ).findFirst().orElse(null)
                );
                if (xeCPR != null) {
                    kafkaService.advisPersonIdUpdated(context, xeCPR.getOid(), xeCPR.getValue());
                }
            }
        }
        if (deletions != null) {
            for (IdentifierWithoutServiceActorVO identifier : deletions) {
                if (identifier.getOid() != null && identifier.getValue() != null) {
                    advisService.advisPersonIdUpdated(context, identifier.getOid(), identifier.getValue());
                    if (identifier.getOid().equals(OID.CPR)) {
                        kafkaService.advisPersonIdDeleted(context, identifier.getOid(), identifier.getValue());
                    }
                }
            }
            if (deletions.stream().anyMatch(id -> id.getOid().equals(OID.CPR) && id.getValidTo().after(Date.from(context.getRequestTimeStamp())))) {
                IdentifierWithoutServiceActorVO xeCPR = deletions.stream().filter(id ->
                        id.getOid().equals(OID.X_ECPR) &&
                                id.getValidTo().after(Date.from(context.getRequestTimeStamp()))
                ).findFirst().orElse(
                        deletions.stream().filter(id ->
                                id.getOid().equals(OID.C_ECPR) &&
                                        id.getValidTo().after(Date.from(context.getRequestTimeStamp()))
                        ).findFirst().orElse(null)
                );
                if (xeCPR != null) {
                    kafkaService.advisPersonIdDeleted(context, xeCPR.getOid(), xeCPR.getValue());
                }
            }
        }
    }

    private void advisDelete(RequestContextVO context, List<IdentifierVO> identifiers) {
        if (identifiers == null) return;
        for (IdentifierVO identifier : identifiers) {
            advisService.advisPersonIdDeleted(context, identifier.getOid(), identifier.getValue());
        }
    }

    public CprPersonVO getCprPerson(KeyIdentifier keyIdentifier) {
        return cprService.getCprPerson(keyIdentifier);
    }
}

