/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service;

import dk.sds.ecpr2.dao.SearchPersonDAO;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.scheduling.annotation.EnableScheduling;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import java.text.SimpleDateFormat;
import java.time.Instant;

@Component
@EnableScheduling
@Log4j2
public class LuceneIndexJobScheduler {

    @Value("${jobs.searchindex.enabled:true}")
    private boolean jobIsEnabled;

    private final int delayWhenErrorMs = 10_000;
    private final int maximumWaitTimeMs = 5 * 60 * 1000;

    @Autowired
    SearchPersonDAO searchPersonDAO;

    Instant lastCompletelySynchronizedTime = null;
    int errorCount = 0;


    @Scheduled(initialDelayString = "${jobs.searchindex.msInitialDelay:2000}", fixedDelayString = "${jobs.searchindex.msDelayBetween:30000}")
    public void runUpdateLuceneIndexes() throws InterruptedException {
        if ( ! jobIsEnabled) {
            log.error("Search index job is disabled. This must not happen in production.");
            return;
        }

        try {
            Instant currentRunStartedTime = Instant.now();
            searchPersonDAO.buildIndex(lastCompletelySynchronizedTime);
            lastCompletelySynchronizedTime = currentRunStartedTime;
            errorCount = 0;
        } catch (Exception e) {
            errorCount++;
            log.error("Lucene index was not updated because of an error.", e);
        } finally {
            // It probably does not help to try again immediately. So wait longer for every error.
            Thread.sleep(Math.min(errorCount * delayWhenErrorMs, maximumWaitTimeMs));
        }
    }

    public String getHealthInformation(){
        String text = "";
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");

        text += jobIsEnabled ? "Enabled" : "Disabled";
        text += ", Last successful indexing: " + (lastCompletelySynchronizedTime == null ? "Never" : lastCompletelySynchronizedTime.toString());
        text += ", Error count: " + errorCount;

        return text;
    }
}
