/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service;


import dk.sds.ecpr2.OID;
import dk.sds.ecpr2.dao.IdentifierDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.Gender;
import dk.sds.ecpr2.domain.IdentifierWithoutServiceActorVO;
import dk.sds.ecpr2.util.FaultCode;
import dk.sds.ecpr2.util.FaultUtil;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.time.LocalDate;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;
import java.util.stream.Collectors;

@Service
public class PersonNumberGeneratorImpl implements PersonNumberGenerator {

    @Autowired
    private final IdentifierDAO identifierDao;

    private final Random rnd = new Random();

    protected final static String VALID_CHARACTERS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    protected final static String VALID_MALE_DIGITS = "13579";
    protected final static String VALID_FEMALE_DIGITS = "02468";
    private final int NUMBER_OF_CHARS_IN_ALPAHBET = VALID_CHARACTERS.length();
    private final int NUMBER_OF_GENDER_DIGITS_PER_GENDER = VALID_MALE_DIGITS.length();

    PersonNumberGeneratorImpl(IdentifierDAO dao) {
        identifierDao = dao;
    }

    /**
     * Generates an unused X-eCPR given a date and gender.
     * Random characters/digits are generated.
     **/
    public String generateXECPR(int day, int month, int year, Gender gender, boolean failOnFutureDate) {
        return generateXECPR(day, month, year, gender, failOnFutureDate, new HashSet<>());
    }

    /**
     * Generates an unused C-eCPR given a date and gender.
     * Random characters/digits are generated.
     **/
    public String generateCECPR(int day, int month, int year, Gender gender, boolean failOnFutureDate) {
        return generateCECPR(day, month, year, gender, failOnFutureDate, new HashSet<>());
    }

    public String generateXECPR(int day, int month, int year, Gender gender, boolean failOnFutureDate, Set<String> alreadyTakenIds) {
        checkDateIsValid(day, month, year, failOnFutureDate);
        checkGenderIsValid(gender);

        String invariantPartOfXECPR = constructInvariantPartOfXECPR(day, month, year);

        String newCECPR = null;
        String searchPatternCECPR = invariantPartOfXECPR + "___"; // Last three characters are undefined.
        try {
            final List<IdentifierWithoutServiceActorVO> unavailableIdentifiers = identifierDao.getMatchingIdentifiersWithoutServiceActor(searchPatternCECPR, OID.X_ECPR);
            List<String> unavailableIds = unavailableIdentifiers.stream().map(IdentifierWithoutServiceActorVO::getValue).collect(Collectors.toList());
            unavailableIds.addAll(alreadyTakenIds);
            newCECPR = findAvailableId(unavailableIds, gender, invariantPartOfXECPR);
        } catch (DAOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.CREATE_NEW_ECPR_ERROR, e);
        }

        if (newCECPR == null) {
            FaultUtil.throwECPR2Fault(FaultCode.CREATE_NEW_ECPR_ERROR, "No available X-eCPR numbers found for the prefix: " + invariantPartOfXECPR);
        }

        return newCECPR;
    }

    public String generateCECPR(int day, int month, int year, Gender gender, boolean failOnFutureDate, Set<String> alreadyTakenIds) {
        checkDateIsValid(day, month, year, failOnFutureDate);
        checkGenderIsValid(gender);

        String invariantPartOfCECPR = constructInvariantPartOfCECPR(day, month, year);

        String newCECPR = null;
        String searchPatternCECPR = invariantPartOfCECPR + "___"; // Last three characters are undefined.
        try {
            final List<IdentifierWithoutServiceActorVO> unavailableIdentifiers = identifierDao.getMatchingIdentifiersWithoutServiceActor(searchPatternCECPR, OID.C_ECPR);
            List<String> unavailableIds = unavailableIdentifiers.stream().map(IdentifierWithoutServiceActorVO::getValue).collect(Collectors.toList());
            unavailableIds.addAll(alreadyTakenIds);
            newCECPR = findAvailableId(unavailableIds, gender, invariantPartOfCECPR);
        } catch (DAOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.CREATE_NEW_ECPR_ERROR, e);
        }

        if (newCECPR == null) {
            FaultUtil.throwECPR2Fault(FaultCode.CREATE_NEW_ECPR_ERROR, "No available C-eCPR numbers found for the prefix: " + invariantPartOfCECPR);
        }

        return newCECPR;
    }

    static String constructInvariantPartOfCECPR(int day, int month, int year) {
        String yearString = String.valueOf(year).substring(2,4);
        int decade = calculateDecadeForCECPR(year);
        return String.format("%02d%02d%s%d", day, month, yearString, decade);
    }

    static String constructInvariantPartOfXECPR(int day, int month, int year) {
        String yearString = String.valueOf(year).substring(2,4);
        int decade = calculateDecadeForXECPR(year);
        return String.format("%02d%02d%s%d", day, month, yearString, decade);
    }

    private void checkDateIsValid(int day, int month, int year, boolean failOnFutureDate) {
        LocalDate date = null;
        try {
            date = LocalDate.of(year, month, day);
        } catch (Exception e) {
            FaultUtil.throwECPR2Fault(FaultCode.CREATE_NEW_ECPR_ERROR, "The specified date is not valid: " + day + "-" + month + "-" + year);
        }
        if (failOnFutureDate && date != null && date.isAfter(LocalDate.now())) {
            FaultUtil.throwECPR2Fault(FaultCode.CREATE_NEW_ECPR_ERROR, "The specified date is in the future: " + day + "-" + month + "-" + year);
        }
    }

    private void checkGenderIsValid(Gender gender) {
        if (gender != Gender.MALE && gender != Gender.FEMALE && gender != Gender.OTHER && gender != Gender.UNKNOWN) {
            FaultUtil.throwECPR2Fault(FaultCode.CREATE_NEW_ECPR_ERROR, "The specified gender is not valid: " + gender);
        }
    }

    private String findAvailableId(List<String> unavailableIds, Gender gender, String idPrefix) {
        String candidateId = null;

        // Total number of possible combinations x 2, to account for same random values being hit multiple times:
        final int maximumNumberOfAttempts = 2 * NUMBER_OF_CHARS_IN_ALPAHBET * NUMBER_OF_CHARS_IN_ALPAHBET * NUMBER_OF_GENDER_DIGITS_PER_GENDER;

        int attemptNumber = 1;
        do {
            attemptNumber++;
            String genderDigtis = null;
            switch (gender) {
                case MALE:
                    genderDigtis = VALID_MALE_DIGITS;
                    break;
                case FEMALE:
                    genderDigtis = VALID_FEMALE_DIGITS;
                    break;
                case OTHER:
                    genderDigtis = VALID_FEMALE_DIGITS + VALID_MALE_DIGITS;
                    break;
                case UNKNOWN:
                    genderDigtis = VALID_FEMALE_DIGITS + VALID_MALE_DIGITS;
                    break;
            }
            char genderDigitCandidate = getRandomCharCandidate(genderDigtis);
            char firstCharCandidate = getRandomCharCandidate(VALID_CHARACTERS);
            char secondCharCandidate = getRandomCharCandidate(VALID_CHARACTERS);
            candidateId = idPrefix + firstCharCandidate + secondCharCandidate + genderDigitCandidate;

            if ( ! unavailableIds.contains(candidateId)) {
                break;
            }
            candidateId = null;
        } while (attemptNumber < maximumNumberOfAttempts);

        return candidateId;
    }

    private char getRandomCharCandidate(String charactersToPickFrom) {
        return charactersToPickFrom.charAt(rnd.nextInt(charactersToPickFrom.length()));
    }

    private static int calculateDecadeForCECPR(int year) {
        int decade;
        if (year < 1900) {
            decade = 5;
        } else if (year > 1999) {
            decade = 6;
        } else {
            decade = 0;
        }
        return decade;
    }

    private static int calculateDecadeForXECPR(int year) {
        int decade;
        if (year > 1999) {
            decade = 7;
        } else {
            decade = 1;
        }
        return decade;
    }

}
