/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service.handler;

import dk.sds.ecpr2.Modified;
import dk.sds.ecpr2.ModifiedBy;
import dk.sds.ecpr2.dao.*;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.SearchCrossTableDAOImpl;
import dk.sds.ecpr2.dao.impl.abstr.AbstractECPRDAOImpl;
import dk.sds.ecpr2.domain.*;
import dk.sds.ecpr2.service.PersonNumberGeneratorImpl;
import org.springframework.beans.factory.annotation.Autowired;

import java.time.LocalDate;
import java.util.*;

public abstract class AbstractHandler {

    @Autowired
    protected PersonNumberGeneratorImpl personNumberGeneratorImpl;

    @Autowired
    protected Validator validator;

    @Autowired
    protected PersonDAO personDAO;

    @Autowired
    protected IdentifierDAO identifierDAO;

    @Autowired
    protected AddressDAO addressDAO;

    @Autowired
    protected ServiceActorDAO serviceActorDAO;

    @Autowired
    protected BirthDateDAO birthDateDAO;

    @Autowired
    protected ContactDAO contactDAO;

    @Autowired
    protected GenderDAO genderDAO;

    @Autowired
    protected NameDAO nameDAO;

    @Autowired
    protected OIDFacade oidFacade;

    @Autowired
    protected SearchCrossTableDAOImpl crossTableDAO;

    @Autowired
    protected SearchPersonDAO searchPersonDAO;

    @Autowired
    protected ModifierDAO modifierDAO;

    protected String createNewCECPR(PersonVO person) {
        return createNewCECPR(person, null);
    }

    protected String createNewCECPR(PersonVO person, PersonVO existingPerson) {
        BirthDateVO newBirthDateVO = person.getBirthDates() != null && person.getBirthDates().size() > 0
                ? person.getBirthDates().get(0)
                : null;
        GenderVO newGenderVo = person.getGenders() != null && person.getGenders().size() > 0
                ? person.getGenders().get(0)
                : null;
        LocalDate today = LocalDate.now();
        if (newBirthDateVO == null) {
            if (existingPerson.getBirthDates() != null) {
                for (BirthDateVO date : existingPerson.getBirthDates()) {
                    if (date.getValidTo() == null || date.getValidTo().after(new Date())) {
                        newBirthDateVO = date;
                        break;
                    }
                }
            }
        }
        if (newBirthDateVO == null) {
            newBirthDateVO = new BirthDateVO();
        }
        if (newGenderVo == null) {
            if (existingPerson.getGenders() != null) {
                for (GenderVO gender : existingPerson.getGenders()) {
                    if (gender.getValidTo() == null || gender.getValidTo().after(new Date())) {
                        newGenderVo = gender;
                        break;
                    }
                }
            }
        }

        if (newBirthDateVO.getYear() == null || newBirthDateVO.getYear() == 0) {
            newBirthDateVO.setYear(today.getYear());
        }
        if (newBirthDateVO.getMonth() == null || newBirthDateVO.getMonth() == 0) {
            newBirthDateVO.setMonth(today.getMonthValue());
        }
        if (newBirthDateVO.getDay() == null || newBirthDateVO.getDay() == 0) {
            // Avoid generating eg the 31. of June
            newBirthDateVO.setDay(Math.min(today.getDayOfMonth(), 30));
        }
        return personNumberGeneratorImpl.generateCECPR(
                newBirthDateVO.getDay(),
                newBirthDateVO.getMonth(),
                newBirthDateVO.getYear(),
                Gender.valueOf(newGenderVo.getValue().toUpperCase()),
                true);
    }

    protected String createNewXECPR(PersonVO person) {
        return createNewXECPR(person, null);
    }

    /**
     * If an existing person is added then, data needs to be concatenated for the fields that are not changed
     * @param person the updated person with the redundant data removed
     * @param existingPerson the existing person found in the database
     * @return a new XeCPR number
     */
    protected String createNewXECPR(PersonVO person, PersonVO existingPerson) {
        BirthDateVO newBirthDateVO = person.getBirthDates() != null && person.getBirthDates().size() > 0
                ? person.getBirthDates().get(0)
                : null;
        GenderVO newGenderVo = person.getGenders() != null && person.getGenders().size() > 0
                ? person.getGenders().get(0)
                : null;
        LocalDate today = LocalDate.now();
        if (newBirthDateVO == null) {
            if (existingPerson.getBirthDates() != null) {
                for (BirthDateVO date : existingPerson.getBirthDates()) {
                    if (date.getValidTo() == null || date.getValidTo().after(new Date())) {
                        newBirthDateVO = date;
                        break;
                    }
                }
            }
        }
        if (newBirthDateVO == null) {
            newBirthDateVO = new BirthDateVO();
        }
        if (newGenderVo == null) {
            if (existingPerson.getGenders() != null) {
                for (GenderVO gender : existingPerson.getGenders()) {
                    if (gender.getValidTo() == null || gender.getValidTo().after(new Date())) {
                        newGenderVo = gender;
                        break;
                    }
                }
            }
        }

        if (newBirthDateVO.getYear() == null || newBirthDateVO.getYear() == 0) {
            newBirthDateVO.setYear(today.getYear());
        }
        if (newBirthDateVO.getMonth() == null || newBirthDateVO.getMonth() == 0) {
            newBirthDateVO.setMonth(today.getMonthValue());
        }
        if (newBirthDateVO.getDay() == null || newBirthDateVO.getDay() == 0) {
            // Avoid generating eg the 31. of June
            newBirthDateVO.setDay(Math.min(today.getDayOfMonth(), 30));
        }
        return personNumberGeneratorImpl.generateXECPR(
                newBirthDateVO.getDay(),
                newBirthDateVO.getMonth(),
                newBirthDateVO.getYear(),
                Gender.valueOf(newGenderVo.getValue().toUpperCase()),
                true);
    }

    protected int createNewVersionOfPerson(String personUUID, ServiceActorVO actor, ModifiedBy modifiedBy, Date now) throws DAOException {
        // ReservedUntil is null'ed on updates
        return createNewVersionOfPerson(personUUID, actor, modifiedBy, null, now);
    }

    protected int createNewVersionOfPerson(String personUUID, ServiceActorVO actor, ModifiedBy modifiedBy, Date reservedUntil, Date now) throws DAOException {

        final PersonVO existingPerson = personDAO.getCurrentPerson(personUUID);
        final int existingVersion = existingPerson == null ? 0 : existingPerson.getVersion();
        final int nextVersion = existingVersion + 1;

        Integer modifierId = modifierDAO.getOrCreateModified(modifiedBy, now).getId();

        final PersonVO newPerson = new PersonVO();
        newPerson.setPersonUUID(personUUID);
        newPerson.setVersion(nextVersion);
        newPerson.setVersionDescription("Version " + nextVersion + ", updated by " + modifiedBy.getPerson().getName());
        newPerson.setModifierPID(modifierId);
        newPerson.setServiceActor(actor);
        newPerson.setReservedUntil(reservedUntil);
        newPerson.setModified(now);

        return createPersonAndActor(newPerson, now);
    }

    protected int createPersonAndActor(PersonVO person, Date now) throws DAOException {

        Integer serviceActorPID = serviceActorDAO.getServiceActorFromHashCode(person.getServiceActor());
        if (serviceActorPID == null) {
            serviceActorPID = serviceActorDAO.createServiceActor(person.getServiceActor(), now);
        }
        person.setServiceActorPID(serviceActorPID);
        person.getServiceActor().setServiceActorPID(serviceActorPID);

        return personDAO.createPerson(person);
    }

    protected String randomUUID() {
        return UUID.randomUUID().toString();
    }

    protected void addModifiers(List<PersonVO> persons, boolean includePersonIdentifier) throws DAOException {
        for (PersonVO person : persons) {
            List<Modified> versions = new ArrayList<>();
            List<PersonVO> personsByUUID = personDAO.getPersonsByUUIDs(Collections.singletonList(person.getPersonUUID()));
            for (int i = 0; i < personsByUUID.size(); i++) {
                PersonVO personVersion = personsByUUID.get(i);
                Modified modified = new Modified();
                ModifiedBy modifiedBy = modifierDAO.getModifier(personVersion.getModifierPID(), includePersonIdentifier);
                modified.setBy(modifiedBy);
                Calendar cal = new GregorianCalendar();
                cal.setTime(personVersion.getModifiedTime());
                modified.setValidFrom(cal);
                if (personsByUUID.size() == (i + 1)) {
                    // last modified, set ValidTo=INF
                    cal = new GregorianCalendar();
                    cal.setTime(AbstractECPRDAOImpl.VALID_TO_DEFAULT);
                    modified.setValidTo(cal);
                } else {
                    cal = new GregorianCalendar();
                    cal.setTime(personsByUUID.get(i + 1).getModifiedTime());
                    modified.setValidTo(cal);
                }
                if (personVersion.getReservedUntil() != null) {
                    cal = new GregorianCalendar();
                    cal.setTime(personVersion.getReservedUntil());
                    modified.setReservedUntil(cal);
                }
                versions.add(modified);
            }
            person.getModifiers().addAll(versions);
        }
    }

    // Merge Person rows into a single PersonVO containing all related entities
    protected PersonVO mergePersons(List<PersonVO> personList) throws DAOException {
        if (personList.isEmpty()) {
            return null;
        }
        PersonVO mergedPerson = new PersonVO();
        PersonVO first = personList.get(0);
        mergedPerson.setPersonUUID(first.getPersonUUID());
        mergedPerson.setReservedUntil(first.getReservedUntil());
        for (PersonVO dbPerson : personList) {
            mergedPerson.getGenders().addAll(dbPerson.getGenders());
            mergedPerson.getIdentifiers().addAll(dbPerson.getIdentifiers());
            mergedPerson.getBirthDates().addAll(dbPerson.getBirthDates());
            mergedPerson.getAddresses().addAll(dbPerson.getAddresses());
            mergedPerson.getContacts().addAll(dbPerson.getContacts());
            mergedPerson.getNames().addAll(dbPerson.getNames());
        }
        addModifiers(Collections.singletonList(mergedPerson), false);
        return mergedPerson;
    }

    protected List<PersonVO> getPersonById(List<IdentifierVO> identifiers) throws DAOException {
        List<PersonIdVO> personIds = getAllPersonIds(identifiers);
        return crossTableDAO.getPersonsWithRelationsByPersonId(personIds);
    }

    private List<PersonIdVO> getAllPersonIds(List<IdentifierVO> identifiers) {
        List<PersonIdVO> personIdVOS = new ArrayList<>();
        for (IdentifierVO identifier : identifiers) {
            PersonIdVO personId = new PersonIdVO();
            personId.setOid(identifier.getOid());
            personId.setValue(identifier.getValue());
            personIdVOS.add(personId);
        }
        return personIdVOS;
    }
}
