/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service.handler;

import dk.sds.ecpr2.KeyIdentifier;
import dk.sds.ecpr2.OID;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.*;
import dk.sds.ecpr2.service.CprService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;

@Component
public class CreatePersonHandler extends AbstractHandler {

    @Autowired
    CprService cprService;

    public PersonVO execute(PersonVO personVO, ServiceActorVO actor, RequestContextVO context) throws DAOException, IOException {
        validator.validateCreatePersonData(personVO);

        Date nowTimestamp = Date.from(context.getRequestTimeStamp());

        List<String> oidsToBeCreated = new ArrayList<>();
        for (IdentifierVO vo : personVO.getIdentifiers()) {
            if (vo.getUuid() == null) {
                if (vo.getValue() == null) {
                    if (validator.oidCreationSupported(vo.getOid())) {
                        oidsToBeCreated.add(vo.getOid());
                    }
                }
            }
        }

        if (oidsToBeCreated.isEmpty()) {
            oidsToBeCreated.add(OID.X_ECPR);
        }

        for (String oid : oidsToBeCreated) {
            assignNewCeCPR(personVO, oid);
        }

        String personUUID = randomUUID();
        final int personId = createNewVersionOfPerson(personUUID, actor, personVO.getModifiers().get(0).getBy(), nowTimestamp);

        for (IdentifierVO identifier : personVO.getIdentifiers()) {
            if (identifier.getOid().equals(OID.CPR)) {
                KeyIdentifier cprIdentifier = new KeyIdentifier();
                cprIdentifier.setValue(identifier.getValue());
                cprIdentifier.setOID(identifier.getOid());
                if (cprService.getCprPerson(cprIdentifier) == null) {
                    throw new IllegalArgumentException("Cpr identifier er ikke valid");
                }
            }
            identifierDAO.createIdentifier(identifier, nowTimestamp, personId);
        }

        if (personVO.getAddresses() != null) {
            for (AddressVO address : personVO.getAddresses()) {
                addressDAO.createAddress(address, nowTimestamp, personId);
            }
        }

        if (personVO.getBirthDates() != null) {
            for (BirthDateVO birthDate : personVO.getBirthDates()) {
                birthDateDAO.createBirthDate(birthDate, nowTimestamp, personId);
            }
        }

        if (personVO.getContacts() != null) {
            for (ContactVO contact : personVO.getContacts()) {
                contactDAO.createContact(contact, nowTimestamp, personId);
            }
        }

        if (personVO.getGenders() != null) {
            for (GenderVO gender : personVO.getGenders()) {
                genderDAO.createGender(gender, nowTimestamp, personId);
            }
        }

        if (personVO.getNames() != null) {
            for (NameVO name : personVO.getNames()) {
                nameDAO.createName(name, nowTimestamp, personId);
            }
        }

        PersonVO createdPerson = crossTableDAO.getPersonsWithRelations(Collections.singletonList(personUUID)).get(0);
        searchPersonDAO.addPersonToIndex(Collections.singletonList(createdPerson));
        addModifiers(Collections.singletonList(createdPerson), false);
        return createdPerson;
    }

    private void assignNewCeCPR(PersonVO person, String eCPRoid) {
        if (person.getIdentifiers() == null) {
            person.setIdentifiers(new ArrayList<>());
        }
        IdentifierVO foundCECPR = null;
        for (IdentifierVO identifier : person.getIdentifiers()) {
            if (eCPRoid.equals(identifier.getOid())) {
                foundCECPR = identifier;
                break;
            }
        }
        String ceCPR = null;
        switch (eCPRoid) {
            case OID.C_ECPR:
                ceCPR = createNewCECPR(person);
                break;
            case OID.X_ECPR:
                ceCPR = createNewXECPR(person);
                break;
        }
        if (foundCECPR != null) {
            foundCECPR.setValue(ceCPR);
            foundCECPR.setValidity(16);
        } else {
            IdentifierVO vo = new IdentifierVO();
            vo.setOid(eCPRoid);
            vo.setValidity(16);
            vo.setValue(ceCPR);
            person.getIdentifiers().add(vo);
        }
    }
}
