/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service.handler;

import dk.sds.ecpr2.KeyIdentifier;
import dk.sds.ecpr2.OID;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.IdentifierVO;
import dk.sds.ecpr2.domain.PersonVO;
import dk.sds.ecpr2.domain.RequestContextVO;
import dk.sds.ecpr2.domain.ServiceActorVO;
import dk.sds.ecpr2.service.CprService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;

@Component
public class CreatePersonIdHandler extends AbstractHandler {

    @Autowired
    CprService cprService;

    public PersonVO execute(PersonVO personVO, ServiceActorVO actor, RequestContextVO context) throws DAOException {
        validator.validateKeyIdentifier(personVO);
        List<PersonVO> personsWithRelations = crossTableDAO.getPersonsWithRelations(Collections.singletonList(personVO.getPersonUUID()));
        PersonVO existingPerson = mergePersons(personsWithRelations);
        personVO.getIdentifiers().addAll(existingPerson.getIdentifiers());

        final Date now = Date.from(context.getRequestTimeStamp());

        List<String> primaryOidsToBeCreated = personVO.getIdentifiers()
                .stream()
                .filter(vo -> vo.getUuid() == null && vo.getValue() == null && validator.oidCreationSupported(vo.getOid()))
                .map(IdentifierVO::getOid)
                .collect(Collectors.toList());

        if (primaryOidsToBeCreated.isEmpty()) {
            primaryOidsToBeCreated.add(OID.X_ECPR);
        }

        for (String oid : primaryOidsToBeCreated) {
            assignNewPrimaryECPR(existingPerson, personVO, oid);
        }

        final int personId = createNewVersionOfPerson(existingPerson.getPersonUUID(), actor, personVO.getModifiers().get(0).getBy(), now);

        for (IdentifierVO vo : personVO.getIdentifiers()) {
            if (vo.getUuid() != null) {
                if (validator.toBeDeleted(vo)) {
                    identifierDAO.delete(vo, now);
                } else {
                    identifierDAO.updateIdentifier(vo, now);
                }
            } else {
                if (vo.getOid().equals(OID.CPR)) {
                    KeyIdentifier cprIdentifier = new KeyIdentifier();
                    cprIdentifier.setValue(vo.getValue());
                    cprIdentifier.setOID(vo.getOid());
                    if (cprService.getCprPerson(cprIdentifier) == null) {
                        throw new IllegalArgumentException("Cpr identifier er ikke valid");
                    }
                }
                identifierDAO.createIdentifier(vo, now, personId);
            }
        }

        List<PersonVO> personList = crossTableDAO.getPersonsWithRelations(Collections.singletonList(personVO.getPersonUUID()));
        addModifiers(personList, false);

        return mergePersons(personList);
    }

    private void assignNewPrimaryECPR(PersonVO existingPerson, PersonVO person, String eCPRoid) {
        if (existingPerson.getIdentifiers() == null) {
            existingPerson.setIdentifiers(new ArrayList<>());
        }
        IdentifierVO foundCECPR = null;
        for (IdentifierVO identifier : person.getIdentifiers()) {
            if (eCPRoid.equals(identifier.getOid())) {
                foundCECPR = identifier;
                break;
            }
        }
        String ceCPR = null;
        switch (eCPRoid) {
            case OID.C_ECPR:
                ceCPR = createNewCECPR(existingPerson);
                break;
            case OID.X_ECPR:
                ceCPR = createNewXECPR(existingPerson);
                break;
        }
        if (foundCECPR != null) {
            foundCECPR.setValue(ceCPR);
            foundCECPR.setValidity(16);
        } else {
            IdentifierVO vo = new IdentifierVO();
            vo.setOid(eCPRoid);
            vo.setValidity(16);
            vo.setValue(ceCPR);
            person.getIdentifiers().add(vo);
        }
    }
}
