/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service.handler;

import dk.sds.ecpr2.Modified;
import dk.sds.ecpr2.OID;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.*;
import dk.sds.ecpr2.webservice.validation.ValidationErrors;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;

@Component
public class MergePersonHandler extends AbstractHandler {

    @Autowired
    DeletePersonHandler deletePersonHandler;

    public PersonVO execute(KeyIdentifierVO from, KeyIdentifierVO to, Modified modified, ServiceActorVO actor, RequestContextVO context) throws DAOException, IOException {

        String personFromUUID = validator.validateKeyIdentifier(from);
        String personToUUID = validator.validateKeyIdentifier(to);

        List<PersonVO> personsFrom = crossTableDAO.getPersonWithRelations(personFromUUID);
        List<PersonVO> personsTo = crossTableDAO.getPersonWithRelations(personToUUID);

        PersonVO personFrom = mergePersons(personsFrom);
        PersonVO personTo = mergePersons(personsTo);

        final Date now = Date.from(context.getRequestTimeStamp());
        checkPersonsCanBeMerged(personFrom, personTo, now);

        PersonVO mergedPerson = mergePersons(personFrom, personTo, modified, actor, now, context);

        List<PersonVO> personList = crossTableDAO.getPersonsWithRelations(Collections.singletonList(mergedPerson.getPersonUUID()));
        addModifiers(personList, false);
        return personList.get(personList.size() - 1);
    }

    private void checkPersonsCanBeMerged(PersonVO personFrom, PersonVO personTo, Date now) {
        List<String> validationErrors = new ArrayList<>();
        List<IdentifierVO> personFromNonKeyIdentifiers = personFrom.getIdentifiers().stream().filter(i -> !i.getOid().equals(OID.X_ECPR)).collect(Collectors.toList());
        List<IdentifierVO> personToNonKeyIdentifiers = personTo.getIdentifiers().stream().filter(i -> !i.getOid().equals(OID.X_ECPR)).collect(Collectors.toList());
        List<IdentifierVO> personFromOverlappingIdentifiers = personFromNonKeyIdentifiers.stream().filter(i -> personToNonKeyIdentifiers.stream().anyMatch(i2 -> i.getOid().equals(i2.getOid()))).collect(Collectors.toList());
        List<IdentifierVO> personToOverlappingIdentifiers = personToNonKeyIdentifiers.stream().filter(i -> personFromNonKeyIdentifiers.stream().anyMatch(i2 -> i.getOid().equals(i2.getOid()))).collect(Collectors.toList());
        List<IdentifierVO> currentIdentifiersFrom = personFromOverlappingIdentifiers.stream().filter(i -> i.getValidTo().after(now)).collect(Collectors.toList());
        List<IdentifierVO> currentIdentifiersTo = personToOverlappingIdentifiers.stream().filter(i -> i.getValidTo().after(now)).collect(Collectors.toList());
        if (!currentIdentifiersFrom.stream().allMatch(iFrom -> currentIdentifiersTo.stream().anyMatch(iFrom::equals))) {
            validationErrors.add(ValidationErrors.NON_UNIQUE_MERGE);
        }
        List<NameVO> currentNamesFrom = personFrom.getNames().stream().filter(n -> n.getValidTo().after(now)).collect(Collectors.toList());
        List<NameVO> currentNamesTo = personTo.getNames().stream().filter(n -> n.getValidTo().after(now)).collect(Collectors.toList());
        if (!currentNamesFrom.stream().allMatch(nFrom -> currentNamesTo.stream().anyMatch(nFrom::equals))) {
            validationErrors.add(ValidationErrors.NON_UNIQUE_MERGE);
        }
        List<GenderVO> currentGendersFrom = personFrom.getGenders().stream().filter(g -> g.getValidTo().after(now)).collect(Collectors.toList());
        List<GenderVO> currentGendersTo = personTo.getGenders().stream().filter(g -> g.getValidTo().after(now)).collect(Collectors.toList());
        if (!currentGendersFrom.stream().allMatch(gFrom -> currentGendersTo.stream().anyMatch(gFrom::equals))) {
            validationErrors.add(ValidationErrors.NON_UNIQUE_MERGE);
        }
        List<BirthDateVO> currentBirthDatesFrom = personFrom.getBirthDates().stream().filter(b -> b.getValidTo().after(now)).collect(Collectors.toList());
        List<BirthDateVO> currentBirthDatesTo = personTo.getBirthDates().stream().filter(b -> b.getValidTo().after(now)).collect(Collectors.toList());
        if (!currentBirthDatesFrom.stream().allMatch(bFrom -> currentBirthDatesTo.stream().anyMatch(bFrom::equals))) {
            validationErrors.add(ValidationErrors.NON_UNIQUE_MERGE);
        }
        List<ContactVO> currentContactsFrom = personFrom.getContacts().stream().filter(c -> c.getValidTo().after(now)).collect(Collectors.toList());
        List<ContactVO> currentContactsTo = personTo.getContacts().stream().filter(c -> c.getValidTo().after(now)).collect(Collectors.toList());
        if (!currentContactsFrom.stream().allMatch(cFrom -> currentContactsTo.stream().anyMatch(cFrom::equals))) {
            validationErrors.add(ValidationErrors.NON_UNIQUE_MERGE);
        }
        List<AddressVO> currentAddressesFrom = personFrom.getAddresses().stream().filter(a -> a.getValidTo().after(now)).collect(Collectors.toList());
        List<AddressVO> currentAddressesTo = personTo.getAddresses().stream().filter(a -> a.getValidTo().after(now)).collect(Collectors.toList());
        if (!currentAddressesFrom.stream().allMatch(aFrom -> currentAddressesTo.stream().anyMatch(aFrom::equals))) {
            validationErrors.add(ValidationErrors.NON_UNIQUE_MERGE);
        }

        validator.throwExceptionInCaseOfValidationErrors(validationErrors);
    }


    public PersonVO mergePersons(PersonVO from, PersonVO to, Modified modified, ServiceActorVO actor, Date now, RequestContextVO context) throws DAOException, IOException {
        PersonVO mergedPerson = new PersonVO();
        mergedPerson.setPersonUUID(to.getPersonUUID());
        List<IdentifierVO> identifiers = from.getIdentifiers().stream().peek(i -> {if (i.getOid().equals(OID.X_ECPR)) { i.setValidTo(now); } }).collect(Collectors.toList());
        identifiers.addAll(to.getIdentifiers());
        List<NameVO> names = from.getNames();
        names.addAll(to.getNames());
        List<GenderVO> genders = from.getGenders();
        genders.addAll(to.getGenders());
        List<BirthDateVO> birthdates = from.getBirthDates();
        birthdates.addAll(to.getBirthDates());
        List<ContactVO> contacts = from.getContacts();
        contacts.addAll(to.getContacts());
        List<AddressVO> addresses = from.getAddresses();
        addresses.addAll(to.getAddresses());
        List<Modified> versions = from.getModifiers();
        versions.addAll(to.getModifiers());

        deletePersonHandler.execute(from.getIdentifiers().stream().filter(i -> i.getOid().equals(OID.X_ECPR)).collect(Collectors.toList()), modified, actor, context);
        personDAO.deletePerson(from.getPersonUUID());

        final int personId = createNewVersionOfPerson(mergedPerson.getPersonUUID(), actor, modified.getBy(), now);
        mergedPerson.setPersonPID(personId);
        mergedPerson.getIdentifiers().addAll(getAllIdentifiers(identifiers, personId, now));
        mergedPerson.getNames().addAll(getAllNames(names, personId, now));
        mergedPerson.getGenders().add(getGender(genders, personId, now));
        mergedPerson.getBirthDates().add(getBirthdate(birthdates, personId, now));
        mergedPerson.getContacts().addAll(getAllContacts(contacts, personId, now));
        mergedPerson.getAddresses().addAll(getAllAddresses(addresses, personId, now));
        mergedPerson.getModifiers().addAll(versions.stream().distinct().collect(Collectors.toList()));

        addModifiers(Collections.singletonList(mergedPerson), false);
        return mergedPerson;
    }

    private List<AddressVO> getAllAddresses(List<AddressVO> addresses, int personId, Date nowTimestamp) throws DAOException {
        if (addresses.isEmpty()) {
            return Collections.emptyList();
        }
        addresses = addresses.stream().distinct().collect(Collectors.toList());
        updateAddresses(addresses, nowTimestamp);
        addresses.forEach(addressPid -> addressPid.setAddressPID(null));
        for (AddressVO address : addresses) {
            addressDAO.createAddress(address, nowTimestamp, personId);
        }
        return addresses;
    }

    private void updateAddresses(List<AddressVO> equalAddresses, Date nowTimestamp) throws DAOException {
        for (AddressVO address : equalAddresses) {
            addressDAO.updateTimeStamp(address, nowTimestamp);
        }
    }

    private List<ContactVO> getAllContacts(List<ContactVO> contacts, int personId, Date nowTimestamp) throws DAOException {
        if (contacts.isEmpty()) {
            return Collections.emptyList();
        }
        contacts = contacts.stream().distinct().collect(Collectors.toList());
        updateContacts(contacts, nowTimestamp);
        contacts.forEach(contactPid -> contactPid.setContactPID(null));
        for (ContactVO contact : contacts) {
            contactDAO.createContact(contact, nowTimestamp, personId);
        }
        return contacts;
    }

    private void updateContacts(List<ContactVO> equalContacts, Date nowTimestamp) throws DAOException {
        for (ContactVO contact : equalContacts) {
            contactDAO.updateTimeStamp(contact, nowTimestamp);
        }
    }

    private BirthDateVO getBirthdate(List<BirthDateVO> birthdates, int personId, Date nowTimestamp) throws DAOException {
        if (birthdates.isEmpty()) {
            return null;
        }
        birthdates = birthdates.stream().distinct().collect(Collectors.toList());
        updateBirthDates(birthdates, nowTimestamp);
        BirthDateVO birthDate = birthdates.get(0);
        birthDate.setBirthDatePID(null);
        birthDateDAO.createBirthDate(birthDate, nowTimestamp, personId);
        return birthDate;
    }

    private void updateBirthDates(List<BirthDateVO> birthdates, Date nowTimestamp) throws DAOException {
        for (BirthDateVO birthdate : birthdates) {
            birthDateDAO.updateTimeStamp(birthdate, nowTimestamp);
        }
    }

    private GenderVO getGender(List<GenderVO> genders, int personId, Date nowTimestamp) throws DAOException {
        if (genders.isEmpty()) {
            return null;
        }
        genders = genders.stream().distinct().collect(Collectors.toList());
        updateGenders(genders, nowTimestamp);
        GenderVO gender = genders.get(0);
        gender.setGenderPID(null);
        genderDAO.createGender(gender, nowTimestamp, personId);
        return gender;
    }

    private void updateGenders(List<GenderVO> genders, Date nowTimestamp) throws DAOException {
        for (GenderVO gender : genders) {
            genderDAO.updateTimeStamp(gender, nowTimestamp);
        }
    }

    private List<NameVO> getAllNames(List<NameVO> names, int personId, Date now) throws DAOException {
        if (names.isEmpty()) {
            return Collections.emptyList();
        }
        names = names.stream().distinct().collect(Collectors.toList());
        updateNames(names, now);
        names.forEach(namePid -> namePid.setNamePID(null));
        for (NameVO name : names) {
            nameDAO.createName(name, now, personId);
        }
        return names;
    }

    private void updateNames(List<NameVO> names, Date now) throws DAOException {
        for (NameVO name : names) {
            nameDAO.updateTimeStamp(name, now);
        }
    }

    public List<IdentifierVO> getAllIdentifiers(List<IdentifierVO> identifiers, int personId, Date now) throws DAOException {
        if (identifiers.isEmpty()) {
            return Collections.emptyList();
        }
        identifiers = identifiers.stream().distinct().collect(Collectors.toList());
        updateIdentifiers(identifiers, now);
        identifiers.forEach(identifierPid -> identifierPid.setIdentifierPID(null));
        for (IdentifierVO identifier : identifiers) {
            identifierDAO.createIdentifier(identifier, now, personId);
        }
        return identifiers;
    }

    private void updateIdentifiers(List<IdentifierVO> xECPRIdentifiers, Date now) throws DAOException {
        for (IdentifierVO identifier : xECPRIdentifiers) {
            identifierDAO.updateTimeStamp(identifier, now);
        }
    }
}
