/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service.handler;

import dk.sds.ecpr2.OID;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.*;
import dk.sds.ecpr2.util.Properties;
import dk.sds.ecpr2.util.TimeHelper;
import dk.sds.ecpr2.webservice.validation.ValidationErrors;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.*;

import static java.time.temporal.ChronoUnit.DAYS;

@Component
public class ReservePersonIdHandler extends AbstractHandler {

    @Autowired
    GetReservedPeopleHandler getReservedPeopleHandler;

    @Autowired
    Properties properties;

    public List<PersonVO> execute(ReserveRequestVO reserveRequest, ServiceActorVO actor, RequestContextVO context) throws DAOException {
        List<PersonVO> persons = new ArrayList<>();

        if (reserveRequest.getNumber() == null) {
            reserveRequest.setNumber(1);
        }
        if (reserveRequest.getReservedUntil() == null) {
            reserveRequest.setReservedUntil(defaultReservedUntil());
        }

        validator.validateReservePersonIdRequest(reserveRequest);

        List<PersonVO> allRecentlyReservedPersons = getReservedPeopleHandler.execute(LocalDateTime.now());
        // Get number of people reserved by reserveRequest.get(0).getModifiedBy().getOrganisation()
        long numberOfRecentlyReservedPersonsByThisOrg = allRecentlyReservedPersons.stream().filter(
                p -> p.getModifiers().get(0).getBy().getOrganisation().getIdentifier().getIdentifier()
                        .equals(reserveRequest.getModifiedBy().getOrganisation().getIdentifier().getIdentifier())
        ).count();
        if (numberOfRecentlyReservedPersonsByThisOrg + reserveRequest.getNumber() > properties.maxReservedPersonIdNumberPrDay) {
            throw new IllegalArgumentException(ValidationErrors.RESERVE_PERSON_ID_TOO_MANY_WITHIN_A_DAY);
        }

        Map<String, Gender> ceCPRs = generateCECPRs(reserveRequest);
        Date nowTimestamp = Date.from(context.getRequestTimeStamp());

        for (String ceCPR : ceCPRs.keySet()) {
            String personUUID = randomUUID();
            int personId = createNewVersionOfPerson(personUUID, actor, reserveRequest.getModifiedBy(), Date.from(reserveRequest.getReservedUntil().toInstant(ZoneOffset.UTC)), nowTimestamp);
            createIdentifier(ceCPR, nowTimestamp, personId, reserveRequest.getOid());
            createGender(personId, nowTimestamp, ceCPRs.get(ceCPR));
            List<PersonVO> createdPerson = crossTableDAO.getPersonsWithRelations(Collections.singletonList(personUUID));
            persons.addAll(createdPerson);
        }

        addModifiers(persons, false);

        return persons;
    }

    private void createGender(int personId, Date nowTimestamp, Gender gender) throws DAOException {
        GenderVO genderVO = new GenderVO();
        genderVO.setValue(gender.toString().toLowerCase());
        genderDAO.createGender(genderVO, nowTimestamp, personId);
    }

    private void createIdentifier(String ceCPR, Date nowTimestamp, int personId, String oid) throws DAOException {
        IdentifierVO identifier = new IdentifierVO();
        identifier.setOid(oid);
        identifier.setValue(ceCPR);
        identifier.setValidity(16);
        identifierDAO.createIdentifier(identifier, nowTimestamp, personId);
    }

    private Map<String, Gender> generateCECPRs(ReserveRequestVO reserveRequest) {
        Map<String, Gender> ceCPRs = new HashMap<>();

        List<LocalDate> dates = generateDatesForCECPRs(reserveRequest);
        for (LocalDate date : dates) {
            Gender actualGender = getGender(reserveRequest.getGender());
            switch (reserveRequest.getOid()) {
                case OID.C_ECPR:
                    String ceCPR = personNumberGeneratorImpl.generateCECPR(date.getDayOfMonth(), date.getMonthValue(), date.getYear(), actualGender, false, ceCPRs.keySet());
                    ceCPRs.put(ceCPR, actualGender);
                    break;
                case OID.X_ECPR:
                    String xeCPR = personNumberGeneratorImpl.generateXECPR(date.getDayOfMonth(), date.getMonthValue(), date.getYear(), actualGender, false, ceCPRs.keySet());
                    ceCPRs.put(xeCPR, actualGender);
                    break;
            }
        }

        return ceCPRs;
    }

    private List<LocalDate> generateDatesForCECPRs(ReserveRequestVO reserveRequest) {
        List<LocalDate> list = new ArrayList<>();
        for (int i=0; i<reserveRequest.getNumber(); i++) {
            LocalDate date = getDateForCECPR(reserveRequest.getForDate(), reserveRequest.getReservedUntil());
            list.add(date);
        }
        return list;
    }

    private Gender getGender(Gender gender) {
        if (gender != null) {
            return gender;
        } else {
            return Math.random() < 0.5 ? Gender.MALE : Gender.FEMALE;
        }
    }

    private LocalDate getDateForCECPR(LocalDate forDate, LocalDateTime reservedUntil) {
        LocalDate dateForCECPR;
        if (forDate != null) {
            dateForCECPR = forDate;
        } else {
            dateForCECPR = randomDateBeforeDate(reservedUntil);
        }
        return dateForCECPR;
    }

    LocalDate randomDateBeforeDate(LocalDateTime date) {
        if (date == null || date.isBefore(LocalDateTime.now())) {
            return LocalDate.now();
        }
        long daysBetween = DAYS.between(LocalDate.now(), date);
        long daysToAdd = (long)(Math.random() * daysBetween + 1);
        return LocalDate.now().plusDays(daysToAdd);
    }

    private LocalDateTime defaultReservedUntil() {
        return LocalDateTime.now().plusDays(2);
    }

}
