/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service.handler;

import dk.sds.ecpr2.KeyIdentifier;
import dk.sds.ecpr2.OID;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.*;
import dk.sds.ecpr2.domain.abstr.PersonDescriptorVO;
import dk.sds.ecpr2.service.CprService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import dk.sds.ecpr2.dao.impl.abstr.AbstractECPRDAOImpl;

import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;

@Component
public class UpdatePersonHandler extends AbstractHandler {

    @Autowired
    CprService cprService;

    public PersonVO execute(PersonVO personVO, ServiceActorVO actor, RequestContextVO context) throws DAOException, IOException {

        validator.validateKeyIdentifier(personVO);
        validator.validateContactHasValue(personVO.getContacts(), new ArrayList<>(), true);

        List<PersonVO> personsWithRelations = crossTableDAO.getPersonsWithRelations(Collections.singletonList(personVO.getPersonUUID()));
        PersonVO existingPerson = mergePersons(personsWithRelations);

        Date now = Date.from(context.getRequestTimeStamp());

        validator.validateUpdatePersonData(personVO, existingPerson, personsWithRelations, now);
        ignoreRedundantRequestData(personVO, existingPerson);

        final int personId = createNewVersionOfPerson(existingPerson.getPersonUUID(), actor, personVO.getModifiers().get(0).getBy(), now);

        for (IdentifierVO vo : personVO.getIdentifiers()) {
            if (vo.getUuid() != null) {
                if (validator.toBeDeleted(vo)) {
                    identifierDAO.delete(vo, now);
                } else {
                    identifierDAO.updateCurrentAndCreateNew(vo, now, personId);
                }
            } else {
                if (vo.getOid().equals(OID.CPR)) {
                    KeyIdentifier cprIdentifier = new KeyIdentifier();
                    cprIdentifier.setValue(vo.getValue());
                    cprIdentifier.setOID(vo.getOid());
                    if (cprService.getCprPerson(cprIdentifier) == null) {
                        throw new IllegalArgumentException("Cpr identifier er ikke valid");
                    }
                }
                identifierDAO.createIdentifier(vo, now, personId);
            }
        }

        if (personVO.getAddresses() != null) {
            for (AddressVO vo : personVO.getAddresses()) {
                if (vo.getUuid() != null) {
                    if (validator.toBeDeleted(vo)) {
                        addressDAO.delete(vo, now);
                    } else {
                        addressDAO.updateCurrentAndCreateNew(vo, now, personId);
                    }
                } else {
                    addressDAO.createAddress(vo, now, personId);
                }
            }
        }

        if (personVO.getBirthDates() != null) {
            for (BirthDateVO vo : personVO.getBirthDates()) {
                if (vo.getUuid() != null) {
                    if (validator.toBeDeleted(vo)) {
                        birthDateDAO.delete(vo, now);
                    } else {
                        birthDateDAO.updateCurrentAndCreateNew(vo, now, personId);
                    }
                } else {
                    birthDateDAO.createBirthDate(vo, now, personId);
                }
            }
        }

        if (personVO.getContacts() != null) {
            for (ContactVO vo : personVO.getContacts()) {
                if (vo.getUuid() != null) {
                    if (validator.toBeDeleted(vo)) {
                        contactDAO.delete(vo, now);
                    } else {
                        contactDAO.updateCurrentAndCreateNew(vo, now, personId);
                    }
                } else {
                    contactDAO.createContact(vo, now, personId);
                }
            }
        }

        if (personVO.getGenders() != null) {
            for (GenderVO vo : personVO.getGenders()) {
                if (vo.getUuid() != null) {
                    if (validator.toBeDeleted(vo)) {
                        genderDAO.delete(vo, now);
                    } else {
                        genderDAO.updateCurrentAndCreateNew(vo, now, personId);
                    }
                } else {
                    genderDAO.createGender(vo, now, personId);
                }
            }
        }

        if (personVO.getNames() != null) {
            for (NameVO vo : personVO.getNames()) {
                if (vo.getUuid() != null) {
                    if (validator.toBeDeleted(vo)) {
                        nameDAO.delete(vo, now);
                    } else {
                        nameDAO.updateCurrentAndCreateNew(vo, now, personId);
                    }
                } else {
                    nameDAO.createName(vo, now, personId);
                }
            }
        }

        List<PersonVO> personList = crossTableDAO.getPersonsWithRelations(Collections.singletonList(personVO.getPersonUUID()));
        PersonVO returnValue = mergePersons(personList);
        searchPersonDAO.updatePersonsInIndex(Collections.singletonList(returnValue));

        return returnValue;
    }

    // Remove request data if it equals the existing data
    private void ignoreRedundantRequestData(PersonVO personVO, PersonVO existingPerson) {
        ignoreRedundantData(personVO.getIdentifiers(), existingPerson.getIdentifiers());
        ignoreRedundantData(personVO.getAddresses(), existingPerson.getAddresses());
        ignoreRedundantData(personVO.getBirthDates(), existingPerson.getBirthDates());
        ignoreRedundantData(personVO.getGenders(), existingPerson.getGenders());
        ignoreRedundantData(personVO.getNames(), existingPerson.getNames());
        ignoreRedundantData(personVO.getContacts(), existingPerson.getContacts());
    }

    private Set<String> getUpdatedIds(List<? extends PersonDescriptorVO> newList, List<? extends PersonDescriptorVO> existingList) {
        Set<String> newIds = newList.stream().map(PersonDescriptorVO::getUuid).collect(Collectors.toSet());
        Set<String> existingIds = existingList.stream().map(PersonDescriptorVO::getUuid).collect(Collectors.toSet());
        newIds.retainAll(existingIds);
        return newIds;
    }

    private void ignoreRedundantData(List<? extends PersonDescriptorVO> newList, List<? extends PersonDescriptorVO> existingList) {
        Set<String> updatedIds = getUpdatedIds(newList, existingList);
        for (String uuid : updatedIds) {
            PersonDescriptorVO newVO = getByUuid(newList, uuid);
            PersonDescriptorVO existingVO = getByUuid(existingList, uuid);
            if (newVO.equals(existingVO)) {
                // Updated value equals existing - skip updating it
                newList.remove(newVO);
            }
        }
    }

    private <T extends PersonDescriptorVO> T getByUuid(List<T> list, String uuid) {
        return list.stream().filter(element -> uuid.equals(element.getUuid())
                && (element.getValidTo() == null || !element.getValidTo().before(new Date()))).findFirst().orElse(null);
    }
}
