/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service.handler;

import dk.sds.ecpr2.Modified;
import dk.sds.ecpr2.OID;
import dk.sds.ecpr2.PredefinedOIDType;
import dk.sds.ecpr2.dao.IdentifierDAO;
import dk.sds.ecpr2.dao.OIDFacade;
import dk.sds.ecpr2.dao.PersonDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.*;
import dk.sds.ecpr2.domain.abstr.PersonDescriptorVO;
import dk.sds.ecpr2.util.Properties;
import dk.sds.ecpr2.util.ValueSets;
import dk.sds.ecpr2.webservice.validation.ValidationErrors;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.util.StringUtils;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

@Log4j2
@Component
public class Validator {

    @Autowired
    private IdentifierDAO identifierDAO;

    @Autowired
    private OIDFacade oidFacade;

    @Autowired
    private PersonDAO personDAO;

    @Autowired
    Properties properties;

    public void validateCreatePersonData(PersonVO person) throws DAOException {
        List<String> validationErrors = new ArrayList<>();

        validatePidNotProvided(person, validationErrors);
        validateOids(person.getIdentifiers(), validationErrors);
        validateOidDuplicates(person.getIdentifiers(), null, validationErrors);
        validateCreateIdentifiers(person.getIdentifiers(), validationErrors);
        validateGenderProvided(person.getGenders(), validationErrors);
        validateContactHasValue(person.getContacts(), validationErrors, false);
        validateModifier(person.getModifiers().get(0), validationErrors); // Always one version in requests

        throwExceptionInCaseOfValidationErrors(validationErrors);
    }

    public void validateUpdatePersonData(PersonVO person, PersonVO existingPerson, List<PersonVO> personsWithRelations, Date nowTimestamp) throws DAOException {
        List<String> validationErrors = new ArrayList<>();

        validateOids(person.getIdentifiers(), validationErrors);
        validateOidDuplicates(person.getIdentifiers(), personsWithRelations, validationErrors);
        validateIdentifiersHaveValue(person.getIdentifiers(), validationErrors);
        validatePidExistence(person, existingPerson, validationErrors);
        validateDuplicateNames(person.getNames(), existingPerson.getNames(), validationErrors, ValidationErrors.DUPLICATE_NAME_USE, nowTimestamp);
        validateDuplicates(person.getBirthDates(), existingPerson.getBirthDates(), validationErrors, ValidationErrors.DUPLICATE_BIRTHDATE, nowTimestamp);
        validateDuplicates(person.getGenders(), existingPerson.getGenders(), validationErrors, ValidationErrors.DUPLICATE_GENDER, nowTimestamp);

        throwExceptionInCaseOfValidationErrors(validationErrors);
    }

    public void validateReservePersonIdRequest(ReserveRequestVO reserveRequest) {
        List<String> validationErrors = new ArrayList<>();
        if (reserveRequest.getForDate() != null && reserveRequest.getForDate().isBefore(LocalDate.now())) {
            validationErrors.add(format(ValidationErrors.FOR_DATE_NON_FUTURE));
        }
        LocalDateTime inTwoYears = LocalDateTime.now().plusYears(2);
        if (reserveRequest.getReservedUntil().isAfter(inTwoYears) || reserveRequest.getReservedUntil().isBefore(LocalDateTime.now())) {
            validationErrors.add(format(ValidationErrors.RESERVED_UNTIL_OVER_THRESHOLD));
        }
        if (reserveRequest.getNumber() > properties.maxReservedPersonIdNumber) {
            validationErrors.add(format(ValidationErrors.RESERVE_PERSON_ID_NUMBER_OVER_THRESHOLD, "" + reserveRequest.getNumber()));
        }
        if (!oidCreationSupported(reserveRequest.getOid())) {
            validationErrors.add(format(ValidationErrors.CREATEPERSONID_INVALID_OID));
        }
        throwExceptionInCaseOfValidationErrors(validationErrors);
    }

    public void validateContactHasValue(List<ContactVO> contacts, List<String> validationErrors, boolean throwInCaseOfError) {
        if (contacts == null) {
            return;
        }
        for (ContactVO contact : contacts) {
            if (contact.getUuid() == null && (contact.getValue() == null || "".equals(contact.getValue()))) {
                validationErrors.add(ValidationErrors.CONTACT_VALUE_EMPTY);
                if (throwInCaseOfError) {
                    this.throwExceptionInCaseOfValidationErrors(validationErrors);
                }
                return;
            }
        }
    }

    public void validateDeletePersonVersion(Modified modified) throws DAOException {
        List<String> validationErrors = new ArrayList<>();
        validateModifier(modified, validationErrors); // Always one version in requests
        throwExceptionInCaseOfValidationErrors(validationErrors);
    }

    private void validateModifier(Modified modified, List<String> validationErrors) {
        if (modified == null) {
            validationErrors.add(ValidationErrors.MODIFIER_REQUIRED);
        } else if (modified.getValidTo() != null || modified.getReservedUntil() != null) {
            validationErrors.add(ValidationErrors.MODIFIER_DATES_NOT_ALLOWED);
        } else if (modified.getBy() == null) {
            validationErrors.add(ValidationErrors.MODIFIED_BY_REQUIRED);
        }
    }

    private void validateGenderProvided(List<GenderVO> genders, List<String> validationErrors) {
        if (genders == null || genders.size() != 1) {
            validationErrors.add(ValidationErrors.GENDER_VALUE_REQUIRED);
            return;
        }
        checkValue(genders.get(0).getValue(), ValueSets.genderValues, ValidationErrors.GENDER_VALUE_REQUIRED, validationErrors);
    }

    private void checkValue(String value, String[] legalValues, String validationError, List<String> validationErrors) {
        for (String legalValue : legalValues) {
            if (legalValue.equals(value)) {
                return;
            }
        }
        validationErrors.add(validationError);
    }

    private void validateCreateIdentifiers(List<IdentifierVO> identifiers, List<String> validationErrors) {
        if (identifiers == null) {
            return;
        }
        for (IdentifierVO identifier : identifiers) {
            if (OID.X_ECPR.equals(identifier.getOid())) {
                if (identifier.getValue() != null) {
                    validationErrors.add(format(ValidationErrors.VALUE_XECPR_FORBIDDEN, identifier.getValue()));
                }
            } else if (OID.C_ECPR.equals(identifier.getOid())) {
                if (identifier.getValue() != null) {
                    validationErrors.add(format(ValidationErrors.VALUE_CECPR_FORBIDDEN, identifier.getValue()));
                }
            } else {
                validateIdentifierHasValue(identifier, validationErrors);
            }
        }
    }

    private void validateOids(List<IdentifierVO> identifiers, List<String> validationErrors) throws DAOException {
        if (identifiers == null) {
            return;
        }
        for (IdentifierVO identifier : identifiers) {
            if (toBeDeleted(identifier)) {
                // Delete situation where only PID/UUID is supplied. Don't validate contents
                continue;
            }
            OIDVO oid = oidFacade.getByOIDValue(identifier.getOid());
            if (oid == null) {
                validationErrors.add(format(ValidationErrors.UNKNOWN_OID, identifier.getOid()));
            } else {
                boolean weSupportCreationOfOID = oidCreationSupported(oid.getOid());
                if (!weSupportCreationOfOID || identifier.getValue() != null) {
                    // Null is ok for OIDs if we support creation
                    checkRegularExpression(oid, identifier.getValue(), validationErrors);
                }
            }
        }
    }

    public boolean oidCreationSupported(String oid) {
        return OID.X_ECPR.equals(oid) || OID.C_ECPR.equals(oid);
    }

    private void checkRegularExpression(OIDVO oid, String value, List<String> validationErrors) {
        if (StringUtils.isEmpty(oid.getRegex())) {
            return;
        }
        Pattern pattern = Pattern.compile(oid.getRegex());
        Matcher matcher = pattern.matcher(value);
        if (!matcher.matches()) {
            validationErrors.add(format(ValidationErrors.BAD_FORMAT_OID, value, oid.getPrimaryOidLabel()));
        }
    }

    // Duplicate OID/value check
    private void validateOidDuplicates(List<IdentifierVO> identifiers, List<PersonVO> existingPersonRecords, List<String> validationErrors) throws DAOException {
        if (identifiers == null) {
            return;
        }
        for (IdentifierVO identifier : identifiers) {
            if (identifier.getUuid() != null) {
                // It already exists in the database - no validation to perform
                continue;
            }
            List<IdentifierWithoutServiceActorVO> existingIdentifiers = identifierDAO.getCurrentIdentifiersByValueAndOID(identifier.getValue(), identifier.getOid());
            if (existingIdentifiers != null && existingIdentifiers.size() > 0) {
                if (existingPersonRecords == null) {
                    // Identifiers must be for another person - not allowed
                    validationErrors.add(format(ValidationErrors.EXISTING_IDENTIFIER, identifier.getValue(), identifier.getOid()));
                } else {
                    for (IdentifierWithoutServiceActorVO existingIdentifier : existingIdentifiers) {
                        for (PersonVO person : existingPersonRecords) {
                            if (existingIdentifier.getPersonPID().equals(person.getPersonPID())) {
                                validationErrors.add(format(ValidationErrors.EXISTING_IDENTIFIER, identifier.getValue(), identifier.getOid()));
                            }
                        }
                    }
                }
            }
        }
    }

    public void throwExceptionInCaseOfValidationErrors(List<String> validationErrors) {
        if (validationErrors.size() > 0) {
            throw new IllegalArgumentException(String.join(", ", validationErrors));
        }
    }

    public String format(String error, String... expressions) {
        if (expressions != null) {
            for (int i = 0; i < expressions.length; i++) {
                if (expressions[i] == null) {
                    expressions[i] = "";
                }
            }
        }
        return String.format(error, expressions);
    }

    public void validatePidNotProvided(PersonVO person, List<String> validationErrors) {
        validatePidNotProvided(person.getAddresses(), validationErrors);
    }

    private void validatePidNotProvided(List<? extends PersonDescriptorVO> list, List<String> validationErrors) {
        if (list == null) {
            return;
        }
        for (PersonDescriptorVO vo : list) {
            if (vo.getUuid() != null) {
                validationErrors.add(format(ValidationErrors.PID_FORBIDDEN, vo.getUuid()));
            }
        }
    }

    private void validateIdentifiersHaveValue(List<IdentifierVO> identifiers, List<String> validationErrors) {
        for (IdentifierVO identifier : identifiers) {
            validateIdentifierHasValue(identifier, validationErrors);
        }
    }

    private void validateIdentifierHasValue(IdentifierVO identifier, List<String> validationErrors) {
        if (!toBeDeleted(identifier)) {
            if (StringUtils.isEmpty(identifier.getValue())) {
                try {
                    OIDVO oid = oidFacade.getByOIDValue(identifier.getOid());
                    if (oid != null) {
                        PredefinedOIDType predefinedOIDType = PredefinedOIDType.fromValue(oid.getType());
                        String errorMessage;
                        if (PredefinedOIDType.EKSTERN_PERSON_IDENTIFIKATION_OG_LAND == predefinedOIDType) {
                            errorMessage = oid.getPrimaryOidLabel() + ", " + oid.getSecondaryOidLabel();
                        } else {
                            errorMessage = oid.getSecondaryOidLabel() != null
                                    ? oid.getSecondaryOidLabel()
                                    : oid.getPrimaryOidLabel();
                        }
                        validationErrors.add(format(ValidationErrors.IDENTIFIER_HAS_NO_VALUE, errorMessage));
                    } else {
                        validationErrors.add(format(ValidationErrors.UNKNOWN_OID, identifier.getOid()));
                    }
                } catch (DAOException | IllegalArgumentException e) {
                    validationErrors.add(format(ValidationErrors.IDENTIFIER_HAS_NO_VALUE, identifier.getOid()));
                    log.warn("Der skete en fejl mens vi prøvede at generere fejlbeskeden", e);
                }
            }
        }
    }

    private void validateDuplicates(List<? extends PersonDescriptorVO> newList, List<? extends PersonDescriptorVO> existingList, List<String> validationErrors, String error, Date now) {
        int activeCount = (int) existingList.stream()
                .filter(b -> !b.getValidFrom().after(now) && b.getValidTo().after(now)).count();
        int newCount = (int) newList.stream().filter(b -> b.getUuid() == null).count();
        if (newCount + activeCount > 1) {
            validationErrors.add(error);
        }
    }

    private void validateDuplicateNames(List<NameVO> newList, List<NameVO> existingList, List<String> validationErrors, String error, Date now) {
        for (String use : ValueSets.nameUseValues) {
            int activeCount = (int) existingList.stream()
                    .filter(b -> !b.getValidFrom().after(now) && (b.getValidTo().after(now) && (b.getExpiry() == null || b.getExpiry().after(now))) && use.equals(b.getUse())).count();
            int newCount = (int) newList.stream().filter(b -> b.getUuid() == null && use.equals(b.getUse())).count();
            if (newCount + activeCount > 1) {
                validationErrors.add(format(error, use));
            }
        }
    }

    private void validatePidExistence(PersonVO newPerson, PersonVO existingPerson, List<String> validationErrors) {
        checkPidExistence(newPerson.getAddresses(), existingPerson.getAddresses(), validationErrors);
        checkPidExistence(newPerson.getBirthDates(), existingPerson.getBirthDates(), validationErrors);
        checkPidExistence(newPerson.getGenders(), existingPerson.getGenders(), validationErrors);
        checkPidExistence(newPerson.getNames(), existingPerson.getNames(), validationErrors);
        checkPidExistence(newPerson.getIdentifiers(), existingPerson.getIdentifiers(), validationErrors);
        checkPidExistence(newPerson.getContacts(), existingPerson.getContacts(), validationErrors);
    }

    private void checkPidExistence(List<? extends PersonDescriptorVO> newList, List<? extends PersonDescriptorVO> existingList, List<String> validationErrors) {
        Set<String> newIds = newList.stream().map(PersonDescriptorVO::getUuid).filter(Objects::nonNull).collect(Collectors.toSet());
        Set<String> existingIds = existingList.stream().map(PersonDescriptorVO::getUuid).filter(Objects::nonNull).collect(Collectors.toSet());
        for (String s : newIds) {
            if (!existingIds.contains(s)) {
                validationErrors.add(format(ValidationErrors.UNKNOWN_PID, s));
            }
        }
    }

    public String validateKeyIdentifier(KeyIdentifierVO keyIdentifier) throws DAOException {
        List<String> validationErrors = new ArrayList<>();
        IdentifierWithoutServiceActorVO identifier;
        if (keyIdentifier.getOid().equals(OID.X_ECPR) || keyIdentifier.getOid().equals(OID.C_ECPR)) {
            // For x-ecpr and c-ecpr identifiers old values are also searched for. If multiple versions are found, the newest should be used
            identifier = identifierDAO.getIdentifiersByValueAndOID(keyIdentifier.getValue(), keyIdentifier.getOid()).stream().max((i1, i2) -> {
                if (i1.getValidTo() == null && i2.getValidTo() == null) {
                    return 0;
                }
                if (i1.getValidTo() == null) {
                    return 1;
                }
                if (i2.getValidTo() == null) {
                    return -1;
                }
                return i1.getValidTo().compareTo(i2.getValidTo());
            }).orElse(null);
        } else {
            // For other identifiers only current values are searched for
            identifier = identifierDAO.getCurrentIdentifierByKeyIdentifier(keyIdentifier);
        }
        if (identifier == null) {
            validationErrors.add(format(ValidationErrors.KEY_IDENTIFIER_INVALID, keyIdentifier == null ? "" : keyIdentifier.getValue()));
        } else {
            PersonVO personFromDatabase = personDAO.getPersonByPID(identifier.getPersonPID());
            if (personFromDatabase == null) {
                validationErrors.add(format(ValidationErrors.KEY_IDENTIFIER_INVALID, keyIdentifier == null ? "" : keyIdentifier.getValue()));
            } else {
                // Side effect of validation: setting UUID on provided PersonVO
                return personFromDatabase.getPersonUUID();
            }
        }
        throwExceptionInCaseOfValidationErrors(validationErrors);
        return null;
    }

    public void validateKeyIdentifier(PersonVO person) throws DAOException {
        List<String> validationErrors = new ArrayList<>();
        KeyIdentifierVO keyIdentifier = person.getKeyIdentifier();
        IdentifierWithoutServiceActorVO identifier;
        if (keyIdentifier.getOid().equals(OID.X_ECPR) || keyIdentifier.getOid().equals(OID.C_ECPR)) {
            // For x-ecpr and c-ecpr identifiers old values are also searched for. If multiple versions are found, the newest should be used
            identifier = identifierDAO.getIdentifiersByValueAndOID(keyIdentifier.getValue(), keyIdentifier.getOid()).stream().max((i1, i2) -> {
                if (i1.getValidTo() == null && i2.getValidTo() == null) {
                    return 0;
                }
                if (i1.getValidTo() == null) {
                    return 1;
                }
                if (i2.getValidTo() == null) {
                    return -1;
                }
                return i1.getValidTo().compareTo(i2.getValidTo());
            }).orElse(null);
        } else {
            // For other identifiers only current values are searched for
            identifier = identifierDAO.getCurrentIdentifierByKeyIdentifier(keyIdentifier);
        }
        if (identifier == null) {
            validationErrors.add(format(ValidationErrors.KEY_IDENTIFIER_INVALID, keyIdentifier == null ? "" : keyIdentifier.getValue()));
        } else {
            PersonVO personFromDatabase = personDAO.getPersonByPID(identifier.getPersonPID());
            if (personFromDatabase == null) {
                validationErrors.add(format(ValidationErrors.KEY_IDENTIFIER_INVALID, keyIdentifier == null ? "" : keyIdentifier.getValue()));
            } else {
                // Side effect of validation: setting UUID on provided PersonVO
                person.setPersonUUID(personFromDatabase.getPersonUUID());
            }
        }
        throwExceptionInCaseOfValidationErrors(validationErrors);
    }

    private boolean onlyUuidIsSet(PersonDescriptorVO vo) {
        return vo.getUuid() != null &&
                vo.getValidTo() == null &&
                vo.getValidFrom() == null;
    }

    public boolean toBeDeleted(IdentifierVO identifier) {
        return onlyUuidIsSet(identifier) &&
                identifier.getOid() == null &&
                (identifier.getValue() == null || identifier.getValue().isEmpty()) &&
                identifier.getValidity() == null &&
                identifier.getExpiry() == null;
    }

    public boolean toBeDeleted(AddressVO address) {
        return onlyUuidIsSet(address) &&
                address.getUse() == null &&
                address.getCountry() == null &&
                address.getExpiry() == null &&
                address.getCity() == null &&
                address.getDistrict() == null &&
                address.getPostalCode() == null &&
                address.getPostBox() == null &&
                address.getState() == null &&
                address.getType() == null;
    }

    public boolean toBeDeleted(BirthDateVO birthDate) {
        return onlyUuidIsSet(birthDate) &&
                birthDate.getYear() == null;
    }

    public boolean toBeDeleted(ContactVO contact) {
        return onlyUuidIsSet(contact) &&
                contact.getUse() == null &&
                contact.getSystem() == null &&
                (contact.getValue() == null || contact.getValue().isEmpty()) &&
                contact.getExpiry() == null;
    }

    public boolean toBeDeleted(GenderVO gender) {
        return onlyUuidIsSet(gender) && (gender.getValue() == null || gender.getValue().isEmpty());
    }

    public boolean toBeDeleted(NameVO name) {
        return onlyUuidIsSet(name) &&
                name.getUse() == null &&
                name.getFamilyName() == null &&
                name.getGivenName() == null &&
                name.getExpiry() == null;
    }
}
