/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service.historyExport;

import dk.sds.ecpr2.dao.ExportHistoryDAO;
import dk.sds.ecpr2.dao.OIDDAO;
import dk.sds.ecpr2.dao.OIDFacade;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.ExportCrossTableDAOImpl;
import dk.sds.ecpr2.domain.*;
import dk.sds.ecpr2.util.FaultCode;
import dk.sds.ecpr2.util.FaultUtil;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

import java.io.File;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;

@Service
@Log4j2
public class ExportHistoryServiceImpl {
    @Autowired
    protected ExportCrossTableDAOImpl crossTableDAO;

    @Autowired
    protected OIDFacade oidFacade;

    @Autowired
    protected OIDDAO oidDao;

    @Autowired
    protected ExportHistoryDAO exportHistoryDAO;

    @Autowired
    protected NspManager nspManager;

    @Value("${jobs.exportHistoryToFile.initialLoad.batchSize:20000}")
    private int batchSize;

    @Value("${jobs.exportHistoryToFile.initialLoad.msBetweenBatches:500}")
    private int SLEEP_TIME_MS_BETWEEN_BATCHES;


    private double nextPersonVersion = 1.0;
    private double nextIdentifierVersion = 1.0;
    private double nextNameVersion = 1.0;
    private double nextGenderVersion = 1.0;
    private double nextBirthdateVersion = 1.0;
    private double nextAddressVersion = 1.0;
    private double nextContactVersion = 1.0;
    private double nextOIDVersion = 1.0;

    private Instant date = null;
    private Integer exportNumber = null;
    private static final String DATETIME_FORMAT = "YYYY-MM-dd HH:mm:ss.SSS";


    public File exportHistoryToFiles(Instant previousExportTime, int exportNumber) throws InterruptedException {
        log.debug("Starting to fetch data for history export... Last fetched timestamp: " + previousExportTime);
        this.exportNumber = exportNumber;
        final boolean isInitialLoad = (previousExportTime == null);
        try {
            if (isInitialLoad) {
                previousExportTime = Instant.now();
            }
            List<PersonVO> personList = loadDatabaseChanges(previousExportTime, isInitialLoad);
            if (CollectionUtils.isEmpty(personList)) {
                log.debug("No changes in database");
                return null;
            }
            List<OIDVO> oids = fetchAllOids();
            log.debug("Done fetching data from database.");
            return exportDatabaseChanges(personList, oids, previousExportTime, exportNumber, isInitialLoad);
        } catch (DAOException | IOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.FECTH_DATABASECHANGES_TO_EXPORT_INTERNAL_ERROR, e);
            return null;
        }
    }

    private List<PersonVO> loadDatabaseChanges(Instant previousExportTime, boolean isInitialLoad) throws DAOException, InterruptedException {
        List<PersonVO> personList;
        if (isInitialLoad) {
            personList = fetchCompleteDatabase();
        } else {
            personList = fetchDatabaseChanges(previousExportTime);
        }
        return personList;
    }

    private List<PersonVO> fetchDatabaseChanges(Instant previousExportTime) throws DAOException {
        List<PersonVO> deltaPersons = crossTableDAO.getPersonsWithRelationsAfterModifiedTime(previousExportTime, false);
        log.debug("Done fetching data incrementally for exporting history (" + deltaPersons.size() + " rows). Exporting to files...");
        return deltaPersons;
    }

    private List<PersonVO> fetchCompleteDatabase() throws DAOException, InterruptedException {
        List<PersonVO> personInfoBatch;
        List<PersonVO> completePersonList = new ArrayList<>();
        int minPersonPID = 0;
        do {
            personInfoBatch = crossTableDAO.getPersonsWithRelationsAfterPID(minPersonPID, batchSize, true, false);
            log.debug("Fetched " + personInfoBatch.size() + " rows for exporting history. First person PID: " + minPersonPID);
            completePersonList.addAll(personInfoBatch);
            minPersonPID = PersonVO.getMaxPersonPID(personInfoBatch) + 1;
            Thread.sleep(SLEEP_TIME_MS_BETWEEN_BATCHES); // Don't overload the machine when loading large amounts of data.
        } while (personInfoBatch.size() > 0);
        return completePersonList;
    }

    private List<OIDVO> fetchAllOids() throws DAOException {
        return oidDao.getAllOIDs();
    }

    private File exportDatabaseChanges(List<PersonVO> personVOList, List<OIDVO> oids, Instant previousExportTime, int exportNumber, boolean isInitialLoad) throws IOException {
        File folder = createPersonFile(personVOList, previousExportTime, exportNumber);
        createIdentifierFile(personVOList, previousExportTime, exportNumber, isInitialLoad);
        createNameFile(personVOList, previousExportTime, exportNumber, isInitialLoad);
        createGenderFile(personVOList, previousExportTime, exportNumber, isInitialLoad);
        createBirthdateFile(personVOList, previousExportTime, exportNumber, isInitialLoad);
        createContactFile(personVOList, previousExportTime, exportNumber, isInitialLoad);
        createAddressFile(personVOList, previousExportTime, exportNumber, isInitialLoad);
        createOIDFile(oids, previousExportTime, exportNumber);

        return createCompletedZipFile(folder);
    }

    private File createPersonFile(List<PersonVO> personVOList, Instant previousExportTime, int exportNumber) {
        List<PersonVO> personList = new ArrayList<>();
        for (PersonVO person : personVOList) {
            PersonVO personVO = new PersonVO();
            personVO.setPersonPID(person.getPersonPID());
            personVO.setPersonUUID(person.getPersonUUID());
            personVO.setVersion(person.getVersion());
            personVO.setModifiedTime(person.getModifiedTime());
            personVO.setVersionDescription(person.getVersionDescription());
            personList.add(personVO);
        }
        ExportHistoryDataSetVO dataSet = exportHistoryDataSet("person", exportNumber, Date.from(previousExportTime), nextPersonVersion);
        File folder = exportBatch(personList, "persons", dataSet, previousExportTime);
        nextPersonVersion++;
        return folder;
    }

    private void createAddressFile(List<PersonVO> personVOList, Instant previousExportTime, int exportNumber, boolean isInitialLoad) {
        List<AddressVO> addressList;
        List<AddressVO> addressVOList = personVOList
                .stream()
                .map(PersonVO::getAddresses)
                .flatMap(Collection::stream)
                .collect(Collectors.toList());
        if (isInitialLoad) {
            addressList = addressVOList;
        } else {
            addressList = addressVOList
                    .stream()
                    .filter(addressVO -> addressVO.getModified().after(Date.from(previousExportTime)))
                    .collect(Collectors.toList());
        }
        if (CollectionUtils.isEmpty(addressList)) {
            log.debug("No changes to address table in database");
        }
        ExportHistoryDataSetVO dataSet = exportHistoryDataSet("address", exportNumber, Date.from(previousExportTime), nextAddressVersion);
        exportBatch(addressList, "addresses", dataSet, previousExportTime);
        nextAddressVersion++;
    }

    private void createContactFile(List<PersonVO> personVOList, Instant previousExportTime, int exportNumber, boolean isInitialLoad) {
        List<ContactVO> contactList;
        List<ContactVO> contactVOList = personVOList
                .stream()
                .map(PersonVO::getContacts)
                .flatMap(Collection::stream)
                .collect(Collectors.toList());
        if (isInitialLoad) {
            contactList = contactVOList;
        } else {
            contactList = contactVOList
                    .stream()
                    .filter(contactVO -> contactVO.getModified().after(Date.from(previousExportTime)))
                    .collect(Collectors.toList());
        }
        if (CollectionUtils.isEmpty(contactList)) {
            log.debug("No changes to contact table in database");
        }
        ExportHistoryDataSetVO dataSet = exportHistoryDataSet("contact", exportNumber, Date.from(previousExportTime), nextContactVersion);
        exportBatch(contactList, "contacts", dataSet, previousExportTime);
        nextContactVersion++;
    }

    private void createBirthdateFile(List<PersonVO> personVOList, Instant previousExportTime, int exportNumber, boolean isInitialLoad) {
        List<BirthDateVO> birthDateList;
        List<BirthDateVO> birthDateVOList = personVOList
                .stream()
                .map(PersonVO::getBirthDates)
                .flatMap(Collection::stream)
                .collect(Collectors.toList());
        if (isInitialLoad) {
            birthDateList = birthDateVOList;
        } else {
            birthDateList = birthDateVOList
                    .stream()
                    .filter(birthDateVO -> birthDateVO.getModified().after(Date.from(previousExportTime)))
                    .collect(Collectors.toList());
        }
        if (CollectionUtils.isEmpty(birthDateList)) {
            log.debug("No changes to birthdate table in database");
        }
        ExportHistoryDataSetVO dataSet = exportHistoryDataSet("birthdate", exportNumber, Date.from(previousExportTime), nextBirthdateVersion);
        exportBatch(birthDateList, "birthdates", dataSet, previousExportTime);
        nextBirthdateVersion++;
    }

    private void createGenderFile(List<PersonVO> personVOList, Instant previousExportTime, int exportNumber, boolean isInitialLoad) {
        List<GenderVO> genderList;
        List<GenderVO> genderVOList = personVOList
                .stream()
                .map(PersonVO::getGenders).flatMap(Collection::stream)
                .collect(Collectors.toList());
        if (isInitialLoad) {
            genderList = genderVOList;
        } else {
            genderList = genderVOList
                    .stream()
                    .filter(birthDateVO -> birthDateVO.getModified().after(Date.from(previousExportTime)))
                    .collect(Collectors.toList());
        }
        if (CollectionUtils.isEmpty(genderList)) {
            log.debug("No changes to gender table in database");
        }
        ExportHistoryDataSetVO dataSet = exportHistoryDataSet("gender", exportNumber, Date.from(previousExportTime), nextGenderVersion);
        exportBatch(genderList, "genders", dataSet, previousExportTime);
        nextGenderVersion++;
    }

    private void createNameFile(List<PersonVO> personVOList, Instant previousExportTime, int exportNumber, boolean isInitialLoad) {
        List<NameVO> nameList;
        List<NameVO> nameVOList = personVOList
                .stream()
                .map(PersonVO::getNames)
                .flatMap(Collection::stream)
                .collect(Collectors.toList());
        if (isInitialLoad) {
            nameList = nameVOList;
        } else {
            nameList = nameVOList
                    .stream()
                    .filter(birthDateVO -> birthDateVO.getModified().after(Date.from(previousExportTime)))
                    .collect(Collectors.toList());
        }
        if (CollectionUtils.isEmpty(nameList)) {
            log.debug("No changes to name table in database");
        }
        ExportHistoryDataSetVO dataSet = exportHistoryDataSet("name", exportNumber, Date.from(previousExportTime), nextNameVersion);
        exportBatch(nameList, "names", dataSet, previousExportTime);
        nextNameVersion++;
    }

    private void createIdentifierFile(List<PersonVO> personVOList, Instant previousExportTime, int exportNumber, boolean isInitialLoad) {
        List<IdentifierVO> identifierList;
        List<IdentifierVO> identifierVOList = personVOList
                .stream()
                .map(PersonVO::getIdentifiers)
                .flatMap(Collection::stream)
                .collect(Collectors.toList());
        if (isInitialLoad) {
            identifierList = identifierVOList;
        } else {
            identifierList = identifierVOList
                    .stream()
                    .filter(identifierVO -> identifierVO.getModified().after(Date.from(previousExportTime)))
                    .collect(Collectors.toList());
        }
        if (CollectionUtils.isEmpty(identifierList)) {
            log.debug("No changes to identifier table in database");
        }
        ExportHistoryDataSetVO dataSet = exportHistoryDataSet("identifier", exportNumber, Date.from(previousExportTime), nextIdentifierVersion);
        exportBatch(identifierList, "identifiers", dataSet, previousExportTime);
        nextIdentifierVersion++;
    }

    private void createOIDFile(List<OIDVO> oids, Instant previousExportTime, int exportNumber) {
        ExportHistoryDataSetVO dataSet = exportHistoryDataSet("oid", exportNumber, Date.from(previousExportTime), nextOIDVersion);
        exportBatch(oids, "oids", dataSet, previousExportTime);
        nextOIDVersion++;
    }

    private ExportHistoryDataSetVO exportHistoryDataSet(String dataType, int exportNumber, Date previousExportTime, double version) {
        SimpleDateFormat dateFormat = new SimpleDateFormat(DATETIME_FORMAT);
        String date = dateFormat.format(previousExportTime);
        ExportHistoryDataSetVO exportHistoryDataSetVO = new ExportHistoryDataSetVO();
        exportHistoryDataSetVO.setDatatype(dataType);
        exportHistoryDataSetVO.setExportNumber(exportNumber);
        exportHistoryDataSetVO.setExportTime(date);
        exportHistoryDataSetVO.setVersion(version);
        return exportHistoryDataSetVO;
    }

    private <T> File exportBatch(List<T> obj, String dataType, ExportHistoryDataSetVO dataSet, Instant previousExportTime) {
        ExportHistoryFileFormat<T> exportHistoryFileFormat = new ExportHistoryFileFormat<T>();
        exportHistoryFileFormat.setData(obj);
        exportHistoryFileFormat.setDataset(dataSet);
        return nspManager.saveToTempFile(exportHistoryFileFormat, dataType, Date.from(previousExportTime));
    }

    private File createCompletedZipFile(File folder) throws IOException {
        return nspManager.createCompletedFile(folder);
    }

    public void updateExportHistoryTable(Instant date) {
        try {
            log.debug("Updating export history table in db.");
            ExportHistoryVO exportHistoryVO = exportHistoryDAO.getLatestExport();
            final int latestExportNumber = (exportHistoryVO == null || exportHistoryVO.getExportNumber() == null) ? 1 : exportHistoryVO.getExportNumber() + 1;
            exportHistoryDAO.dataChangedToExport(latestExportNumber, date);
        } catch (DAOException e) {
            FaultUtil.throwECPR2Fault(FaultCode.EXPORT_HISTORY_INTERNAL_SERVER_ERROR, e);
        }
    }
}
