/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service.historyExport;

import lombok.extern.log4j.Log4j2;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.comparator.LastModifiedFileComparator;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.io.FileSystemResource;
import org.springframework.stereotype.Component;

import java.io.*;
import java.nio.file.*;
import java.nio.file.attribute.BasicFileAttributes;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

@Log4j2
@Component
public class NspManagerSFtp implements NspManager {
    @Value("${temp.dir}")
    String tempDirLocation;
    @Value("${ziptemp.dir}")
    String tempZipDirLocation;
    @Value("${temp.archive.dir}")
    String archiveDirectory;
    @Value("${completedFileName}")
    String completedFileName;
    @Value("${sftp.enabled}")
    boolean uploadEnabled = true;
    @Value("${sftp.remote.path}")
    String remoteFolder;

    @Autowired
    SimpleJsonFormat simpleJsonFormat;

    @Autowired
    public
    SFtpClient sftpClient;

    private File folderName;
    private File zipFileName = null;

    public static List<File> foldersToUpload = new ArrayList<>();

    @Override
    public File saveToTempFile(ExportHistoryFileFormat exportHistoryFileFormat, String dataType, Date startTime) {
        String jsonFormat = simpleJsonFormat.jsonFormat(exportHistoryFileFormat);
        final String fileName = dataType + ".json"; //+ dateFormat.format(startTime) + ".json";
        File folder = writeToTempDir(fileName, jsonFormat, startTime);
        log.debug("Saving " + fileName + " to folder " + folder.getAbsolutePath());
        return folder;
    }

    private File writeToTempDir(String filename, String data, Date startTime) {
        SimpleDateFormat dateFormat = new SimpleDateFormat("'_'yyyy-MM-dd'_'HH-mm-ss");
        folderName = new File(tempDirLocation, "ExportHistory" + dateFormat.format(startTime));
        folderName.mkdir();
        final File file = new File(folderName.getAbsolutePath(), filename);
        try {
            FileUtils.writeStringToFile(file, data, "UTF-8");
        } catch (IOException e) {
            throw new RuntimeException("Could not write to file=" + file.getAbsolutePath(), e);
        }
        return folderName;
    }

    @Override
    public File createCompletedFile(File folder) throws IOException {
        final File completedFile = new File(folder.getAbsolutePath(), completedFileName);
        String zipPathLocation = createCorrectPathFormat(tempDirLocation) + tempZipDirLocation;
        File zipFile = new File(zipPathLocation);
        this.zipFileName = zipFile;
        if (!this.zipFileName.exists()) {
            boolean makeDir = this.zipFileName.mkdir();
            if (makeDir) {
                log.info("Created file " + zipPathLocation);
            }
        }
        try {
            completedFile.createNewFile();
            log.debug("Created completed file " + completedFileName + " to folder " + folder.getAbsolutePath());
            String zipPath = zipPathLocation + File.separator + folder.getName() + ".zip";
            log.debug("Created zip file " + folder.getName() + " to folder " + zipPath);

            writeZipToFolder(Paths.get(folder.getAbsolutePath()), Paths.get(zipPath));
        } catch (Exception e) {
            log.error("Unable to create zip directory " + e.getMessage());
            throw e;
        }
        return zipFile;
    }

    private void writeZipToFolder(Path sourceFolderPath, Path zipPath) throws IOException {
        FileOutputStream fileOutputStream = new FileOutputStream(zipPath.toFile());
        try (ZipOutputStream zipOutputStream = new ZipOutputStream(fileOutputStream)) {
            log.debug("Create zip file of " + sourceFolderPath + " to folder " + zipPath);
            Files.walkFileTree(sourceFolderPath, new SimpleFileVisitor<Path>() {
                public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
                    zipOutputStream.putNextEntry(new ZipEntry(sourceFolderPath.relativize(file).toString()));
                    Files.copy(file, zipOutputStream);
                    zipOutputStream.closeEntry();
                    return FileVisitResult.CONTINUE;
                }
            });
        }
    }

    @Override
    public void uploadHistoryToNSP() {
        if (uploadEnabled) {
            if (this.zipFileName == null) {
                log.error("No zip file to upload");
                return;
            }
            File[] listOfFilesToUpload = this.zipFileName.listFiles();
            if (!this.zipFileName.exists() || listOfFilesToUpload == null || listOfFilesToUpload.length == 0) {
                log.error("No zip file to upload");
                return;
            }
            try {
                Arrays.sort(listOfFilesToUpload, LastModifiedFileComparator.LASTMODIFIED_COMPARATOR);
                for (File file : listOfFilesToUpload) {
                    boolean isZipFile = FilenameUtils.getExtension(file.getName()).equals("zip");
                    if (isZipFile && validateDoneFileExist(file)) {
                        sftpClient.upload(new FileSystemResource(file), remoteFolder);
                        String archiveDirectoryPath = createCorrectPathFormat(tempDirLocation) + archiveDirectory;
                        File archiveDir = new File(archiveDirectoryPath);
                        if (!archiveDir.exists()) {
                            archiveDir.mkdir();
                        }
                        file.renameTo(new File(archiveDir, file.getName()));
                    }
                }
            } catch (Exception e) {
                log.error("Could not upload zip to sftp server" + e.getMessage());
            }
        } else {
            log.info("SFTP export has been disabled. Was supposed to send directory=" + folderName.getAbsolutePath());
        }
    }

    private boolean validateDoneFileExist(File zipFile) throws IOException {
        boolean completedFileExist = false;
        FileInputStream fileInputStream = new FileInputStream(zipFile);
        ZipInputStream zipInputStream = new ZipInputStream(new BufferedInputStream(fileInputStream));
        ZipEntry entry;
        try {
            while ((entry = zipInputStream.getNextEntry()) != null) {
                if (entry.getName().equals(completedFileName)) {
                    completedFileExist = true;
                }
            }
        } finally {
            zipInputStream.close();
            fileInputStream.close();
        }
        return completedFileExist;
    }

    private String createCorrectPathFormat(String folderPath) {
        if (!folderPath.endsWith(File.separator)) {
            return folderPath + File.separator;
        }
        return folderPath;
    }

}
