/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service.historyExport;

import com.jcraft.jsch.*;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.io.Resource;
import org.springframework.stereotype.Component;

import java.io.File;
import java.io.IOException;
@Log4j2
@Component
public class SFtpClientImpl implements SFtpClient {

    private static final JSch jsch = new JSch();
    static {
        JSch.setLogger(new Log4jJSchLogger());
    }

    @Value("${sftp.hostname}")
    public
    String hostName = "localhost";

    @Value("${sftp.port}")
    public
    int port;

    @Value("${sftp.username}")
    public
    String user;

    @Value("${sftp.password}")
    public
    String password;

    @Override
    public void upload(Resource resource, String remoteDir) throws SftpException, JSchException {
        if (remoteDir == null) remoteDir = "";
        try {
            put(resource.getFile(), remoteDir);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
        log.info("Done uploading resource");
    }

    @Override
    public File download(String remoteFile, String localDir) throws JSchException, SftpException {
        if (remoteFile == null) remoteFile = "";
        return get(remoteFile, localDir);
    }

    private File get(String remoteFile, String localDir) throws SftpException, JSchException {
        ChannelSftp c = null;
        Channel channel = null;
        Session session = null;
        try {
            session = getSession();
            channel = session.openChannel("sftp");
            channel.connect();
            c = (ChannelSftp) channel;

            String fsrc = remoteFile;
            String fdest = localDir;
            //Current version of JSch (0.1.49) requires both source and destination to be file paths
            if (remoteFile.endsWith(File.separator) || remoteFile.trim().length() == 0) {
                fdest = fdest + localDir;
            } else {
                fdest = fdest + File.separator + localDir;
            }

            c.get(fsrc, fdest);
            return new File(fdest);
        } catch (JSchException e) {
            log.error("Unable to get session and connect: ", e);
            throw e;
        } catch (SftpException e) {
            log.error("Unable to connect to server: ", e);
            throw e;
        } finally {
            if (c != null) {
                c.disconnect();
            }
            if (session != null) {
                session.disconnect();
            }
        }
    }


    /**
     * Create connection, put one file to remote server in remoteDir location, then disconnect
     * @param srcFile
     * @param remoteDir
     */
    private void put(File srcFile, String remoteDir) throws JSchException, SftpException {
        System.out.println("**********************************" + srcFile.getAbsolutePath());
        ChannelSftp c = null;
        Channel channel = null;
        Session session = null;
        try {
            session = getSession();
            channel = session.openChannel("sftp");
            channel.connect();
            c = (ChannelSftp) channel;

            String fsrc = srcFile.getAbsolutePath();
            String fdest = remoteDir;
            //Current version of JSch (0.1.49) requires both source and destination to be file paths
            if (remoteDir.endsWith(File.separator) || remoteDir.trim().length() == 0) {
                fdest = fdest + srcFile.getName();
            } else {
                fdest = fdest + File.separator + srcFile.getName();
            }

            log.info("Starting file upload of file '" + srcFile.getName() + "' to remote file path '" + fdest + "'");
            c.put(fsrc, fdest);
            log.info("Uploaded file '" + srcFile.getName() + "' successfully");

        } catch (JSchException e) {
            log.error("Unable to get session and connect: ", e);
            throw e;
        } catch (SftpException e) {
            log.error("Unable to connect to server: ", e);
            throw e;
        } finally {
            if (c != null) {
                c.disconnect();
            }
            if (session != null) {
                session.disconnect();
            }
        }
    }

    private Session getSession() throws JSchException {
        Session session = jsch.getSession(user, hostName, port);

        if (password != null) {
            session.setPassword(password);
        }
        session.setDaemonThread(true);
        session.setConfig("StrictHostKeyChecking", "no");
        session.connect();
        return session;
    }

    private static class Log4jJSchLogger implements com.jcraft.jsch.Logger {
        @Override
        public boolean isEnabled(int level) {
            return true;
        }

        @Override
        public void log(int level, String message) {
            System.out.println(message);
            if (level == com.jcraft.jsch.Logger.DEBUG) {
                log.debug(message);
            } else if (level == com.jcraft.jsch.Logger.INFO) {
                log.debug(message);
            } else if (level == com.jcraft.jsch.Logger.WARN) {
                log.warn(message);
            } else if (level == com.jcraft.jsch.Logger.ERROR) {
                log.error(message);
            } else if (level == com.jcraft.jsch.Logger.FATAL) {
                log.fatal(message);
            } else {
                log.warn("Unknown com.jcraft.jsch.Logger log level '" + level + "' - Message: " + message);
            }
        }
    }
}
