/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.util;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.pki.SOSIFederation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.ClasspathCredentialVault;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.FileBasedCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.commons.io.IOUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;

import javax.annotation.PostConstruct;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import java.io.*;
import java.net.HttpURLConnection;
import java.net.URL;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.Date;
import java.util.Properties;

import static dk.sosi.seal.model.SignatureUtil.setupCryptoProviderForJVM;

@Component
public class SosiUtil {

    private static Logger logger = LogManager.getLogger(SosiUtil.class);

    @Value("${sosi.sts.url}")
    private String stsUrl;

    @Value("${sosi.careprovider.name}")
    private String careproviderName;

    @Value("${sosi.careprovider.cvr}")
    private String careproviderCvr;

    @Value("${sosi.production.federation}")
    private Boolean useProductionFederation;

    @Value("${sosi.system.name}")
    private String sosiSystemName;

    @Value("${sosi.keystore.file}")
    private String keyFilePath;

    @Value("${sosi.keystore.password}")
    private String keyStorePassword;

    @Value("${sosi.validate.expiration:true}")
    private boolean validateIDCardExpiration;

    @Value("${sosi.timeout:1000}")
    private int timeout;

    private static final long MAX_IDCARD_ACCEPTANCE_HOURS = 9;

    private final Properties props = setupCryptoProviderForJVM();
    private CredentialVault companyVault;

    private SOSIFactory factory;
    private static ThreadLocal<IDCard> idCard = ThreadLocal.withInitial(() -> null);
    private SOSIFactory testSOSIFactory;

    private RequestSender requestSender = new RequestSender();

    @PostConstruct
    public void init() throws Exception {
        props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_DO_NOT_REGISTER_STR_TRANSFORM, "true");
        String fileOrClasspath = keyFilePath.split(":")[0];
        String fileName = keyFilePath.split(":")[1];
        if (fileOrClasspath.equals("file")) {
            companyVault = new FileBasedCredentialVault(props, new File(fileName), keyStorePassword);
        } else {
            companyVault = new ClasspathCredentialVault(props, fileName, keyStorePassword);
        }
        if (useProductionFederation == null || useProductionFederation) {
            SOSIFederation federation = new SOSIFederation(props);
            factory = new SOSIFactory(federation, companyVault, props);
        } else {
//            SOSITestFederation federation = new SOSITestFederation(props);
            factory = initTestSOSIFactory(companyVault);
        }
    }

    private SOSIFactory initTestSOSIFactory(CredentialVault cv) {
        if (testSOSIFactory == null) {
            SOSITestFederation federation = new SOSITestFederation(props);
            testSOSIFactory = new SOSIFactory(federation, cv, props);
        }
        return testSOSIFactory;
    }

    public IDCard getIdCard() throws IOException {
        if ((idCard.get() != null) && (!validateIDCardExpiration || idCard.get().isValidInTime())) {
            LocalDateTime now = LocalDateTime.now();
            LocalDateTime acceptBefore = LocalDateTime.ofInstant(new Date(idCard.get().getCreatedDate().getTime() + MAX_IDCARD_ACCEPTANCE_HOURS * 3600 * 1000L).toInstant(), ZoneId.of("Europe/Copenhagen"));
            if (now.isBefore(acceptBefore)){
                logger.info("Using cached IDCard");
                return idCard.get();
            }
        }
        CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, careproviderCvr, careproviderName);
        IDCard requestCard = factory.createNewSystemIDCard(
                sosiSystemName,
                careProvider,
                AuthenticationLevel.VOCES_TRUSTED_SYSTEM,
                null,
                null,
                companyVault.getSystemCredentialPair().getCertificate(),
                null);
        if (requestCard == null) {
            throw new RuntimeException("Failed to create a new systemIDCard");
        }
        SecurityTokenRequest request = factory.createNewSecurityTokenRequest();
        request.setIDCard(requestCard);
        String xml = XmlUtil.node2String(request.serialize2DOMDocument(), false, true);
        String response = sendRequest(stsUrl, "", xml, true);
        SecurityTokenResponse securityTokenResponse = factory.deserializeSecurityTokenResponse(response);
        if (securityTokenResponse.isFault()) {
            logger.error("FaultActor: " + securityTokenResponse.getFaultActor());
            logger.error("FaultCode: " + securityTokenResponse.getFaultCode());
            logger.error("FaultString: " + securityTokenResponse.getFaultString());
            throw new RuntimeException("Security token response is faulty: " + securityTokenResponse.getFaultString());
        }
        idCard.set(securityTokenResponse.getIDCard());
        if (idCard == null) {
            throw new RuntimeException("The response from the STS did not contain an IDCard:\n" + response);
        }
        return idCard.get();
    }


    private String sendRequest(String url, String action, String docXml, boolean failOnError) throws IOException {
        return requestSender.sendRequest(url, action, docXml, failOnError, timeout);
    }


    private static Node string2Node(String xmlstr) throws Exception {
        InputSource inStream = new InputSource();
        inStream.setCharacterStream(new StringReader(xmlstr));
        DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
        docFactory.setNamespaceAware(true);
        DocumentBuilder builder = docFactory.newDocumentBuilder();
        Document docBody = builder.parse(inStream);
        return docBody.getFirstChild();
    }

    public String sendServiceRequestStr(String url, String action, String strSoapBodyRequest) throws Exception {
        IDCard card = null;
        try {
            card = getIdCard();
        } catch (Exception e) {
            throw new RuntimeException("Fejl i hentning af ID-kort fra STS", e);
        }

        Element bodyElem = (Element) string2Node(strSoapBodyRequest);


        Request req = factory.createNewRequest(false, null);
        req.setIDCard(card);
        req.setBody(bodyElem);
        Document dom = req.serialize2DOMDocument();
        String strRequest = XmlUtil.node2String(dom, false, true);

        return sendRequest(url, action, strRequest, false);
    }

    public void setRequestSender(RequestSender requestSender) {
        this.requestSender = requestSender;
    }

    public static class RequestSender {

        public String sendRequest(String url, String action, String docXml, boolean failOnError, int timeout) throws IOException {
            HttpURLConnection uc = null;
            OutputStream os = null;
            InputStream is = null;
            try {
                URL u = new URL(url);
                uc = (HttpURLConnection) u.openConnection();
                uc.setDoOutput(true);
                uc.setDoInput(true);
                uc.setRequestMethod("POST");
                uc.setConnectTimeout(timeout);
                uc.setReadTimeout(timeout);
                uc.setRequestProperty("SOAPAction", "\"" + action + "\"");
                uc.setRequestProperty("Content-Type", "text/xml; encoding=utf-8");
                os = uc.getOutputStream();
                IOUtils.write(docXml, os, "UTF-8");
                os.flush();
                if (uc.getResponseCode() != 200) {
                    is = uc.getErrorStream();
                } else {
                    is = uc.getInputStream();
                }
                String res = IOUtils.toString(is, "UTF-8");
                if (uc.getResponseCode() != 200 && (uc.getResponseCode() != 500 || failOnError)) {
                    throw new RuntimeException("Got unexpected response " + uc.getResponseCode() + " from " + url + " : " + res);
                }
                return res;
            } finally {
                if (os != null) IOUtils.closeQuietly(os);
                if (is != null) IOUtils.closeQuietly(is);
                if (uc != null) uc.disconnect();
            }
        }
    }
}
