/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice;

import dk.sds.ecpr2.CprPerson;
import dk.sds.ecpr2.Modified;
import dk.sds.ecpr2.PersonType;
import dk.sds.ecpr2.RolesPermissions;
import dk.sds.ecpr2._2017._10._01.*;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.*;
import dk.sds.ecpr2.security.Permission;
import dk.sds.ecpr2.security.RequestContext;
import dk.sds.ecpr2.service.ECPR2ServiceImpl;
import dk.sds.ecpr2.util.ECPRException;
import dk.sds.ecpr2.util.FaultCode;
import dk.sds.ecpr2.util.FaultUtil;
import dk.sds.ecpr2.util.Properties;
import dk.sds.ecpr2.webservice.mapping.FromXMLMapper;
import dk.sds.ecpr2.webservice.mapping.ToXMLMapper;
import dk.sds.ecpr2.webservice.validation.RequestValidator;
import dk.sds.nsp.security.Security;
import dk.sds.nsp.security.SecurityContext;
import lombok.Getter;
import lombok.Setter;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.ws.server.endpoint.annotation.Endpoint;
import org.springframework.ws.server.endpoint.annotation.RequestPayload;
import org.springframework.ws.server.endpoint.annotation.ResponsePayload;

import java.math.BigInteger;
import java.time.Instant;
import java.util.*;
import java.util.stream.Collectors;


@Endpoint
@Log4j2
public class ECPR2WebServiceImpl extends AbstractWebService implements dk.sds.ecpr2.webservice.ECPR2WebService {

    @Getter
    @Setter
    protected class RequestMetaData {
        private ServiceActorVO actor;
        private RequestContextVO context;
    }

    @Autowired
    private ECPR2ServiceImpl businessService;

    ECPR2ServiceImpl getBusinessService() {
        return businessService;
    }

    @Autowired
    FromXMLMapper requestMapper;

    @Autowired
    ToXMLMapper responseMapper;

    @Value("${admin.service.shouldCheckPermissions:true}")
    private boolean shouldCheckPermissions;

    @Autowired
    Properties properties;

    @Override
    @ResponsePayload
    public SearchPersonResponse searchPerson(@RequestPayload final SearchPersonRequest request) throws Exception {
        checkPermissions(Permission.Read);
        final SearchPersonResponse response = new SearchPersonResponse();
        log.info("Request: " + request.getQuery());
        try {
            new RequestValidator().validate(request);
            String searchQuery = request.getQuery();
            int totalPersonAmount = businessService.countPersons(searchQuery);
            if (request.getMaxNumberOfResults() == null || totalPersonAmount <= request.getMaxNumberOfResults().intValue()) {
                List<PersonVO> personList = businessService.searchPersons(searchQuery);
                response.getPersons().addAll(mapPersons(personList));
            }
            response.setNumberOfPersons(BigInteger.valueOf(totalPersonAmount));
            log.info("Results: " + response.getPersons().size());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.SEARCH_PERSON_VALIDATION_ERROR, FaultCode.SEARCH_PERSON_INTERNAL_ERROR);
        }
        return response;
    }

    @Override
    @ResponsePayload
    public GetPersonByIdResponse getPersonById(@RequestPayload GetPersonByIdRequest request) throws Exception {
        checkPermissions(Permission.Read);
        final GetPersonByIdResponse response = new GetPersonByIdResponse();
        try {
            new RequestValidator().validate(request);
            validateSecurityContext();
            RequestMetaData metaData = getRequestMetadata();
            List<PersonVO> personList = requestMapper.mapIdentifierToPersonFromRequest(request.getKeyIdentifiers());
            List<PersonVO> returnedPersonList = businessService.getPersonsById(personList, metaData.getActor(), metaData.getContext());
            response.getPersons().addAll(mapPersons(returnedPersonList));
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.GET_PERSON_BY_PERSONID_VALIDATION_ERROR, FaultCode.GET_ECPR_INTERNAL_ERROR);
        }
        return response;
    }

    @Override
    @ResponsePayload
    public CreatePersonResponse createPerson(@RequestPayload CreatePersonRequest request) throws Exception {
        checkPermissions(Permission.Write);
        final CreatePersonResponse response = new CreatePersonResponse();
        try {
            new RequestValidator().validate(request);
            PersonVO personVO = requestMapper.mapPersonFromRequest(request.getPerson());
            validateSecurityContext();
            RequestMetaData metaData = getRequestMetadata();
            PersonVO createdPersonVO = businessService.createPerson(personVO, metaData.getActor(), metaData.getContext());
            PersonType createdPerson = responseMapper.mapPerson(createdPersonVO);
            response.setPerson(createdPerson);
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.CREATE_ECPR_VALIDATION_ERROR, FaultCode.CREATE_ECPR_INTERNAL_ERROR);
        }
        return response;
    }

    @Override
    @ResponsePayload
    public CreatePersonIdResponse createPersonId(@RequestPayload CreatePersonIdRequest request) throws Exception {
        checkPermissions(Permission.Write);
        final CreatePersonIdResponse response = new CreatePersonIdResponse();
        PersonVO person;
        try {
            if (request != null) {
                new RequestValidator().validate(request);
                PersonVO createPersonId = requestMapper.mapPersonFromRequest(request.getPerson());
                validateSecurityContext();
                RequestMetaData metaData = getRequestMetadata();
                person = businessService.createPersonId(createPersonId, metaData.actor, metaData.context);
                PersonType createdPersonId = responseMapper.mapPerson(person);
                response.setPerson(createdPersonId);
            }
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.CREATE_PERSONID_VALIDATION_ERROR, FaultCode.CREATE_PERSONID_INTERNAL_ERROR);
        }
        return response;
    }


    @Override
    @ResponsePayload
    public ReservePersonIdResponse reservePersonId(@RequestPayload ReservePersonIdRequest request) throws Exception {
        checkPermissions(Permission.Write);
        final ReservePersonIdResponse response = new ReservePersonIdResponse();

        try {
            new RequestValidator().validate(request);
            List<ReserveRequestVO> reserveRequests = requestMapper.mapReserveRequest(request.getLists());
            reserveRequests.forEach(r -> r.setModifiedBy(request.getModified().getBy()));
            validateSecurityContext();
            RequestMetaData metaData = getRequestMetadata();
            List<List<PersonVO>> personsList = businessService.reservePersonIds(reserveRequests, metaData.getActor(), metaData.getContext());
            responseMapper.mapReserveResponse(response, personsList);
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.CREATE_ECPR_VALIDATION_ERROR, FaultCode.CREATE_ECPR_INTERNAL_ERROR);
        }

        return response;
    }

    @Override
    @ResponsePayload
    public UpdatePersonResponse updatePerson(@RequestPayload UpdatePersonRequest request) throws Exception {
        checkPermissions(Permission.Write);
        final UpdatePersonResponse response = new UpdatePersonResponse();
        try {
            new RequestValidator().validate(request);
            PersonVO personUpdate = requestMapper.mapPersonFromRequest(request.getPerson());
            validateSecurityContext();
            RequestMetaData metaData = getRequestMetadata();
            PersonVO person = businessService.updatePerson(personUpdate, metaData.getActor(), metaData.getContext());
            PersonType updatedPerson = responseMapper.mapPerson(person);
            response.setPerson(updatedPerson);
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.UPDATE_ECPR_VALIDATION_ERROR, FaultCode.UPDATE_ECPR_INTERNAL_ERROR);
        }

        return response;
    }

    @Override
    @ResponsePayload
    public MergePersonsResponse mergePersons(@RequestPayload MergePersonsRequest request) throws Exception {
        checkPermissions(Permission.Write);
        final MergePersonsResponse response = new MergePersonsResponse();
        try {
            new RequestValidator().validate(request);
            KeyIdentifierVO from = requestMapper.mapKeyIdentifier(request.getFrom());
            KeyIdentifierVO to = requestMapper.mapKeyIdentifier(request.getTo());
            validateSecurityContext();
            RequestMetaData metaData = getRequestMetadata();
            Modified modified = request.getModified();
            PersonVO person = businessService.mergePerson(from, to, modified, metaData.actor, metaData.context);
            PersonType mergedPerson = responseMapper.mapPerson(person);
            response.setPerson(mergedPerson);
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.MERGE_PERSON_VALIDATION_ERROR, FaultCode.MERGE_PERSON_INTERNAL_ERROR);
        }
        return response;
    }

    @Override
    @ResponsePayload
    public DeletePersonResponse deletePerson(@RequestPayload DeletePersonRequest request) throws Exception {
        if (properties.isProduction) {
            FaultUtil.throwECPR2Fault(FaultCode.DELETE_PERSON_BY_PERSONID_NOT_ALLOWED_IN_PROD);
        }
        checkPermissions(Permission.Write);
        final DeletePersonResponse response = new DeletePersonResponse();
        try {
            new RequestValidator().validate(request);
            validateSecurityContext();
            RequestMetaData metaData = getRequestMetadata();
            List<IdentifierVO> identifierVOs = requestMapper.mapPersonKeyIdentifier(request.getPersons());
            Modified modified = request.getModified();
            businessService.deletePerson(identifierVOs, modified, metaData.getActor(), metaData.getContext());
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.DELETE_PERSON_BY_PERSONID_VALIDATION_ERROR, FaultCode.DELETE_PERSON_BY_PERSONID_INTERNAL_ERROR);
        }
        return response;
    }

    @Override
    @ResponsePayload
    public GetPermissionsResponse getPermissions(@RequestPayload GetPermissionsRequest request) throws Exception {
        final GetPermissionsResponse response = new GetPermissionsResponse();
        try {
            List<RolesPermissions> responseRolesPermissions = response.getRolesPermissions();
            Map<String, List<Permission>> permissionMap = businessService.getPermissions(request.getGetAllPermissions() != null, request.getGetCallersPermissions() != null);
            permissionMap.forEach((role, permissions) -> {
                RolesPermissions rolesPermissions = new RolesPermissions();
                rolesPermissions.setRequestedRole(role);
                List<String> permissionsList = rolesPermissions.getPermissions();
                permissionsList.addAll(permissions.stream().map(Permission::getName).collect(Collectors.toList()));
                responseRolesPermissions.add(rolesPermissions);
            });
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.GET_PERMISSIONS_VALIDATION_ERROR, FaultCode.GET_PERMISSIONS_INTERNAL_ERROR);
        }
        return response;
    }

    @Override
    @ResponsePayload
    public GetCprPersonResponse getCprPerson(@RequestPayload GetCprPersonRequest request) throws Exception {
        checkPermissions(Permission.Read);
        final GetCprPersonResponse response = new GetCprPersonResponse();
        try {
            CprPersonVO cprPersonVO = businessService.getCprPerson(request.getKeyIdentifier());
            if (cprPersonVO != null) {
                CprPerson cprPerson = new CprPerson();
                String name = (cprPersonVO.getGivenName() == null ? "" : cprPersonVO.getGivenName() + " ") + (cprPersonVO.getMiddleName() == null ? "" : cprPersonVO.getMiddleName() + " ") + (cprPersonVO.getSurName() == null ? "" : cprPersonVO.getSurName());
                cprPerson.setName(name.trim());
                cprPerson.setBirthDate(cprPersonVO.getBirthdate());
                response.setCprPerson(cprPerson);
            }
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.GET_CPR_PERSON_VALIDATION_ERROR, FaultCode.GET_CPR_PERSON_INTERNAL_ERROR);
        }
        return response;
    }

    @Override
    @ResponsePayload
    public CompareNameResponse compareName(@RequestPayload CompareNameRequest request) throws Exception {
        checkPermissions(Permission.Read);
        final CompareNameResponse response = new CompareNameResponse();
        try {
            RequestMetaData metaData = getRequestMetadata();
            Integer score = businessService.compareName(request.getKeyIdentifier(), request.getName(), metaData.getActor(), metaData.getContext());
            if (score != null) {
                response.setScore(score);
            }
        } catch (Exception e) {
            commonExceptionHandler(e, FaultCode.COMPARE_NAME_VALIDATION_ERROR, FaultCode.COMPARE_NAME_INTERNAL_ERROR);
        }
        return response;
    }

    private void checkPermissions(Permission... permissions) {
        if (shouldCheckPermissions) {
            boolean permissionsFound = true;

            try {
                for (Permission permission : permissions) {
                    if (!RequestContext.get().getPermissions().contains(permission)) {
                        permissionsFound = false;
                        break;
                    }
                }
            } catch (Exception ex) {
                log.error("Error checking permissions", ex);
                FaultUtil.throwECPR2Fault(FaultCode.AUTHORIZATION_INTERNAL_ERROR, ex);
            }

            if (!permissionsFound) {
                FaultUtil.throwECPR2Fault(FaultCode.AUTHORIZATION_ERROR);
            }
        }
    }

    private void validateSecurityContext() {
        SecurityContext securityContext = Security.getSecurityContext();
        Optional<SecurityContext.Ticket> ticketOptional = securityContext.getTicket();
        if (!ticketOptional.isPresent()) {
            throw new ECPRException(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
        }
        SecurityContext.Ticket ticket = ticketOptional.get();
        if (!ticket.isValid()) {
            throw new ECPRException(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
        }
        checkDgws(securityContext);
    }

    private void checkDgws(SecurityContext securityContext) {
        Optional<SecurityContext.User> userOptional = securityContext.getActingUser();
        if (!userOptional.isPresent()) {
            return;
        }
        if (userOptional.get().getUserType().equals(SecurityContext.User.UserType.Citizen)) {
            throw new ECPRException(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
        }
        final Optional<SecurityContext.User.Credentials> credentials = userOptional.get().getCredentials();
        if (!credentials.isPresent()) {
            throw new ECPRException(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
        }
    }

    private RequestMetaData getRequestMetadata() {
       final RequestMetaData metaData = new RequestMetaData();
       SecurityContext securityContext = Security.getSecurityContext();
       metaData.setActor(getServiceActor(securityContext));
       metaData.setContext(getRequestContext(securityContext));
       return metaData;
    }

    private RequestContextVO getRequestContext(SecurityContext securityContext) {
        RequestContextVO context = new RequestContextVO();
        if (securityContext.getMessage().isPresent()) {
            SecurityContext.Message message = securityContext.getMessage().get();
            if (message.getConversationIdentifier().isPresent()) {
                context.setFlowId(message.getConversationIdentifier().get());
            }
            if (message.getIdentifier().isPresent()) {
                context.setMessageId(message.getIdentifier().get());
            }
        }
        if (securityContext.getClient().isPresent()) {
            SecurityContext.Client client = securityContext.getClient().get();
            if (client.getName().isPresent()) {
                context.setSystemIdentifier(client.getName().get());
            }
        }
        context.setRequestTimeStamp(Instant.now());
        return context;
    }

    private ServiceActorVO getServiceActor(SecurityContext context) {
        ServiceActorVO actor = new ServiceActorVO();

        Optional<SecurityContext.Organisation> organisation = context.getOrganisation();
        if (organisation.isPresent()) {
            actor.setOrganizationId(organisation.get().getIdentifier());
            if (organisation.get().getName().isPresent()) {
                actor.setOrganizationName(organisation.get().getName().get());
            }
        }
        Optional<SecurityContext.User> user = context.getActingUser();
        if (user.isPresent()) {
            actor.setPersonIdentifier(user.get().getIdentifier());
            if (user.get().getGivenName().isPresent()) {
                actor.setFirstName(user.get().getGivenName().get());
            }
            if (user.get().getSurname().isPresent()) {
                actor.setLastName(user.get().getSurname().get());
            }
            Optional<SecurityContext.User.Credentials> credentials = user.get().getCredentials();
            if (credentials.isPresent()) {
                if (credentials.get().getEducationCode().isPresent()) {
                    actor.setRole(credentials.get().getEducationCode().get()); // Uddannelseskode
                }
                if (credentials.get().getAuthorizationCode().isPresent()) {
                    actor.setAuthorisationNumber(credentials.get().getAuthorizationCode().get().trim());
                }
            }
        }

        return actor;
    }

    private Collection<? extends PersonType> mapPersons(List<PersonVO> personList) throws DAOException {
        List<PersonType> searchPerson = new ArrayList<>();
        for (PersonVO person : personList) {
            searchPerson.add(responseMapper.mapPerson(person));
        }
        return searchPerson;
    }
}
