/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice.mapping;

import dk.sds.ecpr2.*;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.*;
import org.springframework.stereotype.Component;

import javax.xml.datatype.XMLGregorianCalendar;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static dk.sds.ecpr2.webservice.validation.RequestValidator.DATE_REGEX;

@Component
public class FromXMLMapper {

    public PersonVO mapPersonFromRequest(PersonWithKeyIdentifierType personType) {
        PersonVO person = mapPersonFromRequest((PersonType) personType);
        person.setKeyIdentifier(mapKeyIdentifier(personType.getKeyIdentifier()));
        return person;
    }

    public KeyIdentifierVO mapKeyIdentifier(KeyIdentifier keyIdentifier) {
        if (keyIdentifier == null) {
            return null;
        } else {
            KeyIdentifierVO keyIdentifierVO = new KeyIdentifierVO();
            keyIdentifierVO.setOid(keyIdentifier.getOID());
            keyIdentifierVO.setValue(keyIdentifier.getValue());
            return keyIdentifierVO;
        }
    }

    public PersonVO mapPersonFromRequest(PersonType personType) {
        PersonVO person = new PersonVO();

        person.setIdentifiers(mapIdentifiers(personType.getIdentifiers()));
        person.setAddresses(mapAddresses(personType.getAddresses()));
        person.setBirthDates(mapBirthDates(personType.getBirthDates()));
        person.setContacts(mapContacts(personType.getContacts()));
        person.setNames(mapNames(personType.getNames()));
        person.setGenders(mapGenders(personType.getGenders()));
        person.setModifiers(personType.getModifieds());

        return person;
    }

    public PersonVO mapPersonFromRequest(SimplePersonWithKeyIdentifierType keyIdentifierType) {
        PersonVO person = new PersonVO();
        person.setKeyIdentifier(mapKeyIdentifier(keyIdentifierType.getKeyIdentifier()));
        person.setIdentifiers(mapIdentifiers(keyIdentifierType.getIdentifiers()));
        person.setModifiers(keyIdentifierType.getModifieds());
        return person;
    }

    public List<PersonVO> mapIdentifierToPersonFromRequest(List<KeyIdentifier> keyIdentifiers) {
        List<PersonVO> personVOList = new ArrayList<>();
        PersonVO person = new PersonVO();
        KeyIdentifierVO keyIdentifierVO = new KeyIdentifierVO();
        for (KeyIdentifier keyIdentifier : keyIdentifiers) {
            keyIdentifierVO.setOid(keyIdentifier.getOID());
            keyIdentifierVO.setValue(keyIdentifier.getValue());
            person.setKeyIdentifier(keyIdentifierVO);
            personVOList.add(person);
        }
        return personVOList;
    }

    private List<IdentifierVO> mapIdentifiers(List<Identifier> list) {
        if (list == null) {
            return null;
        }
        List<IdentifierVO> mapped = new ArrayList<>();
        for (Identifier identifier : list) {
            mapped.add(mapIdentifier(identifier));
        }
        return mapped;
    }

    private IdentifierVO mapIdentifier(Identifier identifier) {
        IdentifierVO identifierVO = new IdentifierVO();
        identifierVO.setUuid(identifier.getPID());
        identifierVO.setOid(identifier.getOID());
        identifierVO.setValue(identifier.getValue());
        identifierVO.setUuid(identifier.getPID());
        identifierVO.setValidity(identifier.getValidity());
        identifierVO.setExpiry(mapExpiry(identifier.getExpiry()));
        return identifierVO;
    }

    public List<IdentifierVO> mapIdentifier(List<KeyIdentifier> identifiers) {
        List<IdentifierVO> identifierList = new ArrayList<>();
        for (KeyIdentifier keyIdentifier : identifiers) {
            IdentifierVO identifierVO = new IdentifierVO();
            identifierVO.setOid(keyIdentifier.getOID());
            identifierVO.setValue(keyIdentifier.getValue());
            identifierList.add(identifierVO);
        }
        return identifierList;
    }

    public List<IdentifierVO> mapPersonKeyIdentifier(List<PersonIdType> persons) {
        List<IdentifierVO> identifierList = new ArrayList<>();
        for (PersonIdType person : persons) {
            IdentifierVO identifierVO = new IdentifierVO();
            identifierVO.setOid(person.getKeyIdentifier().getOID());
            identifierVO.setValue(person.getKeyIdentifier().getValue());
            identifierList.add(identifierVO);
        }
        return identifierList;
    }

    private List<GenderVO> mapGenders(List<GenderType> list) {
        if (list == null) {
            return null;
        }
        List<GenderVO> mapped = new ArrayList<>();
        for (GenderType gender : list) {
            mapped.add(mapGender(gender));
        }
        return mapped;
    }

    private GenderVO mapGender(GenderType genderType) {
        GenderVO gender = new GenderVO();
        gender.setUuid(genderType.getPID());
        gender.setValue(genderType.getValue());
        return gender;
    }

    private List<NameVO> mapNames(List<NameType> list) {
        if (list == null) {
            return null;
        }
        List<NameVO> mapped = new ArrayList<>();
        for (NameType name : list) {
            mapped.add(mapName(name));
        }
        return mapped;
    }

    private NameVO mapName(NameType nameType) {
        NameVO name = new NameVO();
        name.setUse(nameType.getUse());
        name.setGivenName(nameType.getGivenName());
        name.setFamilyName(nameType.getFamilyName());
        name.setUuid(nameType.getPID());
        name.setExpiry(mapExpiry(nameType.getExpiry()));
        return name;
    }

    private List<ContactVO> mapContacts(List<ContactType> list) {
        if (list == null) {
            return null;
        }
        List<ContactVO> mapped = new ArrayList<>();
        for (ContactType contact : list) {
            mapped.add(mapContact(contact));
        }
        return mapped;
    }

    private ContactVO mapContact(ContactType contactType) {
        ContactVO contact = new ContactVO();
        contact.setUuid(contactType.getPID());
        contact.setSystem(contactType.getSystem());
        contact.setUse(contactType.getUse());
        contact.setValue(contactType.getValue());
        contact.setExpiry(mapExpiry(contactType.getExpiry()));
        return contact;
    }

    private List<BirthDateVO> mapBirthDates(List<BirthDateType> list) {
        if (list == null) {
            return null;
        }
        List<BirthDateVO> mapped = new ArrayList<>();
        for (BirthDateType birthDate : list) {
            mapped.add(mapBirthDate(birthDate));
        }
        return mapped;
    }

    BirthDateVO mapBirthDate(BirthDateType birthDateType) {
        if (birthDateType != null && birthDateType.getValue() != null) {
            BirthDateVO birthDate = new BirthDateVO();
            final Pattern pattern = Pattern.compile(DATE_REGEX);
            Matcher matcher = pattern.matcher(birthDateType.getValue());
            if (!matcher.matches()) {
                throw new IllegalArgumentException("Ulovligt format for fødselsdato: " + birthDateType.getValue());
            }
            if (matcher.group(5) != null) {
                birthDate.setDay(Integer.parseInt(matcher.group(5)));
            } else {
                birthDate.setDay(null);
            }
            if (matcher.group(3) != null) {
                birthDate.setMonth(Integer.parseInt(matcher.group(3)));
            } else {
                birthDate.setMonth(null);
            }
            if (matcher.group(1) != null) {
                birthDate.setYear(Integer.parseInt(matcher.group(1)));
            } else {
                birthDate.setYear(null);
            }
            birthDate.setUuid(birthDateType.getPID());
            return birthDate;
        }
        return new BirthDateVO();
    }

    private List<AddressVO> mapAddresses(List<AddressType> list) {
        if (list == null) {
            return null;
        }
        List<AddressVO> mapped = new ArrayList<>();
        for (AddressType address : list) {
            mapped.add(mapAddress(address));
        }
        return mapped;
    }

    private AddressVO mapAddress(AddressType addressType) {
        AddressVO address = new AddressVO();
        address.setCity(addressType.getCity());
        address.setCountry(addressType.getCountry());
        address.setDistrict(addressType.getDistrict());
        address.setExpiry(mapExpiry(addressType.getExpiry()));
        address.setUse(addressType.getUse());
        address.setPostalCode(addressType.getPostalCode());
        address.setPostBox(addressType.getPostBox());
        address.setState(addressType.getState());
        int count = 0;
        address.setStreetAddressLine1(addressType.getTexts().size() > count ? addressType.getTexts().get(count++) : null);
        address.setStreetAddressLine2(addressType.getTexts().size() > count ? addressType.getTexts().get(count++) : null);
        address.setStreetAddressLine3(addressType.getTexts().size() > count ? addressType.getTexts().get(count++) : null);
        address.setStreetAddressLine4(addressType.getTexts().size() > count ? addressType.getTexts().get(count++) : null);
        address.setStreetAddressLine5(addressType.getTexts().size() > count ? addressType.getTexts().get(count++) : null);
        address.setStreetAddressLine6(addressType.getTexts().size() > count ? addressType.getTexts().get(count++) : null);
        address.setStreetAddressLine7(addressType.getTexts().size() > count ? addressType.getTexts().get(count++) : null);
        address.setStreetAddressLine8(addressType.getTexts().size() > count ? addressType.getTexts().get(count++) : null);
        address.setStreetAddressLine9(addressType.getTexts().size() > count ? addressType.getTexts().get(count++) : null);
        address.setStreetAddressLine10(addressType.getTexts().size() > count ? addressType.getTexts().get(count++) : null);
        address.setType(addressType.getType());
        address.setUuid(addressType.getPID());
        return address;
    }

    Date mapExpiry(Calendar expiry) {
        if (expiry == null) {
            return null;
        }
        return expiry.getTime();
    }

    public Gender map(GenderFemaleOrMaleType gender) {
        if (gender == null) {
            return null;
        }
        switch (gender) {
            case MALE:
                return Gender.MALE;
            case FEMALE:
                return Gender.FEMALE;
            default:
                return null;
        }
    }

    public LocalDate mapDate(XMLGregorianCalendar xmlDate) {
        if (xmlDate == null) {
            return null;
        } else {
            return LocalDate.parse(xmlDate.toXMLFormat());
        }
    }

    public LocalDateTime mapDateTime(Calendar cal) {
        if (cal == null) {
            return null;
        } else {
            return cal.getTime().toInstant().atZone(ZoneId.systemDefault()).toLocalDateTime();
        }
    }

    public List<ReserveRequestVO> mapReserveRequest(List<RequestListType> lists) {
        List<ReserveRequestVO> reserveRequests = new ArrayList<>();
        for (RequestListType req : lists) {
            LocalDate forDate = mapDate(req.getForDate());
            LocalDateTime reservedUntil = mapDateTime(req.getReservedUntil());
            ReserveRequestVO reserveRequest = new ReserveRequestVO();
            reserveRequest.setOid(req.getOID());
            reserveRequest.setForDate(forDate);
            reserveRequest.setGender(map(req.getGender()));
            reserveRequest.setNumber(req.getNumber());
            reserveRequest.setReservedUntil(reservedUntil);
            reserveRequests.add(reserveRequest);
        }
        return reserveRequests;
    }
}
