/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice.mapping;

import dk.sds.ecpr2.*;
import dk.sds.ecpr2._2017._10._01.ReservePersonIdResponse;
import dk.sds.ecpr2.dao.OIDFacade;
import dk.sds.ecpr2.dao.PersonDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.*;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.text.SimpleDateFormat;
import java.util.*;
import java.util.stream.Collectors;

import static dk.sds.ecpr2.dao.impl.abstr.AbstractECPRDAOImpl.VALID_TO_DEFAULT;

@Log4j2
@Component
public class ToXMLMapper {

    @Autowired
    OIDFacade oidFacade;

    @Autowired
    PersonDAO personDao;

    private final static SimpleDateFormat expiryFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ");

    public PersonType mapPerson(PersonVO personVO) throws DAOException {
        if (personVO == null) {
            return null;
        }
        PersonType person = new PersonType();

        person.getAddresses().addAll(mapAddresses(personVO.getAddresses()));
        person.getBirthDates().addAll(mapBirthDates(personVO.getBirthDates()));
        person.getContacts().addAll(mapContacts(personVO.getContacts()));
        person.getGenders().addAll(mapGenders(personVO.getGenders()));
        person.getIdentifiers().addAll(mapIdentifier(personVO.getIdentifiers()));
        person.getNames().addAll(mapNames(personVO.getNames()));
        person.getModifieds().addAll(mapModifieds(personVO.getModifiers()));

        return person;
    }

    private Collection<? extends Modified> mapModifieds(List<Modified> modifiers) {
        Collection<Modified> results = new ArrayList<>();

        for (Modified m : modifiers) {
            Modified result = new Modified();
            result.setBy(m.getBy());
            result.setValidFrom(m.getValidFrom());
            result.setReservedUntil(m.getReservedUntil());

            if (m.getValidTo() == null || VALID_TO_DEFAULT.equals(m.getValidTo().getTime())) {
                result.setValidTo(null);
            } else {
                result.setValidTo(m.getValidTo());
            }

            results.add(result);
        }

        return results;
    }

    private List<AddressType> mapAddresses(List<AddressVO> list) {
        if (list == null || list.size() == 0) {
            return new ArrayList<>();
        } else {
            return list.stream().map(this::map).collect(Collectors.toList());
        }
    }

    private AddressType map(AddressVO entry) {
        AddressType address = new AddressType();
        address.setExpiry(mapExpiry(entry.getExpiry()));
        address.setType(entry.getType());
        address.setUse(entry.getUse());
        address.setValidFrom(mapValidFrom(entry.getValidFrom()));
        address.setValidTo(mapValidTo(entry.getValidTo(), entry.getExpiry()));
        address.setCountry(entry.getCountry());
        address.setState(entry.getState());
        address.setPostBox(entry.getPostBox());
        address.setPostalCode(entry.getPostalCode());
        address.setDistrict(entry.getDistrict());
        address.setCity(entry.getCity());
        address.setCountryName(CountryCode.countryCodeToCountryName(entry.getCountry()));
        address.setCountry(entry.getCountry());
        address.setPID(entry.getUuid());
        addStreetAddressLine(address, entry.getStreetAddressLine1());
        addStreetAddressLine(address, entry.getStreetAddressLine2());
        addStreetAddressLine(address, entry.getStreetAddressLine3());
        addStreetAddressLine(address, entry.getStreetAddressLine4());
        addStreetAddressLine(address, entry.getStreetAddressLine5());
        addStreetAddressLine(address, entry.getStreetAddressLine6());
        addStreetAddressLine(address, entry.getStreetAddressLine7());
        addStreetAddressLine(address, entry.getStreetAddressLine8());
        addStreetAddressLine(address, entry.getStreetAddressLine9());
        addStreetAddressLine(address, entry.getStreetAddressLine10());
        return address;
    }

    private void addStreetAddressLine(AddressType address, String streetAddressLine) {
        if (streetAddressLine != null) {
            address.getTexts().add(streetAddressLine);
        }
    }

    private Calendar mapValidFrom(Date d) {
        if (d == null) {
            return null;
        }
        return toCalendar(d);
    }

    Calendar mapValidTo(Date validTo, Date expiry) {
        if (VALID_TO_DEFAULT.equals(validTo) || (validTo == null)) {
            if (expiryHasPassed(expiry)) {
                return toCalendar(expiry);
            } else {
                return null;
            }
        } else {
            return toCalendar(validTo);
        }
    }

    private boolean expiryHasPassed(Date expiry) {
        return expiry != null && expiry.before(new Date());
    }

    public static Calendar toCalendar(Date date) {
        Calendar cal = Calendar.getInstance();
        cal.setTime(date);
        return cal;
    }

    private Calendar mapExpiry(Date expiry) {
        if (expiry == null) {
            return null;
        }
        return toCalendar(expiry);
    }

    private List<BirthDateType> mapBirthDates(List<BirthDateVO> list) {
        if (list == null || list.size() == 0) {
            return new ArrayList<>();
        } else {
            return list.stream().map(this::map).collect(Collectors.toList());
        }
    }

    private BirthDateType map(BirthDateVO entry) {
        if (entry == null) {
            return null;
        }
        BirthDateType birthDate = new BirthDateType();
        birthDate.setValue(mapBirthDateValue(entry));
        birthDate.setPID(entry.getUuid());
        birthDate.setValidFrom(mapValidFrom(entry.getValidFrom()));
        birthDate.setValidTo(mapValidTo(entry.getValidTo(), null));
        return birthDate;
    }

    private String mapBirthDateValue(BirthDateVO entry) {
        return entry.getYear() +
                (entry.getMonth() != null && entry.getMonth() != 0 ? "-" + (entry.getMonth() < 10 ? "0" + entry.getMonth() : entry.getMonth()) : "") +
                (entry.getDay() != null && entry.getDay() != 0 ? "-" + (entry.getDay() < 10 ? "0" + entry.getDay() : entry.getDay()) : "");
    }

    private List<ContactType> mapContacts(List<ContactVO> list) {
        if (list == null || list.size() == 0) {
            return new ArrayList<>();
        } else {
            return list.stream().map(this::map).collect(Collectors.toList());
        }
    }

    private ContactType map(ContactVO entry) {
        if (entry == null) {
            return null;
        }
        ContactType contact = new ContactType();
        contact.setSystem(entry.getSystem());
        contact.setUse(entry.getUse());
        contact.setValue(entry.getValue());
        contact.setExpiry(mapExpiry(entry.getExpiry()));
        contact.setPID(entry.getUuid());
        contact.setValidFrom(mapValidFrom(entry.getValidFrom()));
        contact.setValidTo(mapValidTo(entry.getValidTo(), entry.getExpiry()));
        return contact;
    }

    private List<GenderType> mapGenders(List<GenderVO> list) {
        if (list == null || list.size() == 0) {
            return new ArrayList<>();
        } else {
            return list.stream().map(this::map).collect(Collectors.toList());
        }
    }

    private GenderType map(GenderVO entry) {
        if (entry == null) {
            return null;
        }
        GenderType gender = new GenderType();
        gender.setValue(entry.getValue());
        gender.setPID(entry.getUuid());
        gender.setValidFrom(mapValidFrom(entry.getValidFrom()));
        gender.setValidTo(mapValidTo(entry.getValidTo(), null));
        return gender;
    }

    private List<Identifier> mapIdentifier(List<IdentifierVO> list) {
        if (list == null || list.size() == 0) {
            return new ArrayList<>();
        } else {
            return list.stream().map(this::map).collect(Collectors.toList());
        }
    }

    private Identifier map(IdentifierVO entry) {
        if (entry == null) {
            return null;
        }
        Identifier identifier = new Identifier();
        identifier.setValidity(entry.getValidity());
        identifier.setOID(entry.getOid());
        identifier.setPrimaryOIDLabel(mapOIDtoPrimaryLabel(entry.getOid()));
        identifier.setSecondaryOIDLabel(mapOIDtoSecondaryLabel(entry.getOid()));
        identifier.setOIDType(mapOIDtoType(entry.getOid()));
        identifier.setValue(entry.getValue());
        identifier.setExpiry(mapExpiry(entry.getExpiry()));
        identifier.setPID(entry.getUuid());
        identifier.setValidFrom(mapValidFrom(entry.getValidFrom()));
        identifier.setValidTo(mapValidTo(entry.getValidTo(), entry.getExpiry()));
        return identifier;
    }

    private String mapOIDtoPrimaryLabel(String oid) {
        try {
            OIDVO value = oidFacade.getByOIDValue(oid);
            if (value != null) {
                return value.getPrimaryOidLabel();
            } else {
                return null;
            }
        } catch (DAOException e) {
            log.warn("Unexpected exception while looking up OID " + oid, e);
            return null;
        }
    }

    private String mapOIDtoSecondaryLabel(String oid) {
        try {
            OIDVO value = oidFacade.getByOIDValue(oid);
            if (value != null) {
                return value.getSecondaryOidLabel();
            } else {
                return null;
            }
        } catch (DAOException e) {
            log.warn("Unexpected exception while looking up OID " + oid, e);
            return null;
        }
    }

    private PredefinedOIDType mapOIDtoType(String oid) {
        try {
            OIDVO value = oidFacade.getByOIDValue(oid);
            if (value != null) {
                return PredefinedOIDType.fromValue(value.getType());
            } else {
                return null;
            }
        } catch (DAOException e) {
            log.warn("Unexpected exception while looking up OID " + oid, e);
            return null;
        }
    }

    private List<NameType> mapNames(List<NameVO> list) {
        if (list == null || list.size() == 0) {
            return new ArrayList<>();
        } else {
            return list.stream().map(this::map).collect(Collectors.toList());
        }
    }

    private NameType map(NameVO entry) {
        if (entry == null) {
            return null;
        }
        NameType name = new NameType();
        name.setPID(entry.getUuid());
        name.setFamilyName(entry.getFamilyName());
        name.setGivenName(entry.getGivenName());
        name.setUse(entry.getUse());
        name.setText(entry.getFullName());
        name.setExpiry(mapExpiry(entry.getExpiry()));
        name.setValidFrom(mapValidFrom(entry.getValidFrom()));
        name.setValidTo(mapValidTo(entry.getValidTo(), entry.getExpiry()));
        return name;
    }

    public List<PersonWithGenderType> mapPersons(List<PersonVO> list) throws DAOException {
        if (list == null || list.size() == 0) {
            return new ArrayList<>();
        } else {
            List<PersonWithGenderType> persons = new ArrayList<>();
            for (PersonVO person : list) {
                PersonWithGenderType mapped = map(person);
                persons.add(mapped);
            }
            return persons;
        }
    }

    private PersonWithGenderType map(PersonVO entry) throws DAOException {
        if (entry == null) {
            return null;
        }
        PersonWithGenderType person = new PersonWithGenderType();
        person.getGenders().addAll(mapGenders(entry.getGenders()));
        person.getIdentifiers().addAll(mapIdentifier(entry.getIdentifiers()));
        person.getModifieds().addAll(mapModifiers(entry.getModifiers()));
        return person;
    }

    private List<Modified> mapModifiers(List<Modified> modifiers) {
        for (Modified modified : modifiers) {
            modified.setValidTo(mapValidTo(modified.getValidTo().getTime(), null));
        }
        return modifiers;
    }

    private Calendar map(Date date) {
        if (date == null) {
            return null;
        }
        Calendar cal = Calendar.getInstance();
        cal.setTime(date);
        return cal;
    }

    public void mapReserveResponse(ReservePersonIdResponse response, List<List<PersonVO>> personsList) throws DAOException {
        for (List<PersonVO> persons : personsList) {
            List<PersonWithGenderType> responsePersons = mapPersons(persons);
            ResponseListType responseListType = new ResponseListType();
            responseListType.getPersons().addAll(responsePersons);
            response.getLists().add(responseListType);
        }
    }

    public List<OIDDefinition> mapOIDs(List<OIDVO> list) {
        List<OIDDefinition> result = new ArrayList<>();
        if (list == null || list.size() == 0) {
            return result;
        } else {
            for (OIDVO oid : list) {
                OIDDefinition mapped = map(oid);
                result.add(mapped);
            }
            return result;
        }
    }

    private OIDDefinition map(OIDVO oid) {
        if (oid == null) {
            return null;
        } else {
            OIDDefinition oidDefinition = new OIDDefinition();
            oidDefinition.setOID(oid.getOid());
            oidDefinition.setPrimaryLabel(oid.getPrimaryOidLabel());
            oidDefinition.setSecondaryLabel(oid.getSecondaryOidLabel());
            oidDefinition.setRegEx(oid.getRegex());
            oidDefinition.setType(PredefinedOIDType.fromValue(oid.getType()));
            return oidDefinition;
        }
    }

}
