/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice.validation;

import com.google.common.base.Splitter;
import dk.sds.ecpr2.*;
import dk.sds.ecpr2._2017._10._01.*;
import dk.sds.ecpr2.domain.CountryCode;
import dk.sds.ecpr2.util.ValueSets;
import lombok.Getter;
import lombok.extern.log4j.Log4j2;
import org.apache.logging.log4j.util.Strings;

import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

@Log4j2
public class RequestValidator {

    public static final String DATE_REGEX = "^([0-9]{4})(-(0[1-9]|1[0-2])(-(0[0-9]|[1-2][0-9]|3[0-1]))?)?$";

    @Getter
    private List<String> validationErrors;

    public RequestValidator() {
        clearValidationErrors();
    }

    public void clearValidationErrors() {
        validationErrors = new LinkedList<>();
    }

    public void validate(CreatePersonRequest request) {
        try {
            checkPerson(request.getPerson());
        } catch (Exception e) {
            log.info("Unexpected error: " + e.getMessage(), e);
            validationErrors.add(buildUnspecificErrorMessage("Opret person", e.getMessage()));
        }
        throwExceptionInCaseOfValidationErrors();
    }

    public void validate(ReservePersonIdRequest request) {
        try {
            checkModifiers(Collections.singletonList(request.getModified()));
        } catch (Exception e) {
            log.info("Unexpected error: " + e.getMessage(), e);
            validationErrors.add(buildUnspecificErrorMessage("Reserver personId", e.getMessage()));
        }
        throwExceptionInCaseOfValidationErrors();
    }

    public void validate(CreatePersonIdRequest request) {
        try {
            checkKeyIdentifiers(Collections.singletonList(request.getPerson().getKeyIdentifier()));
            checkeCPROID(request.getPerson().getKeyIdentifier().getOID());
            checkModifiers(request.getPerson().getModifieds());
        } catch (Exception e) {
            log.info("Unexpected error: " + e.getMessage(), e);
            validationErrors.add(buildUnspecificErrorMessage("Opret personId", e.getMessage()));
        }
        throwExceptionInCaseOfValidationErrors();
    }

    public void validate(UpdatePersonRequest request) {
        try {
            checkPerson(request.getPerson());
        } catch (Exception e) {
            log.info("Unexpected error: " + e.getMessage(), e);
            validationErrors.add(buildUnspecificErrorMessage("Opdater person", e.getMessage()));
        }
        throwExceptionInCaseOfValidationErrors();
    }

    public void validate(MergePersonsRequest request) {
        try {
            checkModifierForDelete(request.getModified());
            checkKeyIdentifiers(Collections.singletonList(request.getFrom()));
            checkKeyIdentifiers(Collections.singletonList(request.getTo()));
        } catch (Exception e) {
            log.info("Unexpected error: " + e.getMessage(), e);
            validationErrors.add(buildUnspecificErrorMessage("Flet personer", e.getMessage()));
        }
        throwExceptionInCaseOfValidationErrors();
    }

    public void validate(SearchPersonRequest request) {
        try {
            validateSearchQuery(request.getQuery());
        } catch (Exception e) {
            log.info("Unexpected error: " + e.getMessage(), e);
            validationErrors.add(buildUnspecificErrorMessage("Søg person", e.getMessage()));
        }
        throwExceptionInCaseOfValidationErrors();
    }

    public void validate(GetPersonByIdRequest request) {
        try {
            checkKeyIdentifiers(request.getKeyIdentifiers());
        } catch (Exception e) {
            log.info("Unexpected error: " + e.getMessage(), e);
            validationErrors.add(buildUnspecificErrorMessage("Indlæs person", e.getMessage()));
        }
        throwExceptionInCaseOfValidationErrors();
    }

    public void validate(DeletePersonRequest request) {
        try {
            checkKeyIdentifiersForDelete(request.getPersons());
            checkModifierForDelete(request.getModified());
        } catch (Exception e) {
            log.info("Unexpected error: " + e.getMessage(), e);
            validationErrors.add(buildUnspecificErrorMessage("Slet person", e.getMessage()));
        }
        throwExceptionInCaseOfValidationErrors();
    }

    private boolean validateSearchQuery(final String query) {
        boolean ok = true;
        ok &= checkIfQueryExist(query);
        ok &= validateQuery(query);
        return ok;
    }

    private boolean validateQuery(String query) {
        boolean ok = true;
        try {
            Map<String, String> r = Splitter.on(" AND ").withKeyValueSeparator(":").split(query);
            for (String key : r.keySet()) {
                ok &= checkFieldName(key.trim(), ValueSets.fieldNameValues);
                ok &= checkValue(r.get(key.trim()));
            }
        } catch (Exception e) {
            ok &= false;
            validationErrors.add(format(ValidationErrors.SEARCH_BAD_QUERY, query));
        }
        return ok;
    }

    private boolean checkFieldName(String value, String[] legalValues) {
        if (Arrays.asList(legalValues).contains(value)) {
            return true;
        }
        validationErrors.add(format(ValidationErrors.SEARCH_BAD_QUERY, value));
        return false;

    }

    private boolean checkIfQueryExist(String query) {
        if(Strings.isBlank(query)) {
            validationErrors.add(ValidationErrors.QUERY_REQUIRED);
            return false;
        }
        return true;
    }

    private void throwExceptionInCaseOfValidationErrors() {
        if (validationErrors.size() > 0) {
            throw new IllegalArgumentException(String.join(", ", validationErrors));
        }
    }

    public boolean checkPerson(PersonType person) {
        boolean ok = true;
        ok &= checkIdentifiers(person.getIdentifiers());
        ok &= checkGenders(person.getGenders());
        ok &= checkModifiers(person.getModifieds());
        ok &= checkContacts(person.getContacts());
        ok &= checkAddresses(person.getAddresses());
        ok &= checkBirthDates(person.getBirthDates());
        ok &= checkNames(person.getNames());
        return ok;
    }

    private String format(String error, String... expressions) {
        try {
            String formatted = String.format(error, expressions);
            return formatted;
        } catch (Exception e) {
            log.warn("Unable to format error: " + error, e);
            return error;
        }
    }

    private boolean checkModifiers(List<Modified> list) {
        if (list == null) {
            validationErrors.add(ValidationErrors.MODIFIER_REQUIRED);
            return false;
        }
        boolean ok = true;
        if (list.size() != 1) {
            validationErrors.add(ValidationErrors.MODIFIER_REQUIRED);
            ok = false;
        }
        if (list.size() > 0) {
            Modified modified = list.get(0);
            if (modified == null || modified.getBy() == null) {
                validationErrors.add(ValidationErrors.MODIFIED_BY_REQUIRED);
                ok = false;
            }
            if (modified != null) {
                ok &= checkValidFrom(modified.getValidFrom());
                ok &= checkValidTo(modified.getValidTo());
            }
        }
        return ok;
    }

    private boolean checkModifierForDelete(Modified modified) {
        if (modified == null) {
            validationErrors.add(ValidationErrors.MODIFIER_REQUIRED);
            return false;
        }
        boolean ok = true;
        if (modified.getBy() == null) {
            validationErrors.add(ValidationErrors.MODIFIED_BY_REQUIRED);
            ok = false;
        }
        ok &= checkValidFrom(modified.getValidFrom());
        ok &= checkValidTo(modified.getValidTo());
        return ok;
    }

    private boolean checkGenders(List<GenderType> list) {
        if (list == null) {
            return true;
        }
        boolean ok = true;
        if (list.size() > 1) {
            validationErrors.add(format(ValidationErrors.DUPLICATE_GENDER));
            ok = false;
        }
        for (GenderType gender : list) {
            ok &= checkGender(gender);
        }
        return ok;
    }

    private boolean checkBirthDates(List<BirthDateType> list) {
        if (list == null) {
            return true;
        }
        boolean ok = true;
        if (list.size() > 1) {
            validationErrors.add(format(ValidationErrors.DUPLICATE_BIRTHDATE));
            ok = false;
        }
        for (BirthDateType birthDate : list) {
            ok &= checkBirthDate(birthDate);
        }
        return ok;
    }

    private boolean checkGender(GenderType gender) {
        if (gender == null || onlyPidIsSupplied(gender)) {
            return true;
        }
        boolean ok = true;
        String value = gender.getValue();
        ok &= checkValue(value, ValueSets.genderValues, format(ValidationErrors.GENDER_VALUE_ILLEGAL, value));
        ok &= checkValidFrom(gender.getValidFrom());
        ok &= checkValidTo(gender.getValidTo());
        return ok;
    }

    private boolean onlyPidIsSupplied(GenderType gender) {
        return onlyPidIsSet(gender) && gender.getValue() == null;
    }

    private boolean checkIdentifiers(List<Identifier> identifiers) {
        if (identifiers == null) {
            return true;
        }
        boolean ok = true;
        Set<String> oidAndValueSet = new HashSet<>();
        for (Identifier identifier : identifiers) {
            ok &= checkKeyIdentifiers(identifier);
            boolean added = oidAndValueSet.add(identifier.getOID() + "," + identifier.getValue());
            if (!added) {
                validationErrors.add(format(ValidationErrors.DUPLICATE_IDENTIFIER, identifier.getValue()));
                ok = false;
            }
        }
        return ok;
    }

    private boolean checkKeyIdentifiers(Identifier identifier) {
        if (identifier == null || onlyPidIsSupplied(identifier) || onlyOIDIsSupplied(identifier)) {
            return true;
        }
        boolean ok = true;
        ok &= checkNull(identifier.getPrimaryOIDLabel(), ValidationErrors.IDENTIFIER_OIDLABEL_FORBIDDEN);
        ok &= checkValidity(identifier.getValidity());
        // OID value checking can not be done generally. See dk.sds.ecpr2.service.handler.Validator
        ok &= checkValidFrom(identifier.getValidFrom());
        ok &= checkValidTo(identifier.getValidTo());
        ok &= checkExpiry(identifier.getExpiry(), identifier.getValidFrom());
        return ok;
    }

    private boolean onlyOIDIsSupplied(Identifier i) {
        return onlyOIDIsSet(i) &&
                i.getPID() == null &&
                i.getValue() == null &&
                i.getPrimaryOIDLabel() == null &&
                i.getValidity() == null &&
                i.getExpiry() == null;
    }

    private boolean onlyOIDIsSet(Identifier e) {
        return e.getOID() != null && e.getValidFrom() == null && e.getValidTo() == null;
    }

    private boolean onlyPidIsSupplied(Identifier i) {
        return onlyPidIsSet(i) &&
                i.getValue() == null &&
                i.getPrimaryOIDLabel() == null &&
                i.getValidity() == null &&
                i.getExpiry() == null;
    }

    private boolean onlyPidIsSet(EntityType e) {
        return e.getPID() != null && e.getValidFrom() == null && e.getValidTo() == null;
    }

    private boolean checkKeyIdentifiers(List<KeyIdentifier> keyIdentifiers) {
        boolean ok = true;
        for (KeyIdentifier keyIdentifier:keyIdentifiers) {
            boolean validateIdentifier = keyIdentifier != null & keyIdentifier.getOID() != null & keyIdentifier.getValue() != null
                    && !keyIdentifier.getOID().isEmpty() && !keyIdentifier.getValue().isEmpty();
                ok &= validateIdentifier;
            if (!ok) {
                validationErrors.add(format(ValidationErrors.PERSONID_VALUES_REQUIRED, keyIdentifier.getValue()));
            }
        }
        return ok;
    }


    private boolean checkKeyIdentifiersForDelete(List<PersonIdType> persons) {
        boolean ok = true;
        for (PersonIdType person : persons) {
            KeyIdentifier keyIdentifier = person.getKeyIdentifier();
            boolean validateIdentifier = keyIdentifier != null & keyIdentifier.getOID() != null & keyIdentifier.getValue() != null
                    && !keyIdentifier.getOID().isEmpty() && !keyIdentifier.getValue().isEmpty();
            ok &= validateIdentifier;
            if (!ok) {
                validationErrors.add(format(ValidationErrors.PERSONID_VALUES_REQUIRED, keyIdentifier.getValue()));
            }
        }
        return ok;
    }


    private boolean checkForMultipleKeyIdentifiers(List<KeyIdentifier> keyIdentifiers) {
        if (keyIdentifiers.size() < 2) {
            //validationErrors.add(format(ValidationErrors.PERSONID_VALUES_REQUIRED, keyIdentifier.getValue()));
            return false;
        }
        return true;
    }

    private boolean checkeCPROID(String oid) {
        if (oid.equals(OID.X_ECPR)) {
            return true;
        }
        validationErrors.add(format(ValidationErrors.CREATEPERSONID_INVALID_OID));
        return false;
    }

    private boolean checkValidity(Integer validity) {
        if (validity == null || !(validity == 0 || validity == 1 || validity == 2 || validity == 4 || validity == 8 || validity == 16)) {
            validationErrors.add(format(ValidationErrors.VALIDITY_REQUIRED, "" + validity));
            return false;
        }
        return true;
    }

    private boolean checkNull(Object s, String errorMessage) {
        if (s != null) {
            validationErrors.add(errorMessage);
            return false;
        } else {
            return true;
        }
    }

    private boolean checkNames(List<NameType> names) {
        if (names == null) {
            return true;
        }
        boolean ok = true;
        Set<String> use = new HashSet<>();
        for (NameType name : names) {
            ok &= checkName(name);
            boolean added = use.add(name.getUse());
            if (!added) {
                validationErrors.add(format(ValidationErrors.DUPLICATE_NAME_USE, name.getUse()));
                ok = false;
            }
        }
        return ok;
    }

    private boolean checkName(NameType name) {
        if (name == null || onlyPidIsSupplied(name)) {
            return true;
        }
        boolean ok = true;
        ok &= checkNameUse(name.getUse());
        ok &= checkValidFrom(name.getValidFrom());
        ok &= checkValidTo(name.getValidFrom());
        ok &= checkNameText(name.getText());
        ok &= checkExpiry(name.getExpiry(), name.getValidFrom());
        // Ingen validering af FamilyName og GivenName udover WSDL
        return ok;
    }

    private boolean onlyPidIsSupplied(NameType name) {
        return onlyPidIsSet(name) &&
                name.getUse() == null &&
                name.getFamilyName() == null &&
                name.getGivenName() == null &&
                name.getText() == null &&
                name.getExpiry() == null;
    }

    private boolean checkNameText(String text) {
        if (text != null) {
            validationErrors.add(ValidationErrors.NAME_TEXT_FORBIDDEN);
            return false;
        }
        return true;
    }

    private boolean checkNameUse(String use) {
        return checkValue(use, ValueSets.nameUseValues, format(ValidationErrors.NAME_USE_INVALID, use));
    }

    private boolean checkBirthDate(BirthDateType birthDate) {
        if (birthDate == null || birthDate.getValue() == null || onlyPidIsSupplied(birthDate)) {
            return true;
        }
        boolean ok = true;
        final Pattern pattern = Pattern.compile(DATE_REGEX);
        final Matcher matcher = pattern.matcher(birthDate.getValue());
        if (!matcher.matches()) {
            validationErrors.add(format(ValidationErrors.BIRTHDATE_BAD_VALUE, birthDate.getValue()));
            ok = false;
        } else if (matcher.group(1) != null && Integer.parseInt(matcher.group(1)) < 1890) {
            validationErrors.add(format(ValidationErrors.BIRTHDATE_BAD_YEAR, matcher.group(1)));
            ok = false;
        }
        ok &= checkValidFrom(birthDate.getValidFrom());
        ok &= checkValidTo(birthDate.getValidFrom());
        return ok;
    }

    private boolean onlyPidIsSupplied(BirthDateType birthDate) {
        return onlyPidIsSet(birthDate) &&
                birthDate.getValue() == null;
    }

    private boolean checkExpiry(Calendar expiry, Calendar validFrom) {
        if (expiry == null) {
            return true;
        }
        if (expiry.before(validFrom)) {
            return false;
        }
        return true;
    }

    private boolean checkContacts(List<ContactType> contacts) {
        if (contacts == null) {
            return true;
        }
        boolean ok = true;
        for (ContactType contact : contacts) {
            ok &= checkContact(contact);
        }
        return ok;
    }

    private boolean checkContact(ContactType contact) {
        if (contact == null || onlyPidIsSupplied(contact)) {
            return true;
        }
        boolean ok = true;
        ok &= checkContactSystem(contact.getSystem());
        ok &= checkContactUse(contact.getUse());
        ok &= checkValidFrom(contact.getValidFrom());
        ok &= checkValidTo(contact.getValidFrom());
        ok &= checkExpiry(contact.getExpiry(), contact.getValidFrom());
        // Ingen validering af PID og Value udover WSDL
        return ok;
    }

    private boolean onlyPidIsSupplied(ContactType contact) {
        return onlyPidIsSet(contact) &&
                contact.getUse() == null &&
                contact.getSystem() == null &&
                contact.getValue() == null &&
                contact.getExpiry() == null;
    }

    private boolean checkContactUse(String use) {
        if (use == null) {
            return true;
        }
        return checkValue(use, ValueSets.contactUseValues, format(ValidationErrors.CONTACT_USE_ILLEGAL, use));
    }

    private boolean checkContactSystem(String system) {
        if (system == null) {
            return true;
        }
        return checkValue(system, ValueSets.contactSystemValues, format(ValidationErrors.CONTACT_TYPE_ILLEGAL, system));
    }

    private boolean checkValue(String value, String[] legalValues, String validationError) {
        for (String legalValue : legalValues) {
            if (legalValue.equals(value)) {
                return true;
            }
        }
        validationErrors.add(validationError);
        return false;
    }

    private boolean checkValue(String value) {
        if (value.isEmpty()) {
            validationErrors.add(format(ValidationErrors.SEARCH_BAD_QUERY, ""));
            return false;
        }
        return true;
    }

    private boolean checkAddresses(List<AddressType> addresses) {
        if (addresses == null) {
            return true;
        }
        boolean ok = true;
        for (AddressType address : addresses) {
            ok &= checkAddress(address);
        }
        return ok;
    }

    public boolean checkAddress(AddressType address) {
        if (address == null || onlyPidIsSupplied(address)) {
            return true;
        }
        boolean ok = true;
        ok &= checkAddressType(address.getType());
        ok &= checkAddressUse(address.getUse());
        ok &= checkCountryName(address.getCountryName());
        ok &= checkValidFrom(address.getValidFrom());
        ok &= checkValidTo(address.getValidFrom());
        ok &= checkCountry(address.getCountry());
        ok &= checkExpiry(address.getExpiry(), address.getValidFrom());
        // Ingen validering af Text, City, PostalCode, District, State og PostBox udover WSDL
        return ok;
    }

    private boolean onlyPidIsSupplied(AddressType address) {
        return onlyPidIsSet(address) &&
                address.getUse() == null &&
                address.getTexts().isEmpty() &&
                address.getCountry() == null &&
                address.getExpiry() == null &&
                address.getCity() == null &&
                address.getCountryName() == null &&
                address.getDistrict() == null &&
                address.getPostalCode() == null &&
                address.getPostBox() == null &&
                address.getState() == null &&
                address.getType() == null;
    }

    private boolean checkCountry(String country) {
        if (country == null) {
            return true;
        }
        if (!CountryCode.isTwoLetterAbbreviationValid(country) &&
            !CountryCode.isThreeLetterAbbreviationValid(country) &&
            !CountryCode.isNumericValueValid(country)) {
                validationErrors.add(format(ValidationErrors.COUNTRY_UNKNOWN, country));
                return false;
        } else {
            return true;
        }
    }

    private boolean checkCountryName(String countryName) {
        if (countryName != null) {
            validationErrors.add(ValidationErrors.COUNTRY_NAME_FORBIDDEN);
            return false;
        }
        return true;
    }

    private boolean checkValidFrom(Calendar validFrom) {
        if (validFrom != null) {
            validationErrors.add(ValidationErrors.VALIDFROM_FORBIDDEN);
            return false;
        }
        return true;
    }

    private boolean checkValidTo(Calendar validTo) {
        if (validTo != null) {
            validationErrors.add(ValidationErrors.VALIDTO_FORBIDDEN);
            return false;
        }
        return true;
    }

    private boolean checkAddressUse(String use) {
        if (use == null) {
            return true;
        }
        return checkValue(use, ValueSets.addressUseValues, format(ValidationErrors.ADDRESS_USE_UNKNOWN, use));
    }

    private boolean checkAddressType(String type) {
        if (type == null) {
            return true;
        }
        return checkValue(type, ValueSets.addressTypeValues, format(ValidationErrors.ADDRESS_TYPE_UNKNOWN, type));
    }

    private String buildUnspecificErrorMessage(String requestName, String errorMessage) {
        return "Parse-fejl ifm. " + requestName + ": " + errorMessage;
    }


}
