/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.DatabaseTest;
import dk.sds.ecpr2.OID;
import dk.sds.ecpr2.dao.DaoTestHelper;
import dk.sds.ecpr2.dao.IdentifierDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.IdentifierVO;
import dk.sds.ecpr2.domain.IdentifierWithoutServiceActorVO;
import dk.sds.ecpr2.util.TimeHelper;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;

import java.util.*;
import java.util.stream.Collectors;

import static org.junit.Assert.*;


public class IdentifierDAOImplTest extends DatabaseTest {


    @Autowired
    private IdentifierDAO dao;

    @Test
    public void createAndGetIdentifier_CreatesCorrectData() throws DAOException {
        // Arrange
        final Date filterDate = TimeHelper.createNew(2018, Calendar.JANUARY, 1);
        List<IdentifierVO> identifiersBefore = dao.getWithServiceActor(DaoTestHelper.PERSON_UUID_FROM_DATA_FILL, filterDate);
        IdentifierVO beforeTest = DaoTestHelper.buildTestIdentifierVO(1, "1234561AA1");

        // Act
        dao.createIdentifier(beforeTest, beforeTest.getValidFrom(), beforeTest.getPersonPID());
        List<IdentifierVO> identifiersAfter = dao.getWithServiceActor(DaoTestHelper.PERSON_UUID_FROM_DATA_FILL, filterDate);

        // Assert
        assertNotNull(identifiersAfter);
        assertEquals(identifiersAfter.size(), identifiersBefore.size() + 1);
        IdentifierVO afterTest = identifiersAfter.get(identifiersAfter.size() - 1);
        assertEquals(beforeTest.getPersonPID(), afterTest.getPersonPID());
        assertEquals(beforeTest.getExpiry(), afterTest.getExpiry());
        assertEquals(beforeTest.getValue(), afterTest.getValue());
        assertEquals(beforeTest.getOid(), afterTest.getOid());
        assertEquals(beforeTest.getValidity(), afterTest.getValidity());
        assertEquals(beforeTest.getValidFrom(), afterTest.getValidFrom());
        // The below statement is only true if we can the rest interface
        // assertEquals(beforeTest.getExpiry(), afterTest.getValidTo());
        assertTrue(afterTest.getModified().before(TimeHelper.getAfterCurrentTime(30)));
        assertNotNull(afterTest.getServiceActor()); // Values are tested in AbstractDAOTest
    }

    @Test
    public void getIdentifiersByPersonUUIDs_FetchesCorrectData() throws DAOException {
        // Arrange
        final List<String> PERSON_UUIDS = Arrays.asList(DaoTestHelper.PERSON_UUID_FROM_DATA_FILL);
        List<IdentifierVO> identifiersBefore = dao.getWithServiceActor(PERSON_UUIDS, Collections.emptyList());
        IdentifierVO beforeTest = DaoTestHelper.buildTestIdentifierVO(1, "1234561AA1");
        dao.createIdentifier(beforeTest, beforeTest.getValidFrom(), beforeTest.getPersonPID());
        IdentifierVO id1 = DaoTestHelper.buildTestIdentifierVO(1, "1234562BB2");
        dao.createIdentifier(id1, id1.getValidFrom(), id1.getPersonPID());
        IdentifierVO id2 = DaoTestHelper.buildTestIdentifierVO(2, "1234563CC3");
        dao.createIdentifier(id2, id2.getValidFrom(), id2.getPersonPID());
        List<IdentifierVO> identifiersAfter = dao.getWithServiceActor(PERSON_UUIDS, Collections.emptyList());

        // Assert
        assertNotNull(identifiersAfter);
        assertEquals(identifiersAfter.size(), identifiersBefore.size() + 3);
        List<String> identifierValuesAfterTest = identifiersAfter.stream().map(IdentifierVO::getValue).collect(Collectors.toList());
        assertTrue(identifierValuesAfterTest.containsAll(Arrays.asList("1234561AA1", "1234562BB2", "1234563CC3")));

        IdentifierVO afterTest = identifiersAfter.get(identifiersAfter.size() - 1);
        assertNotNull(afterTest.getIdentifierPID());
        assertNotNull(afterTest.getPersonPID());
        assertNotNull(afterTest.getServiceActor());
    }

    @Test
    public void getMatchingIdentifierWithoutServiceActor_WithMatch_FetchesCorrectData() throws DAOException {
        IdentifierVO beforeTest = DaoTestHelper.buildTestIdentifierVO(1, "1234561AA1");
        dao.createIdentifier(beforeTest, beforeTest.getValidFrom(), beforeTest.getPersonPID());
        List<IdentifierWithoutServiceActorVO> result = dao.getMatchingIdentifiersWithoutServiceActor("123456%", OID.C_ECPR);

        assertNotNull(result);
        assertEquals(1, result.size());
        IdentifierWithoutServiceActorVO afterTest = result.get(0);
        assertEquals(beforeTest.getIdentifierPID(), afterTest.getIdentifierPID());
        assertEquals(beforeTest.getPersonPID(), afterTest.getPersonPID());
        assertEquals(beforeTest.getExpiry(), afterTest.getExpiry());
        assertEquals(beforeTest.getValue(), afterTest.getValue());
        assertEquals(beforeTest.getOid(), afterTest.getOid());
        assertEquals(beforeTest.getValidity(), afterTest.getValidity());
        assertEquals(beforeTest.getValidFrom(), afterTest.getValidFrom());
        // The below statement is only true if we can the rest interface
        // assertEquals(beforeTest.getExpiry(), afterTest.getValidTo());
        assertTrue(afterTest.getModified().before(TimeHelper.getAfterCurrentTime(30)));
    }

    @Test
    public void getMatchingIdentifierWithoutServiceActor_WithoutMatch_FetchesNothing() throws DAOException {
        IdentifierVO beforeTest = DaoTestHelper.buildTestIdentifierVO(1, "1234561AA1");
        dao.createIdentifier(beforeTest, beforeTest.getValidFrom(), beforeTest.getPersonPID());
        List<IdentifierWithoutServiceActorVO> result = dao.getMatchingIdentifiersWithoutServiceActor("000000%", OID.C_ECPR);

        assertNotNull(result);
        assertEquals(0, result.size());
    }


}
