/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.IntegrationTest;
import dk.sds.ecpr2.OID;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.datagenerator.DataGeneratorApplication;
import dk.sds.ecpr2.domain.Gender;
import dk.sds.ecpr2.domain.IdentifierWithoutServiceActorVO;
import dk.sds.ecpr2.domain.PersonVO;
import dk.sds.ecpr2.service.PersonNumberGeneratorImpl;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;

import java.util.List;
import java.util.Random;

import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

public class PerformanceIT extends IntegrationTest {

    @Autowired
    private PersonDAOImpl personDao;
    @Autowired
    private IdentifierDAOImpl identifierDao;
    @Autowired
    private ServiceActorDAOImpl serviceActorDao;
    @Autowired
    private PersonNumberGeneratorImpl personNumberGenerator;
    @Autowired
    private SearchCrossTableDAOImpl crossTableDAO;

    private static DataGeneratorApplication generator;
    private double firstDurationsAvg;
    private double lastDurationsAvg;

    @Test
    public void getMatchingIdentifierPerformance_WithNumberOfRandomQueries() throws DAOException {
        final int numberOfTestQueries = 1000; // Should be at least 20

        int maxCount = Integer.MIN_VALUE;
        int minCount = Integer.MAX_VALUE;
        int totalCount = 0;
        double totalDuration = 0;
        double maxDuration = Integer.MIN_VALUE;
        double minDuration = Integer.MAX_VALUE;

        for (int i = 0; i < numberOfTestQueries; i++) {
            String idSearchPattern = String.format("0101%02d", i) + "%";

            final double startTime = System.nanoTime();
            List<IdentifierWithoutServiceActorVO> ids = identifierDao.getMatchingIdentifiersWithoutServiceActor(idSearchPattern, OID.C_ECPR);
            final double durationMillis = (System.nanoTime() - startTime) / 1000000;

            int count = ids.size();
            maxCount = Math.max(count, maxCount);
            minCount = Math.min(count, minCount);
            totalCount++;
            maxDuration = Math.max(durationMillis, maxDuration);
            minDuration = Math.min(durationMillis, minDuration);
            totalDuration += durationMillis;
        }

        double averageDuration = totalDuration / numberOfTestQueries;
        double averageCount = totalCount / numberOfTestQueries;

        System.out.println("Average Duration over " + numberOfTestQueries + " calls: " + averageDuration + " ms");
        System.out.println("Smallest count: " + minCount);
        System.out.println("Largest count: " + maxCount);
        System.out.println("Average count: " + averageCount);
        System.out.println("Smallest duration: " + minDuration);
        System.out.println("Largest duration: " + maxDuration);

        assertTrue("The query fetching identifers is way too slow. Some change most likely broke performance.",
                averageDuration < 500);
        if (averageDuration > 50) {
            printVisibleMessage("Potential performance problem with getMatchingIdentifiers");
        }
    }

    @Test
    public void creatingNewCECPRs_ShouldNotGetSlowerOverTime() throws DAOException {
        final int numberOfTests = 1000; // Should be at least 100
        Random rnd = new Random();
        double firstDurations = 0;
        double lastDurations = 0;
        double totalDuration = 0;

        for (int i = 0; i < numberOfTests; i++) {
            Gender randomGender = rnd.nextInt(2) == 0 ? Gender.MALE : Gender.FEMALE;
            final int randomDay = rnd.nextInt(27) + 1;
            final int randomMonth = rnd.nextInt(11) + 1;
            final int randomYear = rnd.nextInt(100) + 1918;

            final double startTime = System.nanoTime();
            String id = personNumberGenerator.generateCECPR(randomDay, randomMonth, randomYear, randomGender, true);
            final double durationMillis = (System.nanoTime() - startTime) / 1000000;
            assertNotNull(id);

            totalDuration += durationMillis;
            if (i < numberOfTests / 10) {  // First 10%
                firstDurations += durationMillis;
            }
            if (i > 9 * numberOfTests / 10) { // Last 10%
                lastDurations += durationMillis;
            }
        }

        final double averageDuration = (totalDuration / numberOfTests);
        firstDurationsAvg = firstDurations / numberOfTests / 10;
        lastDurationsAvg = lastDurations / numberOfTests / 10;

        System.out.printf("Total duration average     : %.3f ms\n", averageDuration);
        System.out.printf("First 10%% duration average : %.3f ms\n", firstDurationsAvg);
        System.out.printf("Last 10%% duration average  : %.3f ms\n", lastDurationsAvg);

        if (firstDurationsAvg < lastDurationsAvg + 0.1 * lastDurationsAvg) {
            printVisibleMessage("Creating new ECPR IDs seems to be getting slower over time");
        }
        assertTrue("Creating new ECPR IDs becomes much slower over time",
                2 * firstDurationsAvg > lastDurationsAvg);

        assertTrue("Creating new ECPR IDs is way too slow. Some change most likely broke performance.",
                averageDuration < 500);
        if (averageDuration > 50) {
            printVisibleMessage("Potential performance problem with creating new ECPR IDs");
        }

    }

    @Test
    public void fetchingDataForLuceneIsFastEnough() throws DAOException {
        final double startTime = System.nanoTime();

        // FIXME: SearchCross
        List<PersonVO> allPeople = crossTableDAO.getPersonsWithRelationsAfterPID(0, 1000, false, false);
        final double durationMillis = (System.nanoTime() - startTime) / 1000000;

        System.out.println("Duration: " + durationMillis);

        assertTrue("Fetching person information for Lucene is way too slow. Some change most likely broke performance.",
                durationMillis < 30000);
        if (durationMillis > 5000) {
            printVisibleMessage("Potential performance problem with getAllPersonInformation");
        }
    }

    private void assertRightNumberOfRowsInTable(int actualRows, int expectedRows, String tableName) {
        assertTrue("There is not a realistic amount of data in the table: " + tableName
                        + " Found only " + actualRows + " and expected at least " + expectedRows,
                actualRows > expectedRows);
    }

    private void printVisibleMessage(String message) {
        System.out.println("#######################################################################################");
        System.out.println("WARNING: " + message);
        System.out.println("#######################################################################################");
    }


}
