/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.IntegrationTest;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.*;
import dk.sds.ecpr2.util.ValueSets;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.mock.mockito.MockBean;

import java.time.Instant;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import static org.junit.Assert.assertEquals;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.when;

public class SearchPersonDAOImplTest extends IntegrationTest {
    @MockBean
    SearchCrossTableDAOImpl crossTableDAOMock;

    @Autowired
    SearchPersonDAOImpl searchPersonDAO;

    @Before
    public void setUpIndex() {
        searchPersonDAO.setCrossTableDAO(crossTableDAOMock);
    }

    @Test
    public void searchPerson_ByPersonUUID_PersonFound() throws Exception {
        // Arrange
        List<PersonVO> expected = Collections.singletonList(CreateFullPerson(1, "123", "Jacob", "Hansen"));
        mockDatabaseAccess(expected);
        searchPersonDAO.buildIndex(null);

        // Act
        List<PersonVO> list = searchPersonDAO.searchPerson(Collections.singletonList("personuuid:123"));

        // Assert
        PersonVO p = list.get(0);
        assertEquals("123", p.getPersonUUID());
        assertEquals("Jacob", p.getNames().get(0).getGivenName());
        assertEquals("Hansen", p.getNames().get(0).getFamilyName());
        assertEquals("Soender Alle", p.getAddresses().get(0).getStreetAddressLine1());
        assertEquals("Aarhus", p.getAddresses().get(0).getCity());
        assertEquals("DK", p.getAddresses().get(0).getCountry());
        assertEquals("Dummy District", p.getAddresses().get(0).getDistrict());
        assertEquals("Dummy State", p.getAddresses().get(0).getState());
        assertEquals("88888888", p.getContacts().get(0).getValue());
        assertEquals("21/11/1990", p.getBirthDates().get(0).getDay() + "/" + p.getBirthDates().get(0).getMonth() + "/" + p.getBirthDates().get(0).getYear());
        assertEquals("male", p.getGenders().get(0).getValue());
    }

    @Test
    public void searchPerson_ByPersonUUID_PersonNotFound() throws Exception {
        // Arrange
        List<PersonVO> persons = Collections.singletonList(CreateFullPerson(2, "234", "Jacob", "Hansen"));
        mockDatabaseAccess(persons);
        searchPersonDAO.buildIndex(null);

        // Act
        List<PersonVO> actual = searchPersonDAO.searchPerson(Collections.singletonList("personuuid:1155"));

        // Assert
        assertEquals(0, actual.size());
    }

    @Test
    public void searchPerson_ByFamilyName_PersonFound() throws Exception {
        // Arrange
        List<PersonVO> expected = Collections.singletonList(CreateFullPerson(2, "234", "Jacob", "Hansen"));
        mockDatabaseAccess(expected);
        searchPersonDAO.buildIndex(null);

        // Act
        List<PersonVO> actual = searchPersonDAO.searchPerson(Collections.singletonList("familyname:Hansen"));

        // Assert
        assertEquals(expected.size(), actual.size());
        assertEquals("234", actual.get(0).getPersonUUID());
    }

    @Test
    public void searchPerson_ByDanishCharacters_PersonFound() throws Exception {
        // Arrange
        List<PersonVO> expected = Collections.singletonList(CreateFullPerson(2, "234", "Jacob", "æøå"));
        mockDatabaseAccess(expected);
        searchPersonDAO.buildIndex(null);

        // Act
        List<PersonVO> actual = searchPersonDAO.searchPerson(Collections.singletonList("familyname:æøå"));

        // Assert
        assertEquals(expected.size(), actual.size());
        assertEquals("234", actual.get(0).getPersonUUID());
    }

    @Test
    public void searchPerson_ByFamilyName_MultipleMatchesFound() throws Exception {
        // Arrange
        List<PersonVO> expected = Arrays.asList(
                CreateFullPerson(1, "123", "Anders", "Hansen"),
                CreateFullPerson(2, "234", "Jacob", "Hansen"));
        mockDatabaseAccess(expected);
        searchPersonDAO.buildIndex(null);

        // Act
        List<PersonVO> actual = searchPersonDAO.searchPerson(Collections.singletonList("familyname:Hansen"));

        // Assert
        assertEquals(expected.size(), actual.size());
    }

    private void mockDatabaseAccess(List<PersonVO> personlist) throws DAOException {
        when(crossTableDAOMock.getPersonsWithRelationsAfterPID(anyInt(), anyInt(), anyBoolean(), anyBoolean())).thenReturn(personlist).thenReturn(new ArrayList<>());
        when(crossTableDAOMock.getPersonsWithRelations(any())).thenReturn(personlist);
    }

    @Test
    public void updateIndex_UpdateSucces_SearchPersonByPersonUUID() throws Exception {
        // Arrange
        List<PersonVO> initialVersion = Collections.singletonList(CreateFullPerson(1, "123", "Jens", "Hansen"));
        List<PersonVO> updatedVersion = Collections.singletonList(CreateFullPerson(1, "123", "Jens-Peter", "Hansen"));

        mockDatabaseAccess(initialVersion);
        searchPersonDAO.buildIndex(null);
        List<PersonVO> firstSearchResult = searchPersonDAO.searchPerson(Collections.singletonList("personuuid:123"));
        assertEquals("Jens", firstSearchResult.get(0).getNames().get(0).getGivenName());

        // Act
        mockDatabaseAccess(updatedVersion);
        searchPersonDAO.buildIndex(Instant.now());
        List<PersonVO> secondSearchResult = searchPersonDAO.searchPerson(Collections.singletonList("personuuid:123"));

        // Assert
        assertEquals(updatedVersion.size(), secondSearchResult.size());
        assertEquals("Jens-Peter", secondSearchResult.get(0).getNames().get(0).getGivenName());
    }

    @Ignore
    @Test
    public void deleteIndex_DeleteSucces_SearchPersonNotFound() throws Exception {
        // Arrange
        List<PersonVO> expected = Collections.singletonList(CreateFullPerson(2, "234", "Jacob", "Hansen"));
        mockDatabaseAccess(expected);
        searchPersonDAO.buildIndex(null);
        List<PersonVO> firstSearchResult = searchPersonDAO.searchPerson(Collections.singletonList("familyname:Hansen"));
        assertEquals(expected.size(), firstSearchResult.size());

        // Act
        searchPersonDAO.deleteIndex(expected);
        List<PersonVO> secondSearchResult = searchPersonDAO.searchPerson(Collections.singletonList("familyname:Hansen"));

        // Assert
        assertEquals(0, secondSearchResult.size());
    }


    private PersonVO CreateFullPerson(Integer personPID, String personUUID, String givenName, String familyName) {
        PersonVO person = new PersonVO();
        person.setPersonPID(personPID);
        person.setPersonUUID(personUUID);
        person.getAddresses().add(CreateAddress());
        person.getBirthDates().add(createBirthDate());
        person.getContacts().add(createContact());
        person.getNames().add(createName(givenName, familyName));
        person.getGenders().add(createGender());

        return person;
    }

    private AddressVO CreateAddress() {
        AddressVO address = new AddressVO();
        address.setStreetAddressLine1("Soender Alle");
        address.setCity("Aarhus");
        address.setCountry("DK");
        address.setDistrict("Dummy District");
        address.setPostalCode("8000");
        address.setPostBox("Postbox 1000");
        address.setState("Dummy State");
        address.setType(ValueSets.addressTypeValues[0]);
        address.setUse(ValueSets.addressUseValues[0]);
        return address;
    }

    private GenderVO createGender() {
        GenderVO gender = new GenderVO();
        gender.setValue(ValueSets.genderValues[0]);
        return gender;
    }

    private NameVO createName(String givenName, String familyName) {
        NameVO name = new NameVO();
        name.setGivenName(givenName);
        name.setFamilyName(familyName);
        name.setUse(ValueSets.nameUseValues[0]);
        return name;
    }

    private ContactVO createContact() {
        ContactVO contact = new ContactVO();
        contact.setSystem(ValueSets.contactSystemValues[0]);
        contact.setUse(ValueSets.contactUseValues[0]);
        contact.setValue("88888888");
        return contact;
    }

    private BirthDateVO createBirthDate() {
        BirthDateVO birthDate = new BirthDateVO();
        birthDate.setDay(21);
        birthDate.setMonth(11);
        birthDate.setYear(1990);

        return birthDate;
    }
}
