/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.DatabaseTest;
import dk.sds.ecpr2.dao.DaoTestHelper;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.ServiceActorVO;
import dk.sds.ecpr2.util.TimeHelper;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;

import java.util.Arrays;
import java.util.Date;

import static org.junit.Assert.*;

public class ServiceActorDAOImplTest extends DatabaseTest {

    @Autowired
    private ServiceActorDAOImpl dao;

    @Test
    public void getServiceActorById_UsingPrefilledTestData_FetchesCorrectServiceActor() throws DAOException {
        ServiceActorVO sa = dao.getServiceActor(1);
        assertEquals("12345", sa.getAuthorisationNumber());
        assertEquals("Max ÆØÅæøå", sa.getFirstName());
        assertEquals("Test ÆØÅæøå Berggren", sa.getLastName());
        assertTrue(sa.getModifiedTime().before(TimeHelper.getCurrentTime()));
        assertEquals("8001149", sa.getOrganizationId());
        assertEquals("Regionshospital Randers ÆØÅæøå", sa.getOrganizationName());
        assertEquals("0107729995", sa.getPersonIdentifier());
        assertEquals("1.2.208.176.1.2", sa.getPersonIdentifierOID());
        assertEquals("7170", sa.getRole());
    }

    @Test
    public void createServiceActor_InsertsServiceActorThatIsEqualToOriginal() throws DAOException {
        final ServiceActorVO serviceActor = DaoTestHelper.buildTestServiceActorVO("0102030123");

        Date now = new Date();
        final int pid = dao.createServiceActor(serviceActor, now);
        assertNotNull(pid);

        final ServiceActorVO insertedServiceActor = dao.getServiceActor(pid);
        assertNotNull(insertedServiceActor);
        assertEquals("service actors must be equal", serviceActor, insertedServiceActor);
        assertTrue("Should be able to find inserted actor by hashcode", dao.exists(serviceActor));
    }

    @Test
    public void createServiceActorsBatch_InsertsRightNumberOfRows() throws DAOException {
        final ServiceActorVO serviceActor1 = DaoTestHelper.buildTestServiceActorVO("0102030123");
        final ServiceActorVO serviceActor2 = DaoTestHelper.buildTestServiceActorVO("0102030123");

        int rowCountBefore = dao.rowCount();
        dao.createServiceActors(Arrays.asList(serviceActor1, serviceActor2));

        int rowCountAfter = dao.rowCount();
        assertEquals(2, rowCountAfter - rowCountBefore);
    }

    @Test
    public void exists_WhenItExists_ReturnsTrue() throws DAOException {
        final ServiceActorVO sa = dao.getServiceActor(1);
        assertTrue("Did not find any service actor rows with the hash value: " + sa.hashCode(), dao.exists(sa));
    }

    @Test
    public void exists_WhenItDoesntExist_ReturnsFalse() throws DAOException {
        final ServiceActorVO sa = DaoTestHelper.buildTestServiceActorVO("DoesNotExist");
        assertFalse(dao.exists(sa));
    }

}
