/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service.ExportHistory;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import dk.sds.ecpr2.AddressType;
import dk.sds.ecpr2.NameType;
import dk.sds.ecpr2.PersonIdentifierType;
import dk.sds.ecpr2.PersonType;
import dk.sds.ecpr2.dao.DaoTestHelper;
import dk.sds.ecpr2.dao.ExportHistoryDAO;
import dk.sds.ecpr2.dao.IdentifierDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.AddressVO;
import dk.sds.ecpr2.domain.ExportHistoryVO;
import dk.sds.ecpr2.domain.IdentifierVO;
import dk.sds.ecpr2.domain.NameVO;
import dk.sds.ecpr2.service.historyExport.ExportHistoryDataSetVO;
import dk.sds.ecpr2.service.historyExport.ExportHistoryFileFormat;
import dk.sds.ecpr2.service.historyExport.ExportHistoryServiceImpl;
import dk.sds.ecpr2.service.historyExport.NspManager;
import dk.sds.ecpr2.util.TimeHelper;
import dk.sds.ecpr2.webservice.WebServiceTestAbstract;
import dk.sds.ecpr2.webservice.mapping.FromXMLMapper;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;

import java.io.*;
import java.text.SimpleDateFormat;
import java.time.Instant;
import java.util.*;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import static org.junit.Assert.*;


public class ExportHistoryImplTest extends WebServiceTestAbstract {
    @Autowired
    FromXMLMapper requestMapper;

    @Autowired
    private IdentifierDAO daos;

    @Autowired
    NspManager nspManager;

    @Autowired
    ExportHistoryServiceImpl ecpr2ExportService;

    @Autowired
    ExportHistoryDAO exportHistoryDAO;

    @Value("${completedFileName}")
    String completedFileName;

    private String destination = null;
    private Instant startTime = null;

    private List<PersonType> persons = null;

    @Before
    public void onSetUp() throws Exception {
        this.startTime = Instant.now().minusSeconds(1);
        this.persons = createPersons();
    }

    @After
    public void tearDown() throws Exception {
        if (this.destination != null) {
            FileUtils.forceDelete(new File(this.destination));
            this.destination = null;
        }
    }

    @Test
    public void TestJSONFormat() throws DAOException {
        final Date filterDate = TimeHelper.createNew(2018, Calendar.JANUARY, 1);
        IdentifierVO beforeTest = DaoTestHelper.buildTestIdentifierVO(1, "1234561AA1");
        IdentifierVO beforeBeforeTest = DaoTestHelper.buildTestIdentifierVO(1, "1234561BB1");
        beforeBeforeTest.setUuid(beforeTest.getUuid());
        beforeBeforeTest.setValidFrom(Date.from(beforeBeforeTest.getValidFrom().toInstant().minusSeconds(1000)));
        beforeBeforeTest.setValidTo(beforeTest.getValidFrom());

        // Act
        daos.createIdentifier(beforeBeforeTest, beforeBeforeTest.getValidFrom(), beforeBeforeTest.getPersonPID());
        daos.createIdentifier(beforeTest, beforeTest.getValidFrom(), beforeTest.getPersonPID());
        List<IdentifierVO> identifiersAfter = daos.getWithServiceActor(DaoTestHelper.PERSON_UUID_FROM_DATA_FILL, filterDate);
        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
        String date = dateFormat.format(new Date());
        ExportHistoryDataSetVO historyDataSet = new ExportHistoryDataSetVO();
        historyDataSet.setDatatype("identifier");
        historyDataSet.setExportNumber(10);
        historyDataSet.setExportTime(date);
        historyDataSet.setVersion(1.0);

        ExportHistoryFileFormat<IdentifierVO> exportHistoryFormat = new ExportHistoryFileFormat<>();
        exportHistoryFormat.setData(identifiersAfter);
        exportHistoryFormat.setDataset(historyDataSet);

        ObjectMapper mapper = new ObjectMapper();
        mapper.enable(SerializationFeature.INDENT_OUTPUT);
        try {
            File tmpFile = nspManager.saveToTempFile(exportHistoryFormat, "identifiers", new Date());
            tmpFile = Objects.requireNonNull(tmpFile.listFiles((dir, name) -> name.equals("identifiers.json")))[0];
            BufferedReader buf = new BufferedReader(new InputStreamReader(new FileInputStream(tmpFile)));
            Iterator<String> template = getIdentifierJsonTemplate().iterator();
            String fileLine = buf.readLine();
            while (fileLine != null) {
                String templateLine = template.next();
                assertNotNull(templateLine);

                assertTrue(fileLine, fileLine.contains(templateLine));

                fileLine = buf.readLine();
            }
            assertFalse(template.hasNext());
            buf.close();

        } catch (Exception e) {
            e.printStackTrace();
            fail();
        }
    }

    private List<String> getIdentifierJsonTemplate() {
        return Arrays.asList(
                "{",
                "\"dataset\" : {",
                "\"datatype\" : \"identifier\",",
                "\"version\" : 1.0,",
                "\"exportTime\" : \"",
                "\"exportNumber\" : 10",
                "},",
                "\"data\" : [ {",
                "\"modified\" : ",
                "\"validFrom\" : ",
                "\"validTo\" : ",
                "\"uuid\" : \"",
                "\"value\" : \"1234561AA1\",",
                "\"oid\" : \"",
                "\"expiry\" : ",
                "\"validity\" : 16",
                "\"personUUID\" : \"",
                "} ]",
                "}"
        );
    }

    @Test
    public void testCompletedExportHistory() throws Exception {
        Date now = TimeHelper.getCurrentTime();
        ecpr2ExportService.exportHistoryToFiles(null, 1);
        ecpr2ExportService.updateExportHistoryTable(now.toInstant());
        ExportHistoryVO latestExport = exportHistoryDAO.getLatestExport();
        // assertEquals(latestExport.getExportTime().getTime(), truncate(now));
        assertEquals((long) latestExport.getExportNumber(), 1);
    }

    private long truncate(Date date) {
        Calendar cal = Calendar.getInstance();
        cal.setTime(date);
        cal.set(Calendar.HOUR_OF_DAY, 0);
        cal.set(Calendar.MINUTE, 0);
        cal.set(Calendar.SECOND, 0);
        cal.set(Calendar.MILLISECOND, 0);
        return cal.getTimeInMillis();
    }

    @Test
    public void exportHistory_exportToFiles_zipFileToUploadCreated() throws Exception {
        int exportNumber = 1;
        File exportHistoryFolder = ecpr2ExportService.exportHistoryToFiles(this.startTime, exportNumber);
        this.destination = exportHistoryFolder.getAbsolutePath();
        boolean validateZipFileExist = validateZipIsCreated(exportHistoryFolder);
        assertTrue(exportHistoryFolder.exists());
        assertTrue(validateZipFileExist);
    }

    private boolean validateZipIsCreated(File exportHistoryFolder) {
        File[] listOfFiles = exportHistoryFolder.listFiles();
        if (listOfFiles != null && listOfFiles.length > 0) {
            for (File file : listOfFiles) {
                String fileName = file.getName();
                if (file.isFile() && fileName.endsWith(".zip")) {
                    return true;
                }
            }
        }
        return false;
    }

    @Test
    public void exportHistory_exportToFiles_emptyDoneFileCreated() throws Exception {
        int exportNumber = 1;

        File exportHistoryFolder = ecpr2ExportService.exportHistoryToFiles(this.startTime, exportNumber);
        this.destination = exportHistoryFolder.getAbsolutePath();
        File[] listOfFiles = exportHistoryFolder.listFiles();
        if (listOfFiles != null && listOfFiles.length > 0) {
            for (File file : listOfFiles) {
                if ("zip".equals(FilenameUtils.getExtension(file.getName()))) {
                    boolean validateDoneFile = checkDoneFileExist(file);
                    assertTrue(validateDoneFile);
                }
            }
        }
    }

    @Test
    public void exportHistory_exportToFiles_validateFilesContent() throws Exception {
        int exportNumber = 1;

        File exportHistoryFolder = ecpr2ExportService.exportHistoryToFiles(this.startTime, exportNumber);
        this.destination = exportHistoryFolder.getAbsolutePath();
        File[] listOfFiles = exportHistoryFolder.listFiles();
        if (listOfFiles != null) {
            for (File file : listOfFiles) {
                validateContentFile(file);
            }
        }
    }

    @Test
    public void exportHistory_exportToFiles_UpdatedAddress() throws Exception {
        int exportNumber = 1;
        ecpr2ExportService.exportHistoryToFiles(this.startTime, exportNumber);

        PersonType person = this.persons.get(0);
        AddressType address = person.getAddresses().get(0);
        address.setCity("Testbyen");

        updatePerson();
        exportNumber = 2;
        this.startTime = Instant.now().minusSeconds(1);
        File deltaFolder = ecpr2ExportService.exportHistoryToFiles(this.startTime, exportNumber);
        this.destination = deltaFolder.getAbsolutePath();
        File[] zipFiles = deltaFolder.listFiles();
        assertNotNull(zipFiles);
        Arrays.sort(zipFiles, Comparator.comparingLong(File::lastModified));
        if (zipFiles.length == 2) {
            File deltaZipFile = zipFiles[1];
            validateUpdatedAddress(deltaZipFile, address.getPID(), address.getCity());
        }
    }

    @Test
    public void exportHistory_exportToFiles_CreateAndUpdateWithSameExport() throws Exception {
        int exportNumber = 1;

        PersonType person = this.persons.get(0);
        NameType nameType = person.getNames().get(0);
        nameType.setGivenName("John");

        updatePerson();
        File deltaFolder = ecpr2ExportService.exportHistoryToFiles(this.startTime, exportNumber);
        this.destination = deltaFolder.getAbsolutePath();
        File[] zipFiles = deltaFolder.listFiles();
        assertNotNull(zipFiles);
        if (zipFiles.length == 1) {
            File deltaZipFile = zipFiles[0];
            validateNames(deltaZipFile);
        }
    }

    @Test
    public void exportHistory_exportToFiles_DeletedAddress() throws Exception {
        int exportNumber = 1;
        ecpr2ExportService.exportHistoryToFiles(this.startTime, exportNumber);

        PersonType person = this.persons.get(0);
        AddressType address = person.getAddresses().get(0);

        AddressType addressType = new AddressType();
        addressType.setPID(address.getPID());
        person.getAddresses().remove(address);
        person.getAddresses().add(addressType);

        PersonType updatedPersonWithDeletedAddress = updatePerson();
        String pid = updatedPersonWithDeletedAddress.getAddresses().get(0).getPID();
        long newValidTo = updatedPersonWithDeletedAddress.getAddresses().get(0).getValidTo().getTimeInMillis();

        this.startTime = Instant.now().minusSeconds(1);
        exportNumber = 2;
        File deltaFolder = ecpr2ExportService.exportHistoryToFiles(this.startTime, exportNumber);

        this.destination = deltaFolder.getAbsolutePath();
        File[] zipFiles = deltaFolder.listFiles();
        Arrays.sort(zipFiles, Comparator.comparingLong(File::lastModified));
        if (zipFiles != null) {
            if (zipFiles.length == 2) {
                File deltaZipFile = zipFiles[1];
                validateDeletedAddress(deltaZipFile, pid, newValidTo);
            }
        }
    }

    private void validateContentFile(File file) throws IOException {
        Map<String, String> expectedFiles = new HashMap<String, String>() {{
            put("persons.json", "person");
            put("identifiers.json", "identifier");
            put("names.json", "name");
            put("genders.json", "gender");
            put("birthdates.json", "birthdate");
            put("contacts.json", "contact");
            put("addresses.json", "address");
            put("oids.json", "oid");
            put("DONE", "");
        }};
        FileInputStream fileInputStream = new FileInputStream(file);
        ZipInputStream zipInputStream = new ZipInputStream(new BufferedInputStream(fileInputStream));
        ZipEntry entry;
        try {
            while ((entry = zipInputStream.getNextEntry()) != null) {
                assertTrue(entry.getName() + " is not supposed to be included", expectedFiles.containsKey(entry.getName()));
                HashMap<String, Object> keyValuePairs = new HashMap<>();
                keyValuePairs.put("datatype", expectedFiles.get(entry.getName()));
                if (!"DONE".equals(entry.getName())) {
                    boolean hasDataType = hasKeyValuePair(zipInputStream, keyValuePairs);
                    assertTrue(hasDataType);
                }
                expectedFiles.remove(entry.getName());
            }
        } finally {
            zipInputStream.close();
            fileInputStream.close();
        }
        assertTrue(expectedFiles.isEmpty());
    }

    private void validateNames(File file) throws IOException {
        FileInputStream fileInputStream = new FileInputStream(file);
        ZipInputStream zipInputStream = new ZipInputStream(new BufferedInputStream(fileInputStream));
        ZipEntry entry;
        try {
            while ((entry = zipInputStream.getNextEntry()) != null) {
                if ("names.json".equals(entry.getName())) {
                    ExportHistoryFileFormat<NameVO> exportHistoryFormat = this.getExportHistoryFormat(
                            zipInputStream,
                            new TypeReference<ExportHistoryFileFormat<NameVO>>() {
                            });
                    List<NameVO> data = exportHistoryFormat.getData();
                    assertEquals("There should be two names in the dataset", 3, data.size());
                    assertEquals("Benny", data.get(0).getGivenName());
                    assertEquals("Benny", data.get(1).getGivenName());
                    assertEquals("John", data.get(2).getGivenName());
                    break;
                }
            }
        } finally {
            zipInputStream.close();
            fileInputStream.close();
        }
    }

    private void validateUpdatedAddress(File file, String uuid, String city) throws IOException {
        FileInputStream fileInputStream = new FileInputStream(file);
        ZipInputStream zipInputStream = new ZipInputStream(new BufferedInputStream(fileInputStream));
        ZipEntry entry;
        try {
            while ((entry = zipInputStream.getNextEntry()) != null) {
                if ("addresses.json".equals(entry.getName())) {
                    ExportHistoryFileFormat<AddressVO> exportHistoryFormat = this.getExportHistoryFormat(
                            zipInputStream,
                            new TypeReference<ExportHistoryFileFormat<AddressVO>>() {
                            });
                    List<AddressVO> data = exportHistoryFormat.getData();
                    boolean hasUuid = false;
                    boolean hasCity = false;
                    for (AddressVO datum : data) {
                        if (datum.getUuid().equals(uuid)) {
                            hasUuid = true;
                        }
                        if (datum.getCity().equals(city)) {
                            hasCity = true;
                        }
                    }
                    assertTrue(hasUuid && hasCity);
                    break;
                }
            }
        } finally {
            zipInputStream.close();
            fileInputStream.close();
        }
    }

    private void validateDeletedAddress(File file, String uuid, long validToTimeInMillis) throws IOException {
        FileInputStream fileInputStream = new FileInputStream(file);
        ZipInputStream zipInputStream = new ZipInputStream(new BufferedInputStream(fileInputStream));
        ZipEntry entry;
        try {
            while ((entry = zipInputStream.getNextEntry()) != null) {
                if ("addresses.json".equals(entry.getName())) {
                    HashMap<String, Object> keyValuePairs = new HashMap<>();
                    keyValuePairs.put("uuid", uuid);
                    keyValuePairs.put("validTo", validToTimeInMillis);
                    boolean addressExists = hasKeyValuePair(zipInputStream, keyValuePairs);
                    assertTrue(addressExists);
                }
            }
        } finally {
            zipInputStream.close();
            fileInputStream.close();
        }
    }

    private boolean hasKeyValuePair(ZipInputStream zipInputStream, HashMap<String, Object> keyValuePairs) throws IOException {
        if (keyValuePairs.size() == 0) {
            return true;
        }
        int foundMatches = 0;
        Scanner scanner = new Scanner(zipInputStream);
        while (scanner.hasNextLine()) {
            String nextLine = scanner.nextLine();
            for (Map.Entry<String, Object> entry : keyValuePairs.entrySet()) {
                if (entry.getValue() instanceof String) {
                    String keyValueFormat = String.format("\"%s\" : \"%s\"", entry.getKey(), entry.getValue());
                    if (nextLine.contains(keyValueFormat)) {
                        foundMatches++;
                    }
                } else if (entry.getValue() instanceof Long) {
                    String keyValueFormat = String.format("\"%s\" : %d", entry.getKey(), entry.getValue());
                    if (nextLine.contains(keyValueFormat)) {
                        foundMatches++;
                    }
                }
                // Add more instanceof if there are other types that need to be checked
            }
        }
        return foundMatches == keyValuePairs.size();
    }

    private boolean checkDoneFileExist(File file) throws IOException {
        boolean completedFileExist = false;
        FileInputStream fileInputStream = new FileInputStream(file);
        ZipInputStream zis = new ZipInputStream(new BufferedInputStream(fileInputStream));
        ZipEntry entry;
        try {
            while ((entry = zis.getNextEntry()) != null) {
                if (entry.getName().equals(completedFileName)) {
                    completedFileExist = true;
                }
            }
            return completedFileExist;
        } finally {
            zis.close();
            fileInputStream.close();
        }
    }

    private List<PersonType> createPersons() throws Exception {
        List<PersonType> createdPersons = new ArrayList<>();
        PersonType p1 = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        PersonType p2 = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        createdPersons.add(p1);
        createdPersons.add(p2);
        return createPersonsInDb(createdPersons);
    }

    private PersonType updatePerson() throws Exception {
        PersonType personType = persons.get(0);
        PersonIdentifierType personId = new PersonIdentifierType();
        personId.setPersonIdentifier("1111111118");
        personId.setSource("CPR");
        personType.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personId);
        return updatePersonInDb(personType);
    }

    private <T> ExportHistoryFileFormat<T> getExportHistoryFormat(ZipInputStream inputStream, TypeReference<ExportHistoryFileFormat<T>> typeReference) throws IOException {
        ObjectMapper objectMapper = new ObjectMapper();
        return objectMapper.readValue(inputStream, typeReference);
    }
}
