/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service;

import dk.sds.ecpr2.UnitTest;
import dk.sds.ecpr2.dao.IdentifierDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.Gender;
import dk.sds.ecpr2.domain.IdentifierWithoutServiceActorVO;
import dk.sds.ecpr2.util.ECPRException;
import dk.sds.ecpr2.util.FaultCode;
import org.junit.Test;

import java.time.LocalDate;
import java.util.ArrayList;
import java.util.List;

import static org.junit.Assert.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class PersonNumberGeneratorImplTest extends UnitTest {

    private IdentifierDAO daoMock = mock(IdentifierDAO.class);
    private PersonNumberGenerator generator = new PersonNumberGeneratorImpl(daoMock);

    @Test
    public void generateCECPR_WithValidInputAfter1999_ReturnsValidId() {
        String id = generator.generateCECPR(1, 1, 2000, Gender.FEMALE, true);
        assertCorrectCECPR(id, "010100", "6", PersonNumberGeneratorImpl.VALID_FEMALE_DIGITS);
    }

    @Test
    public void generateCECPR_WithValidInputBetween1900And2000_ReturnsValidId() {
        String id = generator.generateCECPR(1, 1, 1900, Gender.MALE, true);
        assertCorrectCECPR(id, "010100", "0", PersonNumberGeneratorImpl.VALID_MALE_DIGITS);
    }

    @Test
    public void generateCECPR_WithValidInputBefore1900_ReturnsValidId() {
        String id = generator.generateCECPR(1, 1, 1899, Gender.MALE, true);
        assertCorrectCECPR(id, "010199", "5", PersonNumberGeneratorImpl.VALID_MALE_DIGITS);
    }

    @Test
    public void generateCECPR_WithInvalidDay_ThrowsException() {
        generateECPRAndExpectException(32, 1, 2018, Gender.MALE, FaultCode.CREATE_NEW_ECPR_ERROR, "date");
    }

    @Test
    public void generateCECPR_WithInvalidMonth_ThrowsException() {
        generateECPRAndExpectException(31, 13, 2018, Gender.MALE, FaultCode.CREATE_NEW_ECPR_ERROR, "date");
    }

    @Test
    public void generateCECPR_WithInvalidDayYearCombination_ThrowsException() {
        generateECPRAndExpectException(30, 2, 2018, Gender.MALE, FaultCode.CREATE_NEW_ECPR_ERROR, "date");
    }

    @Test
    public void generateCECPR_WithDateToday_ReturnsValidId() {
        LocalDate today = LocalDate.now();
        String id = generator.generateCECPR(today.getDayOfMonth(), today.getMonthValue(), today.getYear(), Gender.MALE, true);
        assertNotNull(id);
    }


    @Test
    public void generateCECPR_WithDateInTheFuture_ThrowsException() {
        generateECPRAndExpectException(1, 1, 2040, Gender.MALE, FaultCode.CREATE_NEW_ECPR_ERROR, "future");
    }

    @Test
    public void generateCECPR_WithInvalidGenderInput_ThrowsException() {
        generateECPRAndExpectException(1, 1, 2018, null, FaultCode.CREATE_NEW_ECPR_ERROR, "gender");
    }

    @Test
    public void generateCECPR_WithValidInputButNoAvailableNumber_ThrowsException() throws DAOException {
        // Arrange
        int day = 1;
        int month = 1;
        int year = 2018;
        Gender gender = Gender.FEMALE; // Don't just change this. Tied to the gender digit below.
        String idPrefix = PersonNumberGeneratorImpl.constructInvariantPartOfCECPR(day, month, year);

        final List<IdentifierWithoutServiceActorVO> allIdCombinations = buildListWithNoAvailableIds(idPrefix);
        when(daoMock.getMatchingIdentifiersWithoutServiceActor(any(), any())).thenReturn(allIdCombinations);

        // Act & Assert
        generateECPRAndExpectException(day, month, year, gender, FaultCode.CREATE_NEW_ECPR_ERROR, "available");
    }

    @Test
    public void generateCECPR_WithDatabaseError_ThrowsException() throws DAOException {
        when(daoMock.getMatchingIdentifiersWithoutServiceActor(any(), any())).thenThrow(DAOException.class);
        generateECPRAndExpectException(1, 1, 2018, Gender.MALE, FaultCode.CREATE_NEW_ECPR_ERROR, "");
    }

    protected List<IdentifierWithoutServiceActorVO> buildListWithNoAvailableIds(String idPrefix) {
        List<IdentifierWithoutServiceActorVO> allIDs = new ArrayList<>();
        for (int firstCharacter = 0; firstCharacter < PersonNumberGeneratorImpl.VALID_CHARACTERS.length(); firstCharacter++) {
            for (int secondCharacter = 0; secondCharacter < PersonNumberGeneratorImpl.VALID_CHARACTERS.length(); secondCharacter++) {
                for (int genderDigit = 0; genderDigit < 10; genderDigit = genderDigit + 2) {
                    IdentifierWithoutServiceActorVO identifier = new IdentifierWithoutServiceActorVO();
                    String id = idPrefix
                            + PersonNumberGeneratorImpl.VALID_CHARACTERS.charAt(firstCharacter)
                            + PersonNumberGeneratorImpl.VALID_CHARACTERS.charAt(secondCharacter)
                            + genderDigit;
                    identifier.setValue(id);
                    allIDs.add(identifier);
                }
            }
        }
        return allIDs;
    }

    protected void generateECPRAndExpectException(int day, int month, int year, Gender gender,
                                                  FaultCode expectedFaultCode,
                                                  String expectedMessage) {
        try {
            String id = generator.generateCECPR(day, month, year, gender, true);
            fail("Did not throw an exception as expected");
        } catch (ECPRException e) {
            assertEquals(expectedFaultCode, e.getFaultCode());
            assertTrue("Exception messge did not contain the string " + expectedMessage,
                    e.getMessage().toLowerCase().contains(expectedMessage.toLowerCase()));
        } catch (Exception e) {
            fail("Did not throw the right exception type");
        }
    }

    private void assertCorrectCECPR(String id, String expectedDateString, String expectedDecadeString, String validGenderChars) {
        assertEquals(10, id.length());
        assertEquals(expectedDateString, id.substring(0, 6));
        assertEquals(expectedDecadeString, id.substring(6, 7));
        assertTrue(PersonNumberGeneratorImpl.VALID_CHARACTERS.contains(id.substring(7, 8)));
        assertTrue(PersonNumberGeneratorImpl.VALID_CHARACTERS.contains(id.substring(8, 9)));
        assertTrue(validGenderChars.contains(id.substring(9, 10)));
    }

}
