/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.service.admin.impl;

import dk.nsi.fmk.admin.Admin;
import dk.sds.ecpr2.AbstractTest;
import dk.sds.ecpr2.service.admin.AdminModuleClientFactoryMock;
import dk.sds.ecpr2.service.admin.AdministrationFacade;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;

import static org.hamcrest.CoreMatchers.is;
import static org.hamcrest.CoreMatchers.notNullValue;
import static org.junit.Assert.assertThat;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;

public class AdministrationFacadeWrapperTest extends AbstractTest {
    @Configuration
    static class ContextConfiguration {
        @Bean
        public AdminModuleClientFactory adminModuleClientFactory() {
            AdminModuleClientFactory adminModuleClientFactory = new AdminModuleClientFactory(true);
            return adminModuleClientFactory;
        }

        @Bean
        public AdministrationFacade adminModuleAdministrationFacade() {
            return new AdminModuleAdministrationFacadeImpl();
        }

        @Bean
        public AdministrationFacade administrationFacade() {
            return new AdministrationFacadeWrapper();
        }

        @Bean
        public AdminModuleClientFactoryMock adminModuleClientFactoryMock() {
            return new AdminModuleClientFactoryMock(false);
        }

        @Bean
        public Admin adminClient() {
            return adminModuleClientFactoryMock().getAdminServiceClient();
        }
    }

    @Autowired
    @Qualifier("administrationFacade")
    private AdministrationFacade adminFacade;

    @Autowired
    private Admin adminMock;

    private static final String SYSTEM_NAME_ECPR = "eCPR";
    private static final String ECPR_SUPPORTER = "0201112345";
    private static final String ECPR_NOT_SUPPORTER = "0301112345";


    @Before
    public void onSetUp() throws Exception {
        assertThat("Injected 'AdministrationFacade' cannot be NULL", adminFacade, is(notNullValue()));
        addMockFindAdminUserResponse(ECPR_SUPPORTER, true);
        addMockFindAdminUserResponse(ECPR_NOT_SUPPORTER, false);
    }

    @After
    public void tearDown() throws Exception {
        // reset invocation count before next iteration, not required ?
        Mockito.reset(adminMock);
        // restore the initial value of the feature switch...
    }

    @Test
    public void isSupporterUser_ecprAdmin() {
        assertThat("User should be 'ECPRCoordinator'", adminFacade.isSupporterUser(ECPR_SUPPORTER), is(true));
        assertThat("User should NOT be 'ECPRCoordinator'", adminFacade.isSupporterUser(ECPR_NOT_SUPPORTER), is(false));

        // verify that no calls are made towards the module
        verifyFindAdminUserInvocation(ECPR_SUPPORTER);
        verifyFindAdminUserInvocation(ECPR_NOT_SUPPORTER);

        Mockito.verifyNoMoreInteractions(adminMock);
    }

    private void addMockFindAdminUserResponse(String cpr, boolean isSupporter) {
        dk.nsi.fmk.admin.admin.data.proto.Admin.AdminUserVOResponse.Builder responseBuilder = dk.nsi.fmk.admin.admin.data.proto.Admin.AdminUserVOResponse.newBuilder();
        if (isSupporter) {
            responseBuilder.setAdminUserVO(
                    // since we only check for existence none of the other fields are set
                    dk.nsi.fmk.admin.admin.data.proto.Admin.AdminUserVO.newBuilder()
                            .setCpr(cpr)
                            .build()
            );
        }
        // the ClientName 'DDV' should match the one in AdminModuleAdministrationFacadeImpl
        Mockito.when(adminMock.findAdminUserWithWildcardCVR(any(), eq(toProtocString(cpr)), any(), eq(toProtocString(SYSTEM_NAME_ECPR)))
        ).thenReturn(responseBuilder.build());
    }

    private dk.nsi.fmk.admin.admin.data.proto.Admin.ProtocString toProtocString(String s) {
        return dk.nsi.fmk.admin.admin.data.proto.Admin.ProtocString.newBuilder().setValue(s).build();
    }

    private void verifyFindAdminUserInvocation(String cpr) {
        Mockito.verify(adminMock).findAdminUserWithWildcardCVR(any(), eq(toProtocString(cpr)), any(), eq(toProtocString(SYSTEM_NAME_ECPR)));
    }
}
