/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice;

import dk.sds.ecpr2.IntegrationTest;
import dk.sds.ecpr2.domain.BlacklistingVO;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.ResponseEntity;

import java.util.Arrays;
import java.util.Collections;

import static org.junit.Assert.assertEquals;

public class BlacklistingTest extends IntegrationTest {
    @Autowired
    ECPRRestService ecprRestService;

    @Test
    public void testPutSingleBlacklistingSucceeds() {
        BlacklistingVO blacklistingVO = new BlacklistingVO("1.2.208.176.1.2000", Collections.singletonList("1234560AA0"));
        ResponseEntity response = ecprRestService.putBlacklisting(blacklistingVO);
        assertEquals(201, response.getStatusCode().value());
    }

    @Test
    public void testPostSingleBlacklistingSucceeds() {
        BlacklistingVO blacklistingVO = new BlacklistingVO("1.2.208.176.1.2000", Collections.singletonList("1234560AA0"));
        ResponseEntity response = ecprRestService.postBlacklisting(blacklistingVO);
        assertEquals(201, response.getStatusCode().value());
    }

    @Test
    public void testPutTwoSeperateBlacklistingsSucceeds() {
        BlacklistingVO blacklistingVO = new BlacklistingVO("1.2.208.176.1.2000", Arrays.asList("1234560AA0", "1234561BB1"));
        ResponseEntity response = ecprRestService.putBlacklisting(blacklistingVO);
        assertEquals(201, response.getStatusCode().value());
    }

    @Test
    public void testPostTwoSeperateBlacklistingsSucceeds() {
        BlacklistingVO blacklistingVO = new BlacklistingVO("1.2.208.176.1.2000", Arrays.asList("1234560AA0", "1234561BB1"));
        ResponseEntity response = ecprRestService.postBlacklisting(blacklistingVO);
        assertEquals(201, response.getStatusCode().value());
    }

    @Test
    public void testPutDuplicateBlacklistingsInOneCallFails() {
        BlacklistingVO blacklistingVO = new BlacklistingVO("1.2.208.176.1.2000", Arrays.asList("1234560AA0", "1234560AA0"));
        ResponseEntity response = ecprRestService.putBlacklisting(blacklistingVO);
        assertEquals(200, response.getStatusCode().value());
    }

    @Test
    public void testPostDuplicateBlacklistingsInOneCallFails() {
        BlacklistingVO blacklistingVO = new BlacklistingVO("1.2.208.176.1.2000", Arrays.asList("1234560AA0", "1234560AA0"));

        ResponseEntity response = ecprRestService.postBlacklisting(blacklistingVO);

        assertEquals(409, response.getStatusCode().value());
        assertEquals("{\"Blacklist\":[{\"D-eCPR\":\"1234560AA0\",\"OIDs\":[\"1.2.208.176.1.2000\"]}]}", response.getBody().toString()); //JSON content of response
    }

    @Test
    public void testPutDuplicateBlacklistingsInMultipleCallsFails() {
        BlacklistingVO blacklistingVO = new BlacklistingVO("1.2.208.176.1.2000", Collections.singletonList("1234560AA0"));
        ResponseEntity response = ecprRestService.putBlacklisting(blacklistingVO);
        assertEquals(201, response.getStatusCode().value());

        blacklistingVO = new BlacklistingVO("1.2.208.176.1.2000", Collections.singletonList("1234560AA0"));
        response = ecprRestService.putBlacklisting(blacklistingVO);
        assertEquals(200, response.getStatusCode().value());
    }

    @Test
    public void testPostDuplicateBlacklistingsInMultipleCallsFails() {
        BlacklistingVO blacklistingVO = new BlacklistingVO("1.2.208.176.1.2000", Collections.singletonList("1234560AA0"));

        ResponseEntity response = ecprRestService.postBlacklisting(blacklistingVO);

        assertEquals(201, response.getStatusCode().value());

        blacklistingVO = new BlacklistingVO("1.2.208.176.1.2000", Collections.singletonList("1234560AA0"));

        response = ecprRestService.postBlacklisting(blacklistingVO);

        assertEquals(409, response.getStatusCode().value());
        assertEquals("{\"Blacklist\":[{\"D-eCPR\":\"1234560AA0\",\"OIDs\":[\"1.2.208.176.1.2000\"]}]}", response.getBody().toString()); //JSON content of response
    }

    @Test
    public void testPutDuplicatedBlacklistingWrites() {
        BlacklistingVO blacklistingVO = new BlacklistingVO("1.2.208.176.1.2000", Collections.singletonList("1234560AA0"));

        ResponseEntity response = ecprRestService.putBlacklisting(blacklistingVO);

        assertEquals(201, response.getStatusCode().value());

        blacklistingVO = new BlacklistingVO("1.2.208.176.1.2001", Collections.singletonList("1234560AA0"));

        response = ecprRestService.putBlacklisting(blacklistingVO);

        assertEquals(200, response.getStatusCode().value()); // Fails, but blacklists anyway

        blacklistingVO = new BlacklistingVO("1.2.208.176.1.2000", Collections.singletonList("1234560AA0"));

        response = ecprRestService.postBlacklisting(blacklistingVO);

        assertEquals(409, response.getStatusCode().value());
        assertEquals("{\"Blacklist\":[{\"D-eCPR\":\"1234560AA0\",\"OIDs\":[\"1.2.208.176.1.2000\",\"1.2.208.176.1.2001\"]}]}", response.getBody().toString()); //JSON content of response
    }
}
