/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice;

import dk.sds.ecpr2.*;
import dk.sds.ecpr2._2017._10._01.CompareNameRequest;
import dk.sds.ecpr2._2017._10._01.CompareNameResponse;
import dk.sds.ecpr2._2017._10._01.CreatePersonRequest;
import dk.sds.ecpr2._2017._10._01.CreatePersonResponse;
import dk.sds.ecpr2.domain.CountryCode;
import dk.sds.ecpr2.util.ECPRException;
import dk.sds.ecpr2.util.ValueSets;
import dk.sds.ecpr2.webservice.validation.ValidationErrors;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.test.context.junit4.SpringRunner;

import java.util.ArrayList;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.UUID;

import static org.junit.Assert.*;

@RunWith(SpringRunner.class)
public class CompareNameTest extends WebServiceTestAbstract {

    @Test
    public void testNoNameComparedToActualName() throws Exception {
        PersonType personToBeCreated = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        PersonType createdPerson = createPersonInDb(personToBeCreated);
        KeyIdentifier key = new KeyIdentifier();
        key.setValue(createdPerson.getIdentifiers().get(0).getValue());
        key.setOID(createdPerson.getIdentifiers().get(0).getOID());

        CompareNameRequest request = new CompareNameRequest();
        request.setName("");
        request.setKeyIdentifier(key);
        CompareNameResponse response = service.compareName(request);

        assertEquals(0, response.getScore());
    }

    @Test
    public void testNoActualNameComparedToName() throws Exception {
        PersonType personToBeCreated = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        personToBeCreated.getNames().clear(); // Remove name from person before creating person
        PersonType createdPerson = createPersonInDb(personToBeCreated);
        KeyIdentifier key = new KeyIdentifier();
        key.setValue(createdPerson.getIdentifiers().get(0).getValue());
        key.setOID(createdPerson.getIdentifiers().get(0).getOID());

        CompareNameRequest request = new CompareNameRequest();
        request.setName("Benny Olsen");
        request.setKeyIdentifier(key);
        CompareNameResponse response = service.compareName(request);

        assertEquals(0, response.getScore());
    }

    @Test
    public void testPreciseName() throws Exception {
        PersonType personToBeCreated = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        PersonType createdPerson = createPersonInDb(personToBeCreated);
        KeyIdentifier key = new KeyIdentifier();
        key.setValue(createdPerson.getIdentifiers().get(0).getValue());
        key.setOID(createdPerson.getIdentifiers().get(0).getOID());

        CompareNameRequest request = new CompareNameRequest();
        request.setName("Benny Olsen");
        request.setKeyIdentifier(key);
        CompareNameResponse response = service.compareName(request);

        assertEquals(100, response.getScore());
    }

    @Test
    public void testOldName() throws Exception {
        PersonType personToBeCreated = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        PersonType createdPerson = createPersonInDb(personToBeCreated);
        KeyIdentifier key = new KeyIdentifier();
        key.setValue(createdPerson.getIdentifiers().get(0).getValue());
        key.setOID(createdPerson.getIdentifiers().get(0).getOID());

        createdPerson.getNames().get(0).setFamilyName("Jensen");
        PersonType updatedPerson = updatePersonInDb(createdPerson);

        CompareNameRequest request = new CompareNameRequest();
        request.setName("Benny Olsen");
        request.setKeyIdentifier(key);
        CompareNameResponse response = service.compareName(request);

//        assertEquals(67, response.getScore()); // JaccardSimilarity
        assertEquals(75, response.getScore()); // LevenshteinDistance
    }

    @Test
    public void testGivenNameOnly() throws Exception {
        PersonType personToBeCreated = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        PersonType createdPerson = createPersonInDb(personToBeCreated);
        KeyIdentifier key = new KeyIdentifier();
        key.setValue(createdPerson.getIdentifiers().get(0).getValue());
        key.setOID(createdPerson.getIdentifiers().get(0).getOID());

        CompareNameRequest request = new CompareNameRequest();
        request.setName("Benny");
        request.setKeyIdentifier(key);
        CompareNameResponse response = service.compareName(request);

//        assertEquals(50, response.getScore()); // JaccardSimilarity
        assertEquals(45, response.getScore()); // LevenshteinDistance
    }

    @Test
    public void testMisspelledName() throws Exception {
        PersonType personToBeCreated = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        PersonType createdPerson = createPersonInDb(personToBeCreated);
        KeyIdentifier key = new KeyIdentifier();
        key.setValue(createdPerson.getIdentifiers().get(0).getValue());
        key.setOID(createdPerson.getIdentifiers().get(0).getOID());

        CompareNameRequest request = new CompareNameRequest();
        request.setName("Bangy Aulsen");
        request.setKeyIdentifier(key);
        CompareNameResponse response = service.compareName(request);

//        assertEquals(58, response.getScore()); // JaccardSimilarity
        assertEquals(67, response.getScore()); // LevenshteinDistance
    }

    @Test
    public void testTotallyOtherName() throws Exception {
        PersonType personToBeCreated = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        PersonType createdPerson = createPersonInDb(personToBeCreated);
        KeyIdentifier key = new KeyIdentifier();
        key.setValue(createdPerson.getIdentifiers().get(0).getValue());
        key.setOID(createdPerson.getIdentifiers().get(0).getOID());

        CompareNameRequest request = new CompareNameRequest();
        request.setName("Jakob Finnerup");
        request.setKeyIdentifier(key);
        CompareNameResponse response = service.compareName(request);

//        assertEquals(17, response.getScore()); // JaccardSimilarity
        assertEquals(14, response.getScore()); // LevenshteinDistance
    }

    @Test
    public void testLongName() throws Exception {
        PersonType personToBeCreated = createPerson("Margrethe Alexandrine Þórhildur", "Ingrid", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        PersonType createdPerson = createPersonInDb(personToBeCreated);
        KeyIdentifier key = new KeyIdentifier();
        key.setValue(createdPerson.getIdentifiers().get(0).getValue());
        key.setOID(createdPerson.getIdentifiers().get(0).getOID());

        CompareNameRequest request = new CompareNameRequest();
        request.setName("Margrethe Alexandrine Þórhildur");
        request.setKeyIdentifier(key);
        CompareNameResponse response = service.compareName(request);

//        assertEquals(94, response.getScore()); // JaccardSimilarity
        assertEquals(82, response.getScore()); // LevenshteinDistance
    }

    @Test
    public void testMuhammadName() throws Exception {
        PersonType personToBeCreated = createPerson("Muhammad", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        PersonType createdPerson = createPersonInDb(personToBeCreated);
        KeyIdentifier key = new KeyIdentifier();
        key.setValue(createdPerson.getIdentifiers().get(0).getValue());
        key.setOID(createdPerson.getIdentifiers().get(0).getOID());

        CompareNameRequest request = new CompareNameRequest();
        request.setName("Mohamed Olsen");
        request.setKeyIdentifier(key);
        CompareNameResponse response = service.compareName(request);

//        assertEquals(84, response.getScore()); // JaccardSimilarity
        assertEquals(79, response.getScore()); // LevenshteinDistance
    }
}
