/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice;

import dk.sds.ecpr2.*;
import dk.sds.ecpr2._2017._10._01.CreatePersonIdRequest;
import dk.sds.ecpr2._2017._10._01.CreatePersonIdResponse;
import dk.sds.ecpr2.util.ECPRException;
import dk.sds.ecpr2.webservice.validation.ValidationErrors;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.test.context.junit4.SpringRunner;

import java.util.List;
import java.util.stream.Collectors;

import static org.junit.Assert.*;

@RunWith(SpringRunner.class)
public class CreatePersonIdTest extends WebServiceTestAbstract {

    @Test
    public void createPersonId_ValidRequest_newCeCPRCreated() throws Exception {
        PersonType newPerson = createPerson();
        Identifier cecprIdentifier = new Identifier();
        cecprIdentifier.setOID(OID.C_ECPR);
        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        newPerson.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        SimplePersonWithKeyIdentifierType personForNewPersonId = createNewPersonId(newPerson, cecprIdentifier);
        CreatePersonIdRequest request = new CreatePersonIdRequest();

        request.setPerson(personForNewPersonId);
        CreatePersonIdResponse createdPersonIdResponse = service.createPersonId(request);

        List<Identifier> ceCPRs = createdPersonIdResponse.getPerson().getIdentifiers().stream().filter(i -> OID.C_ECPR.equals(i.getOID())).collect(Collectors.toList());
        List<Identifier> regions = createdPersonIdResponse.getPerson().getIdentifiers().stream().filter(i -> OID.ECPR_REGION_SOUTH.equals(i.getOID())).collect(Collectors.toList());
        assertNotNull(createdPersonIdResponse);
        assertEquals(3, createdPersonIdResponse.getPerson().getIdentifiers().size());
        assertEquals(1, ceCPRs.size());
        assertEquals(1, regions.size());
        assertEquals(1, ceCPRs.stream().filter(i -> i.getValidTo() == null).count());
        Identifier firstCeCPR = ceCPRs.stream().filter(i -> i.getValidTo() == null).findFirst().get();
        assertTrue(firstCeCPR.getValue().charAt(6) == '0' || firstCeCPR.getValue().charAt(6) == '5' || firstCeCPR.getValue().charAt(6) == '6');
    }

    @Test
    public void createPersonId_ValidRequest_newXeCPRCreated() throws Exception {
        PersonType newPerson = createPerson();
        Identifier cecprIdentifier = new Identifier();
        cecprIdentifier.setOID(OID.X_ECPR);
        PersonIdentifierType personId = new PersonIdentifierType();
        personId.setPersonIdentifier("1111111118");
        personId.setSource("CPR");
        newPerson.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personId);
        SimplePersonWithKeyIdentifierType personForNewPersonId = createNewPersonId(newPerson, cecprIdentifier);

        CreatePersonIdRequest request = new CreatePersonIdRequest();
        request.setPerson(personForNewPersonId);
        CreatePersonIdResponse createdPersonIdResponse = service.createPersonId(request);

        List<Identifier> xeCPRs = createdPersonIdResponse.getPerson().getIdentifiers().stream().filter(i -> OID.X_ECPR.equals(i.getOID())).collect(Collectors.toList());
        List<Identifier> regions = createdPersonIdResponse.getPerson().getIdentifiers().stream().filter(i -> OID.ECPR_REGION_SOUTH.equals(i.getOID())).collect(Collectors.toList());
        assertNotNull(createdPersonIdResponse);
        assertEquals(3, createdPersonIdResponse.getPerson().getIdentifiers().size());
        assertEquals(2, xeCPRs.size());
        assertEquals(1, regions.size());
        assertEquals(1, xeCPRs.stream().filter(i -> i.getValidTo() == null).count());
        assertEquals(1, xeCPRs.stream().filter(i -> i.getValidFrom() != null && i.getValidTo() != null).count());
        Identifier firstXeCPR = xeCPRs.stream().filter(i -> i.getValidTo() == null).findFirst().get();
        assertTrue(firstXeCPR.getValue().charAt(6) == '1' || firstXeCPR.getValue().charAt(6) == '7');
    }

    @Test
    public void createPersonId_BadKeyIdentifier_ValidationError() throws Exception {
        Identifier cecprIdentifier = new Identifier();
        cecprIdentifier.setOID(OID.C_ECPR);
        SimplePersonWithKeyIdentifierType personForNewPersonId = createNewPersonId(createPerson(), cecprIdentifier);
        personForNewPersonId.getKeyIdentifier().setOID("1.2.208.176.1.1111");
        expectError(personForNewPersonId, ValidationErrors.CREATEPERSONID_INVALID_OID);
    }

    @Test
    public void createPersonId_BadKeyIdentifierPersonID_Failure() throws Exception {
        Identifier cecprIdentifier = new Identifier();
        cecprIdentifier.setOID(OID.X_ECPR);
        SimplePersonWithKeyIdentifierType personForNewPersonId = createNewPersonId(createPerson(), cecprIdentifier);
        personForNewPersonId.getKeyIdentifier().setValue("0101010AA0"); // Non-existing
        expectError(personForNewPersonId, ValidationErrors.KEY_IDENTIFIER_INVALID);
    }

    @Test
    public void createPerson_PassportNumber() throws Exception {
        Identifier passportNumber = new Identifier();
        passportNumber.setOID(OID.PASSPORT + ".208");
        passportNumber.setValue("12345678");
        passportNumber.setValidity(4);
        PersonType person = createPerson();
        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        person.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        SimplePersonWithKeyIdentifierType personForNewPersonId = createNewPersonId(person, passportNumber);
        CreatePersonIdRequest request = new CreatePersonIdRequest();

        request.setPerson(personForNewPersonId);
        CreatePersonIdResponse createPersonIdResponse = service.createPersonId(request);

        Identifier passIdentifier = createPersonIdResponse.getPerson().getIdentifiers().stream().filter(i -> i.getOID().startsWith(OID.PASSPORT)).findFirst().get();
        assertEquals(OID.PASSPORT + ".208", passIdentifier.getOID());
        assertEquals("Pasnummer", passIdentifier.getPrimaryOIDLabel());
        assertEquals("Danmark", passIdentifier.getSecondaryOIDLabel());
    }

    @Test
    public void createPerson_ValidCpr() throws Exception {
        Identifier passportNumber = new Identifier();
        passportNumber.setOID(OID.CPR);
        passportNumber.setValue("1111111118");
        passportNumber.setValidity(4);
        PersonType person = createPerson();
        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        person.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        SimplePersonWithKeyIdentifierType personForNewPersonId = createNewPersonId(person, passportNumber);
        CreatePersonIdRequest request = new CreatePersonIdRequest();

        request.setPerson(personForNewPersonId);
        CreatePersonIdResponse createPersonIdResponse = service.createPersonId(request);

        Identifier cprIdentifier = createPersonIdResponse.getPerson().getIdentifiers().stream().filter(i -> i.getOID().equals(OID.CPR)).findFirst().get();
        assertEquals(OID.CPR, cprIdentifier.getOID());
        assertEquals("1111111118", cprIdentifier.getValue());
    }

    @Test
    public void createPerson_InvalidCpr() throws Exception {
        Identifier passportNumber = new Identifier();
        passportNumber.setOID(OID.CPR);
        passportNumber.setValue("0000000000");
        passportNumber.setValidity(4);
        PersonType person = createPerson();
        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        person.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        SimplePersonWithKeyIdentifierType personForNewPersonId = createNewPersonId(person, passportNumber);
        CreatePersonIdRequest request = new CreatePersonIdRequest();

        request.setPerson(personForNewPersonId);
        try {
            CreatePersonIdResponse createPersonIdResponse = service.createPersonId(request);
        } catch (ECPRException e) {
            assertEquals(e.getMessage(), "Fejl i request ifm. oprettelse af nyt eCPR, Cpr identifier er ikke valid");
        }
    }

    private void expectError(SimplePersonWithKeyIdentifierType personForNewPersonId, String expectedError) {
        CreatePersonIdRequest request = new CreatePersonIdRequest();
        request.setPerson(personForNewPersonId);
        try {
            debug(request);
            CreatePersonIdResponse response = service.createPersonId(request);
            debug(response);
            fail("Expected validation error. Response above");
        } catch (Exception e) {
            checkError(e, expectedError);
        }
    }
}
