/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice;

import dk.sds.ecpr2.BirthDateType;
import dk.sds.ecpr2.Identifier;
import dk.sds.ecpr2.OID;
import dk.sds.ecpr2.PersonType;
import dk.sds.ecpr2._2017._10._01.CreatePersonRequest;
import dk.sds.ecpr2._2017._10._01.CreatePersonResponse;
import dk.sds.ecpr2.domain.CountryCode;
import dk.sds.ecpr2.util.ECPRException;
import dk.sds.ecpr2.util.ValueSets;
import dk.sds.ecpr2.webservice.validation.ValidationErrors;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.test.context.junit4.SpringRunner;

import java.util.ArrayList;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.UUID;

import static org.junit.Assert.*;

@RunWith(SpringRunner.class)
public class CreatePersonTest extends WebServiceTestAbstract {

    @Test
    public void testNameNullVoid() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        PersonType reqPerson = createDefaultPerson();
        reqPerson.getNames().clear();
        reqPerson.getNames().add(factory.createName("Null", "Void", ValueSets.nameUseValues[0]));
        request.setPerson(reqPerson);
        CreatePersonResponse response = service.createPerson(request);
        PersonType person = response.getPerson();
        assertNotNull(person);
        assertEquals(2, person.getIdentifiers().size());
        assertEquals("Null", person.getNames().get(0).getGivenName());
        assertEquals("Void", person.getNames().get(0).getFamilyName());
        checkDatePartOfCECPRMatchesBirthDate(person);
    }

    @Test
    public void testValidRequest() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        request.setPerson(createDefaultPerson());
        CreatePersonResponse response = service.createPerson(request);
        PersonType person = response.getPerson();
        assertNotNull(person);
        assertEquals(2, person.getIdentifiers().size());
        checkDatePartOfCECPRMatchesBirthDate(person);
    }

    @Test
    public void testIllegalBirthdate() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        PersonType defaultPerson = factory.createMinimalPerson("Günther", "Weber", "1840");
        request.setPerson(defaultPerson);
        try {
            CreatePersonResponse response = service.createPerson(request);
            fail("Should have caused error message: " + ValidationErrors.BIRTHDATE_BAD_YEAR);
        } catch (Exception e) {
            checkError(e, ValidationErrors.BIRTHDATE_BAD_YEAR);
        }
    }

    @Test
    public void testMinimalRequest() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        PersonType defaultPerson = factory.createMinimalPerson("Günther", "Weber", "1940");
        request.setPerson(defaultPerson);
        CreatePersonResponse response = service.createPerson(request);
        PersonType person = response.getPerson();
        assertNotNull(person);
        assertEquals(1, person.getIdentifiers().size());
        assertEquals(1, person.getModifieds().size());
        assertEquals("12345678901", person.getModifieds().get(0).getBy().getOrganisation().getIdentifier().getIdentifier());
        checkDatePartOfCECPRMatchesBirthDate(person);
    }

    @Test
    public void testDuplicateIdentifers() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        PersonType defaultPerson = createDefaultPerson();
        defaultPerson.getIdentifiers().add(defaultPerson.getIdentifiers().get(1)); // Use the same identifier twice
        request.setPerson(defaultPerson);
        try {
            service.createPerson(request);
            fail("Should have caused error message: " + ValidationErrors.DUPLICATE_IDENTIFIER);
        } catch (Exception ex) {
            checkError(ex, ValidationErrors.DUPLICATE_IDENTIFIER);
        }
    }

    @Test
    public void testDuplicateNameUse() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        PersonType defaultPerson = createDefaultPerson();
        defaultPerson.getNames().add(defaultPerson.getNames().get(0)); // Use the same name use twice
        request.setPerson(defaultPerson);
        try {
            service.createPerson(request);
            fail("Should have caused error message: " + ValidationErrors.DUPLICATE_NAME_USE);
        } catch (Exception ex) {
            checkError(ex, ValidationErrors.DUPLICATE_NAME_USE);
        }
    }

    private void checkDatePartOfCECPRMatchesBirthDate(PersonType person) {
        for (Identifier identifier : person.getIdentifiers()) {
            if (OID.C_ECPR.equals(identifier.getOID())) {
                assertNotNull(identifier.getValue());
                assertEquals(10, identifier.getValue().length());
                assertBirthDateMatchesCECPR(person.getBirthDates().get(0), identifier.getValue());
            }
        }
    }

    @Test
    public void testXECPRIsCreatedWhenNotProvidedInRequest() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        PersonType defaultPerson = createDefaultPerson();
        defaultPerson.getIdentifiers().get(0).setValue(null); // Remove C-eCPR entry. The service must add it then
        request.setPerson(defaultPerson);
        CreatePersonResponse response = service.createPerson(request);
        PersonType person = response.getPerson();
        assertNotNull(person);
        assertEquals(2, person.getIdentifiers().size());
        assertTrue(person.getIdentifiers().stream().anyMatch(i -> i.getOID().equals(OID.X_ECPR) && i.getValidity().equals(16)));
        checkDatePartOfCECPRMatchesBirthDate(person);
    }

    private void assertBirthDateMatchesCECPR(BirthDateType birthDateType, String ceCpr) {
        String birthDate = birthDateType.getValue();
        assertEquals(birthDate.substring(2, 4), ceCpr.substring(4, 6)); // year
        assertEquals(birthDate.substring(5, 7), ceCpr.substring(2, 4)); // month
        assertEquals(birthDate.substring(8, 10), ceCpr.substring(0, 2)); // date
    }

    @Test
    public void testDuplicateIdentifierIsRejected() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        request.setPerson(createDefaultPerson());
        CreatePersonResponse response = service.createPerson(request);
        assertNotNull(response.getPerson());
        // Use the same request once again - this must cause a validation error due to duplicate identifier
        try {
            service.createPerson(request);
            fail("Should have caused duplicate error message: " + ValidationErrors.EXISTING_IDENTIFIER);
        } catch (Exception ex) {
            checkError(ex, ValidationErrors.EXISTING_IDENTIFIER);
        }
    }

    @Test
    public void testGermanPassportIsGermanPassport() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        request.setPerson(createDefaultPerson());
        request.getPerson().getIdentifiers().get(0).setOID(OID.PASSPORT + "." + CountryCode.DEU.getNumericValue());
        request.getPerson().getIdentifiers().get(0).setValue("PASSNUMMER");
        CreatePersonResponse response = service.createPerson(request);
        assertEquals(3, response.getPerson().getIdentifiers().size());
        assertTrue(response.getPerson().getIdentifiers().stream().anyMatch(i -> i.getOID().equalsIgnoreCase(OID.PASSPORT + "." + String.format("%03d", CountryCode.DEU.getNumericValue()))));
    }

    @Test
    public void testBadOIDFormatIsRejected() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        request.setPerson(createDefaultPerson());
        request.getPerson().getIdentifiers().get(0).setOID(OID.CPR); // Use CPR-format for C-eCPR (which must fail)
        request.getPerson().getIdentifiers().get(0).setValue("1111110AB2"); // Bad value for CPR
        try {
            service.createPerson(request);
            fail("Should have caused duplicate error message: " + ValidationErrors.BAD_FORMAT_OID);
        } catch (Exception ex) {
            checkError(ex, ValidationErrors.BAD_FORMAT_OID);
        }
    }

    @Test
    public void testBadGenderIsRejected() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        request.setPerson(createDefaultPerson());
        request.getPerson().getGenders().get(0).setValue("something wrong");
        try {
            service.createPerson(request);
            fail("Should have caused error message: " + ValidationErrors.GENDER_VALUE_ILLEGAL);
        } catch (Exception ex) {
            checkError(ex, ValidationErrors.GENDER_VALUE_ILLEGAL);
        }
    }

    @Test
    public void testBirthDateFormat() throws Exception {
        assertBirthDateValidationFailure("foo");
        assertBirthDateValidationFailure("2018-18");
        assertBirthDateValidationFailure("2018-11-24-");
        assertBirthDateValidationFailure("2018-11-");
        assertBirthDateValidationFailure("2018-");
        assertBirthDateValidationFailure("2018-1");
        assertBirthDateValidationOk("2018-01-13");
        assertBirthDateValidationOk("2018-01");
        assertBirthDateValidationOk("2018");
    }

    @Test
    public void testResponseEnrichment() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        PersonType fullPerson = factory.createFullPerson2();
        request.setPerson(fullPerson);
        CreatePersonResponse response = service.createPerson(request);
        PersonType person = response.getPerson();
        assertEquals(2, person.getAddresses().size());
        assertEquals(1, person.getBirthDates().size());
        assertEquals(2, person.getContacts().size());
        assertEquals(2, person.getNames().size());
        assertEquals(1, person.getIdentifiers().size());
        assertEquals(1, person.getGenders().size());
        Identifier identifier = person.getIdentifiers().get(0);
        assertNotNull(identifier.getPID());
    }


    @Test
    public void testEmptyBirthdate() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        request.setPerson(createDefaultPerson());
        request.getPerson().getBirthDates().clear();
        request.getPerson().getBirthDates().add(new BirthDateType());
        CreatePersonResponse response = service.createPerson(request);
        PersonType person = response.getPerson();
        assertNotNull(person);
        assertEquals(2, person.getIdentifiers().size());
        assertEquals(1, person.getBirthDates().size());
        checkDatePartOfCECPRMatchesBirthDate(person);
    }

    private void assertBirthDateValidationFailure(String birthDate) throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        PersonType person = createDefaultPerson();
        person.getBirthDates().get(0).setValue(birthDate);
        request.setPerson(person);
        try {
            service.createPerson(request);
            fail("Expected IllegalArgumentException");
        } catch (Exception ex) {
            IllegalArgumentException e = (IllegalArgumentException) ex.getCause();
            assertMessageContainsError(e.getMessage(), ValidationErrors.BIRTHDATE_BAD_VALUE);
        }
    }

    private void assertBirthDateValidationOk(String birthDate) throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        PersonType person = createDefaultPerson();
        person.getBirthDates().get(0).setValue(birthDate);
        request.setPerson(person);
        service.createPerson(request); // No validation error expected
    }

    @Test
    public void testVeryInvalidRequest() throws Exception {
        // Provoke each known error by changing a valid request
        List<String> expected = new ArrayList<>();
        PersonType person = createDefaultPerson();

        expected.add(ValidationErrors.GENDER_VALUE_ILLEGAL);
        person.getGenders().get(0).setValue("foo");

        expected.add(ValidationErrors.IDENTIFIER_OIDLABEL_FORBIDDEN);
        person.getIdentifiers().get(0).setPrimaryOIDLabel("foo");

        expected.add(ValidationErrors.VALIDITY_REQUIRED);
        person.getIdentifiers().get(0).setValidity(42);

        expected.add(ValidationErrors.NAME_TEXT_FORBIDDEN);
        person.getNames().get(0).setText("foo");

        expected.add(ValidationErrors.NAME_USE_INVALID);
        person.getNames().get(0).setUse("foo");

        expected.add(ValidationErrors.BIRTHDATE_BAD_VALUE);
        person.getBirthDates().get(0).setValue("foo");

        expected.add(ValidationErrors.CONTACT_USE_ILLEGAL);
        person.getContacts().get(0).setUse("foo");

        expected.add(ValidationErrors.CONTACT_TYPE_ILLEGAL);
        person.getContacts().get(0).setSystem("foo");

        expected.add(ValidationErrors.COUNTRY_UNKNOWN);
        person.getAddresses().get(0).setCountry("foo");

        expected.add(ValidationErrors.COUNTRY_NAME_FORBIDDEN);
        person.getAddresses().get(0).setCountryName("foo");

        expected.add(ValidationErrors.VALIDFROM_FORBIDDEN);
        person.getAddresses().get(0).setValidFrom(new GregorianCalendar());

        expected.add(ValidationErrors.VALIDTO_FORBIDDEN);
        person.getAddresses().get(0).setValidTo(new GregorianCalendar());

        expected.add(ValidationErrors.ADDRESS_USE_UNKNOWN);
        person.getAddresses().get(0).setUse("foo");

        expected.add(ValidationErrors.ADDRESS_TYPE_UNKNOWN);
        person.getAddresses().get(0).setType("foo");

        checkCreatePersonValidationErrors(expected, person);
    }

    private void checkCreatePersonValidationErrors(List<String> expected, PersonType person) {
        CreatePersonRequest request = new CreatePersonRequest();
        request.setPerson(person);
        try {
            service.createPerson(request);
            fail("Expected Exception");
        } catch (Exception ex) {
            Exception e = (Exception) ex.getCause();
            String message = e.getMessage();
            for (String error : expected) {
                assertMessageContainsError(message, error);
            }
        }
    }

    // Check at PID ikke må angives - hverken på identifiers, adresser osv.
    @Test
    public void checkPidInRequestsIsRejected() {
        // Provoke each known error by changing a valid request
        List<String> expected = new ArrayList<>();
        PersonType person = createDefaultPerson();

        expected.add(ValidationErrors.PID_FORBIDDEN);
        person.getGenders().get(0).setPID(UUID.randomUUID().toString());

        expected.add(ValidationErrors.PID_FORBIDDEN);
        person.getIdentifiers().get(0).setPID(UUID.randomUUID().toString());

        expected.add(ValidationErrors.PID_FORBIDDEN);
        person.getNames().get(0).setPID(UUID.randomUUID().toString());

        expected.add(ValidationErrors.PID_FORBIDDEN);
        person.getBirthDates().get(0).setPID(UUID.randomUUID().toString());

        expected.add(ValidationErrors.PID_FORBIDDEN);
        person.getContacts().get(0).setPID(UUID.randomUUID().toString());

        expected.add(ValidationErrors.PID_FORBIDDEN);
        person.getAddresses().get(0).setPID(UUID.randomUUID().toString());

        checkCreatePersonValidationErrors(expected, person);
    }

    @Test
    public void testValidCPRNumber() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        request.setPerson(createDefaultPerson());
        request.getPerson().getIdentifiers().add(factory.createCprIdentifier("1111111118"));
        CreatePersonResponse response = service.createPerson(request);
        PersonType person = response.getPerson();
        assertNotNull(person);
        assertEquals(3, person.getIdentifiers().size());
        assertTrue(person.getIdentifiers().stream().anyMatch(i -> i.getOID().equals(OID.CPR) && i.getValue().equals("1111111118")));
        checkDatePartOfCECPRMatchesBirthDate(person);
    }

    @Test
    public void testInvalidCPRNumber() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        request.setPerson(createDefaultPerson());
        request.getPerson().getIdentifiers().add(factory.createCprIdentifier("0000000000"));
        try {
            CreatePersonResponse response = service.createPerson(request);
        } catch (ECPRException e) {
            assertEquals(e.getMessage(), "Fejl i request ifm. oprettelse af eCPR-data, Cpr identifier er ikke valid");
        }
    }
}
