/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice;

import dk.sds.ecpr2.*;
import dk.sds.ecpr2._2017._10._01.DeletePersonRequest;
import dk.sds.ecpr2._2017._10._01.DeletePersonResponse;
import dk.sds.ecpr2.advis.AsyncAdvisFacade;
import dk.sds.ecpr2.domain.PersonVO;
import dk.sds.ecpr2.util.FaultCode;
import dk.sds.ecpr2.util.Properties;
import dk.sds.ecpr2.webservice.mapping.FromXMLMapper;
import dk.sds.ecpr2.webservice.validation.ValidationErrors;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.test.context.junit4.SpringRunner;

import static org.junit.Assert.*;
import static org.mockito.Mockito.*;


@RunWith(SpringRunner.class)
public class DeletePersonByPersonIdTest extends WebServiceTestAbstract {

    AsyncAdvisFacade advisMock = mock(AsyncAdvisFacade.class);

    @Before
    public void setup() {
        getBusinessService().setAdvisService(advisMock);
    }

    @Autowired
    FromXMLMapper requestMapper;

    @Autowired
    JdbcTemplate jdbcTemplate;

    @Autowired
    Properties properties;

    @Test
    public void deleteByPersonId_DeactivatedInProd() throws Exception {
        boolean oldIsProduction = properties.isProduction;
        properties.isProduction = true;
        PersonVO person = requestMapper.mapPersonFromRequest(createPerson());
        DeletePersonRequest deletePersonRequest = setRequest(person, false);
        try {
            DeletePersonResponse response = service.deletePerson(deletePersonRequest);
        } catch (Exception e) {
            assertEquals(FaultCode.DELETE_PERSON_BY_PERSONID_NOT_ALLOWED_IN_PROD.getMessage(), e.getMessage());
        } finally {
            properties.isProduction = oldIsProduction;
        }
    }

    @Test
    public void deletePersonByPersonID_WithValidRequest_PersonDeleted() throws Exception {
        PersonVO person = requestMapper.mapPersonFromRequest(createPerson());
        DeletePersonRequest deletePersonRequest = setRequest(person, false);
        DeletePersonResponse response = service.deletePerson(deletePersonRequest);
        assertNotNull(response);
        verify(advisMock, times(0)).advisPersonIdUpdated(any(), any(), any());
        verify(advisMock, times(1)).advisPersonIdDeleted(any(), any(), any());
    }

    @Test
    public void deleteByPersonID_WithInvalidRequest_ValidationError() throws Exception {
        PersonVO person = requestMapper.mapPersonFromRequest(createPerson());
        DeletePersonRequest deletePersonRequest = setRequest(person, true);

        try {
            service.deletePerson(deletePersonRequest);
            fail("Should have caused error message: " + ValidationErrors.PERSONID_VALUES_REQUIRED);
        } catch (Exception e) {
            checkError(e, ValidationErrors.PERSONID_VALUES_REQUIRED);
        }
        // Failed request should cause no advis to be sent
        verify(advisMock, times(0)).advisPersonIdUpdated(any(), any(), any());
        verify(advisMock, times(0)).advisPersonIdDeleted(any(), any(), any());
    }

    @Test
    public void deleteByPersonId_InvalidPersonId_ValidationError() throws Exception {
        PersonVO person = requestMapper.mapPersonFromRequest(createPerson());
        DeletePersonRequest deletePersonRequest = new DeletePersonRequest();
        PersonIdType personIdType = new PersonIdType();
        KeyIdentifier keyIdentifier = new KeyIdentifier();
        Modified modified = factory.createModified();

        keyIdentifier.setOID(OID.C_ECPR);
        keyIdentifier.setValue("0101010AA0"); // Non-existing

        personIdType.setKeyIdentifier(keyIdentifier);
        deletePersonRequest.getPersons().add(personIdType);
        deletePersonRequest.setModified(modified);

        try {
            service.deletePerson(deletePersonRequest);
            fail("Should have caused error message: " + ValidationErrors.KEY_IDENTIFIER_INVALID);
        } catch (Exception e) {
            checkError(e, ValidationErrors.KEY_IDENTIFIER_INVALID);
        }
    }

    @Test
    public void deleteByPersonId_validToSetOnEverything() throws Exception {
        int initialAddressCount = jdbcTemplate.queryForObject("SELECT count(1) FROM Address WHERE ValidTo > now()", Integer.class);
        int initialBirthDateCount = jdbcTemplate.queryForObject("SELECT count(1) FROM BirthDate WHERE ValidTo > now()", Integer.class);
        int initialContactCount = jdbcTemplate.queryForObject("SELECT count(1) FROM Contact WHERE ValidTo > now()", Integer.class);
        int initialGenderCount = jdbcTemplate.queryForObject("SELECT count(1) FROM Gender WHERE ValidTo > now()", Integer.class);
        int initialIdentifierCount = jdbcTemplate.queryForObject("SELECT count(1) FROM Identifier WHERE ValidTo > now()", Integer.class);
        int initialNameCount = jdbcTemplate.queryForObject("SELECT count(1) FROM Name WHERE ValidTo > now()", Integer.class);

        PersonVO person = requestMapper.mapPersonFromRequest(createPerson());

        assertEquals((Integer) (initialAddressCount + 1), jdbcTemplate.queryForObject("SELECT count(1) FROM Address WHERE ValidTo > now()", Integer.class));
        assertEquals((Integer) (initialBirthDateCount + 1), jdbcTemplate.queryForObject("SELECT count(1) FROM BirthDate WHERE ValidTo > now()", Integer.class));
        assertEquals((Integer) (initialContactCount + 1), jdbcTemplate.queryForObject("SELECT count(1) FROM Contact WHERE ValidTo > now()", Integer.class));
        assertEquals((Integer) (initialGenderCount + 1), jdbcTemplate.queryForObject("SELECT count(1) FROM Gender WHERE ValidTo > now()", Integer.class));
        assertEquals((Integer) (initialIdentifierCount + 2), jdbcTemplate.queryForObject("SELECT count(1) FROM Identifier WHERE ValidTo > now()", Integer.class));
        assertEquals((Integer) (initialNameCount + 1), jdbcTemplate.queryForObject("SELECT count(1) FROM Name WHERE ValidTo > now()", Integer.class));

        DeletePersonRequest deletePersonRequest = setRequest(person, false);
        DeletePersonResponse response = service.deletePerson(deletePersonRequest);

        assertEquals((Integer) (initialAddressCount), jdbcTemplate.queryForObject("SELECT count(1) FROM Address WHERE ValidTo > now()", Integer.class));
        assertEquals((Integer) (initialBirthDateCount), jdbcTemplate.queryForObject("SELECT count(1) FROM BirthDate WHERE ValidTo > now()", Integer.class));
        assertEquals((Integer) (initialContactCount), jdbcTemplate.queryForObject("SELECT count(1) FROM Contact WHERE ValidTo > now()", Integer.class));
        assertEquals((Integer) (initialGenderCount), jdbcTemplate.queryForObject("SELECT count(1) FROM Gender WHERE ValidTo > now()", Integer.class));
        assertEquals((Integer) (initialIdentifierCount), jdbcTemplate.queryForObject("SELECT count(1) FROM Identifier WHERE ValidTo > now()", Integer.class));
        assertEquals((Integer) (initialNameCount), jdbcTemplate.queryForObject("SELECT count(1) FROM Name WHERE ValidTo > now()", Integer.class));
    }
}
