/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice;

import dk.sds.ecpr2.*;
import dk.sds.ecpr2._2017._10._01.GetPersonByIdRequest;
import dk.sds.ecpr2._2017._10._01.GetPersonByIdResponse;
import dk.sds.ecpr2._2017._10._01.UpdatePersonRequest;
import dk.sds.ecpr2._2017._10._01.UpdatePersonResponse;
import dk.sds.ecpr2.domain.PersonVO;
import dk.sds.ecpr2.service.handler.DeletePersonHandler;
import dk.sds.ecpr2.webservice.mapping.FromXMLMapper;
import dk.sds.ecpr2.webservice.validation.ValidationErrors;
import org.junit.Before;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;

import java.util.Calendar;
import java.util.stream.Collectors;

import static org.junit.Assert.*;

public class GetPersonByPersonIDTest extends WebServiceTestAbstract {
    GetPersonByIdRequest getPersonByIdRequest;
    KeyIdentifier keyIdentifier;

    @Autowired DeletePersonHandler deletePersonHandler;

    @Autowired FromXMLMapper requestMapper;

    @Before
    public void setUp() throws Exception {
        getPersonByIdRequest = new GetPersonByIdRequest();
        PersonType person = createPerson();
        PersonVO per = requestMapper.mapPersonFromRequest(person);
        keyIdentifier = new KeyIdentifier();
        keyIdentifier.setOID(per.getIdentifiers().get(0).getOid());
        keyIdentifier.setValue(per.getIdentifiers().get(0).getValue());
        getPersonByIdRequest.getKeyIdentifiers().add(keyIdentifier);

    }

    @Test
    public void getPersonByPersonID_WithValidRequest_PersonLoaded() throws Exception {
        GetPersonByIdResponse response = service.getPersonById(getPersonByIdRequest);
        assertNotNull(response);
        assertNotNull(response.getPersons());
        assertEquals(1, response.getPersons().size());
        assertEquals("Hansen", response.getPersons().get(0).getNames().get(0).getFamilyName());
    }

    @Test
    public void getPersonByPersonID_PersonUpdated_PersonLoaded() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = setupPersonForUpdate();
        ContactType addedEntity = factory.createContact("12345678");
        personForUpdate.getContacts().add(addedEntity);
        UpdatePersonResponse updatedPerson = callUpdatePerson(personForUpdate, null);

        PersonVO per = requestMapper.mapPersonFromRequest(personForUpdate);
        keyIdentifier = new KeyIdentifier();
        keyIdentifier.setOID(per.getKeyIdentifier().getOid());
        keyIdentifier.setValue(per.getKeyIdentifier().getValue());
        GetPersonByIdRequest getPersonByIdRequest = new GetPersonByIdRequest();
        getPersonByIdRequest.getKeyIdentifiers().add(keyIdentifier);
        GetPersonByIdResponse response = service.getPersonById(getPersonByIdRequest);
        assertNotNull(response);
        assertNotNull(response.getPersons());
        assertEquals(1, response.getPersons().size());
        assertEquals("Hansen", response.getPersons().get(0).getNames().get(0).getFamilyName());
    }

    /*
    The other methods that use requestContext aren't tested either
    @Test
    public void getPersonByPersonID_DeletePerson_EmptyList() throws Exception {
        List<IdentifierVO> identifierVO = requestMapper.mapIdentifier(getPersonByIdRequest.getKeyIdentifiers());
        ServiceActorVO actor = new ServiceActorVO();
        Modified modified = factory.createModified();
        deletePersonHandler.execute(identifierVO, modified, actor, context);
        try {
            assertEquals(service.getPersonById(getPersonByIdRequest).getPersons().size(), 0);
        } catch (Exception e) {
            checkError(e, ValidationErrors.KEY_IDENTIFIER_INVALID);
        }
    }*/

    @Test
    public void getPersonByPersonID_WithInvalidRequest_ValidationError() throws Exception {
        keyIdentifier.setOID("");
        keyIdentifier.setValue(""); //NON EXISTING
        GetPersonByIdRequest request = new GetPersonByIdRequest();
        request.getKeyIdentifiers().add(keyIdentifier);
        try {
            service.getPersonById(getPersonByIdRequest);
            fail("Should have caused error message: " + ValidationErrors.PERSONID_VALUES_REQUIRED);
        } catch (Exception e) {
            checkError(e, ValidationErrors.PERSONID_VALUES_REQUIRED);
        }
    }

    @Test
    public void getPersonByPersonID_PersonIDNotExist_EmptyList() throws Exception {
        keyIdentifier.setOID("1.2.208.176.1.1001");
        keyIdentifier.setValue("2106176SM0"); //NON EXISTING
        GetPersonByIdRequest request = new GetPersonByIdRequest();
        request.getKeyIdentifiers().add(keyIdentifier);
        try {
            assertEquals(service.getPersonById(getPersonByIdRequest).getPersons().size(), 0);
        } catch (Exception e) {
            checkError(e, ValidationErrors.KEY_IDENTIFIER_INVALID);
        }
    }

    @Test
    public void getPersonByOldIdentifier() throws Exception {
        PersonType personType = createPerson();
        PersonVO personVO = requestMapper.mapPersonFromRequest(personType);
        KeyIdentifier oldKeyIdentifier = new KeyIdentifier();
        oldKeyIdentifier.setOID(personVO.getIdentifiers().get(0).getOid());
        oldKeyIdentifier.setValue(personVO.getIdentifiers().get(0).getValue());

        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(personType);
        personForUpdate.getBirthDates().get(0).setValue("2001-01-01");
        personForUpdate.getBirthDates().get(0).setValidFrom(null);
        personForUpdate.getBirthDates().get(0).setValidTo(null);
        UpdatePersonResponse updatePersonResponse = callUpdatePerson(personForUpdate, null);
//
//        personForUpdate = factory.createPersonForUpdate(updatePersonResponse.getPerson());
//        BirthDateType birthDateType = new BirthDateType();
//        birthDateType.setValue("2001-01-01");
//        personForUpdate.getBirthDates().add(birthDateType);
//        updatePersonResponse = callUpdatePerson(personForUpdate, null);

        GetPersonByIdRequest request = new GetPersonByIdRequest();
        request.getKeyIdentifiers().add(oldKeyIdentifier);
        GetPersonByIdResponse response = service.getPersonById(request);
        assertEquals(1, response.getPersons().size());
        assertEquals(1, response.getPersons().get(0).getIdentifiers().stream().filter(i -> i.getOID().equals(OID.X_ECPR)).count());
        assertNull(response.getPersons().get(0).getIdentifiers().stream().filter(i -> i.getOID().equals(OID.X_ECPR)).collect(Collectors.toList()).get(0).getValidTo());
        assertEquals(oldKeyIdentifier.getValue(), response.getPersons().get(0).getIdentifiers().stream().filter(i -> i.getOID().equals(OID.X_ECPR)).collect(Collectors.toList()).get(0).getValue());
    }

    @Test
    public void getPersonByRenewedIdentifier() throws Exception {
        PersonType personType = createPerson();
        PersonVO personVO = requestMapper.mapPersonFromRequest(personType);
        KeyIdentifier oldDecprIdentifier = new KeyIdentifier();
        oldDecprIdentifier.setOID(personVO.getIdentifiers().get(1).getOid());
        oldDecprIdentifier.setValue(personVO.getIdentifiers().get(1).getValue());

        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(personType);
        personForUpdate.getIdentifiers().get(0).setExpiry(null);
        personForUpdate.getIdentifiers().get(0).setOID(null);
        personForUpdate.getIdentifiers().get(0).setOIDType(null);
        personForUpdate.getIdentifiers().get(0).setPrimaryOIDLabel(null);
        personForUpdate.getIdentifiers().get(0).setSecondaryOIDLabel(null);
        personForUpdate.getIdentifiers().get(0).setValidFrom(null);
        personForUpdate.getIdentifiers().get(0).setValidTo(null);
        personForUpdate.getIdentifiers().get(0).setValidity(null);
        personForUpdate.getIdentifiers().get(0).setValue(null);
        UpdatePersonResponse updatePersonResponse = callUpdatePerson(personForUpdate, null);

        personForUpdate = factory.createPersonForUpdate(updatePersonResponse.getPerson());
        Modified modified = personForUpdate.getModifieds().get(0);
        personForUpdate.getModifieds().clear();
        personForUpdate.getModifieds().add(modified);
        Identifier identifier = new Identifier();
        identifier.setValue(keyIdentifier.getValue());
        identifier.setOID(keyIdentifier.getOID());
        identifier.setValidity(0);
        personForUpdate.getIdentifiers().add(identifier);
        updatePersonResponse = callUpdatePerson(personForUpdate, null);

        GetPersonByIdRequest request = new GetPersonByIdRequest();
        request.getKeyIdentifiers().add(oldDecprIdentifier);
        GetPersonByIdResponse response = service.getPersonById(request);
        assertEquals(1, response.getPersons().size());
        assertEquals(2, response.getPersons().get(0).getIdentifiers().stream().filter(i -> !i.getOID().equals(OID.X_ECPR)).count());
        assertEquals(2, response.getPersons().get(0).getIdentifiers().stream().filter(i -> i.getValue().equals(oldDecprIdentifier.getValue())).count());
    }

    private PersonWithKeyIdentifierType setupPersonForUpdate() throws Exception {
        PersonType newPerson = createPerson();
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(newPerson);
        factory.clearAllExceptKeyIdentifier(personForUpdate);
        return personForUpdate;
    }

    private UpdatePersonResponse callUpdatePerson(PersonWithKeyIdentifierType personForUpdate, String orgId) throws Exception {
        if (orgId != null) {
            personForUpdate.getModifieds().get(0).getBy().getOrganisation().getIdentifier().setIdentifier(orgId);
        }
        UpdatePersonRequest request = new UpdatePersonRequest();

        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        request.setPerson(personForUpdate);
        return service.updatePerson(request);
    }
}
