/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice;

import dk.sds.ecpr2.*;
import dk.sds.ecpr2._2017._10._01.*;
import dk.sds.ecpr2.domain.PersonVO;
import dk.sds.ecpr2.util.ValueSets;
import dk.sds.ecpr2.webservice.mapping.FromXMLMapper;
import dk.sds.ecpr2.webservice.validation.ValidationErrors;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.stream.Collectors;

import static org.junit.Assert.*;

@Component
public class MergePersonsTest extends WebServiceTestAbstract {
    @Autowired
    FromXMLMapper requestMapper;

    @Test
    public void mergePersons_twoEqualPersonsWithDifferentKeyIdentifiers_foundOneUniquePerson() throws Exception {
        PersonType personToBeCreatedFrom = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        PersonType personToBeCreatedTo = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");

        PersonType createdPersonFrom = createPersonInDb(personToBeCreatedFrom);
        PersonType createdPersonTo = createPersonInDb(personToBeCreatedTo);
        KeyIdentifier keyFrom = new KeyIdentifier();
        keyFrom.setValue(createdPersonFrom.getIdentifiers().get(0).getValue());
        keyFrom.setOID(createdPersonFrom.getIdentifiers().get(0).getOID());
        KeyIdentifier keyTo = new KeyIdentifier();
        keyTo.setValue(createdPersonTo.getIdentifiers().get(0).getValue());
        keyTo.setOID(createdPersonTo.getIdentifiers().get(0).getOID());

        MergePersonsResponse mergedPerson = callMergePerson(keyFrom, keyTo);

        assertNotNull(mergedPerson);
        assertEquals(1, mergedPerson.getPerson().getAddresses().size());
        assertEquals(1, mergedPerson.getPerson().getNames().size());
        assertEquals(1, mergedPerson.getPerson().getBirthDates().size());
        assertEquals(1, mergedPerson.getPerson().getContacts().size());
        assertEquals(2, mergedPerson.getPerson().getIdentifiers().size());
        assertEquals(marshal(createdPersonTo.getNames().get(0)), marshal(mergedPerson.getPerson().getNames().get(0)));
        assertEquals(marshal(createdPersonTo.getAddresses().get(0)), marshal(mergedPerson.getPerson().getAddresses().get(0)));
        assertEquals(marshal(createdPersonTo.getBirthDates().get(0)), marshal(mergedPerson.getPerson().getBirthDates().get(0)));
        assertEquals(marshal(createdPersonTo.getContacts().get(0)), marshal(mergedPerson.getPerson().getContacts().get(0)));
        isRecentlyInvalidated(mergedPerson.getPerson().getIdentifiers().get(0), 1);
        createdPersonFrom.getIdentifiers().get(0).setValidTo(mergedPerson.getPerson().getIdentifiers().get(0).getValidTo());
        assertEquals(marshal(createdPersonFrom.getIdentifiers().get(0)), marshal(mergedPerson.getPerson().getIdentifiers().get(0)));
        assertEquals(marshal(createdPersonTo.getIdentifiers().get(0)), marshal(mergedPerson.getPerson().getIdentifiers().get(1)));
    }

    @Test
    public void mergePersons_DifferentNamesFails() throws Exception {
        PersonType personToBeCreatedFrom = createPerson("Benny", "Jensen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        PersonType personToBeCreatedTo = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");

        PersonType createdPersonFrom = createPersonInDb(personToBeCreatedFrom);
        PersonType createdPersonTo = createPersonInDb(personToBeCreatedTo);
        KeyIdentifier keyFrom = new KeyIdentifier();
        keyFrom.setValue(createdPersonFrom.getIdentifiers().get(0).getValue());
        keyFrom.setOID(createdPersonFrom.getIdentifiers().get(0).getOID());
        KeyIdentifier keyTo = new KeyIdentifier();
        keyTo.setValue(createdPersonTo.getIdentifiers().get(0).getValue());
        keyTo.setOID(createdPersonTo.getIdentifiers().get(0).getOID());

        try {
            callMergePerson(keyFrom, keyTo);
            fail("Should have caused error message: " + ValidationErrors.NON_UNIQUE_MERGE);
        } catch (Exception ex) {
            checkError(ex, ValidationErrors.NON_UNIQUE_MERGE);
        }
    }

    @Test
    public void mergePersons_DifferentButSimilarNamesFails() throws Exception {
        PersonType personToBeCreatedFrom = createPerson("Benny Olsen", "", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        PersonType personToBeCreatedTo = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");

        PersonType createdPersonFrom = createPersonInDb(personToBeCreatedFrom);
        PersonType createdPersonTo = createPersonInDb(personToBeCreatedTo);
        KeyIdentifier keyFrom = new KeyIdentifier();
        keyFrom.setValue(createdPersonFrom.getIdentifiers().get(0).getValue());
        keyFrom.setOID(createdPersonFrom.getIdentifiers().get(0).getOID());
        KeyIdentifier keyTo = new KeyIdentifier();
        keyTo.setValue(createdPersonTo.getIdentifiers().get(0).getValue());
        keyTo.setOID(createdPersonTo.getIdentifiers().get(0).getOID());

        try {
            callMergePerson(keyFrom, keyTo);
            fail("Should have caused error message: " + ValidationErrors.NON_UNIQUE_MERGE);
        } catch (Exception ex) {
            checkError(ex, ValidationErrors.NON_UNIQUE_MERGE);
        }
    }

    @Test
    public void mergePersons_DifferentAddressFails() throws Exception {
        PersonType personToBeCreatedFrom = createPerson("Benny", "Olsen", "Testvej 16", "Teststrup", "8000", "1981-05-19", "DK");
        PersonType personToBeCreatedTo = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");

        PersonType createdPersonFrom = createPersonInDb(personToBeCreatedFrom);
        PersonType createdPersonTo = createPersonInDb(personToBeCreatedTo);
        KeyIdentifier keyFrom = new KeyIdentifier();
        keyFrom.setValue(createdPersonFrom.getIdentifiers().get(0).getValue());
        keyFrom.setOID(createdPersonFrom.getIdentifiers().get(0).getOID());
        KeyIdentifier keyTo = new KeyIdentifier();
        keyTo.setValue(createdPersonTo.getIdentifiers().get(0).getValue());
        keyTo.setOID(createdPersonTo.getIdentifiers().get(0).getOID());

        try {
            callMergePerson(keyFrom, keyTo);
            fail("Should have caused error message: " + ValidationErrors.NON_UNIQUE_MERGE);
        } catch (Exception ex) {
            checkError(ex, ValidationErrors.NON_UNIQUE_MERGE);
        }
    }

    @Test
    public void mergePersons_DifferentBirthdateFails() throws Exception {
        PersonType personToBeCreatedFrom = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "2001-05-19", "DK");
        PersonType personToBeCreatedTo = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");

        PersonType createdPersonFrom = createPersonInDb(personToBeCreatedFrom);
        PersonType createdPersonTo = createPersonInDb(personToBeCreatedTo);
        KeyIdentifier keyFrom = new KeyIdentifier();
        keyFrom.setValue(createdPersonFrom.getIdentifiers().get(0).getValue());
        keyFrom.setOID(createdPersonFrom.getIdentifiers().get(0).getOID());
        KeyIdentifier keyTo = new KeyIdentifier();
        keyTo.setValue(createdPersonTo.getIdentifiers().get(0).getValue());
        keyTo.setOID(createdPersonTo.getIdentifiers().get(0).getOID());

        try {
            callMergePerson(keyFrom, keyTo);
            fail("Should have caused error message: " + ValidationErrors.NON_UNIQUE_MERGE);
        } catch (Exception ex) {
            checkError(ex, ValidationErrors.NON_UNIQUE_MERGE);
        }
    }

    @Test
    public void mergePersons_DifferentGenderFails() throws Exception {
        PersonType personToBeCreatedFrom = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        personToBeCreatedFrom.getGenders().clear();
        personToBeCreatedFrom.getGenders().add(factory.createGender(ValueSets.genderValues[1]));
        PersonType personToBeCreatedTo = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");

        PersonType createdPersonFrom = createPersonInDb(personToBeCreatedFrom);
        PersonType createdPersonTo = createPersonInDb(personToBeCreatedTo);
        KeyIdentifier keyFrom = new KeyIdentifier();
        keyFrom.setValue(createdPersonFrom.getIdentifiers().get(0).getValue());
        keyFrom.setOID(createdPersonFrom.getIdentifiers().get(0).getOID());
        KeyIdentifier keyTo = new KeyIdentifier();
        keyTo.setValue(createdPersonTo.getIdentifiers().get(0).getValue());
        keyTo.setOID(createdPersonTo.getIdentifiers().get(0).getOID());

        try {
            callMergePerson(keyFrom, keyTo);
            fail("Should have caused error message: " + ValidationErrors.NON_UNIQUE_MERGE);
        } catch (Exception ex) {
            checkError(ex, ValidationErrors.NON_UNIQUE_MERGE);
        }
    }

    @Test
    public void mergePersons_MissingNameSucceedsWithAvailableName() throws Exception {
        PersonType personToBeCreatedFrom = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        personToBeCreatedFrom.getNames().clear();
        PersonType personToBeCreatedTo = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");

        PersonType createdPersonFrom = createPersonInDb(personToBeCreatedFrom);
        PersonType createdPersonTo = createPersonInDb(personToBeCreatedTo);
        KeyIdentifier keyFrom = new KeyIdentifier();
        keyFrom.setValue(createdPersonFrom.getIdentifiers().get(0).getValue());
        keyFrom.setOID(createdPersonFrom.getIdentifiers().get(0).getOID());
        KeyIdentifier keyTo = new KeyIdentifier();
        keyTo.setValue(createdPersonTo.getIdentifiers().get(0).getValue());
        keyTo.setOID(createdPersonTo.getIdentifiers().get(0).getOID());

        MergePersonsResponse mergedPerson = callMergePerson(keyFrom, keyTo);

        assertNotNull(mergedPerson);
        assertEquals(1, mergedPerson.getPerson().getAddresses().size());
        assertEquals(1, mergedPerson.getPerson().getNames().size());
        assertEquals(1, mergedPerson.getPerson().getBirthDates().size());
        assertEquals(1, mergedPerson.getPerson().getContacts().size());
        assertEquals(2, mergedPerson.getPerson().getIdentifiers().size());
        assertEquals(marshal(createdPersonTo.getNames().get(0)), marshal(mergedPerson.getPerson().getNames().get(0)));
        assertEquals(marshal(createdPersonTo.getAddresses().get(0)), marshal(mergedPerson.getPerson().getAddresses().get(0)));
        assertEquals(marshal(createdPersonTo.getBirthDates().get(0)), marshal(mergedPerson.getPerson().getBirthDates().get(0)));
        assertEquals(marshal(createdPersonTo.getContacts().get(0)), marshal(mergedPerson.getPerson().getContacts().get(0)));
        assertEquals(marshal(createdPersonTo.getIdentifiers().get(0)), marshal(mergedPerson.getPerson().getIdentifiers().get(1)));
        isRecentlyInvalidated(mergedPerson.getPerson().getIdentifiers().get(0), 1);
        createdPersonFrom.getIdentifiers().get(0).setValidTo(mergedPerson.getPerson().getIdentifiers().get(0).getValidTo());
        assertEquals(marshal(createdPersonFrom.getIdentifiers().get(0)), marshal(mergedPerson.getPerson().getIdentifiers().get(0)));
    }

    @Test
    public void mergePersons_ExtraIdentifierOnPerson1Succeeds() throws Exception {
        PersonType personToBeCreatedFrom = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        personToBeCreatedFrom.getIdentifiers().add(factory.createCprIdentifier("1111111118"));
        PersonType personToBeCreatedTo = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");

        PersonType createdPersonFrom = createPersonInDb(personToBeCreatedFrom);
        PersonType createdPersonTo = createPersonInDb(personToBeCreatedTo);
        KeyIdentifier keyFrom = new KeyIdentifier();
        keyFrom.setValue(createdPersonFrom.getIdentifiers().get(0).getValue());
        keyFrom.setOID(createdPersonFrom.getIdentifiers().get(0).getOID());
        KeyIdentifier keyTo = new KeyIdentifier();
        keyTo.setValue(createdPersonTo.getIdentifiers().get(0).getValue());
        keyTo.setOID(createdPersonTo.getIdentifiers().get(0).getOID());

        MergePersonsResponse mergedPerson = callMergePerson(keyFrom, keyTo);

        assertNotNull(mergedPerson);
        assertEquals(1, mergedPerson.getPerson().getAddresses().size());
        assertEquals(1, mergedPerson.getPerson().getNames().size());
        assertEquals(1, mergedPerson.getPerson().getBirthDates().size());
        assertEquals(1, mergedPerson.getPerson().getContacts().size());
        assertEquals(3, mergedPerson.getPerson().getIdentifiers().size());
        assertEquals(marshal(createdPersonTo.getNames().get(0)), marshal(mergedPerson.getPerson().getNames().get(0)));
        assertEquals(marshal(createdPersonTo.getAddresses().get(0)), marshal(mergedPerson.getPerson().getAddresses().get(0)));
        assertEquals(marshal(createdPersonTo.getBirthDates().get(0)), marshal(mergedPerson.getPerson().getBirthDates().get(0)));
        assertEquals(marshal(createdPersonTo.getContacts().get(0)), marshal(mergedPerson.getPerson().getContacts().get(0)));
        isRecentlyInvalidated(mergedPerson.getPerson().getIdentifiers().get(0), 1);
        createdPersonFrom.getIdentifiers().get(0).setValidTo(mergedPerson.getPerson().getIdentifiers().get(0).getValidTo());
        assertEquals(marshal(createdPersonFrom.getIdentifiers().get(0)), marshal(mergedPerson.getPerson().getIdentifiers().get(0)));
        assertEquals(marshal(createdPersonFrom.getIdentifiers().get(1)), marshal(mergedPerson.getPerson().getIdentifiers().get(1)));
        assertEquals(marshal(createdPersonTo.getIdentifiers().get(0)), marshal(mergedPerson.getPerson().getIdentifiers().get(2)));
    }

    @Test
    public void mergePersons_ExtraIdentifierOnPerson2Succeeds() throws Exception {
        PersonType personToBeCreatedFrom = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        PersonType personToBeCreatedTo = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        personToBeCreatedTo.getIdentifiers().add(factory.createCprIdentifier("1111111118"));

        PersonType createdPersonFrom = createPersonInDb(personToBeCreatedFrom);
        PersonType createdPersonTo = createPersonInDb(personToBeCreatedTo);
        KeyIdentifier keyFrom = new KeyIdentifier();
        keyFrom.setValue(createdPersonFrom.getIdentifiers().get(0).getValue());
        keyFrom.setOID(createdPersonFrom.getIdentifiers().get(0).getOID());
        KeyIdentifier keyTo = new KeyIdentifier();
        keyTo.setValue(createdPersonTo.getIdentifiers().get(0).getValue());
        keyTo.setOID(createdPersonTo.getIdentifiers().get(0).getOID());

        MergePersonsResponse mergedPerson = callMergePerson(keyFrom, keyTo);

        assertNotNull(mergedPerson);
        assertEquals(1, mergedPerson.getPerson().getAddresses().size());
        assertEquals(1, mergedPerson.getPerson().getNames().size());
        assertEquals(1, mergedPerson.getPerson().getBirthDates().size());
        assertEquals(1, mergedPerson.getPerson().getContacts().size());
        assertEquals(3, mergedPerson.getPerson().getIdentifiers().size());
        assertEquals(marshal(createdPersonTo.getNames().get(0)), marshal(mergedPerson.getPerson().getNames().get(0)));
        assertEquals(marshal(createdPersonTo.getAddresses().get(0)), marshal(mergedPerson.getPerson().getAddresses().get(0)));
        assertEquals(marshal(createdPersonTo.getBirthDates().get(0)), marshal(mergedPerson.getPerson().getBirthDates().get(0)));
        assertEquals(marshal(createdPersonTo.getContacts().get(0)), marshal(mergedPerson.getPerson().getContacts().get(0)));
        isRecentlyInvalidated(mergedPerson.getPerson().getIdentifiers().get(0), 1);
        createdPersonFrom.getIdentifiers().get(0).setValidTo(mergedPerson.getPerson().getIdentifiers().get(0).getValidTo());
        assertEquals(marshal(createdPersonFrom.getIdentifiers().get(0)), marshal(mergedPerson.getPerson().getIdentifiers().get(0)));
        assertEquals(marshal(createdPersonTo.getIdentifiers().get(0)), marshal(mergedPerson.getPerson().getIdentifiers().get(1)));
        assertEquals(marshal(createdPersonTo.getIdentifiers().get(1)), marshal(mergedPerson.getPerson().getIdentifiers().get(2)));
    }

    @Test
    public void mergePersons_HistoryOnFromSucceeds() throws Exception {
        PersonType personToBeCreatedFrom = createPerson("Benny", "Jensen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        PersonType personToBeCreatedTo = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");

        PersonType createdPersonFrom = createPersonInDb(personToBeCreatedFrom);
        PersonType createdPersonTo = createPersonInDb(personToBeCreatedTo);
        KeyIdentifier keyFrom = new KeyIdentifier();
        keyFrom.setValue(createdPersonFrom.getIdentifiers().get(0).getValue());
        keyFrom.setOID(createdPersonFrom.getIdentifiers().get(0).getOID());
        KeyIdentifier keyTo = new KeyIdentifier();
        keyTo.setValue(createdPersonTo.getIdentifiers().get(0).getValue());
        keyTo.setOID(createdPersonTo.getIdentifiers().get(0).getOID());

        createdPersonFrom.getNames().get(0).setFamilyName("Olsen");
        PersonType updatedPersonFrom = updatePersonInDb(createdPersonFrom);

        MergePersonsResponse mergedPerson = callMergePerson(keyFrom, keyTo);

        assertNotNull(mergedPerson);
        assertEquals(1, mergedPerson.getPerson().getAddresses().size());
        assertEquals(2, mergedPerson.getPerson().getNames().size());
        assertEquals(1, mergedPerson.getPerson().getBirthDates().size());
        assertEquals(1, mergedPerson.getPerson().getContacts().size());
        assertEquals(2, mergedPerson.getPerson().getIdentifiers().size());
        assertEquals(marshal(updatedPersonFrom.getNames().get(0)), marshal(mergedPerson.getPerson().getNames().get(0)));
        assertEquals(marshal(createdPersonTo.getNames().get(0)), marshal(mergedPerson.getPerson().getNames().get(1)));
        assertEquals(marshal(createdPersonTo.getAddresses().get(0)), marshal(mergedPerson.getPerson().getAddresses().get(0)));
        assertEquals(marshal(createdPersonTo.getBirthDates().get(0)), marshal(mergedPerson.getPerson().getBirthDates().get(0)));
        assertEquals(marshal(createdPersonTo.getContacts().get(0)), marshal(mergedPerson.getPerson().getContacts().get(0)));
        isRecentlyInvalidated(mergedPerson.getPerson().getIdentifiers().get(0), 1);
        updatedPersonFrom.getIdentifiers().get(0).setValidTo(mergedPerson.getPerson().getIdentifiers().get(0).getValidTo());
        assertEquals(marshal(updatedPersonFrom.getIdentifiers().get(0)), marshal(mergedPerson.getPerson().getIdentifiers().get(0)));
        assertEquals(marshal(createdPersonTo.getIdentifiers().get(0)), marshal(mergedPerson.getPerson().getIdentifiers().get(1)));
    }

    @Test
    public void mergePersons_HistoryOnToSucceeds() throws Exception {
        PersonType personToBeCreatedFrom = createPerson("Benny", "Olsen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");
        PersonType personToBeCreatedTo = createPerson("Benny", "Jensen", "Testvej 16", "Testby", "8000", "1981-05-19", "DK");

        PersonType createdPersonFrom = createPersonInDb(personToBeCreatedFrom);
        PersonType createdPersonTo = createPersonInDb(personToBeCreatedTo);
        KeyIdentifier keyFrom = new KeyIdentifier();
        keyFrom.setValue(createdPersonFrom.getIdentifiers().get(0).getValue());
        keyFrom.setOID(createdPersonFrom.getIdentifiers().get(0).getOID());
        KeyIdentifier keyTo = new KeyIdentifier();
        keyTo.setValue(createdPersonTo.getIdentifiers().get(0).getValue());
        keyTo.setOID(createdPersonTo.getIdentifiers().get(0).getOID());

        createdPersonTo.getNames().get(0).setFamilyName("Olsen");
        PersonType updatedPersonTo = updatePersonInDb(createdPersonTo);

        MergePersonsResponse mergedPerson = callMergePerson(keyFrom, keyTo);

        assertNotNull(mergedPerson);
        assertEquals(1, mergedPerson.getPerson().getAddresses().size());
        assertEquals(2, mergedPerson.getPerson().getNames().size());
        assertEquals(1, mergedPerson.getPerson().getBirthDates().size());
        assertEquals(1, mergedPerson.getPerson().getContacts().size());
        assertEquals(2, mergedPerson.getPerson().getIdentifiers().size());
        assertEquals(marshal(updatedPersonTo.getNames().get(0)), marshal(mergedPerson.getPerson().getNames().get(1)));
        assertEquals(marshal(updatedPersonTo.getNames().get(1)), marshal(mergedPerson.getPerson().getNames().get(0)));
        assertEquals(marshal(updatedPersonTo.getAddresses().get(0)), marshal(mergedPerson.getPerson().getAddresses().get(0)));
        assertEquals(marshal(updatedPersonTo.getBirthDates().get(0)), marshal(mergedPerson.getPerson().getBirthDates().get(0)));
        assertEquals(marshal(updatedPersonTo.getContacts().get(0)), marshal(mergedPerson.getPerson().getContacts().get(0)));
        isRecentlyInvalidated(mergedPerson.getPerson().getIdentifiers().get(0), 1);
        createdPersonFrom.getIdentifiers().get(0).setValidTo(mergedPerson.getPerson().getIdentifiers().get(0).getValidTo());
        assertEquals(marshal(createdPersonFrom.getIdentifiers().get(0)), marshal(mergedPerson.getPerson().getIdentifiers().get(0)));
        assertEquals(marshal(updatedPersonTo.getIdentifiers().get(0)), marshal(mergedPerson.getPerson().getIdentifiers().get(1)));
    }

    private MergePersonsResponse callMergePerson(KeyIdentifier from, KeyIdentifier to) throws Exception {
        MergePersonsRequest request = new MergePersonsRequest();
        request.setModified(factory.createModified("12345", "Jens Jensen", "1111111118", "CPR", "Læge", "Hospitalsvej 1", "hospital@hospital.dk", "Hospitalets hospital", "12345678", "hospital", "12345678901", "SOR"));
        request.setFrom(from);
        request.setTo(to);
        return service.mergePersons(request);
    }
}
