/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice;

import dk.sds.ecpr2.*;
import dk.sds.ecpr2._2017._10._01.ReservePersonIdRequest;
import dk.sds.ecpr2._2017._10._01.ReservePersonIdResponse;
import dk.sds.ecpr2.webservice.validation.ValidationErrors;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.test.context.junit4.SpringRunner;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.Calendar;
import java.util.List;

import static org.junit.Assert.*;

@RunWith(SpringRunner.class)
public class ReservePersonIdTest extends WebServiceTestAbstract {

    @Autowired
    JdbcTemplate jdbcTemplate;

    @Test
    public void testReservedUntilIllegal() throws Exception {
        Calendar reservedUntil = getReservedUntil(LocalDate.now().getYear() + 3);
        XMLGregorianCalendar forDate = getForDate(LocalDate.now().getYear() + 1);
        ReservePersonIdRequest request = createReserveRequest(1, forDate, reservedUntil, GenderFemaleOrMaleType.MALE);

        try {
            service.reservePersonId(request);
            fail("Should have caused error message: " + ValidationErrors.RESERVED_UNTIL_OVER_THRESHOLD);
        } catch (Exception ex) {
            checkError(ex, ValidationErrors.RESERVED_UNTIL_OVER_THRESHOLD);
        }
    }

    @Test
    public void testForDateNonFuture() throws Exception {
        Calendar reservedUntil = getReservedUntil(LocalDate.now().getYear() + 1);
        XMLGregorianCalendar forDate = getForDate(LocalDate.now().getYear() - 1);
        ReservePersonIdRequest request = createReserveRequest(1, forDate, reservedUntil, GenderFemaleOrMaleType.MALE);

        try {
            service.reservePersonId(request);
            fail("Should have caused error message: " + ValidationErrors.FOR_DATE_NON_FUTURE);
        } catch (Exception ex) {
            checkError(ex, ValidationErrors.FOR_DATE_NON_FUTURE);
        }
    }

    @Test
    public void testNumberTooLarge() throws Exception {
        Calendar reservedUntil = getReservedUntil(LocalDate.now().getYear() + 1);
        XMLGregorianCalendar forDate = getForDate(LocalDate.now().getYear() + 1);
        ReservePersonIdRequest request = createReserveRequest(200, forDate, reservedUntil, GenderFemaleOrMaleType.MALE);

        try {
            service.reservePersonId(request);
            fail("Should have caused error message: " + ValidationErrors.RESERVE_PERSON_ID_NUMBER_OVER_THRESHOLD);
        } catch (Exception ex) {
            checkError(ex, ValidationErrors.RESERVE_PERSON_ID_NUMBER_OVER_THRESHOLD);
        }
    }

    @Test
    public void testNotTooManyWithinTwoDays() throws Exception {
        Calendar reservedUntil = getReservedUntil(LocalDate.now().getYear() + 1);
        XMLGregorianCalendar forDate = getForDate(LocalDate.now().getYear() + 1);
        ReservePersonIdRequest request = createReserveRequest(51, forDate, reservedUntil, GenderFemaleOrMaleType.MALE);
        service.reservePersonId(request);

        jdbcTemplate.update("UPDATE Person SET ModifiedTime = ?", LocalDateTime.now().minusDays(1));

        ReservePersonIdResponse response = service.reservePersonId(request);
        assertEquals(1, response.getLists().size());
        assertEquals(51, response.getLists().get(0).getPersons().size());
    }

    @Test
    public void testTooManyWithinADay() throws Exception {
        Calendar reservedUntil = getReservedUntil(LocalDate.now().getYear() + 1);
        XMLGregorianCalendar forDate = getForDate(LocalDate.now().getYear() + 1);
        ReservePersonIdRequest request = createReserveRequest(51, forDate, reservedUntil, GenderFemaleOrMaleType.MALE);
        service.reservePersonId(request);

        try {
            service.reservePersonId(request);
            fail("Should have caused error message: " + ValidationErrors.RESERVE_PERSON_ID_TOO_MANY_WITHIN_A_DAY);
        } catch (Exception ex) {
            checkError(ex, ValidationErrors.RESERVE_PERSON_ID_TOO_MANY_WITHIN_A_DAY);
        }
    }

    @Test
    public void testRequestWithoutReservedUntil() throws Exception {
        XMLGregorianCalendar forDate = getForDate(LocalDate.now().getYear() + 1);
        ReservePersonIdRequest request = createReserveRequest(1, forDate, null, GenderFemaleOrMaleType.MALE);

        ReservePersonIdResponse response = service.reservePersonId(request);
        List<ResponseListType> lists = response.getLists();
        assertEquals(1, lists.size());
        List<PersonWithGenderType> persons = lists.get(0).getPersons();
        assertEquals(1, persons.size());
        assertStructureOk(persons, false, OID.C_ECPR);
        // ReservedUntil default to +48 hours when not provided in request (48h = 2 days)
        persons.forEach(p -> assertEquals(LocalDate.now().plusDays(2).getDayOfMonth(), p.getModifieds().get(0).getReservedUntil().get(Calendar.DAY_OF_MONTH)));
    }

    @Test
    public void testRequestWithoutForDate() throws Exception {
        Calendar reservedUntil = getReservedUntil(LocalDate.now().getYear() + 1);
        ReservePersonIdRequest request = createReserveRequest(1, null, reservedUntil, GenderFemaleOrMaleType.MALE);

        ReservePersonIdResponse response = service.reservePersonId(request);
        List<ResponseListType> lists = response.getLists();
        assertEquals(1, lists.size());
        List<PersonWithGenderType> persons = lists.get(0).getPersons();
        assertEquals(1, persons.size());
        // C-eCPR date is tested in ReservePersonIdHandlerTest
        persons.forEach(p -> assertEquals(10, p.getIdentifiers().get(0).getValue().length()));
        assertStructureOk(persons, false, OID.C_ECPR);
    }

    public static ReservePersonIdRequest createReserveRequest(int number, XMLGregorianCalendar forDate, Calendar reservedUntil, GenderFemaleOrMaleType gender) {
        ReservePersonIdRequest request = new ReservePersonIdRequest();
        RequestListType entry = new RequestListType();
        entry.setOID(OID.C_ECPR);
        entry.setForDate(forDate);
        entry.setGender(gender);
        entry.setNumber(number);
        entry.setReservedUntil(reservedUntil);
        request.getLists().add(entry);
        Modified modified = new Modified();
        ModifiedBy modifiedBy = new ModifiedBy();
        modifiedBy.setRole("Læge");
        ModificatorPersonType person = new ModificatorPersonType();
        person.setName("Lars Larsen");
        person.setSpecialityCode("1234");
        person.setAuthorisationIdentifier("1234KJ");
        PersonIdentifierType personId = new PersonIdentifierType();
        personId.setPersonIdentifier("1111111118");
        personId.setSource("CPR");
        person.setPersonIdentifier(personId);
        modifiedBy.setPerson(person);
        OrganisationType organisation = new OrganisationType();
        organisation.setAddressLine("Vejnavn 24");
        organisation.setTelephoneNumber("12345678");
        organisation.setType("Hospital");
        organisation.setName("Stedets hospital");
        organisation.setEmailAddress("hospital@hospital.dk");
        OrganisationIdentifierType orgId = new OrganisationIdentifierType();
        orgId.setIdentifier("12341234");
        orgId.setSource("SOR");
        organisation.setIdentifier(orgId);
        modifiedBy.setOrganisation(organisation);
        modified.setBy(modifiedBy);
        request.setModified(modified);
        return request;
    }

    @Test
    public void testRequestWithoutGender() throws Exception {
        XMLGregorianCalendar forDate = getForDate(LocalDate.now().getYear() + 1);
        Calendar reservedUntil = getReservedUntil(LocalDate.now().getYear() + 1);
        ReservePersonIdRequest request = createReserveRequest(1, forDate, reservedUntil, null);

        ReservePersonIdResponse response = service.reservePersonId(request);
        List<ResponseListType> lists = response.getLists();
        assertEquals(1, lists.size());
        List<PersonWithGenderType> persons = lists.get(0).getPersons();
        assertEquals(1, persons.size());
        persons.forEach(p -> assertTrue(
                GenderFemaleOrMaleType.MALE.value().equals(p.getGenders().get(0).getValue()) ||
                        GenderFemaleOrMaleType.FEMALE.value().equals(p.getGenders().get(0).getValue())
        ));
        assertStructureOk(persons, true, OID.C_ECPR);
    }

    @Test
    public void testValidRequestForCeCPR() throws Exception {
        ReservePersonIdRequest request = new ReservePersonIdRequest();
        RequestListType entry = new RequestListType();
        entry.setOID(OID.C_ECPR);
        int nextYear = LocalDate.now().getYear() + 1;
        entry.setForDate(getForDate(nextYear));
        entry.setGender(GenderFemaleOrMaleType.MALE);
        int number = 100;
        entry.setNumber(number);
        entry.setReservedUntil(getReservedUntil(nextYear));
        request.setModified(factory.createModified());
        request.getLists().add(entry);

        ReservePersonIdResponse response = service.reservePersonId(request);

        List<ResponseListType> lists = response.getLists();
        assertEquals(1, lists.size());
        List<PersonWithGenderType> persons = lists.get(0).getPersons();
        assertEquals(number, persons.size());
        assertStructureOk(persons, false, OID.C_ECPR);
    }

    @Test
    public void testValidRequestForXeCPR() throws Exception {
        ReservePersonIdRequest request = new ReservePersonIdRequest();
        RequestListType entry = new RequestListType();
        entry.setOID(OID.X_ECPR);
        int nextYear = LocalDate.now().getYear() + 1;
        entry.setForDate(getForDate(nextYear));
        entry.setGender(GenderFemaleOrMaleType.MALE);
        int number = 100;
        entry.setNumber(number);
        entry.setReservedUntil(getReservedUntil(nextYear));
        request.setModified(factory.createModified());
        request.getLists().add(entry);

        ReservePersonIdResponse response = service.reservePersonId(request);

        List<ResponseListType> lists = response.getLists();
        assertEquals(1, lists.size());
        List<PersonWithGenderType> persons = lists.get(0).getPersons();
        assertEquals(number, persons.size());
        assertStructureOk(persons, false, OID.X_ECPR);
    }

    private void assertStructureOk(List<PersonWithGenderType> persons, boolean ignoreGender, String ecprOid) {
        // Gender asserts
        persons.forEach(p -> assertEquals(1, p.getGenders().size()));
        if (ignoreGender) {
            persons.forEach(p -> assertNotNull(p.getGenders().get(0).getValue()));
        } else {
            persons.forEach(p -> assertEquals(GenderFemaleOrMaleType.MALE.value(), p.getGenders().get(0).getValue()));
        }
        persons.forEach(p -> assertNotNull(p.getGenders().get(0).getPID()));
        persons.forEach(p -> assertNotNull(p.getGenders().get(0).getValidFrom()));
        persons.forEach(p -> assertNull(p.getGenders().get(0).getValidTo()));

        // Identifier asserts
        persons.forEach(p -> assertEquals(1, p.getIdentifiers().size()));
        persons.forEach(p -> assertEquals(16, p.getIdentifiers().get(0).getValidity().intValue()));
        persons.forEach(p -> assertEquals(ecprOid, p.getIdentifiers().get(0).getOID()));
        persons.forEach(p -> assertNotNull(p.getIdentifiers().get(0).getPrimaryOIDLabel()));
        persons.forEach(p -> assertNotNull(p.getIdentifiers().get(0).getPID()));
        persons.forEach(p -> assertNotNull(p.getIdentifiers().get(0).getValidFrom()));
        persons.forEach(p -> assertEquals(10, p.getIdentifiers().get(0).getValue().length()));
        persons.forEach(p -> assertNull(p.getIdentifiers().get(0).getValidTo()));
        persons.forEach(p -> assertNull(p.getIdentifiers().get(0).getExpiry()));

        // Version asserts
        persons.forEach(p -> assertEquals(1, p.getModifieds().size()));
        persons.forEach(p -> assertNotNull(p.getModifieds().get(0).getValidFrom()));
        persons.forEach(p -> assertNotNull(p.getModifieds().get(0).getReservedUntil()));
        persons.forEach(p -> assertNull(p.getModifieds().get(0).getValidTo()));
    }

    private Calendar getReservedUntil(int nextYear) {
        Calendar cal = Calendar.getInstance();
        cal.set(Calendar.YEAR, nextYear);
        cal.set(Calendar.MONTH, 11);
        cal.set(Calendar.DAY_OF_MONTH, 24);
        return cal;
    }

    private XMLGregorianCalendar getForDate(int nextYear) throws DatatypeConfigurationException {
        XMLGregorianCalendar forDate = DatatypeFactory.newInstance().newXMLGregorianCalendar();
        forDate.setYear(nextYear);
        forDate.setMonth(5);
        forDate.setDay(11);
        return forDate;
    }
}
