/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice;

import dk.sds.ecpr2.IntegrationTest;
import dk.sds.ecpr2.security.AssistantAuthenticator;
import dk.sds.ecpr2.security.RoleValidationInterceptor;
import dk.sds.ecpr2.security.SecurityCredentials;
import org.junit.Before;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.ws.context.MessageContext;
import org.springframework.ws.soap.SoapHeader;
import org.springframework.ws.soap.SoapMessage;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.dom.DOMSource;
import java.io.StringReader;
import java.util.ArrayList;

import static org.junit.Assert.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class RoleValidationTest extends IntegrationTest {

    private static final String DOCTOR_SOAP_HEADER =
            "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" +
                    "<soap:Header xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\">" +
                    "<Security xmlns=\"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd\">" +
                    "<Timestamp xmlns=\"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd\">" +
                    "<Created>2021-07-01T21:57:33.755+02:00</Created>" +
                    "</Timestamp>" +
                    "<saml:Assertion xmlns:saml=\"urn:oasis:names:tc:SAML:2.0:assertion\" IssueInstant=\"2021-07-01T18:49:04Z\">" +
                    "Version=\"2.0\" id=\"IDCard\">" +
                    "<saml:AttributeStatement id=\"IDCardData\">" +
                    "<saml:Attribute Name=\"sosi:IDCardID\">" +
                    "<saml:AttributeValue>qg0054ZapUXEp8jwoJaAQQ==</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"sosi:IDCardVersion\">" +
                    "<saml:AttributeValue>1.0.1</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"sosi:IDCardType\">" +
                    "<saml:AttributeValue>user</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"sosi:AuthenticationLevel\">" +
                    "<saml:AttributeValue>4</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "</saml:AttributeStatement>" +
                    "<saml:AttributeStatement id=\"UserLog\">" +
                    "<saml:Attribute Name=\"medcom:UserCivilRegistrationNumber\">" +
                    "<saml:AttributeValue>2511718360</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:UserGivenName\">" +
                    "<saml:AttributeValue>Mette Frost</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:UserSurName\">" +
                    "<saml:AttributeValue>Pedersen</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:UserEmailAddress\">" +
                    "<saml:AttributeValue>min.email@adatatest.com</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:UserRole\">" +
                    "<saml:AttributeValue>7170</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:UserAuthorizationCode\">" +
                    "<saml:AttributeValue>J0184</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:UserOccupation\">" +
                    "<saml:AttributeValue>Overtester</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "</saml:AttributeStatement>" +
                    "<saml:AttributeStatement id=\"SystemLog\">" +
                    "<saml:Attribute Name=\"medcom:ITSystemName\">" +
                    "<saml:AttributeValue>SOSITEST</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:CareProviderID\" NameFormat=\"medcom:cvrnumber\">" +
                    "<saml:AttributeValue>20921897</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:CareProviderName\">" +
                    "<saml:AttributeValue>TRIFORK A/S</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "</saml:AttributeStatement>" +
                    "</saml:Assertion>" +
                    "</Security>" +
                    "<ns3:WhitelistingHeader xmlns:ns2=\"http://www.sdsd.dk/dgws/2010/08\" xmlns:ns3=\"http://www.sdsd.dk/dgws/2012/06\">" +
                    "<ns2:SystemOwnerName>NSI</ns2:SystemOwnerName>" +
                    "<ns2:SystemName>FMK-Online-ECPR2</ns2:SystemName>" +
                    "<ns2:SystemVersion>TODO</ns2:SystemVersion>" +
                    "<ns2:OrgResponsibleName>NSI</ns2:OrgResponsibleName>" +
                    "<ns2:OrgUsingName>Aalborg Universitetshospital, Alb Nuklearmed. Ambulatorium.</ns2:OrgUsingName>" +
                    "<ns2:OrgUsingID NameFormat=\"medcom:skscode\">8001149</ns2:OrgUsingID>" +
                    "<ns2:RequestedRole>Læge</ns2:RequestedRole>" +
                    "</ns3:WhitelistingHeader>" +
                    "</soap:Header>";

    private static final String ON_BEHALF_OF_AUTHORISATION_SOAP_HEADER =
            "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" +
                    "<soap:Header xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\">" +
                    "<Security xmlns=\"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd\">" +
                    "<Timestamp xmlns=\"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd\">" +
                    "<Created>2021-07-01T21:57:33.755+02:00</Created>" +
                    "</Timestamp>" +
                    "<saml:Assertion xmlns:saml=\"urn:oasis:names:tc:SAML:2.0:assertion\" IssueInstant=\"2021-07-01T18:49:04Z\">" +
                    "Version=\"2.0\" id=\"IDCard\">" +
                    "<saml:AttributeStatement id=\"IDCardData\">" +
                    "<saml:Attribute Name=\"sosi:IDCardID\">" +
                    "<saml:AttributeValue>qg0054ZapUXEp8jwoJaAQQ==</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"sosi:IDCardVersion\">" +
                    "<saml:AttributeValue>1.0.1</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"sosi:IDCardType\">" +
                    "<saml:AttributeValue>user</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"sosi:AuthenticationLevel\">" +
                    "<saml:AttributeValue>4</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "</saml:AttributeStatement>" +
                    "<saml:AttributeStatement id=\"UserLog\">" +
                    "<saml:Attribute Name=\"medcom:UserCivilRegistrationNumber\">" +
                    "<saml:AttributeValue>2511718360</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:UserGivenName\">" +
                    "<saml:AttributeValue>Mette Frost</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:UserSurName\">" +
                    "<saml:AttributeValue>Pedersen</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "</saml:AttributeStatement>" +
                    "<saml:AttributeStatement id=\"SystemLog\">" +
                    "<saml:Attribute Name=\"medcom:ITSystemName\">" +
                    "<saml:AttributeValue>SOSITEST</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:CareProviderID\" NameFormat=\"medcom:cvrnumber\">" +
                    "<saml:AttributeValue>20921897</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:CareProviderName\">" +
                    "<saml:AttributeValue>TRIFORK A/S</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "</saml:AttributeStatement>" +
                    "</saml:Assertion>" +
                    "</Security>" +
                    "<ns3:WhitelistingHeader xmlns:ns2=\"http://www.sdsd.dk/dgws/2010/08\" xmlns:ns3=\"http://www.sdsd.dk/dgws/2012/06\">" +
                    "<ns2:SystemOwnerName>NSI</ns2:SystemOwnerName>" +
                    "<ns2:SystemName>FMK-Online-ECPR2</ns2:SystemName>" +
                    "<ns2:SystemVersion>TODO</ns2:SystemVersion>" +
                    "<ns2:OrgResponsibleName>NSI</ns2:OrgResponsibleName>" +
                    "<ns2:OrgUsingName>Aalborg Universitetshospital, Alb Nuklearmed. Ambulatorium.</ns2:OrgUsingName>" +
                    "<ns2:OrgUsingID NameFormat=\"medcom:skscode\">8001149</ns2:OrgUsingID>" +
                    "<ns2:RequestedRole>Assistent for Læge</ns2:RequestedRole>" +
                    "</ns3:WhitelistingHeader>" +
                    "<ns2:OnBehalfOf xmlns:ns2=\"urn:sds-dk:ecpr2\">" +
                    "<ns2:AuthorisationIdentifier>J0184</ns2:AuthorisationIdentifier>" +
                    "</ns2:OnBehalfOf>" +
                    "</soap:Header>";

    private static final String ON_BEHALF_OF_CPR_SOAP_HEADER =
            "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" +
                    "<soap:Header xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\">" +
                    "<Security xmlns=\"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd\">" +
                    "<Timestamp xmlns=\"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd\">" +
                    "<Created>2021-07-01T21:57:33.755+02:00</Created>" +
                    "</Timestamp>" +
                    "<saml:Assertion xmlns:saml=\"urn:oasis:names:tc:SAML:2.0:assertion\" IssueInstant=\"2021-07-01T18:49:04Z\">" +
                    "Version=\"2.0\" id=\"IDCard\">" +
                    "<saml:AttributeStatement id=\"IDCardData\">" +
                    "<saml:Attribute Name=\"sosi:IDCardID\">" +
                    "<saml:AttributeValue>qg0054ZapUXEp8jwoJaAQQ==</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"sosi:IDCardVersion\">" +
                    "<saml:AttributeValue>1.0.1</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"sosi:IDCardType\">" +
                    "<saml:AttributeValue>user</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"sosi:AuthenticationLevel\">" +
                    "<saml:AttributeValue>4</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "</saml:AttributeStatement>" +
                    "<saml:AttributeStatement id=\"UserLog\">" +
                    "<saml:Attribute Name=\"medcom:UserCivilRegistrationNumber\">" +
                    "<saml:AttributeValue>2511718360</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:UserGivenName\">" +
                    "<saml:AttributeValue>Mette Frost</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:UserSurName\">" +
                    "<saml:AttributeValue>Pedersen</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "</saml:AttributeStatement>" +
                    "<saml:AttributeStatement id=\"SystemLog\">" +
                    "<saml:Attribute Name=\"medcom:ITSystemName\">" +
                    "<saml:AttributeValue>SOSITEST</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:CareProviderID\" NameFormat=\"medcom:cvrnumber\">" +
                    "<saml:AttributeValue>20921897</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "<saml:Attribute Name=\"medcom:CareProviderName\">" +
                    "<saml:AttributeValue>TRIFORK A/S</saml:AttributeValue>" +
                    "</saml:Attribute>" +
                    "</saml:AttributeStatement>" +
                    "</saml:Assertion>" +
                    "</Security>" +
                    "<ns3:WhitelistingHeader xmlns:ns2=\"http://www.sdsd.dk/dgws/2010/08\" xmlns:ns3=\"http://www.sdsd.dk/dgws/2012/06\">" +
                    "<ns2:SystemOwnerName>NSI</ns2:SystemOwnerName>" +
                    "<ns2:SystemName>FMK-Online-ECPR2</ns2:SystemName>" +
                    "<ns2:SystemVersion>TODO</ns2:SystemVersion>" +
                    "<ns2:OrgResponsibleName>NSI</ns2:OrgResponsibleName>" +
                    "<ns2:OrgUsingName>Aalborg Universitetshospital, Alb Nuklearmed. Ambulatorium.</ns2:OrgUsingName>" +
                    "<ns2:OrgUsingID NameFormat=\"medcom:skscode\">8001149</ns2:OrgUsingID>" +
                    "<ns2:RequestedRole>Assistent for Læge</ns2:RequestedRole>" +
                    "</ns3:WhitelistingHeader>" +
                    "<ns2:OnBehalfOf xmlns:ns2=\"urn:sds-dk:ecpr2\">" +
                    "<ns2:PersonIdentifierType><ns2:PersonIdentifier>0501792275</ns2:PersonIdentifier><ns2:Source>0501792275</ns2:Source></ns2:PersonIdentifierType>" +
                    "</ns2:OnBehalfOf>" +
                    "</soap:Header>";


    @Autowired
    private RoleValidationInterceptor roleValidationInterceptor;

    @Autowired
    private AssistantAuthenticator assistantAuthenticator;

    private MessageContext messageContext;
    private SoapHeader soapHeader;

    @Before
    public void setUp() {
        assistantAuthenticator.skipDelegationCheckInTest();
    }

    @Test
    public void testOnBehalfOf() {
        setupContext(ON_BEHALF_OF_AUTHORISATION_SOAP_HEADER);
        try {
            assertEquals(11, roleValidationInterceptor.getRoleToAuthenticatorMap().size());
            boolean handled = roleValidationInterceptor.handleRequest(messageContext, null);
            assertTrue(handled);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Test
    public void testSecurityCredentialsForDoctor() {
        setupContext(DOCTOR_SOAP_HEADER);
        SecurityCredentials securityCredentials = roleValidationInterceptor.getSecurityCredentials((DOMSource) soapHeader.getSource(), new ArrayList<>());
        assertEquals("20921897", securityCredentials.getCvr());
        assertEquals("2511718360", securityCredentials.getUserCpr());
        assertEquals("user", securityCredentials.getIdCardType());
        assertEquals("J0184", securityCredentials.getAuthorisationId());
        assertEquals("Læge", securityCredentials.getRequestedRole());
        assertEquals("FMK-Online-ECPR2", securityCredentials.getSystemName());
        assertNull(securityCredentials.getOnBehalfOfAuthorisationNo());
        assertNull(securityCredentials.getOnBehalfOfPersonIdentifier());
    }

    @Test
    public void testSecurityCredentialsForAuthorisationDelegation() {
        setupContext(ON_BEHALF_OF_AUTHORISATION_SOAP_HEADER);
        SecurityCredentials securityCredentials = roleValidationInterceptor.getSecurityCredentials((DOMSource) soapHeader.getSource(), new ArrayList<>());
        assertEquals("20921897", securityCredentials.getCvr());
        assertEquals("2511718360", securityCredentials.getUserCpr());
        assertEquals("user", securityCredentials.getIdCardType());
        assertNull(securityCredentials.getAuthorisationId());
        assertEquals("Assistent for Læge", securityCredentials.getRequestedRole());
        assertEquals("FMK-Online-ECPR2", securityCredentials.getSystemName());
        assertEquals("J0184", securityCredentials.getOnBehalfOfAuthorisationNo());
        assertNull(securityCredentials.getOnBehalfOfPersonIdentifier()); // OnBehalfOf is a choice, so make sure both aren't set
    }

    @Test
    public void testSecurityCredentialsForCPRDelegation() {
        setupContext(ON_BEHALF_OF_CPR_SOAP_HEADER);
        SecurityCredentials securityCredentials = roleValidationInterceptor.getSecurityCredentials((DOMSource) soapHeader.getSource(), new ArrayList<>());
        assertEquals("20921897", securityCredentials.getCvr());
        assertEquals("2511718360", securityCredentials.getUserCpr());
        assertEquals("user", securityCredentials.getIdCardType());
        assertNull(securityCredentials.getAuthorisationId());
        assertEquals("Assistent for Læge", securityCredentials.getRequestedRole());
        assertEquals("FMK-Online-ECPR2", securityCredentials.getSystemName());
        assertNull(securityCredentials.getOnBehalfOfAuthorisationNo());
        assertEquals("0501792275", securityCredentials.getOnBehalfOfPersonIdentifier());
    }

    @Test
    public void testAssistantForDoctorWithDelegationSkipped() {
        setupContext(ON_BEHALF_OF_AUTHORISATION_SOAP_HEADER);
        try {
            assertEquals(11, roleValidationInterceptor.getRoleToAuthenticatorMap().size());
            boolean handled = roleValidationInterceptor.handleRequest(messageContext, null);
            assertTrue(handled);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private void setupContext(String xml) {
        try {
            DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
            documentBuilderFactory.setNamespaceAware(true);
            DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();

            InputSource inputSource = new InputSource(new StringReader(xml));
            Document document = documentBuilder.parse(inputSource);
            Node node = document.getFirstChild();

            soapHeader = mock(SoapHeader.class);
            when(soapHeader.getSource()).thenReturn(new DOMSource(node));

            SoapMessage requestMessage = mock(SoapMessage.class);
            when(requestMessage.getSoapHeader()).thenReturn(soapHeader);

            messageContext = mock(MessageContext.class);
            when(messageContext.getRequest()).thenReturn(requestMessage);
        } catch (Exception ignored) { /* Won't fail */ }
    }
}
