/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice;

import dk.sds.ecpr2.*;
import dk.sds.ecpr2._2017._10._01.*;
import dk.sds.ecpr2.advis.AsyncAdvisFacade;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.datagenerator.DataFactory;
import dk.sds.ecpr2.util.ECPRException;
import dk.sds.ecpr2.util.ValueSets;
import dk.sds.ecpr2.webservice.validation.ValidationErrors;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.stream.Collectors;

import static java.lang.Thread.sleep;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;

public class UpdatePersonTest extends WebServiceTestAbstract {

    AsyncAdvisFacade advisMock = mock(AsyncAdvisFacade.class);

    @Before
    public void setup() {
        getBusinessService().setAdvisService(advisMock);
    }

    @Rule
    public ExpectedException expectedEx = ExpectedException.none();

    @Test
    public void testUpdateOnReservedPersonId() throws Exception {
        ReservePersonIdRequest reserveRequest = ReservePersonIdTest.createReserveRequest(1, null, null, GenderFemaleOrMaleType.MALE);
        ReservePersonIdResponse response = service.reservePersonId(reserveRequest);
        List<PersonWithGenderType> persons = response.getLists().get(0).getPersons();

        UpdatePersonRequest request = new UpdatePersonRequest();
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(map(persons.get(0)));
        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        // Changing gender doesn't change identifier
        personForUpdate.getGenders().get(0).setValue("female");
        request.setPerson(personForUpdate);
        service.updatePerson(request);
        verify(advisMock, times(1)).advisPersonIdUpdated(any(), any(), any());
        verify(advisMock, times(0)).advisPersonIdDeleted(any(), any(), any());
    }

    private PersonType map(PersonWithGenderType personWithGender) {
        PersonType person = new PersonType();

        person.getGenders().addAll(personWithGender.getGenders());
        person.getIdentifiers().addAll(personWithGender.getIdentifiers());
        person.getModifieds().addAll(personWithGender.getModifieds());

        return person;
    }

    @Test
    public void testRequestNewPrimaryIdentifierException() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = setupPersonForUpdate();

        expectedEx.expect(ECPRException.class);
        expectedEx.expectMessage("Fejl i request ifm. opdatering af eCPR-data, C-eCPR har ingen værdi");

        Identifier identifier = new Identifier();
        identifier.setOID(OID.C_ECPR);
        personForUpdate.getIdentifiers().add(identifier);
        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        callUpdatePerson(personForUpdate);
    }

    @Test
    public void testUpdatePersonWithFakeOIDException() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = setupPersonForUpdate();

        expectedEx.expect(ECPRException.class);
        expectedEx.expectMessage("Fejl i request ifm. opdatering af eCPR-data, Ukendt OID: [1.2.3.4], Ukendt OID: [1.2.3.4]");

        Identifier identifier = new Identifier();
        identifier.setOID("1.2.3.4"); // This is not a real oid
        identifier.setValidity(4);
        personForUpdate.getIdentifiers().add(identifier);
        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        callUpdatePerson(personForUpdate);
    }

    @Test
    public void testSettingExpiry() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = setupPersonForUpdate();

        ContactType addedEntity = factory.createContact("12345678");
        Calendar cal = new GregorianCalendar();
        cal.set(2018, 12, 12, 18, 0, 0);
        cal.set(Calendar.MILLISECOND, 0);
        addedEntity.setExpiry(cal);
        personForUpdate.getContacts().add(addedEntity);

        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        UpdatePersonResponse updatedPerson = callUpdatePerson(personForUpdate);
        assertEquals(2, updatedPerson.getPerson().getContacts().size());
        ContactType returnedEntity = updatedPerson.getPerson().getContacts().get(1);
        assertNotNull(returnedEntity.getPID());
        addedEntity.setValidTo(addedEntity.getExpiry());
        String returned = marshal(returnedEntity);
        String added = marshal(addedEntity);
        assertEquals(returned, added);
        verify(advisMock, times(2)).advisPersonIdUpdated(any(), any(), any());
    }

    @Test
    public void testAddIdentifier() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = setupPersonForUpdate();

        Identifier addedEntity = factory.createMigrantIdentifier();
        personForUpdate.getIdentifiers().add(addedEntity);

        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        UpdatePersonResponse updatedPerson = callUpdatePerson(personForUpdate);
        assertEquals(3, updatedPerson.getPerson().getIdentifiers().size());
        Identifier returnedEntity = updatedPerson.getPerson().getIdentifiers().get(2);
        returnedEntity.setPrimaryOIDLabel(null); //Set returned to null because added did not have it
        returnedEntity.setOIDType(null); //Set returned to null because added did not have it
        assertNotNull(returnedEntity.getPID());
        assertEquals(marshal(returnedEntity), marshal(addedEntity));
    }

    @Test
    public void testAddContact() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = setupPersonForUpdate();

        ContactType addedEntity = factory.createContact("12345678");
        personForUpdate.getContacts().add(addedEntity);

        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        UpdatePersonResponse updatedPerson = callUpdatePerson(personForUpdate);
        assertEquals(2, updatedPerson.getPerson().getContacts().size());
        ContactType returnedEntity = updatedPerson.getPerson().getContacts().get(1);
        assertNotNull(returnedEntity.getPID());
        assertEquals(marshal(returnedEntity), marshal(addedEntity));
    }

    @Test
    public void testAddAddress() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = setupPersonForUpdate();

        AddressType addedEntity = factory.createAddress("Gadevej 44", "Aarhus", "8000", "DK");
        Calendar cal = new GregorianCalendar();
        cal.set(2018, 12, 12, 18, 0, 0);
        cal.set(Calendar.MILLISECOND, 0);
        addedEntity.setExpiry(cal);
        personForUpdate.getAddresses().add(addedEntity);

        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        UpdatePersonResponse updatedPerson = callUpdatePerson(personForUpdate);
        assertEquals(2, updatedPerson.getPerson().getAddresses().size());
        AddressType returnedEntity = updatedPerson.getPerson().getAddresses().get(1);
        assertNotNull(returnedEntity.getPID());
        returnedEntity.setCountryName(null);
        addedEntity.setValidTo(addedEntity.getExpiry());
        assertEquals(marshal(returnedEntity), marshal(addedEntity));
    }

    @Test
    public void testAddNameWithDifferentUseOk() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = setupPersonForUpdate();
        sleep(1000);
        NameType addedEntity = factory.createName("Hansen", "Hans", ValueSets.nameUseValues[1]);
        Calendar cal = new GregorianCalendar();
        cal.set(2018, 12, 12, 18, 0, 0);
        cal.set(Calendar.MILLISECOND, 0);
        addedEntity.setExpiry(cal);
        personForUpdate.getNames().add(addedEntity);
        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        UpdatePersonResponse updatedPerson = callUpdatePerson(personForUpdate);
        assertEquals(2, updatedPerson.getPerson().getNames().size());
        NameType returnedEntity = updatedPerson.getPerson().getNames().get(1);
        assertNotNull(returnedEntity.getPID());
        assertNotNull(returnedEntity.getText());
        returnedEntity.setText(null);
        addedEntity.setValidTo(cal);
        assertEquals(marshal(returnedEntity), marshal(addedEntity));
        assertEquals(2, updatedPerson.getPerson().getModifieds().size());
    }

    @Test
    public void testVersionIsUpdatedOk() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = setupPersonForUpdate();
        sleep(1000);
        NameType addedEntity = factory.createName("Hansen", "Hans", ValueSets.nameUseValues[1]);
        personForUpdate.getNames().add(addedEntity);

        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        UpdatePersonResponse updatedPerson = callUpdatePerson(personForUpdate, "1234123412");
        assertEquals(2, updatedPerson.getPerson().getModifieds().size());
        System.out.println(DataFactory.asString(updatedPerson));
        assertEquals("12345678901", updatedPerson.getPerson().getModifieds().get(0).getBy().getOrganisation().getIdentifier().getIdentifier());
        assertEquals("1234123412", updatedPerson.getPerson().getModifieds().get(1).getBy().getOrganisation().getIdentifier().getIdentifier());
    }

    @Test
    public void testAddNameWithSameUseMustFail() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = setupPersonForUpdate();
        NameType addedEntity = factory.createName("Hansen", "Hans", ValueSets.nameUseValues[0]);
        personForUpdate.getNames().add(addedEntity);
        sleep(1000); // MySQL may cause ValidFrom to be in the future on existing record
        expectError(personForUpdate, ValidationErrors.DUPLICATE_NAME_USE);
    }

    @Test
    public void testAddBirthDateMustFail() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = setupPersonForUpdate();
        BirthDateType addedEntity = factory.createBirthDate("1998");
        personForUpdate.getBirthDates().add(addedEntity);
        sleep(1000); // MySQL may cause ValidFrom to be in the future on existing record
        expectError(personForUpdate, ValidationErrors.DUPLICATE_BIRTHDATE);
    }

    @Test
    public void testAddGenderMustFail() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = setupPersonForUpdate();

        GenderType addedEntity = factory.createGender("male");
        personForUpdate.getGenders().add(addedEntity);
        sleep(1000); // MySQL may cause ValidFrom to be in the future on existing record
        expectError(personForUpdate, ValidationErrors.DUPLICATE_GENDER);
        // A validation error should mean that no advis has been sent
        verify(advisMock, times(0)).advisPersonIdUpdated(any(), any(), any());
        verify(advisMock, times(0)).advisPersonIdDeleted(any(), any(), any());
    }

    @Test
    public void testSimultaneousAddAndRemoveGenderMustFail() throws Exception {
        // We don't allow simultaneous remove and add. In that case an update should be used
        PersonType newPerson = createPerson();
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(newPerson);

        GenderType gender = personForUpdate.getGenders().get(0);
        gender.setValue(null);
        gender.setValidTo(null);
        gender.setValidFrom(null);
        GenderType addedEntity = factory.createGender("male");
        factory.clearAllExceptKeyIdentifier(personForUpdate);

        personForUpdate.getGenders().add(gender);
        personForUpdate.getGenders().add(addedEntity);

        expectError(personForUpdate, ValidationErrors.DUPLICATE_GENDER);
    }

    @Test
    public void testDeleteAddress() throws Exception {
        PersonType newPerson = createPerson();
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(newPerson);
        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);

        AddressType address = new AddressType();
        // Setting only PID means delete
        address.setPID(newPerson.getAddresses().get(0).getPID());
        factory.clearAllExceptKeyIdentifier(personForUpdate);
        personForUpdate.getAddresses().add(address);

        UpdatePersonResponse updatedPerson = callUpdatePerson(personForUpdate);

        assertEquals(1, updatedPerson.getPerson().getAddresses().size());
        AddressType updatedAddress = updatedPerson.getPerson().getAddresses().get(0);
        // Address must have ValidTo set now
        assertNotNull(updatedAddress.getValidTo());
    }

    @Test
    public void testUpdateWithIllegalBirthdate() throws Exception {
        PersonType newPerson = createPerson();
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(newPerson);
        UpdatePersonRequest request = new UpdatePersonRequest();
        personForUpdate.getBirthDates().get(0).setValue("1840");
        request.setPerson(personForUpdate);
        try {
            UpdatePersonResponse updatedPerson = service.updatePerson(request);
            fail("Should have caused error message: " + ValidationErrors.BIRTHDATE_BAD_YEAR);
        } catch (Exception e) {
            checkError(e, ValidationErrors.BIRTHDATE_BAD_YEAR);
        }
    }

    @Test
    public void testUpdateIdentifier() throws Exception {
        PersonType newPerson = createPerson();
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(newPerson);
        UpdatePersonRequest request = new UpdatePersonRequest();
        Identifier identifier = personForUpdate.getIdentifiers().get(0);
        String newRegionECPR = identifierGenerator.getRandomXECPR();
        identifier.setValue(newRegionECPR);
        request.setPerson(personForUpdate);
        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        UpdatePersonResponse updatedPerson = service.updatePerson(request);
        assertEquals(3, updatedPerson.getPerson().getIdentifiers().size());
        List<Identifier> xeCPRs = updatedPerson.getPerson().getIdentifiers().stream().filter(i -> OID.X_ECPR.equals(i.getOID())).collect(Collectors.toList());
        List<Identifier> regions = updatedPerson.getPerson().getIdentifiers().stream().filter(i -> OID.ECPR_REGION_SOUTH.equals(i.getOID())).collect(Collectors.toList());
        // Check that a new "Region" identifier was added with only ValidFrom set and the new value. The old identifier must have ValidTo set
        assertEquals(1, xeCPRs.size());
        assertEquals(2, regions.size());
        assertEquals(1, regions.stream().filter(i -> i.getValidTo() == null).collect(Collectors.toList()).size());
        assertEquals(1, regions.stream().filter(i -> i.getValidFrom() != null && i.getValidTo() != null).collect(Collectors.toList()).size());
        assertEquals(newRegionECPR, regions.stream().filter(i -> i.getValidTo() == null).collect(Collectors.toList()).get(0).getValue());
    }

    @Test
    public void testUpdateXeCPR() throws Exception {
        PersonType newPerson = createPerson();
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(newPerson);
        UpdatePersonRequest request = new UpdatePersonRequest();
        Identifier xeCPRIdentifier = newPerson.getIdentifiers().get(0);
        String newECPR = identifierGenerator.getRandomXECPR();
        xeCPRIdentifier.setValue(newECPR);
        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        personForUpdate.getIdentifiers().add(xeCPRIdentifier); // Update the C-eCPR
        request.setPerson(personForUpdate);
        UpdatePersonResponse updatedPerson = service.updatePerson(request);
        assertEquals(3, updatedPerson.getPerson().getIdentifiers().size());
        List<Identifier> xeCPRs = updatedPerson.getPerson().getIdentifiers().stream().filter(i -> OID.X_ECPR.equals(i.getOID())).collect(Collectors.toList());
        List<Identifier> regions = updatedPerson.getPerson().getIdentifiers().stream().filter(i -> OID.ECPR_REGION_SOUTH.equals(i.getOID())).collect(Collectors.toList());
        // Check that a new "Region" identifier was added with only ValidFrom set and the new value. The old identifier must have ValidTo set
        assertEquals(2, xeCPRs.size());
        assertEquals(1, regions.size());
        assertEquals(1, xeCPRs.stream().filter(i -> i.getValidTo() == null).collect(Collectors.toList()).size());
        assertEquals(1, xeCPRs.stream().filter(i -> i.getValidFrom() != null && i.getValidTo() != null).collect(Collectors.toList()).size());
        assertEquals(newECPR, xeCPRs.stream().filter(i -> i.getValidTo() == null).collect(Collectors.toList()).get(0).getValue());
    }

    @Test
    public void testDeleteIdentifier() throws Exception {
        PersonType newPerson = createPerson();
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(newPerson);
        UpdatePersonRequest request = new UpdatePersonRequest();
        Identifier identifier = personForUpdate.getIdentifiers().get(0); // "Region" identifier
        cleanIdentifierDataExceptPID(identifier);
        request.setPerson(personForUpdate);
        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        UpdatePersonResponse updatedPerson = service.updatePerson(request);
        debug(updatedPerson);
        assertEquals(2, updatedPerson.getPerson().getIdentifiers().size());
        // Only one may have validTo set
        assertEquals(1, updatedPerson.getPerson().getIdentifiers().stream().filter(i -> i.getValidTo() != null).collect(Collectors.toList()).size());
        List<Identifier> xeCPRs = updatedPerson.getPerson().getIdentifiers().stream().filter(i -> OID.X_ECPR.equals(i.getOID())).collect(Collectors.toList());
        List<Identifier> regions = updatedPerson.getPerson().getIdentifiers().stream().filter(i -> OID.ECPR_REGION_SOUTH.equals(i.getOID())).collect(Collectors.toList());
        // Check that the "Region" identifier was removed
        assertEquals(1, xeCPRs.size());
        assertEquals(1, regions.size());
        assertNotNull(regions.get(0).getValidTo()); // Region identifier must be closed
        // Both identifiers must be notified even though one is deleted
        verify(advisMock, times(2)).advisPersonIdUpdated(any(), any(), any());
        verify(advisMock, times(0)).advisPersonIdDeleted(any(), any(), any());
    }

    @Test
    public void testUpdateWithSameDataDoesNotProduceNewVersions() throws Exception {
        PersonType newPerson = createPerson();
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(newPerson);
        UpdatePersonRequest request = new UpdatePersonRequest();
        request.setPerson(personForUpdate);
        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        PersonType updatedPerson1 = service.updatePerson(request).getPerson();
        PersonType updatedPerson2 = service.updatePerson(request).getPerson();

        assertEquals(updatedPerson1.getBirthDates().size(), updatedPerson2.getBirthDates().size());
        assertEquals(updatedPerson1.getNames().size(), updatedPerson2.getNames().size());
        assertEquals(updatedPerson1.getContacts().size(), updatedPerson2.getContacts().size());
        assertEquals(updatedPerson1.getGenders().size(), updatedPerson2.getGenders().size());
        assertEquals(updatedPerson1.getIdentifiers().size(), updatedPerson2.getIdentifiers().size());
        assertEquals(updatedPerson1.getAddresses().size(), updatedPerson2.getAddresses().size());
    }

    private void cleanIdentifierDataExceptPID(Identifier identifier) {
        // Clear all data except PID
        identifier.setValue(null);
        identifier.setOID(null);
        identifier.setValidity(null);
        identifier.setExpiry(null);
    }

    @Test
    public void testFailureOnBadKeyIdentifier() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(createPerson());
        personForUpdate.getKeyIdentifier().setValue("0101010AA0"); // Non-existing
        expectError(personForUpdate, ValidationErrors.KEY_IDENTIFIER_INVALID);
    }

    @Test
    public void testFailureOnDuplicateName() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(createPerson());
        personForUpdate.getNames().add(personForUpdate.getNames().get(0));
        expectError(personForUpdate, ValidationErrors.DUPLICATE_NAME_USE);
    }

    @Test
    public void testFailureForDuplicateGender() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(createPerson());
        personForUpdate.getGenders().add(personForUpdate.getGenders().get(0));
        expectError(personForUpdate, ValidationErrors.DUPLICATE_GENDER);
    }

    @Test
    public void testFailureForDuplicateBirthDate() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(createPerson());
        personForUpdate.getBirthDates().add(personForUpdate.getBirthDates().get(0));
        expectError(personForUpdate, ValidationErrors.DUPLICATE_BIRTHDATE);
    }

    @Test
    public void testFailureOnUnknownAddressPID() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(createPerson());
        personForUpdate.getAddresses().get(0).setPID("unknown-pid");
        expectError(personForUpdate, ValidationErrors.UNKNOWN_PID);
    }

    @Test
    public void testFailureOnUnknownNamePID() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(createPerson());
        personForUpdate.getNames().get(0).setPID("unknown-pid");
        expectError(personForUpdate, ValidationErrors.UNKNOWN_PID);
    }

    @Test
    public void testFailureOnUnknownIdentifierPID() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(createPerson());
        personForUpdate.getIdentifiers().get(0).setPID("unknown-pid");
        expectError(personForUpdate, ValidationErrors.UNKNOWN_PID);
    }

    @Test
    public void testFailureOnUnknownGenderPID() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(createPerson());
        personForUpdate.getGenders().get(0).setPID("unknown-pid");
        expectError(personForUpdate, ValidationErrors.UNKNOWN_PID);
    }

    @Test
    public void testFailureOnUnknownContactPID() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(createPerson());
        personForUpdate.getContacts().get(0).setPID("unknown-pid");
        expectError(personForUpdate, ValidationErrors.UNKNOWN_PID);
    }

    @Test
    public void testFailureOnUnknownBirthDatePID() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(createPerson());
        personForUpdate.getBirthDates().get(0).setPID("unknown-pid");
        expectError(personForUpdate, ValidationErrors.UNKNOWN_PID);
    }

    @Test
    public void test_AddAndSearchIdentifier() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = setupPersonForUpdate();

        Identifier addedEntity = createMigrantIdentifier();
        personForUpdate.getIdentifiers().add(addedEntity);

        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        personForUpdate.getModifieds().get(0).getBy().getPerson().setPersonIdentifier(personIdentifier);
        UpdatePersonResponse updatedPerson = callUpdatePerson(personForUpdate);
        assertEquals(3, updatedPerson.getPerson().getIdentifiers().size());
        Identifier returnedEntity = updatedPerson.getPerson().getIdentifiers().get(2);

        GetPersonByIdRequest request = new GetPersonByIdRequest();
        KeyIdentifier keyIdentifier = new KeyIdentifier();
        keyIdentifier.setOID(returnedEntity.getOID());
        keyIdentifier.setValue(returnedEntity.getValue());
        request.getKeyIdentifiers().add(keyIdentifier);

        GetPersonByIdResponse response = service.getPersonById(request);
        assertNotNull(response);
    }

    @Test
    public void testValidCPRNumber() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(createPerson());
        personForUpdate.getIdentifiers().add(factory.createCprIdentifier("1111111118"));
        UpdatePersonRequest request = new UpdatePersonRequest();
        request.setPerson(personForUpdate);
        UpdatePersonResponse response = service.updatePerson(request);
        PersonType person = response.getPerson();
        assertNotNull(person);
        assertEquals(3, person.getIdentifiers().size());
        assertTrue(person.getIdentifiers().stream().anyMatch(i -> OID.CPR.equals(i.getOID()) && "1111111118".equals(i.getValue())));
    }

    @Test
    public void testInvalidCPRNumber() throws Exception {
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(createPerson());
        personForUpdate.getIdentifiers().add(factory.createCprIdentifier("0000000000"));
        UpdatePersonRequest request = new UpdatePersonRequest();
        request.setPerson(personForUpdate);
        try {
            service.updatePerson(request);
        } catch (ECPRException e) {
            assertEquals(e.getMessage(), "Fejl i request ifm. opdatering af eCPR-data, Cpr identifier er ikke valid");
        }
    }

    private Identifier createMigrantIdentifier() {
        Identifier identifier = new Identifier();
        identifier.setValidity(0);
        identifier.setOID(OID.MIGRANT_IDENTIFIER);
        identifier.setValue("1234567891");
        return identifier;
    }

    private void expectError(PersonWithKeyIdentifierType personForUpdate, String expectedError) {
        UpdatePersonRequest request = new UpdatePersonRequest();
        request.setPerson(personForUpdate);
        try {
            debug(request);
            UpdatePersonResponse response = service.updatePerson(request);
            debug(response);
            fail("Expected validation error. Response above");
        } catch (Exception e) {
            checkError(e, expectedError);
        }
    }

    private PersonWithKeyIdentifierType setupPersonForUpdate() throws Exception {
        PersonType newPerson = createPerson();
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(newPerson);
        factory.clearAllExceptKeyIdentifier(personForUpdate);
        return personForUpdate;
    }

    private UpdatePersonResponse callUpdatePerson(PersonWithKeyIdentifierType personForUpdate) throws Exception {
        return callUpdatePerson(personForUpdate, null);
    }

    private UpdatePersonResponse callUpdatePerson(PersonWithKeyIdentifierType personForUpdate, String orgId) throws Exception {
        if (orgId != null) {
            personForUpdate.getModifieds().get(0).getBy().getOrganisation().getIdentifier().setIdentifier(orgId);
        }
        UpdatePersonRequest request = new UpdatePersonRequest();
        request.setPerson(personForUpdate);
        return service.updatePerson(request);
    }

}
