/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.webservice;

import dk.sds.ecpr2.*;
import dk.sds.ecpr2._2017._10._01.*;
import dk.sds.ecpr2.datagenerator.DataFactory;
import dk.sds.ecpr2.datagenerator.IdentifierGenerator;
import dk.sds.ecpr2.domain.PersonVO;
import dk.sds.ecpr2.service.ECPR2ServiceImpl;
import dk.sds.ecpr2.util.ECPRException;
import org.springframework.beans.factory.annotation.Autowired;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.namespace.QName;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;

import static java.lang.Thread.sleep;
import static org.hamcrest.CoreMatchers.containsString;
import static org.junit.Assert.assertThat;
import static org.junit.Assert.assertTrue;

public abstract class WebServiceTestAbstract extends IntegrationTest {

    @Autowired
    IdentifierGenerator identifierGenerator;

    @Autowired
    DataFactory factory;

    @Autowired
    protected ECPR2WebServiceImpl service;

    @Autowired
    protected ECPR2MasterDataWebServiceImpl masterDataservice;

    protected PersonType createPerson(String givenName, String familyName, String streetAddressLine, String city, String postalCode, String birthDate, String countryCode) {
        return factory.createPerson(givenName, familyName, streetAddressLine, city, postalCode, birthDate, countryCode);
    }

    protected SimplePersonWithKeyIdentifierType createNewPersonId(PersonType createdPerson, Identifier toBeCreated) {
        String oid = createdPerson.getIdentifiers().get(0).getOID();
        String value = createdPerson.getIdentifiers().get(0).getValue();
        SimplePersonWithKeyIdentifierType personKeyIdentifier = new SimplePersonWithKeyIdentifierType();
        personKeyIdentifier.setKeyIdentifier(factory.createKeyIdentifier(oid, value));
        personKeyIdentifier.getIdentifiers().add(toBeCreated);
        createdPerson.getModifieds().forEach(m -> m.setValidTo(null));
        createdPerson.getModifieds().forEach(m -> m.setValidFrom(null));
        personKeyIdentifier.getModifieds().addAll(createdPerson.getModifieds());
        return personKeyIdentifier;
    }

    protected PersonType createDefaultPerson() {
        PersonType person = new PersonType();
        factory.populatePersonType(person);
        return person;
    }

    protected void assertMessageContainsError(String message, String error) {
        if (error.contains("%")) {
            String errorWithoutWildcard = error.substring(0, error.indexOf("%"));
            assertThat(message, containsString(errorWithoutWildcard));
        } else {
            assertThat(message, containsString(error));
        }
    }

    protected void checkError(Exception ex, String error) {
        Exception e = (Exception) ex.getCause();
        String message = e.getMessage();
        assertMessageContainsError(message, error);
    }

    protected void checkECPRExceptionError(Exception e, String error) {
        ECPRException exception = (ECPRException) e.getCause();
        String message = exception.getMessage();
        assertMessageContainsError(message, error);
    }

    protected void debug(Object request) throws JAXBException {
        System.out.println(DataFactory.asString(request));
    }

    protected PersonType createPerson() throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        request.setPerson(createDefaultPerson());
        CreatePersonResponse response = service.createPerson(request);
        return response.getPerson();
    }

    protected PersonType updatePersonInDb(PersonType person) throws Exception {
        UpdatePersonRequest request = new UpdatePersonRequest();
        PersonWithKeyIdentifierType personForUpdate = factory.createPersonForUpdate(person);
        request.setPerson(personForUpdate);
        sleep(1000);
        UpdatePersonResponse response = service.updatePerson(request);
        return response.getPerson();
    }

    protected String marshal(EntityType entityType) throws JAXBException {
        clearGeneratedFields(entityType);
        StringWriter stringWriter = new StringWriter();
        JAXBContext jaxbContext = JAXBContext.newInstance(EntityType.class);
        Marshaller jaxbMarshaller = jaxbContext.createMarshaller();
        jaxbMarshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
        JAXBElement<EntityType> root = new JAXBElement<>(new QName("", "foo"), EntityType.class, entityType);
        jaxbMarshaller.marshal(root, stringWriter);
        return stringWriter.toString();
    }

    protected void clearGeneratedFields(EntityType entity) {
        entity.setPID(null);
        entity.setValidFrom(null);
//        entity.setValidTo(null);
    }

    protected Calendar now() {
        Calendar now = Calendar.getInstance();
        now.set(Calendar.MILLISECOND, 0);
        return now;
    }

    protected ECPR2ServiceImpl getBusinessService() {
        return service.getBusinessService();
    }

    protected DeletePersonRequest setRequest(PersonVO person, boolean invalidRequest) {
        DeletePersonRequest deletePersonRequest = new DeletePersonRequest();
        PersonIdType personIdType = new PersonIdType();
        KeyIdentifier keyIdentifier = checkValidRequest(person, invalidRequest);
        personIdType.setKeyIdentifier(keyIdentifier);
        deletePersonRequest.getPersons().add(personIdType);
        deletePersonRequest.setModified(factory.createModified("12345", "Jens Jensen", "1111111118", "CPR", "Læge", "Hospitalsvej 1", "hospital@hospital.dk", "Hospitalets hospital", "12345678", "hospital", "12345678901", "SOR"));
        return deletePersonRequest;
    }

    private KeyIdentifier checkValidRequest(PersonVO person, boolean invalidRequest) {
        KeyIdentifier keyIdentifier = new KeyIdentifier();
        if (invalidRequest) {
            keyIdentifier.setOID(null);
            keyIdentifier.setValue(null);
        } else {
            keyIdentifier.setOID(person.getIdentifiers().get(0).getOid());
            keyIdentifier.setValue(person.getIdentifiers().get(0).getValue());
        }
        return keyIdentifier;
    }

    protected List<PersonType> createPersonsInDb(List<PersonType> personTypeList) throws Exception {
        List<PersonType> personList = new ArrayList<>();
        for (PersonType person : personTypeList) {
            CreatePersonRequest request = new CreatePersonRequest();
            request.setPerson(person);
            sleep(1000);
            CreatePersonResponse response = service.createPerson(request);
            personList.add(response.getPerson());
        }
        return personList;
    }

    protected PersonType createPersonInDb(PersonType person) throws Exception {
        CreatePersonRequest request = new CreatePersonRequest();
        request.setPerson(person);
        CreatePersonResponse response = service.createPerson(request);
        sleep(1000);
        return response.getPerson();
    }

    protected void isRecentlyInvalidated(EntityType entityType, int seconds) {
        Calendar now = now();
        Calendar recently = now();
        recently.set(Calendar.SECOND, recently.get(Calendar.SECOND) - seconds);
        assertTrue(!recently.after(entityType.getValidTo()) && !now.before(entityType.getValidTo()));
    }

}
