/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2;

import dk.nsp.test.idp.NspSoapClient;
import dk.sosi.seal.xml.XmlUtil;
import lombok.extern.log4j.Log4j2;
import org.apache.commons.io.IOUtils;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.protocol.BasicHttpContext;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.*;
import java.io.*;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.Properties;

import static dk.nsp.test.idp.EmployeeIdentities.lægeCharlesBabbage;

@Log4j2
public abstract class AbstractITCase {
    public static final String TESTFILE_PATH = "src/test/resources/testfiles/";
    protected final static String HOST = "ecpr2";
    protected final static int PORT = 8380;
    protected final static String ECPR_URL_PREFIX = "http://" + HOST + ":" + PORT + "/ecpr2/";
    protected final static String SOAP_NAMESPACE = "http://sds.dk/ecpr2/2017/10/01/";

    protected HttpResponse httpGet(String urlMapping) throws IOException {
        HttpClient httpClient = getHttpClient();
        log.info("Http Get : " + ECPR_URL_PREFIX + urlMapping + "/");
        return httpClient.execute(new HttpGet(ECPR_URL_PREFIX + urlMapping + "/"), new BasicHttpContext());
    }

    private HttpClient getHttpClient() {
        return HttpClientBuilder.create().build();
    }

    protected String makeDGWSHttpCall(String requestMessage, String SOAP_ACTION, String requestedRole) throws Exception {

        Element request = XmlUtil.readXml(new Properties(), requestMessage, false).getDocumentElement();

        String whitelistingHeader = "<sdsd201206:WhitelistingHeader xmlns:sdsd=\"http://www.sdsd.dk/dgws/2010/08\" xmlns:sdsd201206=\"http://www.sdsd.dk/dgws/2012/06\">" +
                "   <sdsd:SystemOwnerName>SDS</sdsd:SystemOwnerName>" +
                "   <sdsd:SystemName>FMK-Online</sdsd:SystemName>" +
                "   <sdsd:SystemVersion>0.5</sdsd:SystemVersion>" +
                "   <sdsd:OrgResponsibleName>TestOrg</sdsd:OrgResponsibleName>" +
                "   <sdsd:OrgUsingName>TestOrgName</sdsd:OrgUsingName>" +
                "   <sdsd:OrgUsingId>1234</sdsd:OrgUsingId>" +
                "   <sdsd:RequestedRole>" + requestedRole + "</sdsd:RequestedRole>" +
                "</sdsd201206:WhitelistingHeader>";

        Element whitelistingHeaderElement = XmlUtil.readXml(new Properties(), whitelistingHeader, false).getDocumentElement();

        try (NspSoapClient client = new NspSoapClient(log::info)) {
            try (NspSoapClient.NspSoapResponse response = client.request(URI.create(ECPR_URL_PREFIX + "services/ecpr2/"), SOAP_ACTION).as(lægeCharlesBabbage()).execute(request, whitelistingHeaderElement)) {
                return IOUtils.toString(response.getResponse(), StandardCharsets.UTF_8);
            }
        }
    }

    protected String insertTestData(String xml, String tokenName, String value) {
        if (xml.contains(tokenName)) {
            xml = xml.replaceAll(tokenName, value);
            log.info(" Inserted " + tokenName + " value '" + value + "' into request");
        }
        return xml;
    }

    protected String getRequestFileName(String requestName) {
        return new File(TESTFILE_PATH + requestName + "_request.xml").getAbsolutePath();
    }

    protected String readFile(String path) throws IOException {
        byte[] encoded = Files.readAllBytes(Paths.get(path));
        return new String(encoded, StandardCharsets.UTF_8);
    }

    protected String getXmlValue(String documentText, String parentNode, String childNode, String attributeName) {
        String result = null;

        // Currently supported: //parentNode//childNode[attributeName]
        String xPathString = "//*[local-name()='" + parentNode + "']//*[local-name()='" + childNode
                + "']";
        if (attributeName != null) {
            xPathString += "/@*[local-name()='" + attributeName + "']";
        }

        try {
            DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
            DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
            Document doc = docBuilder.parse(new InputSource(new StringReader(documentText)));
            XPathFactory xPathfactory = XPathFactory.newInstance();
            XPath xpath = xPathfactory.newXPath();
            XPathExpression expr = xpath.compile(xPathString);
            result = expr.evaluate(doc, XPathConstants.STRING).toString();
        } catch (IOException | SAXException | XPathExpressionException | ParserConfigurationException e) {
            e.printStackTrace();
        }
        return result;
    }

    // Source: https://stackoverflow.com/questions/25864316/pretty-print-xml-in-java-8/33541820#33541820
    protected static String toPrettyString(String xml) {
        try {
            // Turn xml string into a document
            Document document = DocumentBuilderFactory.newInstance()
                    .newDocumentBuilder()
                    .parse(new InputSource(new ByteArrayInputStream(xml.getBytes(StandardCharsets.UTF_8))));

            // Remove whitespaces outside tags
            document.normalize();
            XPath xPath = XPathFactory.newInstance().newXPath();
            NodeList nodeList = (NodeList) xPath.evaluate("//text()[normalize-space()='']",
                    document,
                    XPathConstants.NODESET);

            for (int i = 0; i < nodeList.getLength(); ++i) {
                Node node = nodeList.item(i);
                node.getParentNode().removeChild(node);
            }

            // Setup pretty print options
            TransformerFactory transformerFactory = TransformerFactory.newInstance();
            Transformer transformer = transformerFactory.newTransformer();
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
            transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");

            // Return pretty print xml string
            StringWriter stringWriter = new StringWriter();
            transformer.transform(new DOMSource(document), new StreamResult(stringWriter));
            return stringWriter.toString();
        } catch (Exception e) {
            log.error(e);
            return "Error: Not valid XML: " + xml;
        }
    }
}
