/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2;

import dk.nsp.test.idp.NspSoapClient;
import lombok.extern.log4j.Log4j2;
import org.junit.Test;

import javax.xml.soap.SOAPException;
import java.io.IOException;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;

import static org.junit.Assert.*;

/**
 * Test service by executing a series of create, update and delete requests
 */
@Log4j2
public class FlowTestITCase extends AbstractITCase {

    private static final String CREATE_PERSON = "CreatePerson";
    private static final String GET_PERSON_BY_ID = "GetPersonById";
    private static final String DELETE_PERSON = "DeletePerson";
    private static final String SEARCH_PERSON = "SearchPerson";
    private static final String RESERVE_PERSON_ID = "ReservePersonId";
    private static final String UPDATE_PERSON = "UpdatePerson";

    @Test
    public void typicalCRUDFlow() throws Exception {

        String response = makeAndValidateHttpRequest(CREATE_PERSON);
        String identifierValue = getXmlValue(response, "Identifier", "Value", null);
        String pid = getXmlValue(response, "Identifier", "PID", null);

        makeAndValidateHttpRequest(GET_PERSON_BY_ID, identifierValue, null,
                Arrays.asList(":Modified", ":Gender", ":Name", ":BirthDate", ":Contact", ":Address"));

        makeAndValidateHttpRequest(UPDATE_PERSON, identifierValue, pid);

        makeAndValidateHttpRequest(GET_PERSON_BY_ID, identifierValue, null,
                Arrays.asList(":Modified", ":Gender", ":Name", ":BirthDate", ":Contact", ":Address"));

        makeAndValidateHttpRequest(DELETE_PERSON, identifierValue, null);
    }

    @Test
    public void updateAndSearch() throws Exception {
        String response = makeAndValidateHttpRequest(CREATE_PERSON);
        String nameValue = getXmlValue(response, "Name", "GivenName", null);
        String identifierValue = getXmlValue(response, "Identifier", "Value", null);
        String pid = getXmlValue(response, "Identifier", "PID", null);

        makeAndValidateHttpRequest(UPDATE_PERSON, identifierValue, pid);

        String searchResponse = makeAndValidateHttpRequest(SEARCH_PERSON, "givenname:" + nameValue, null);
        String searchResult = getXmlValue(searchResponse, "SearchPersonResponse", "NumberOfPersons", null);
        assertEquals(searchResult, "2"); // Number of Persons is arbitrary, it needs to be fixed
    }

    @Test
    public void search() throws Exception {
        makeAndValidateHttpRequest(CREATE_PERSON);
        makeAndValidateHttpRequest(SEARCH_PERSON, "givenname:WillNotBeFound", null);
    }


    @Test
    public void reserveIds() throws Exception {
        String forDate = new Date().getYear() + 1902 + "-12-24";
        makeAndValidateHttpRequest(RESERVE_PERSON_ID, forDate, null,
                Arrays.asList(":Modified", ":Gender", ":Identifier"));
    }



    private String makeAndValidateHttpRequest(String requestName) throws Exception {
        return makeAndValidateHttpRequest(requestName, null, null, new ArrayList<>());
    }

    private String makeAndValidateHttpRequest(String requestName, String replacementValue, String pidValue) throws Exception {
        return makeAndValidateHttpRequest(requestName, replacementValue, pidValue, new ArrayList<>());
    }

    private String makeAndValidateHttpRequest(String requestName, List<String> validationValues) throws Exception {
        return makeAndValidateHttpRequest(requestName, null, null, validationValues);
    }

    private String makeAndValidateHttpRequest(String requestName, String replacementValue, String pidValue, List<String> validationValues) throws Exception {
        log.info("############################# Testing request: " + requestName + " at " + Instant.now() + " ################################");

        String requestMessage = readFile(getRequestFileName(requestName));

        if (replacementValue != null) {
            requestMessage = insertTestData(requestMessage, "PLACEHOLDER", replacementValue);
        }

        if (pidValue != null) {
            requestMessage = insertTestData(requestMessage, "IDENTIFIER_PID", pidValue);
        }

        String responseMessage = makeDGWSHttpCall(requestMessage, getSOAPAction(requestName), "Læge");
        log.info("Called service.");
        log.info("REQUEST: \r\n" + toPrettyString(requestMessage));
        assertNotNull(responseMessage);
        log.info("RESPONSE: \r\n" + responseMessage);
        log.info("pretty printed response: \r\n" + toPrettyString(responseMessage));

        assertDoesNotContain(responseMessage, "fault");
        assertDoesNotContain(responseMessage, "error");

        for (String validationValue : validationValues) {
            assertContains(responseMessage, validationValue);
        }

        System.out.println("SUCCESSFUL: " + requestName + " at " + Instant.now());
        return responseMessage;
    }

    private void assertDoesNotContain(String message, String text) {
        assertFalse("The message contained the forbidden text: " + text, message.toLowerCase().contains(text.toLowerCase()));
    }

    private void assertContains(String message, String text) {
        assertTrue("The message did not contain the required text: " + text, message.toLowerCase().contains(text.toLowerCase()));
    }

    private String getSOAPAction(String requestName) {
        return SOAP_NAMESPACE + "#" + requestName;
    }
}
